/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.collection;

/**
 * This class provides a low-synchronization list of objects.  It is
 * designed to allow proper traversal of the list while objects are
 * being added and/or removed, and to allow simultaneous adds and
 * removals.
 * <p>
 * The design is to have a list of <code>FastList.Node</code> nodes, each of
 * which has a <code>next</code> reference and a <code>removed</code>
 * flag.  New nodes are added at the tail, requiring synchronization.
 * When a node is removed from the list, its <code>removed</code> flag
 * is set to true in synchronized code.  These are two of the three
 * points of synchronization (see below for the third).
 * <p>
 * Nodes are not directly removed from the list.  When a reader of the
 * list traverses the list, invoking <code>FastList.Node.next</code>,
 * removed nodes are skipped over, and the node's <code>next</code>
 * reference is updated to skip over them.  Thus, removed nodes are
 * progressively orphaned as the node connects around them, leaving them
 * unreachable, and hence garbage.  
 * <p>
 * All this means that users of the list will occasionally see a
 * removed node in the list, because it may be marked removed after it
 * has been reached while traversing the list.  This will be rare, but
 * it is up to the user of the list to skip over nodes that have been
 * removed if necessary, and to handle or avoid attempted re-removal of
 * a node.  This is why <code>remove</code> returns a boolean -- if the
 * node has already been removed by the time the attempt to remove it
 * happens, <code>remove</code> will return <code>false</code>.  This
 * is done synchronizing on the node itself.  A typical piece of code
 * to remove would look like this:
 *
 * <pre>
 *  public synchronized Value findMatch(Value matchFor) {
 *	for (Value val = list.head(); val != null; val = val.next()) {
 *	    if (matchFor.matches(val)) {
 *		if (!list.remove(n))	// oh, well -- someone got to it first
 *		    continue;		// keep looking
 *		return val;		// it's ours to take!
 *	    }
 *	}
 *	return null;
 *  }
 * </pre>
 *
 * This code assumes that getting a removed node will be rare enough
 * that there is no point in testing for it until it's found a match.
 * If testing for a match was expensive, one might
 * <code>synchronize</code> on the node object and testing for it being
 * removed and then trying for matching, thus preventing any other
 * thread from attempting a match at the same time.  In any case, the
 * <code>remove</code> test is itself atomic -- either it succeeds, or
 * it fails because someone got there first.
 *
 * @author Ken Arnold
 */

// NOTE: if you change the implementation, remember that any
// modificaiton of the head reference requires synchronization

public class FastList {

    /**
     * A single node in a <code>FastList</code>.
     */
    public static class Node {
	public  Node	next;
	private Node	prev;
	private boolean	removed;	// has this node been removed?

	/**
	 * Return the next node, skipping over removed ones.  Note that
	 * between the time that the next node is returned and the time you
	 * look at it someone may have removed it.
	 */
	public final Node next() {

            // Check if pruning needs to be done
            // Note: There is the potential for next to become null
            // between the check and it's usage (w/o sync).
            if (next != null && next.removed) {
    
                // Lock "this" node 
                synchronized (this) {
    
                    // Verify that state hasn't changed since we last checked
                    if (next != null) {
                        Node n = next;

                        // We have to be careful to always leave the tail 
                        // node, even if it is removed, since new nodes 
                        // are added to the tail
                        while (n != null && n.removed && n.next != null) {
                            n = n.next;
                        }
    
                        // If necessary, update next to orphan any removed 
                        // node(s)
                        if (next != n) 
                            next = n;
                    }
                }
            }
            return next;
	}

	/**
	 * Return the prev node, skipping over removed ones.  Note that
	 * between the time that the prev node is returned and the time you
	 * look at it someone may have removed it.
	 */
	public final Node prev() {
    
            // Check if pruning needs to be done
            // Note: There is the potential for prev to become null
            // between the check and it's usage (w/o sync).
            if (prev != null && prev.removed) {
    
                // Lock "this" node 
                synchronized (this) {
    
                    // Verify that state hasn't changed since we last checked
                    if (prev != null) {
    
                        Node p = prev;
                        while (p != null && p.removed) {
                            p = p.prev;
                        }
    
                        // If necessary, update next to orphan any 
                        // removed node(s) 
                        if (prev != p) 
                            prev = p;
                    }
                }
            }
            return prev;
	}

	/**
	 * Return <code>true</code> if the object has been removed from the list
	 */
	public final boolean removed() {

	    return removed;
	}

	/**
	 * Set the <code>next</code> reference for this node.
	 */
	private final void next(Node n) {

	    next = n;
	}

	/**
	 * Set the <code>prev</code> reference for this node.
	 */
	private final void prev(Node p) {

	    prev = p;
	}

	/**
	 * Remove the node from the list.  This will be unsuccessful if the
	 * node was removed by someone else after you got your hands on
	 * it.  This must be synchronized because the test to check if it
	 * is already removed and the following decision to remove it if it
	 * hasn't been removed must be unitary.
	 */
	private synchronized final boolean remove() {

	    if (removed)
		return false;

	    return (removed = true);
	}

	final void dump(java.io.PrintWriter out) { //DEBUG
	  if (removed) //DEBUG
	      out.print('!'); //DEBUG
	  out.print(toString()); //DEBUG
	  if (next != null) { //DEBUG
	      out.print(' '); //DEBUG
	      next.dump(out); //DEBUG
	  } //DEBUG
	} //DEBUG
    }

    private Node	head;			// the head of the list
    private Node	tail;			// the tail of the list
    private boolean	reapable;		// modified since last reap?

    /**
     * Create an new empty list.
     */
    public FastList() {
	// Do nothing.
    }

    // the outer this is the only interesting data in this class
    private class FastListNode extends Node {
	void reap() { FastList.this.reap(); }
    }

    /**
     * Add a given node to the end of the list.
     */
    public void add(Node node) {

	synchronized (this) {
	    if (tail == null) {
		tail = head = node;
		node.next(null);
		node.prev(null);
	    }
	    else {
		node.next(null);
		synchronized (tail) {
		    tail.next(node);
		    node.prev(tail);
		    tail = node;
		}
	    }
	}

    }

    /**
     * Add a given node to the list after the specified node.  The
     * specified node must not be removed, or an
     * <code>IllegalArgumentException</code> is thrown.  If
     * <code>after</code> is <code>null</code>, <code>addAfter</code>
     * is equivalent to <code>add</code>.
     */
    public void addAfter(Node after, Node node) {
	if (after == null) {
	    // no need for debug code here -- add() has its own
	    add(node);
	    return;
	}

	/*
	 * We must synchronize so that the node "after" isn't orphaned while
	 * we adding this.
	 */

	synchronized (after) {

	    if (after.removed()) {
		throw new IllegalArgumentException(
		    "Adding node after a removed node");
	    }

	    synchronized (node) {
		node.next(after.next());
		node.prev(after);

                Node n = node.next();
		if (n != null) {
                    synchronized (n) {
		        n.prev(node);
	            }
	        }
	    }
	    after.next(node);
	}

	synchronized(this) {
	    if (tail == after)
	        tail = node;
	}
    }

    /**
     * Return the start of the list.  This is unsynchronized unless the
     * existing head has been removed, when it must then synchronize on
     * the list to update the head.
     */
    // We could leave this optimization to the reap, but this seems better
    public Node head() {

	if (head != null && head.removed()) {
	    synchronized (this) {
		// retest -- it may have changed between the test and here
		if (head != null && head.removed())
		    head = head.next();
		if (head == null)
		    tail = null;
	    }
	}
	return head;
    }

    /**
     * Return the tail of the list.
     */
    // we could leave this optimization to the reap, but this seems better
    public Node tail() {

	if (tail != null) {
	    synchronized (this) {
		// retest -- it may have changed between the test and here
		if (tail == null)
		    return null;
		synchronized (tail) {
		    if (tail != null && tail.removed())
			tail = tail.prev();
		    if (tail == null)
			head = null;
		}
	    }
	}
	return tail;
    }

    /**
     * Remove the given node from this list.  
     */
    public boolean remove(Node node) {

	if (!node.remove()) {
	    return false;
	}
	reapable = true;
	return true;
    }

    /**
     * Reap the removed nodes of this list.  You should need this rarely,
     * if ever.
     *
     */
    public void reap() {

	if (reapable) {	// does this list need reaping?
	    reapCnt++;//DEBUG

	    reapable = false;	// not any more

	    // iteration happens to reap, so that's all we must do

	    synchronized (this) {
		if (head == null)
		    return;
		if (head.removed())
		    head = head.next();
		if (head == null)
		    tail = null;
		else if (tail.removed())
		    tail = tail.prev();
	    }

	    for (Node n = head; n != null; n = n.next()) {

		continue;
	    }
	    for (Node n = tail; n != null; n = n.prev()) {

		continue;
	    }
	}
    }

    int reapCnt = 0; //DEBUG
    public final int reapCnt() { return reapCnt; } //DEBUG


    public void dump(java.io.PrintWriter out) {
	if (head == null) //DEBUG
	    out.print("null"); //DEBUG
	else //DEBUG
	    head.dump(out); //DEBUG
    }
}
