/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;



import net.jini.admin.JoinAdmin;
import com.sun.jini.admin.StorageLocationAdmin;
import com.sun.jini.admin.DestroyAdmin;

import java.rmi.RemoteException;

/**
 * An administrative interface for the com.sun.jini.fiddler implementation
 * of the lookup discovery service.
 * <p>
 * The comments that follow describe Fiddler with respect to the following
 * set of method pairs:
 * <ul>
 *  <li> setLogFileWeight()          and getLogFileWeight()
 *  <li> setSnapshotWeight()         and getSnapshotWeight()
 *  <li> setLogToSnapshotThreshold() and getLogToSnapshotThreshold()
 * </ul>
 * The intent of the information contained in this note is to 
 * clarify the use of these methods and the effect that use will have
 * on a lookup discovery service's configuration.
 * <p>
 * While a lookup discovery service is running, clients register and
 * un-register (cancel their lease); leases on the registrations created
 * for a client by the lookup discovery service are granted, renewed,
 * cancelled and expired; managed sets of groups, locators and lookup
 * services associated with the registrations are added, modified, and
 * removed from the lookup discovery service. In order to make the lookup
 * discovery service's state persistent across system crashes or network
 * outages, each of the state changes just described are written to a file
 * referred to as the "log file". Thus, the log file records, over
 * time, the incremental changes -- or "deltas" -- made to the lookup
 * discovery service's state.
 * <p>
 * To prevent the log file from growing indefinitely, the lookup
 * discovery service's complete state is intermittenly written to another
 * file referred to as a "snapshot file". When the current state is
 * written to the snapshot file, the log file is cleared and the
 * incremental logging of deltas begins again. The period prior to
 * the creation of the first snapshot is referred to as the "system
 * ramp-up"; this is the only period where a log file exists without
 * a corresponding snapshot file. 
 * <p>
 * When recovering the system's state after a crash or network outage
 * (or after the lookup discovery service or its ActivationGroup has been 
 * un-registered and then re-registered through the Activation daemon),
 * a "base state" is first recovered by retrieving and applying the
 * contents of the snapshot file (if it exists). Then, if the log
 * file has length greater than zero, its contents are retrieved
 * and applied in order to incrementally recover the state 
 * changes that occurred from the point of the base state.
 * <p>
 * The criteria used by the lookup discovery service to determine
 * exactly when to "take a snapshot" depends on the current size of
 * the log file relative to the size that the snapshot file will be.
 * Note that whereas, the size of the log file depends on the total number 
 * of changes to the lookup discovery service's state, the size of the
 * snapshot file depends on the number of registrations that are 
 * currently active; that is, the more registrations that have been 
 * created and which have valid leases, the larger the snapshot.
 * For example, if only 10 registrations are active, the snapshot 
 * file will be relatively small; but lease renewals may be regularly
 * requested on some of those registrations. The regular lease renewals
 * will result in a very large log file.
 * <p>
 * A balance must be maintained between how large the log file is
 * allowed to get and how often a snapshot is taken; taking snapshots
 * too often can slow down processing significantly. The lookup 
 * discovery service is initially configured with a threshold value
 * and a weighting factor that are employed in a computation that 
 * determines when to take a snapshot. The methods referenced above
 * (and declared below) provide ways to retrieve and change these
 * values. Thus, based on a particular lookup discovery service's
 * makeup, these methods can be used by a lookup discovery service's
 * administrative client to "tune" performance with respect to logging
 * the persistent state of the lookup discovery service.
 * <pre>
 * The following comparison is made to determine when to take a snapshot:
 * 
 *   if ( logSize >= (W*snapshotsize) && (snapshotSize >= T) {
 *       take a snapshot;
 *   }
 *       where W = snapshotWeight
 *             T = logToSnapshotThreshold
 * </pre>
 * The first comparison is used to ensure that the log file does not
 * grow too large. The second comparison ensures that snapshots are
 * not taken too often.
 * <p>
 * The Administrative client of a lookup discovery service should make
 * note of these relationships when using the methods below to tune the
 * persistence mechanism.
 */
public interface FiddlerAdmin
                 extends JoinAdmin, StorageLocationAdmin, DestroyAdmin
{
    /**
     * Changes the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     * <p>
     * When a client registers with the lookup discovery service, it
     * requests a desired duration for the lease that the lookup discovery
     * service grants on the registration. As stated in the lookup discovery
     * service specification, the actual duration granted is guaranteed
     * to never be greater than the requested duration. But the Fiddler
     * implementation of the lookup discovery service applies an additional
     * restriction on the duration of the lease that is ultimately granted
     * to the client: when determining the actual duration to grant,
     * Fiddler applies a bound to the duration request. That is, whenever
     * a client requests a lease duration that is greater than the value of
     * this bound, the value of the <em>actual</em> duration assigned by
     * Fiddler will not only be less than the <em>requested</em> value,
     * it will also be less than or equal to the value of the bound.
     * <p>
     * The bound satisfies the definition of a <em>least upper bound</em>
     * on the set of all possible <em>granted</em> durations because the
     * durations granted by Fiddler can be arbitrarily close to the bound,
     * but will never be greater than the bound.
     * <p>
     * Thus, this method is a mechanism for an entity with the appropriate
     * privileges to administratively change the value of the least upper
     * bound that will be applied by the Fiddler implementation of the lookup
     * discovery service when determining the duration to assign to the lease
     * on a requested registration.
     *
     * @param newBound <code>long</code> value representing the new least
     *        upper bound (in milliseconds) on the set of all possible
     *        lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         bound value may or may not have been changed successfully.
     */
    void setLeaseDurationBound(long newBound) throws RemoteException;

    /**
     * Retrieves the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     *
     * @return <code>long</code> value representing the current least
     *         upper bound (in milliseconds) on the set of all possible
     *         lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see setLeaseDurationBound
     */
    long getLeaseDurationBound() throws RemoteException;

    /**
     * Change the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not
     * to take a "snapshot" of the system state.
     *
     * @param weight weight factor for snapshot size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         weight factor may or may not have been changed successfully.
     */
    void setSnapshotWeight(float weight) throws RemoteException;

    /**
     * Retrieve the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not to
     * take a "snapshot" of the system state.
     * 
     * @return float value corresponding to the weight factor for snapshot
     *         size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     */
    float getSnapshotWeight() throws RemoteException;

    /**
     * Change the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     *
     * @param threshold size threshold for taking a snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         threshold may or may not have been changed successfully.
     */
    void setLogToSnapshotThreshold(int threshold) throws RemoteException;

    /**
     * Retrieve the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     * 
     * @return int value corresponding to the size threshold of the snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     */
    int getLogToSnapshotThreshold() throws RemoteException;

    /**
     * Changes the value of the <code>com.sun.jini.fiddler.server.debug</code>
     * system property. The value of this property configures the debug
     * facility of the fiddler implementation of the lookup discovery service
     * to write debug information (to a file or standard output) selectively;
     * depending on the values the debug property is given through this
     * method.
     * 
     * Note that the input <code>String</code> parameter must adhere to a
     * particular format, otherwise the debug property value may be set to
     * a value which causes the debug facility to become "silent". That
     * format requires that the input parameter consist of a comma-separated
     * set of valid debug property values.
     * <p>
     * For example: "startup,events,registration,log"
     * <p>
     * Currently, the valid debug property values (tokens) that can be 
     * included in the input parameter are: 
     * <ul>
     * <li> all          - log all debug information
     * <li> startup      - log designated startup information
     * <li> tasks        - log designated information about the various tasks
     * <li> events       - log designated event information
     * <li> groups       - log info about the groups over all registrations
     * <li> locators     - log info about the locators over all registrations
     * <li> discard      - log designated information about discard processing
     * <li> lease        - log designated information about lease processing
     * <li> registration - log designated information about registrations
     * <li> log          - log designated information about persistence logging
     * </ul>
     *
     * <p>
     * Also note that in order to query, modify and/or use the debug facility
     * described above, the policy file of the fiddler implementation of the
     * lookup discovery service must grant the following permissions to the
     * appropriate codebases; otherwise, a SecurityException will result and
     * this method returns <code>false</code>.
     *
     * <pre>
     * grant codebase "file:/<directory_path>/sun-util.jar" { 
     *    . . .
     *    permission java.util.PropertyPermission
     *                               "com.sun.jini.fiddler.server.*", "read";
     *    permission java.io.FilePermission "./*", "read,write,execute,delete";
     *    . . .
     * }
     * grant codebase "file:/<directory_path>/jars/fiddler.jar" { 
     *    . . .
     *    permission java.util.PropertyPermission "*", "read,write"; 
     *    . . .
     * }
     * </pre>
     *
     * @param newValues <code>String</code> consisting of a comma-separated set
     *                  of  valid debug property values
     *
     * @return <code>boolean</code> value that indicates whether or not the
     *         debug property was successfully set
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         debug property may or may not have been changed successfully.
     */
    boolean setDebugProperty(String newValues) throws RemoteException;

    /**
     * Returns the current value of the system property used to configure the
     * debug facility of the fiddler implementation of the lookup discovery
     * service (<code>com.sun.jini.fiddler.server.debug</code>).
     *
     * @return <code>String</code> consisting of a comma-separated set
     *         of valid debug property values
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     */
    String getDebugProperty() throws RemoteException;
}
