/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.FiddlerLeaseMap;

import com.sun.jini.lease.AbstractLease;
import com.sun.jini.proxy.UUID;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMap;
import net.jini.core.lease.UnknownLeaseException;

import java.rmi.RemoteException;

/**
 * When the Fiddler implementation of the lookup discovery service grants
 * a lease on a registration requested by a client, a proxy to that lease
 * is provided to allow the client to interact with the granted lease.
 * This class is the implementation of that proxy.
 * <p>
 * Clients only see instances of this class via the <code>Lease</code>
 * interface.
 */
class FiddlerLease extends AbstractLease {

    private static final long serialVersionUID = 8763201085020941535L;

    /**
     * The reference to the back-end server of the lookup discovery service
     * that granted this lease (the granting entity).
     *
     * @serial
     */
    protected final Fiddler server;
    /**
     * The unique identifier assigned by the lookup discovery service to
     * the registration to which this lease corresponds.
     *
     * @serial
     */
    protected final UUID registrationID;
    /**
     * The internal identifier assigned to this lease by the granting entity.
     *
     * @serial
     */
    protected final long leaseID;

    /**
     * Constructs a proxy to the lease the Fiddler implementation of the
     * lookup discovery service places on a client's requested registration.
     * The lease corresponding to the constructed proxy will have as its
     * expiration time, the time input to the <code>expiration</code>
     * parameter.
     *
     * @param server         reference to the back-end server of the lookup
     *                       discovery service that granted this lease (the
     *                       granting entity)
     * @param registrationID unique identifier assigned to the registration
     *                       to which this lease corresponds
     * @param leaseID        identifier assigned to this lease by the
     *                       granting entity
     * @param expiration     the time of expiration of the lease being granted
     */
    protected FiddlerLease(Fiddler server,
                           UUID registrationID,
                           long leaseID,
                           long expiration) {
        super(expiration);
        this.server         = server;
        this.registrationID = registrationID;
        this.leaseID        = leaseID;
    }

    /**
     * Creates a <code>LeaseMap</code> object that can contain leases whose
     * renewal or cancellation can be batched. Additionally, upon creating
     * the map the current lease is placed in the map as a key, and
     * the value of the <code>duration</code> parameter is placed in
     * the map as the lease's corresponding mapped value of type
     * <code>Long</code>.
     *
     * @param duration the amount of time (in milliseconds) during which
     *                 the lease being placed in the map will remain in
     *                 effect. This value will be converted to an instance
     *                 of <code>Long</code> prior to being placed in the
     *                 map with its associated lease.
     *                 the lease in the map.
     *                      
     * @return an instance of <code>LeaseMap</code> that contains as its
     *         first mapping, the ordered pair consisting of this lease
     *         and the value of the <code>duration</code> parameter.
     *
     * @see net.jini.core.lease.Lease#createLeaseMap
     */
    public LeaseMap createLeaseMap(long duration) {
        return new FiddlerLeaseMap(this, duration);
    }

    /**
     * Examines the input parameter to determine if that parameter, along
     * with the current lease (the current instance of this class), can
     * be batched in a <code>LeaseMap</code>.
     * <p>
     * For the Fiddler implementation of the lookup discovery service, two
     * leases can be batched (placed in the same </code>FiddlerLeaseMap</code>)
     * if they are both instances of <code>FiddlerLease</code> and they
     * were both granted by the same Fiddler implementation of the lookup
     * discovery service. That is, they are the same type, and they were
     * granted by the same server.
     *                      
     * @param lease reference to the <code>Lease</code> object that this
     *              method examines for batching compatibility with the
     *              the current lease
     *                      
     * @return true if the input parameter is compatible for batching with
     *         the current lease, false otherwise
     *
     * @see net.jini.core.lease.Lease#canBatch
     */
    public boolean canBatch(Lease lease) {
        return (    (lease instanceof FiddlerLease)
                 && (server.equals(((FiddlerLease)lease).server))  );
    }

    /**
     * Returns a reference to the back-end server of the lookup discovery
     * service that granted this lease.
     *                      
     * @return an instance of the Fiddler implementation of the lookup
     *         discovery service.
     */
    Fiddler getServer() {
        return server;
    }

    /**
     * Returns the unique identifier assigned by the lookup discovery 
     * service to the registration to which the current lease corresponds.
     *                      
     * @return a <code>com.sun.jini.proxy.UUID</code> value corresponding to
     *         the unique ID assigned to the registration associated with
     *         the current lease.
     */
    UUID getRegistrationID() {
        return registrationID;
    }

    /**
     * Returns the identifier assigned to the current lease by the entity
     * that granted it.
     *                      
     * @return a <code>long</code> value corresponding to the ID assigned
     *         to the current lease.
     */
    long getLeaseID() {
        return leaseID;
    }

    /**
     * Replaces the current value of the <code>expiration</code> field 
     * (defined in the <code>AbstractLease</code> super class) with the
     * value contained in the input parameter. The value contained in the
     * <code>expiration</code> field represents the absolute (non-relative)
     * time (in milliseconds) at which the current lease will expire.
     *
     * @param expiration the new value of the <code>expiration</code> field
     */
    void setExpiration(long expiration) {
        this.expiration = expiration;
    }

    /**
     * This method allows the entity to which the current lease is granted
     * (the lease holder) to indicate that it is no longer interested
     * in the resources provided to the entity by the lookup discovery
     * service. When an entity invokes this method, the overall effect 
     * is the same as if the lease expired, except that expiration occurs
     * immediately instead of at the end of a pre-agreed duration.
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being cancelled is unknown to the lease grantor.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been cancelled successfully.
     *
     * @see net.jini.core.lease.Lease#cancel
     */
    public void cancel() throws UnknownLeaseException, RemoteException {
        server.cancelLease(registrationID, leaseID);
    }

    /** 
     * This method allows the entity to which the current lease is granted
     * (the lease holder) to indicate that it is still interested in the
     * resources of the lookup discovery service, and to request continued
     * access to those resources for an amount of time (in milliseconds)
     * relative to the current time. That is, the duration is not added
     * added to the current expiration, but is added to the current time.
     * <p>
     * Note that the duration of the renewed lease will be no greater than
     * the requested duration, and may be less than that duration.
     *
     * @param duration the requested duration for the lease being renewed
     *
     * @return <code>long</code> value representing the actual duration that
     *         was granted for the renewed lease. Note that the actual
     *         duration granted and returned by this method may be less than
     *         the duration requested.
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed does not exist, or is unknown
     *         to the lease grantor; typically because the lease has expired.
     *         
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been renewed successfully.
     *
     * @see net.jini.core.lease.Lease#renew
     * @see com.sun.jini.lease.AbstractLease#renew
     * @see com.sun.jini.lease.AbstractLease#doRenew
     */
    protected long doRenew(long duration)
                                 throws UnknownLeaseException, RemoteException
    {
        return server.renewLease(registrationID, leaseID, duration);
    }

    /** 
     * Returns a hash code value of the current instance of this class
     * based on the content of the instance.
     *
     * @return <code>int</code> value representing the hash code for an
     *         instance of this class.
     */
    public int hashCode() {
        return server.hashCode() ^ registrationID.hashCode() ^ (int)leaseID;
    }

    /** 
     * Two instances of this lease class are defined to be equal if they
     * satisfy the following conditions:
     * <ul>
     *   <li> both leases were generated by the same back-end server
     *   <li> both leases are associated with the same registration
     *   <li> the IDs assigned to each lease are equal (==)
     * </ul>
     */
    public boolean equals(Object obj) {
        if ( !(obj instanceof FiddlerLease) ) {
            return false;
        }
        FiddlerLease ls = (FiddlerLease)obj;
        return (    (server.equals(ls.server))
                 && (registrationID.equals(ls.registrationID) )
                 && (leaseID == ls.leaseID) );
    }
}
