/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.FiddlerLease;
import com.sun.jini.fiddler.FiddlerRenewResults;
import com.sun.jini.lease.AbstractLeaseMap;
import com.sun.jini.proxy.UUID;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMapException;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

/**
 * When clients request a registration with the Fiddler implementation of
 * the lookup discovery service, leases of type FiddlerLease are granted
 * on those registrations. Under certain circumstances it may be desirable
 * to collect multiple granted leases in a set which implements the
 * <code>net.jini.core.lease.LeaseMap</code> interface. 
 * <p>
 * This class is the implementation class of the <code>LeaseMap</code>
 * interface that is employed by the Fiddler implementation of the lookup
 * discovery service. When a client wishes to "batch" leases granted by
 * that service, the are placed in an instance of this class.
 * <p>
 * Clients only see instances of this class via the  <code>LeaseMap</code>
 * interface.
 */
class FiddlerLeaseMap extends AbstractLeaseMap {

    private static final long serialVersionUID = -1654776531916717026L;

    /**
     * The reference to the back-end server of the lookup discovery service
     *
     * @serial
     */
    private final Fiddler server;

    /**
     * Constructs a new instance of FiddlerLeaseMap.
     *
     * @param lease    reference to a lease to add to the map as a key value
     * @param duration the duration of the corresponding lease. This value
     *                 is the "mapped" value corresponding to the lease key.
     */
    public FiddlerLeaseMap(FiddlerLease lease, long duration) {
        super(lease, duration);
        this.server = lease.getServer();
    }

    /**
     * Examines the input parameter to determine if that parameter will be
     * accepted or rejected by this map as a "legal" key value.
     * <p>
     * This method will return true if the <code>key</code> parameter is 
     * the type of lease which can be "batch-wise" renewed and cancelled 
     * along with all of the other leases in the map.
     * <p>
     * For the Fiddler implementation of the lookup discovery service, two
     * leases can be batched (placed in the same </code>FiddlerLeaseMap</code>)
     * if they are both instances of <code>FiddlerLease</code> and they
     * were both granted by the same Fiddler implementation of the lookup
     * discovery service. That is, they are the same type, and they were
     * granted by the same server.
     *                      
     * @param key reference to the object that this method examines to
     *        determine if this map will accept or reject it as a key
     *                      
     * @return true if this map will accept the <code>key</code> parameter,
     *         false otherwise
     *
     * @see net.jini.core.lease.Lease#canBatch
     */
    public boolean canContainKey(Object key) {
        return (    (key instanceof FiddlerLease)
                 && (server.equals(((FiddlerLease)key).getServer()))  );
    }

    /**
     * Renews all leases in this map. For each lease (key) in the map, the
     * duration used to renew the lease is the lease's corresponding map
     * value. If all renewal attempts are successful, this method returns
     * normally; otherwise, this method removes from this map all leases
     * that could not be renewed, and throws a <code>LeaseMapException</code>.
     *
     * @throws net.jini.core.lease.LeaseMapException this exception is thrown
     *         when one or more leases in the map could not be renewed.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the leases in the map
     *         may or may not have been renewed successfully.
     */
    public void renewAll() throws LeaseMapException, RemoteException {
        int size = map.size();
        if (size == 0) {
            return;
        }
        UUID[] registrationIDs = new UUID[size];
        long[] leaseIDs = new long[size];
        long[] durations = new long[size];
        int i = 0;
        for (Iterator iter = map.entrySet().iterator(); iter.hasNext(); i++) {
            Map.Entry e = (Map.Entry)iter.next();
            FiddlerLease ls = (FiddlerLease)e.getKey();
            registrationIDs[i] = ls.getRegistrationID();
            leaseIDs[i] = ls.getLeaseID();
            durations[i] = ((Long)e.getValue()).longValue();
        }
        FiddlerRenewResults results = server.renewLeases(registrationIDs,
                                                         leaseIDs,
                                                         durations);
        long now = System.currentTimeMillis();
        HashMap emap = (results.exceptions != null) ?
                         new HashMap(2 * results.exceptions.length + 1) : null;
        i = 0;
        int j = 0;
        for (Iterator iter = map.entrySet().iterator(); iter.hasNext(); i++) {
            Map.Entry e = (Map.Entry)iter.next();
            long duration = results.durations[i];
            if (duration >= 0) {
                ((FiddlerLease)e.getKey()).setExpiration(duration + now);
            } else {
                emap.put(e.getKey(), results.exceptions[j++]);
                iter.remove();
            }
        }
        if (emap != null) {
            throw new LeaseMapException("lease renewal failures", emap);
        }
    }

    /**
     * Cancels all leases in this map. If all cancellation attempts are
     * successful, this method returns normally; otherwise, this method
     * removes from this map all leases that could not be cancelled, and
     * throws a <code>LeaseMapException</code>.
     * <p>
     * Note that a lease is removed from this map only when an attempt to
     * cancel it fails; that is, every lease that is successfully cancelled
     * is left in the map.
     *
     * @throws net.jini.core.lease.LeaseMapException this exception is thrown
     *         when one or more leases in the map could not be cancelled.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the leases in the map
     *         may or may not have been cancelled successfully.
     */
    public void cancelAll() throws LeaseMapException, RemoteException {
        int size = map.size();
        if (size == 0) {
            return;
        }
        UUID[] registrationIDs = new UUID[size];
        long[] leaseIDs = new long[size];
        int i = 0;
        for (Iterator iter = map.keySet().iterator(); iter.hasNext(); i++) {
            FiddlerLease ls = (FiddlerLease)iter.next();
            registrationIDs[i] = ls.getRegistrationID();
            leaseIDs[i] = ls.getLeaseID();
        }
        Exception[] exceptions = server.cancelLeases(registrationIDs,leaseIDs);
        if (exceptions == null) {
            return;
        }
        i = 0;
        HashMap emap = new HashMap(13);
        for (Iterator iter = map.keySet().iterator(); iter.hasNext(); i++) {
            FiddlerLease ls = (FiddlerLease)iter.next();
            Exception ex = exceptions[i];
            if (ex != null) {
                emap.put(ls, ex);
                iter.remove();
            }
        }
        throw new LeaseMapException("lease cancellation failures", emap);
    }
}
