/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.proxy.UUID;

import net.jini.discovery.LookupDiscoveryRegistration;
import net.jini.discovery.LookupUnmarshalException;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.event.EventRegistration;
import net.jini.core.lease.Lease;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.IOException;
import java.io.Serializable;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;

/**
 * This class is an implementation of the LookupDiscoveryRegistration
 * interface.
 * <p>
 * When a client requests a registration with a lookup discovery service,
 * an instance of this class is returned. This class is used by the client
 * as a proxy to the registration object created by the lookup discovery
 * service for the client. The remote methods of this class each have a
 * counterpart on the back-end server of the Fiddler implementation of the
 * lookup discovery service. The client can use the methods implemented in
 * this class to manage the parameters of its registration with the lookup
 * discovery service.
 * 
 * @see net.jini.discovery.LookupDiscoveryRegistration
 */
class FiddlerRegistration implements LookupDiscoveryRegistration, Serializable
{

    private static final long serialVersionUID = 1025105263973750839L;

    /**
     * The reference through which communication occurs between the
     * client-side and the server-side of the lookup discovery service
     *
     * @serial
     */
    private final Fiddler server;
    /** 
     * The unique identifier assigned to the current instance of this 
     * registration proxy class by the lookup discovery service. This 
     * ID is used to determine equality between registrations (proxies),
     * as well as to index into the various managed sets maintained by
     * the back-end server.
     *
     * @serial
     */
    private final UUID registrationID;
    /**
     * The object which encapsulates the information used by the client
     * to identify a notification sent by the lookup discovery service
     * to the listener registered with the lookup discovery service by
     * the client's registration, for which the instance of this class
     * serves as proxy.
     *
     * @serial
     */
    private final EventRegistration eventReg;
    /**
     * The lease object granted by the lookup discovery service on the
     * current instance of a client's registration with the lookup discovery
     * service (for which the instance of this class serves as proxy).
     * <p>
     * It is through this object that the client requests the renewal or
     * cancellation of the registration with the lookup discovery service
     * corresponding to the current instance of this class.
     *
     * @serial
     */
    private final Lease regLease;

    /**
     * Constructs a new instance of FiddlerRegistration.
     *
     * @param server         reference to the server object through which 
     *                       communication occurs between the client-side and
     *                       server-side of the lookup discovery service.
     * @param registrationID the unique identifier assigned by the lookup
     *                       discovery service to the current instance of 
     *                       this proxy
     * @param eventReg       object which encapsulates the information used
     *                       by the client to identify a notification sent by
     *                       the lookup discovery service to the listener
     *                       registered with the lookup discovery service by
     *                       the client's registration, for which the instance
     *                       of this class serves as proxy .
     * @param regLease       lease object granted by the lookup discovery
     *                       service
     *                       on the current instance of the client's
     *                       registration with the lookup discovery service.
     * <p>
     *                       It is through this object that the client requests
     *                       the renewal or cancellation of the registration 
     *                       being constructed.
     */
    public FiddlerRegistration(Fiddler server,
                               UUID registrationID,
                               EventRegistration eventReg,
                               Lease regLease)
    {
	this.server         = server;
	this.registrationID = registrationID;
	this.eventReg       = eventReg;
	this.regLease       = regLease;
    }

    /* *** Methods of net.jini.discovery.LookupDiscoveryRegistration *** */

    /**
     * Returns an EventRegistration object that encapsulates the information
     * needed by the client to identify a notification sent by the lookup
     * discovery service to the registration's listener. This method is
     * not remote and takes no arguments.
     *
     * @return the EventRegistration for this registration.
     *
     * @see 
     * net.jini.discovery.LookupDiscoveryRegistration#getEventRegistration
     */
    public EventRegistration getEventRegistration() {
        return eventReg;
    }

    /**
     * Returns the Lease object that controls a client's registration with 
     * the lookup discovery service. It is through the object returned by
     * this method that the client can request the renewal or cancellation
     * of the registration with the lookup discovery service. This method is
     * not remote and takes no arguments.
     * 
     * @return the Lease on this registration.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#getLease
     */
    public Lease getLease() {
        return regLease;
    }

    /**
     * Returns an array consisting of instances of the ServiceRegistrar
     * interface. Each element in the returned set is a proxy to one of
     * lookup service(s) that have already been discovered for the
     * registration corresponding to the current instance of this class.
     * The contents of the returned set make up the 'remote state' of that
     * registration's currently discovered lookup service(s).
     * <p>
     * To obtain the desired lookup service proxies, this method sends a
     * request to the the lookup discovery service. Upon receiving the
     * request, the lookup discovery service sends the requested set of
     * proxies as a set of marshalled instances of the ServiceRegistrar
     * interface. Thus, in order to construct the return set, this method
     * attempts to unmarshal each element of the set received from the
     * lookup discovery service. Should a failure occur while attempting
     * to unmarshal any of the elements of the received set of marshalled
     * proxy objects, this method will throw an exception of type
     * LookupUnmarshalException. 
     * 
     * @return an array of ServiceRegistrar objects.
     * 
     * @throws net.jini.discovery.LookupUnmarshalException this exception
     *         is thrown when failure occurs while attempting to unmarshal
     *         one or more of the marshalled instances of ServiceRegistrar
     *         received from the lookup discovery service.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#getRegistrars
     */
    public ServiceRegistrar[] getRegistrars()
                              throws LookupUnmarshalException, RemoteException
    {
	MarshalledObject[] marshalledRegs 
                                       = server.getRegistrars(registrationID);

        if (marshalledRegs == null) return null;
        ServiceRegistrar[] regs = new ServiceRegistrar[marshalledRegs.length];
        for(int i=0; i<marshalledRegs.length; i++) {
            try {
                regs[i] = (ServiceRegistrar)(marshalledRegs[i].get());
            } catch (IOException e) {
                throw new UnmarshalException("error unmarshalling return", e);
            } catch (ClassNotFoundException e) {
                throw new UnmarshalException("error unmarshalling return", e);
            }
        }
        return regs;
    }

    /**
     * Returns an array consisting of the names of the groups whose members
     * are lookup services the lookup discovery service will attempt to
     * discover for the registration corresponding to the current instance
     * of this class. This set of group names is referred to as the
     * registration's 'managed set of groups'.
     * <p>
     * If the registration's managed set of groups is currently empty, then
     * the empty array is returned. If the lookup discovery service currently
     * has no managed set of groups for the registration through which the
     * request is being made, then null will be returned.
     * 
     * @return a String array containing the elements of the managed set of
     *         groups for the registration.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#getGroups
     */
    public String[] getGroups()  throws RemoteException {
        return server.getGroups(registrationID);
    }

    /**
     * Returns an array consisting of the the LookupLocator objects
     * corresponding to specific lookup services the lookup discovery
     * service will attempt to discover for for the registration
     * corresponding to the current instance of this class. This set of
     * locators is referred to as the registration's 'managed set of locators'.
     * <p>
     * If the registration's managed set of locators is currently empty, then
     * the empty array is returned. If the lookup discovery service currently
     * has no managed set of locators for the registration through which the
     * request is being made, then null will be returned.
     * 
     * @return array consisting of net.jini.core.discovery.LookupLocator
     *         objects corresponding to the elements of the managed set of
     *         locators for the registration.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#getLocators
     */
    public LookupLocator[] getLocators()  throws RemoteException {
        return server.getLocators(registrationID);
    }

    /**
     * Adds a set of group names to the managed set of groups associated
     * with the registration corresponding to the current instance of
     * this class.
     * 
     * @param groups a String array, none of whose elements may be null,
     *               consisting of the group names with which to augment
     *               the registration's managed set of groups.
     * <p>
     *               If any element of this parameter duplicates any other
     *               element of this parameter, the duplicate will be ignored.
     *               If any element of this parameter duplicates any element
     *               of the registration's current managed set of groups, the
     *               duplicate will be ignored.
     * <p>
     *               If the empty set is input, then the registration's
     *               managed set of groups will not change. If null is
     *               input, this method will throw a NullPointerException.
     * 
     * @throws java.lang.UnsupportedOperationException this exception occurs
     *         when the lookup discovery service has no managed set of groups
     *         associated with the registration.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the groups parameter, or one or more
     *         of the elements of the groups parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully augmented.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#addGroups
     */
    public void addGroups(String[] groups)  throws RemoteException {
        server.addGroups(registrationID,groups);
    }

    /**
     * Replaces all of the group names in the managed set of groups
     * associated with the registration corresponding to the current
     * instance of this class.
     * 
     * @param groups a String array, none of whose elements may be null,
     *               consisting of the group names with which to replace the
     *               names in this registration's managed set of groups.
     * <p>
     *               If any element of this parameter duplicates any other
     *               element of this parameter, the duplicate will be ignored.
     * <p>
     *               If the empty set is input, then group discovery for
     *               the registration will cease. If null is input, the
     *               lookup discovery service will attempt to discover all
     *               as yet undiscovered lookup services located within its
     *               multicast radius and, upon discovery of any such lookup
     *               service, will send to the registration's listener an
     *               event signaling that discovery.
     * 
     * @throws java.lang.NullPointerException this exception occurs when one
     *         or more of the elements of the groups parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully replaced.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#setGroups
     */
    public void setGroups(String[] groups)  throws RemoteException {
        server.setGroups(registrationID,groups);
    }

    /**
     * Deletes a set of group names from the managed set of groups
     * associated with the registration corresponding to the current
     * instance of this class.
     * 
     * @param groups a String array, none of whose elements may be null,
     *               consisting of the group names to delete from the
     *               registration's managed set of groups.
     * <p>
     *               If any element of this parameter duplicates any other
     *               element of this parameter, the duplicate will be ignored.
     * <p>
     *               If the empty set is input, the registration's managed
     *               set of groups will not change. If null is input, this
     *               method will throw a NullPointerException.
     * 
     * @throws java.lang.UnsupportedOperationException this exception occurs
     *         when the lookup discovery service has no managed set of groups
     *         associated with the registration.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the groups parameter, or one or more
     *         of the elements of the groups parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully modified.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeGroups
     */
    public void removeGroups(String[] groups)  throws RemoteException {
        server.removeGroups(registrationID,groups);
    }

    /**
     * Adds a set of LookupLocator objects to the managed set of locators
     * associated with the registration corresponding to the current
     * instance of this class.
     * 
     * @param locators an array, none of whose elements may be null, consisting
     *                 of the LookupLocator objects with which to augment
     *                 the registration's managed set of locators.
     * <p>
     *                 If any element of this parameter duplicates any other
     *                 element of this parameter, the duplicate will be
     *                 ignored. If any element of this parameter duplicates
     *                 any element of the registration's managed set of
     *                 locators, the duplicate will be ignored.
     * <p>
     *                 If the empty array is input, then the registration's
     *                 managed set of locators will not change. If null is
     *                 input, this method will throw a NullPointerException.
     * 
     * @throws java.lang.UnsupportedOperationException this exception occurs
     *         when the lookup discovery service has no managed set of
     *         locators associated with the registration.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the locators parameter, or one or
     *         more of the elements of the locators parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully augmented.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#addLocators
     */
    public void addLocators(LookupLocator[] locators) throws RemoteException {
        server.addLocators(registrationID,locators);
    }

    /**
     * Replaces with a new set of LookupLocator objects, all of the
     * elements in the managed set of locators associated with the
     * registration corresponding to the current instance of this class.
     * 
     * @param locators an array, none of whose elements may be null, consisting
     *                 of the LookupLocator objects with which to replace the
     *                 locators in the registration's managed set of locators.
     * <p>
     *                 If any element of this parameter duplicates any other
     *                 element of this parameter, the duplicate will be
     *                 ignored.
     * <p>
     *                 If the empty array is input, then locator discovery for
     *                 the registration will cease. If null is input, this
     *                 method will throw a NullPointerException.
     * 
     * @throws java.lang.NullPointerException this exception occurs when one
     *         or more of the elements of the locators parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully replaced.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#setLocators
     */
    public void setLocators(LookupLocator[] locators) throws RemoteException {
        server.setLocators(registrationID,locators);
    }

    /**
     * Deletes a set of LookupLocator objects from the managed set of
     * locators associated with the registration corresponding to the
     * current instance of this class.
     * 
     * @param locators an array, none of whose elements may be null, consisting
     *                 of the LookupLocator objects to remove from the
     *                 registration's managed set of locators.
     * <p>
     *                 If any element of this parameter duplicates any other
     *                 element of this parameter, the duplicate will be
     *                 ignored.
     * <p>
     *                 If the empty set is input, the managed set of locators
     *                 will not change. If null is input, this method will
     *                 throw a NullPointerException.
     * 
     * @throws java.lang.UnsupportedOperationException this exception occurs
     *         when the lookup discovery service has no managed set of
     *         locators associated with the registration.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the locators parameter, or one or
     *         more of the elements of the locators parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully modified.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeLocators
     */
    public void removeLocators(LookupLocator[] locators) throws RemoteException
    {
        server.removeLocators(registrationID,locators);
    }

    /**
     * Informs the lookup discovery service of the existence of an 
     * unavailable lookup service and requests that the lookup discovery
     * service discard the unavailable lookup service.
     * 
     * @param registrar a reference to the lookup service that the lookup
     *                  discovery service is being asked to discard.
     * <p>
     *                  If this parameter equals none of the lookup services
     *                  contained in the managed set of lookup services for
     *                  this registration, no action will be taken.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         null is input to the registrar parameter.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, 
     *         the lookup service may or may not have been successfully
     *         discarded.
     *
     * @see net.jini.discovery.LookupDiscoveryRegistration#discard
     */
    public void discard(ServiceRegistrar registrar)  throws RemoteException {
	server.discard(registrationID,registrar);
    }

    /* *** hashCode and Equals for this class *** */

    /** 
     * For any instance of this class, returns the hashcode value generated
     * by the hashCode method of the proxy ID associated with the current
     * instance of this proxy.
     *
     * @return an int value representing the hashcode of the object on
     *         which this method is invoked.
     */
    public int hashCode() {
	return registrationID.hashCode();
    }

    /** 
     * For any instance of this class, indicates whether the object input
     * to this method is equal to the instance; where equality of proxies
     * to a lookup discovery service is defined by equality between
     * the unique identifiers assigned by the lookup discovery service
     * to the respective proxies.
     *
     * @param obj reference to the object that is to be compared to the
     *            object on which this method is invoked.
     *
     * @return true if the object input is equal to the object on which
     *         this method is invoked; false otherwise.
     */
    public boolean equals(Object obj) {
	return (    (obj instanceof FiddlerRegistration)
		 && (registrationID.equals
                              (((FiddlerRegistration)obj).registrationID))  );
    }
}
