/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.Fiddler;
import com.sun.jini.fiddler.FiddlerImpl;
import com.sun.jini.fiddler.FiddlerProxy;
import com.sun.jini.fiddler.ParsedArgs;

import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryService;

import net.jini.core.discovery.LookupLocator;

import java.io.IOException;

import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationException;
import java.rmi.MarshalledObject;

import java.util.Properties;

public class StartFiddler {

    /* **************************** Inner Classes ************************** */
    /** 
     * Static class that acts as a data structure for holding the multiple
     * return values from the create methods.
     */
    public static class Created {
        /** The activation group ID of the created lookup discovery service */
        public final ActivationGroupID gid;
        /** The proxy to the created lookup discovery service */
        public final LookupDiscoveryService proxy;
        /**
         * Constructs an instance of this class. Simply stores the values
         * returned by the create methods.
         *
         * @param gid   the activation group ID of the created lookup
         *              discovery service
         * @param proxy the proxy to the created lookup discovery service
         */
        public Created(ActivationGroupID gid, LookupDiscoveryService proxy) {
            this.gid   = gid;
            this.proxy = proxy;
        }
    }//end class Created
    /* ************************** END Inner Classes ************************ */

    private static StartUtl startUtl = new StartUtl
                                     ("com.sun.jini.fiddler.resources.fiddler",
                                      "fiddler");
    /** No-arg constructor that creates an instance this class */
    private StartFiddler() { }

    public static void main(String[] args) {
        ParsedArgs parsedArgs = startUtl.parseActivatableArgs(args);
        if(parsedArgs == null) {
            startUtl.printUsage();
            return;
        }
        String codebase          = parsedArgs.getCodebase();
        String policyFile        = parsedArgs.getPolicyFile();
        String logDir            = parsedArgs.getLogDir();
        String[] groups          = parsedArgs.getGroups();
        LookupLocator[] locators = parsedArgs.getLocators();
        Properties properties    = parsedArgs.getProperties();
        String javaProgram       = parsedArgs.getJavaProgram();
        String[] options         = parsedArgs.getOptions();
        if( (codebase != null) && (policyFile != null) && (logDir != null) ) {
	    String classpath  = System.getProperty("java.class.path");

            /* Test for an existing Activation System */
            if (startUtl.activationUp(10)) {
                /* If Activation is up, create & register with Activation */
                try {
                    LookupDiscoveryService fiddler = (create(classpath, 
                                                             codebase, 
                                                             policyFile,
                                                             logDir, 
                                                             groups, 
                                                             locators,
                                                             properties,
                                                             javaProgram,
                                                             options)).proxy;
                } catch (Exception e) {
                    e.printStackTrace(System.out);
                }
            } else {
	        startUtl.print("fiddler.activation.notUp", null);
            }
        }
        System.exit(0);
    }//end main

    /**
     * This method creates an activatable instance of the Fiddler
     * implementation of the lookupdiscovery service and returns a 
     * proxy that provides client-side access to that instance.
     * <p>
     * If no group names are entered on the command line, then the default
     * action that the lookup discovery service takes with respect to groups
     * is to join the un-named public group ("").
     * <p>
     * Note that when the lookup discovery service is destroyed (through
     * the <code>com.sun.jini.admin.DestroyAdmin</code> interface provided
     * by the <code>Fiddler</code> implementation), it will also destroy
     * its activation group, regardless of whether or not there are other
     * activatable objects still registered in the same group. This is due
     * to the following reason:
     * <p>
     * When the <code> ActivationDesc</code> is created below, the 
     * <code>location</code> parameter to the constructor is not set.
     * If it were to be set, then both the import codebase (the classpath)
     * and the export codebase of this class would be the same; that is,
     * the activation system would use that location to retrieve classes
     * when activating the server-side, and would also annotate with that
     * location all remote classes that are downloaded to the client-side.
     * Since this class wishes to distinguish the import codebase from the
     * export codebase, it must communicate each codebase's value to the
     * activation system through a different mechanism than the
     * <code>location</code> parameter of the constructor of
     * <code>ActivationDesc</code>. That alternate mechanism consists
     * of creating an activation group descriptor
     * (<code>ActivationGroupDesc</code>) in which the properties are
     * over-ridden to include the desired import codebase as the new
     * classpath property ("-cp <import codebase>"); and the export 
     * codebase as the new <code>java.rmi.server.codebase</code> property.
     * This class' activation descriptor (<code>ActivationDesc</code>) is
     * the created using the same group ID (<code>ActivationGroupID</code>)
     * as that generated by the creation of the activation group descriptor
     * with the desired import and export codebases.
     * <p>
     * Because the activation group that is created and registered with the
     * activation system is useful only to entities that have the same
     * import and export codebases as those registered with the group,
     * (and because such entities wishing to register with that activation
     * group would also have to obtain the activation group ID from the
     * service registered by this class), it is not expected that when
     * the service is destroyed through the <code>DestroyAdmin</code> that
     * any other entities will be registered in the activation group. Being
     * the only entity registered in the activation group means that when
     * it is destroyed, it is also the last entity registered in the group.
     * Thus, upon destroying the service, the activation group created here
     * is also ultimately unregistered (removed) from the activation system
     * because once that service is destroyed, there is no way to access
     * the activation group ID for future registrations (or even for 
     * destroying).
     *
     * @param gid        ID of the activation group to use
     * @param logDirname directory in which the state of the lookup discovery
     *                   service will be recorded for persistence
     * @param groups     <code>String</code> array containing the names of
     *                   the groups whose members are the lookup services that
     *                   should be discovered and joined
     * @param locators   String array containing the names of the hosts
     *                   running the specific lookup services to discover
     *                   and join (these host names will eventually be used
     *                   to build corresponding instances of the class
     *                   net.jini.core.discovery.LookupLocator)
     * 
     * @return an instance of the LookupDiscoveryService interface which is
     *         a proxy that provides client-side access to the created
     *         lookup discovery service
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     */
    public static LookupDiscoveryService create(ActivationGroupID gid,
				                String            logDirname,
				                String[]          groups,
				                LookupLocator[]   locators)
        throws ActivationException, IOException
    {
        FiddlerStartParams params = new FiddlerStartParams(logDirname,
                                                           groups,
                                                           locators);
	ActivationDesc desc =
	    new ActivationDesc(gid, "com.sun.jini.fiddler.FiddlerImpl", null,
                               new MarshalledObject(params), true);
	Fiddler server = (Fiddler)Activatable.register(desc);
	return new FiddlerProxy(server, server.getProxyID());
    }//end create


    /**
     * This method creates an activatable instance of the Fiddler
     * implementation of the lookupdiscovery service and returns a 
     * proxy that provides client-side access to that instance.
     * <p>
     * The lookup discovery service is initially a member of no groups.
     * <p>
     * Note that when the lookup discovery service is destroyed (through
     * the com.sun.jini.admin.DestroyAdmin interface provided by the
     * Fiddler implementation), it will also destroy its activation group,
     * regardless of whether or not there are other activatable objects
     * still registered in the same group.
     *
     * @param classpath classpath for the lookup discovery service
     * @param codebase  codebase for the lookup discovery service
     * @param policy    the lookup discovery service's security policy file
     * @param logdir    directory in which the lookup discovery service's
     *                  state will be recorded for persistence
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     */
    public static Created create(String classpath,
				 String codebase,
				 String policy,
				 String logdir)
                                      throws ActivationException, IOException
    {
	return create(classpath, codebase, policy, 
                      logdir, new String[]{""}, null,
                      null, null, null);
    }//end create

    /**
     * This method creates an activatable instance of the Fiddler
     * implementation of the lookupdiscovery service and returns a 
     * proxy that provides client-side access to that instance.
     * <p>
     * The lookup discovery service is initially a member of no groups.
     * <p>
     * Note that when the lookup discovery service is destroyed (through
     * the com.sun.jini.admin.DestroyAdmin interface provided by the
     * Fiddler implementation), it will also destroy its activation group,
     * regardless of whether or not there are other activatable objects
     * still registered in the same group.
     *
     * @param classpath classpath for the lookup discovery service
     * @param codebase  codebase for the lookup discovery service
     * @param policy    the lookup discovery service's security policy file
     * @param logdir    directory in which the lookup discovery service's
     *                  state will be recorded for persistence
     * @param groups    String array containing the names of the groups
     *                  whose members are the lookup services that should
     *                  be discovered and joined
     * @param locators   String array containing the names of the hosts
     *                   running the specific lookup services to discover
     *                   and join (these host names will eventually be used
     *                   to build corresponding instances of the class
     *                   net.jini.core.discovery.LookupLocator)
     * @param props     properties for the activation group, or null
     * @param cmdpath   command program for the activation group, or null
     * @param argv      command options for the activation group, or null
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     */
    public static Created create(String          classpath,
                                 String          codebase,
                                 String          policy,
                                 String          logdir,
                                 String[]        groups,
                                 LookupLocator[] locators,
                                 Properties      props,
                                 String          cmdpath,
                                 String[]        argv)
                                       throws ActivationException, IOException
    {
        ActivationGroupID gid = startUtl.createGroup(classpath,
                                                     codebase,
                                                     policy,
                                                     props, 
                                                     cmdpath, 
                                                     argv);
        try {
            return new Created(gid, create(gid, logdir,groups,locators));
        } catch (Exception e) {
            try {
                ActivationGroup.getSystem().unregisterGroup(gid);
            } catch (Exception ee) { }
            if (e instanceof ActivationException) {
                throw (ActivationException)e;
            } else if (e instanceof IOException) {
                throw (IOException)e;
            }
            throw (RuntimeException)e;
        }
    }//end create

}//end class StartFiddler
