/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.ParsedArgs;

import net.jini.discovery.DiscoveryGroupManagement;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.File;
import java.io.IOException;

import java.net.MalformedURLException;
import java.net.URL;

import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroupID;

import java.rmi.RemoteException;

import java.text.MessageFormat;

import java.util.ArrayList;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

/** 
 * This class provides utilities that can be useful when starting clients
 * or services. In particular, this class can be useful when starting either
 * the version of a service that is activatable and persistent (referred to
 * as the <i>activatable</i> version) or the version of the service that is
 * non-activatable and non-persistent (referred to as the <i>transient</i>
 * version). The particular version of the service that is started as a result
 * of employing this facility depends on the values and format of the arguments
 * entered on the command line and contained in the <code>args</code>
 * parameter. A particular syntax and format for the command line arguments
 * is required when starting a service using this facility.
 * <p>
 * To start an activatable service:
 * <p>
 * If the service being started is an activatable service, then not only
 * are the first three arguments required, but they must be input in a
 * pre-defined order. The remaining arguments are optional and can occur
 * in any order. Although the number of arguments that follow the first
 * three required arguments has no limit, they fall into one of four
 * categories: the "groups" argument, the "locators" argument, the
 * "java program" argument, and the "properties & options" argument. The
 * optional arguments can be input to the command line in one of two
 * formats. The first format imposes no order on the optional arguments.
 * The second format is allowed so that scripts written to start services
 * that follow the command line syntax from previous releases, remain
 * compatible. This second format of the command line imposes a particular
 * order and format on the optional arguments; whereas the new format is
 * less restrictive.
 * <P>
 * The required format of the command line is as follows:
 * <p>
 * <pre>
 *   The new, less restrictive format -
 * <p>
 *   codebase policyfile logdir [ [group0] [group1] [group2] ... ]
 *                              [ [locator0] [locator1] [locator2] ... ]
 *                              [javaprogram] 
 *                              [ [property0] [option0] [property1] ... ]
 * <p>
 *    where the first three arguments are required and the remaining
 *    arguments are optional and can appear in any order
 * </pre>
 * <p>
 * <pre>
 *   The original format -
 * <p>
 *   codebase policyfile logdir
 *                         [group0,locator0,locator1,group1,locator2 ... ]
 *                         [javaprogram] 
 *                         [ [property0] [option0] [property1] ... ]
 * <p>
 *    where the first three arguments are required and the remaining
 *    arguments are optional, but must appear in the order shown above.
 *    Note that although the fourth argument above contains both groups
 *    and locators, it is actually a single argument; with the groups
 *    and locators delimited by commas (with no spaces). Note also that
 *    in earlier releases, locators were not recognized. 
 * </pre>
 * <p>
 * For each of the arguments, both required and optional, the following
 * formats are imposed:
 * <p>
 * <pre>
 *   codebase:    '"http://<hostname>:<port>/jarFile.jar"'
 * <p>              or
 *                '"http://<hostname>:<port>/<directoryPath>/"'
 * <p>
 *   policyfile:  '<directoryPath>/<policyFilename>'
 * <p>
 *   logdir:      '<directoryPath>/<logFilename>'
 * <p>
 *   group:       'any String not in the jini protocol format'
 * <p>
 *   locator:     'String jini protocol format
 * <p>             'jini://<hostname>:[<port>]' (the port is optional)
 * <p>
 *   javaprogram: '<directoryPath>/<nameOfFileToExecute>'
 * <p>
 *   property:    '-Dmy.property0=propVal0 -Dmy.property1=propVal1 ...'
 * <p>
 *   option:      '-classpath <classpath value> ...'
 * </pre>
 * <p>
 * An example command line might be something like the following (note
 * that this example is broken into separate lines and annotated
 * for readability):
 * <p>
 * java -jar myService.jar  // the service to start using jar execution
 * <p>          "http:myHost:8081/myService-dl.jar"    // codebase
 * <p>          /home/myDir/myService.policy           // policy
 * <p>          /tmp/myServiceLog0                     // log dir
 * <p>          com.my.service.group jini://yourHost   // group & locator
 * <p>          /export/progs/jdk1.3/java              // java prog to run
 * <p>          -classpath /files/jdk1.2/lib           // option0
 * <p>          -Dcom.sun.jini.discovery.debug=true    // property0
 * <p>          -Dcom.sun.jini.reggie.proxy.debug=true // property1
 * <p>          -Djava.compiler=                       // property2
 * <p>          
 * To start a transient service:
 * <p>
 * If the service being started is a transient service, then the command
 * line requires only the "log directory" argument; and accepts only the 
 * following optional arguments: the "groups" argument, and the
 * "locators" argument.
 * <p>
 * If the codebase, the policy file, and/or any options are desired to
 * be set for the VM of a transient service started by this facility,
 * those items should be set through arguments input directly to the
 * java command used to start the service, (for the case of the transient
 * service, this class and the service share the same VM). For example,
 * <p>
 * /export/progs/jdk1.3/java 
 * <p>  -cp /jars/jini-core.jar:/jars/jini-ext.jar:/jars/myService.jar
 * <p>  -Dcom.sun.jini.discovery.debug=true
 * <p>  -Dcom.sun.jini.reggie.proxy.debug=true
 * <p>  -Djava.compiler=
 * <p>  -Djava.rmi.server.codebase=""http:myHost:8081/myService-dl.jar"
 * <p>  -Djava.security.policy=/home/myDir/myService.policy
 * <p>  my.service.package.myService    // class file that's executed
 * <p>      /tmp/myServiceLog0                       // log dir argument
 * <p>      group0 jini://host0 group1 jini://host1  // groups & locators
 * <p>          
 * This method uses the particular format of the argument list input
 * to the command to determine whether the service being started is
 * intended to be started with the argument format for an activatable
 * service, or a transient service. This determination is made by
 * examining the first argument in the <code>args</code> parameter. If
 * the first argument is an http or a file url (http:// or file://),
 * then the activatable format is assumed. If the first argument is an
 * existing directory, then the transient format is assumed.
 *
 * @param args <code>String</code> array containing the arguments input
 *        to the command line
 *
 * @see com.sun.jini.fiddler.ParsedArgs
 */
class StartUtl {

    public static class ParsedListResult {
        public int      nArgsParsed;  //total number of args that were examined
        public Object[] extractedArgs;//groups or locs extracted from the list
        public ParsedListResult(int nArgsParsed, Object[] extractedArgs) {
            this.nArgsParsed   = nArgsParsed;
            this.extractedArgs = extractedArgs;
        }
    }//end class ParsedListResult

    /** Fully-qualified package name from which resource bundle retrieved */
    private String resourcePackage;
    /** name of particular item in resource bundle that will be printed */
    private String resourceName;
    /** The set of resources used to enable localization */
    private ResourceBundle resources;
    /** The default set of groups to discover and join */
    private String[] defaultGroups = new String[]{""}; //public group

    /**
     * Constructs a new instance of this utility class, initializing the
     * information about the resource bundle that is used to display usage
     * and error information.
     * 
     * @param ResourcePackage the fully-qualified package name from which the
     *                        resource bundle can be located
     * @param ResourceName    the name used to identify the particular item
     *                        in the resource bundle that will be printed
     */
    StartUtl(String resourcePackage, String resourceName) {
        this.resourcePackage = resourcePackage;
        this.resourceName    = resourceName;
	try {
	    resources = ResourceBundle.getBundle(resourcePackage);
        } catch (MissingResourceException e) {
            e.printStackTrace();
        }
    }//end startUtl constructor

    /**
     * Parses the elements contained in the <code>args</code> parameter, 
     * which should contain the arguments input on the command line
     * when requesting that a program be started. This method assumes a
     * particular syntax and format for that command line (see this class'
     * description for a complete description of the expected syntax).
     * <p>
     *
     * @param args <code>String</code> array containing the arguments input
     *        to the command line minus any initial "required" arguments
     *        (all required parameters are assumed to have already been
     *        parsed before this method is invoked).
     *
     * @return instance of the class <code>ParsedArgs</code>
     *         containing the values of the arguments that were parsed.
     *         Will return <code>null</code> if the arguments cannot be
     *         parsed for any reason (for example, the argument list format
     *         is invalid).
     *
     * @see ParsedArgs
     */
    public ParsedArgs parseArgs(String[] args) {
        return parseArgs(args,0);
    }//end parseArgs

    /**
     * Beginning with element at the index correpsonding to the value of the
     * <code>startIndx</code> parameter, parses the elements contained in
     * the <code>args</code> parameter, which should contain the arguments
     * input on the command line when requesting that a program be started.
     * This method assumes a particular syntax and format for that command
     * line (see this class' description for a complete description of the
     * expected syntax).
     * <p>
     *
     * @param args      <code>String</code> array containing the arguments
     *                  input to the command line.
     * @param startIndx <code>int</code> value indicating the index of the
     *                  element of the <code>args</code> parameter where
     *                  parsing should begin.
     *
     * @return instance of the class <code>ParsedArgs</code>
     *         containing the values of the arguments that were parsed.
     *         Will return <code>null</code> if the arguments cannot be
     *         parsed for any reason (for example, the argument list format
     *         is invalid).
     *
     * @see ParsedArgs
     */
    public ParsedArgs parseArgs(String[] args, int startIndx) {
        boolean argsValid = true;
        /* Parse the remaining arguments, looking for groups, locators, etc. */
        String[] groups          = defaultGroups;
        LookupLocator[] locators = new LookupLocator[]{};
        String javaProgram       = null;
        Properties properties    = null;
        String[] options         = null;

        ArrayList groupList      = new ArrayList();
        ArrayList locatorList    = new ArrayList();
        ArrayList propertyList   = new ArrayList();
        ArrayList optionList     = new ArrayList();

        for(int i=startIndx;i<args.length; ) {
            if(isClasspath(args[i])) {
                if(args.length <= (i+1)) {
                    print(resourceName+".classpath.invalid", null);
                    argsValid = false;
                    break;
                }
                optionList.add(args[i]);
                optionList.add(args[i+1]);
                i = i+2; //step over -classpath and arg just set
                continue;
            }//endif(-classpath arg)
            if(isProperty(args[i])) {
                propertyList.add(args[i]);
                i = i+1; //step over arg just set
                continue;
            }//endif(-DpropKey0.propKey1.propKey2=propertyValue)
            if(isProgram(args[i])) {
                javaProgram = args[i];
                i = i+1; //step over arg just set
                continue;
            }//endif(/dir0/dir1/dir2/javaProgramToExecute)
            LookupLocator loc = getLocator(args[i]);
            if(loc != null) {
                locatorList.add(loc);
                i = i+1; //step over arg just set
                continue;
            }//endif(jini://host:[port])
            if(isCommaSeparatedList(args[i])) {
                String[] grps = getGroupsFromCommaSeparatedList(args[i]);
                /* If groups has already been set to ALL_GROUPS because of
                 * some other argument, that value takes precedence
                 */
                if(groups != DiscoveryGroupManagement.ALL_GROUPS) {
                    /* If grps is NO_GROUPS, that takes precedence over all
                     * other groups, except ALL_GROUPS
                     */
                    if( (grps == DiscoveryGroupManagement.ALL_GROUPS)) {
                        groups = DiscoveryGroupManagement.ALL_GROUPS;
                    } else if( (grps == DiscoveryGroupManagement.NO_GROUPS)) {
                        groups = DiscoveryGroupManagement.NO_GROUPS;
                    } else { //add the groups found to the current ArrayList
                        for(int j=0;j<grps.length;j++) {
                            groupList.add(grps[j]);
                        }
                    }
                }//endif(group != ALL_GROUPS)
                LookupLocator[] locs
                                = getLocatorsFromCommaSeparatedList(args[i]);
                for(int j=0;j<locs.length;j++) {
                    locatorList.add(locs[j]);
                }
                i = i+1; //step over the single arg just processed
                continue;
            }//endif(g0,jini://host0,g1,jini://host1:port,g2)
            if( (args[i].equals("-groups")) || (args[i].equals("-group")) ) {
                if((i+1) == args.length) {
                    print(resourceName+".groups.invalid", null);
                    argsValid = false;
                    break;
                }
                int srcOffset = i+1;
                int count = args.length - srcOffset;
                String[] dst = new String[count];
                System.arraycopy(args, srcOffset, dst, 0, count);
                ParsedListResult grpsResult = parseGroupList(dst);
                for(int j=0;j<(grpsResult.extractedArgs).length;j++) {
                    groupList.add(grpsResult.extractedArgs[j]);
                }
                i = i+grpsResult.nArgsParsed+1; //step over the arg SET
                continue;
            }//endif(-groups g0 g1 g2)
            if(    (args[i].equals("-locators")) || (args[i].equals("-locs"))
                || (args[i].equals("-locator"))  || (args[i].equals("-loc")) )
            {
                if((i+1) == args.length) {
                    print(resourceName+".locators.invalid", null);
                    argsValid = false;
                    break;
                }
                int srcOffset = i+1;
                int count = args.length - srcOffset;
                String[] dst = new String[count];
                System.arraycopy(args, srcOffset, dst, 0, count);
                ParsedListResult locsResult = parseLocatorList(dst);
                for(int j=0;j<(locsResult.extractedArgs).length;j++) {
                    locatorList.add(locsResult.extractedArgs[j]);
                }
                i = i+locsResult.nArgsParsed+1; //step over the arg SET
                continue;
            }//endif(-locators host0 host1 host2)

            /* If the current argument satisfies none of the above
             * characteristics, assume it must be a group name
             */
            groupList.add(args[i]);
            i = i+1; //step over arg just set

        }//end loop(i)

        if(!argsValid) return null;

        /* Prepare the groups for return */
        if(!groupList.isEmpty()) {
            groups = (String[])(groupList.toArray
                                           (new String[groupList.size()]));
        }
        /* Prepare the locators for return */
        if(!locatorList.isEmpty()) {
            locators = (LookupLocator[])(locatorList.toArray
                                      (new LookupLocator[locatorList.size()]));
        }
        /* Prepare the properties for return */
        if(!propertyList.isEmpty()) {
            properties = new Properties();
            for(int i=0;i<propertyList.size();i++) {
                String prop = (String)(propertyList.get(i));
                int k = prop.indexOf("=");
                if (k < 0) {
                    properties.put(prop.substring(2), "");
                } else {
                    properties.put(prop.substring(2, k),  // property key
                                   prop.substring(k + 1));//property value
                }
            }
        }
        /* Prepare the options for return */
        if(!optionList.isEmpty()) {
            options = (String[])(optionList.toArray
                                           (new String[optionList.size()]));
        }
        return new ParsedArgs(null,null,null,
                              groups,
                              locators,
                              javaProgram,
                              properties,
                              options);
    }//end parseArgs

    /**
     * Parses the elements contained in the <code>args</code> parameter, 
     * which should contain the arguments input on the command line
     * when requesting that a program be started. This method assumes a
     * particular syntax and format for that command line: that of an
     * activatable service.
     * <p>
     *
     * @param args <code>String</code> array containing the arguments input
     *        to the command line
     *
     * @return instance of the class <code>ParsedArgs</code>
     *         containing the values of the arguments that were parsed.
     *         Will return <code>null</code> if the arguments cannot be
     *         parsed for any reason (for example, the argument list format
     *         is invalid).
     *
     * @see ParsedArgs
     */
    public ParsedArgs parseActivatableArgs(String[] args) {
        boolean argsValid = true;
	String classpath  = System.getProperty("java.class.path");
	String codebase   = args[0];
	String policyFile = args[1];
	String logDir     = args[2];
        /* Verify codebase is a valid URL that can be used as a codebase */
        argsValid = isValidCodebaseUrl(codebase);
        /* Verify policyFile is a valid file that can be used as a policy */
        argsValid = isValidPolicyFile(policyFile);
        /* Verify lodDir does not already exist */
        argsValid = isValidLogDirectory(logDir);

        if(!argsValid) return null;

        ParsedArgs parsedArgs = parseArgs(args,3);
        return new ParsedArgs(codebase,
                              policyFile,
                              logDir,
                              parsedArgs.getGroups(),
                              parsedArgs.getLocators(),
                              parsedArgs.getJavaProgram(),
                              parsedArgs.getProperties(),
                              parsedArgs.getOptions());
    }//end parseActivatableArgs

    /**
     * Parses the elements contained in the <code>args</code> parameter, 
     * which should contain the arguments input on the command line
     * when requesting that a program be started. This method assumes a
     * particular syntax and format for that command line: that of a
     * transient service.
     * <p>
     *
     * @param args <code>String</code> array containing the arguments input
     *        to the command line
     *
     * @return instance of the class <code>ParsedArgs</code>
     *         containing the values of the arguments that were parsed.
     *         Will return <code>null</code> if the arguments cannot be
     *         parsed for any reason (for example, the argument list format
     *         is invalid).
     *
     * @see ParsedArgs
     */
    public ParsedArgs parseTransientArgs(String[] args) {
        /* Verify lodDir does not already exist */
        if(!isValidLogDirectory(args[0])) return null;
        ParsedArgs parsedArgs = parseArgs(args,1);
        return new ParsedArgs(null,  //codebase,
                              null,  //policyFile,
                              args[0],
                              parsedArgs.getGroups(),
                              parsedArgs.getLocators(),
                              parsedArgs.getJavaProgram(),
                              parsedArgs.getProperties(),
                              parsedArgs.getOptions());
    }//end parseTransientArgs

    /** Prints a general usage message extracted from the resource bundle */
    public void printUsage() {
        print(resourceName+".usage", null);
        print(resourceName+".groups.usage", null);
        print(resourceName+".locators.usage", null);
        print(resourceName+".javaprog.usage", null);
        print(resourceName+".properties.usage", null);
        print(resourceName+".options.usage", null);
    }//end printUsage

    /** Prints a usage message, specific to a transient service, extracted 
     *  from the resource bundle
     */
    public void printTransientUsage() {
        print(resourceName+".transient.usage", null);
        print(resourceName+".groups.usage", null);
        print(resourceName+".locators.usage", null);
    }//end printTransientUsage

    /**
     * Tests for the existence of the ActivationSystem (RMID). This method
     * will always make at least one attempt to verify the existence of a
     * a running ActivationSystem. 
     *
     * @param n int value representing the number of additional attempts
     *        (beyond the initial attempt) as well as the number of seconds
     *        to wait for the ActivationSystem to come up.
     *
     * @return true if the ActivationSystem exists, false otherwise
     */
    public static boolean activationUp(int n) {
        /* First attempt */
        try {
            ActivationGroup.getSystem();
            return true;
        } catch (ActivationException e) { }
        /* Make a new attempt every second for n seconds */
        for(int i=0; i<n; i++) {
            try {
                ActivationGroup.getSystem();
                return true;
            } catch (ActivationException e) { }
            try {
                Thread.sleep(1000); // wait 1 second
            } catch (InterruptedException e) { }
        }
        return false;
    }//end activationUp

    /**
     * Creates an activation group suitable for a lookup discovery service.
     *
     * @param classpath classpath for the lookup discovery service
     * @param codebase  codebase for the lookup discovery service
     * @param policy    the lookup discovery service's security policy file
     * @param props     properties for the activation group, or null
     * @param cmdpath   command program for the activation group, or null
     * @param argv      command options for the activation group, or null
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the activation 
     *         system's VM and the VM in which this class is instantiated.
     *         When this exception does occur, the ActivationGroup may
     *         or may not have been created successfully.
     */
    public static ActivationGroupID createGroup(String classpath,
                                                String codebase,
                                                String policy,
                                                Properties props,
                                                String cmdpath,
                                                String[] argv)
                                    throws ActivationException, RemoteException
    {
        if (props == null) {
            props = new Properties();
        } else {
            props = (Properties)props.clone();
        }
        props.put("java.security.policy", policy);
        props.put("java.rmi.server.codebase", codebase);
        if (argv == null) {
            argv = new String[]{"-cp", classpath};
        } else {
            String[] nargv = new String[argv.length + 2];
            System.arraycopy(argv, 0, nargv, 0, argv.length);
            nargv[argv.length] = "-cp";
            nargv[argv.length + 1] = classpath;
            argv = nargv;
        }
        CommandEnvironment cmd = new CommandEnvironment(cmdpath, argv);
        return ActivationGroup.getSystem().registerGroup
                                         (new ActivationGroupDesc(props, cmd));
    }//end createGroup

    /**
     * Displays the resource String corresponding to a particular key, but 
     * with a given value substituted in the appropriate location in the
     * String.
     *
     * @param key a non-null case-sensitive String containing the name of 
     *            the resource whose String is to be printed.
     * @param val the value to substitute in the resource String before
     *            displaying 
     */
    public void print(String key, String val) {
	String fmt = getString(key);
	if (fmt == null)
	    fmt = "no text found: \"" + key + "\" {0}";
	System.out.println(MessageFormat.format(fmt, new String[]{val}));
    }//end print

    /**
     * Displays the contents of each field of the input parameter. Useful
     * for debugging.
     *
     * @param parsedArgs data structure continuing values of arguments input
     *                   to the command line and parsed prior to invoking
     *                   this method
     */
    public static void printParsedArgs(ParsedArgs parsedArgs) {
        System.out.println("codebase    = "+parsedArgs.getCodebase());
        System.out.println("policyFile  = "+parsedArgs.getPolicyFile());
        System.out.println("logDir      = "+parsedArgs.getLogDir());

        String[] groups = parsedArgs.getGroups();
        if(groups == DiscoveryGroupManagement.ALL_GROUPS) {
            System.out.println("groups      = ALL_GROUPS");
        }else if(groups.length == 0){
            System.out.println("groups      = NO_GROUPS");
        }else{
            for(int i=0;i<groups.length;i++) {
                if(groups[i].equals("")) {
                    System.out.println("  groups["+i+"] = The PUBLIC Group");
                } else {
                    System.out.println("  groups["+i+"] = "+groups[i]);
                }
            }//end loop
        }//endif(groups)

        LookupLocator[] locators = parsedArgs.getLocators();
        if(locators == null) {
            System.out.println("locators    = null");
        }else if(locators.length == 0){
            System.out.println("locators    = NO_LOCATORS");
        }else{
            for(int i=0;i<locators.length;i++) {
                System.out.println("  locators["+i+"] = "+locators[i]);
            }//end loop
        }//endif(locators)

        System.out.println("javaProgram = "+parsedArgs.getJavaProgram());
        Properties properties = parsedArgs.getProperties();
        if(properties == null) {
            System.out.println("properties  = null");
        }else {
            properties.list(System.out);
        }//endif

        String[] options = parsedArgs.getOptions();
        if(options == null) {
            System.out.println("options  = null");
        }else {
            for(int k=0;k<options.length;k++) {
                System.out.println("  options["+k+"] = "+options[k]);
            }//end loop
        }//endif
    }//end printParsedArgs

    /**
     * Retrieves a String resource corresponding to a particular key from the
     * resource bundle. If the resource bundle does not yet exist, this
     * method will first load and instantiate the resource bundle.
     *
     * @param key a non-null case-sensitive String containing the name of 
     *            the resource whose String is to be retrieved.
     *
     * @return a non-null String containing the value of the desired resource
     */
    private String getString(String key) {
	try {
	    return resources.getString(key);
	} catch (MissingResourceException e) {
	    return null;
	}
    }//end getString

    /** Determines if the input String is a valid URL */
    private boolean isValidCodebaseUrl(String str) {
        return isValidCodebaseUrl(resourceName,str);
    }//end isValidCodebaseUrl

    /** Determines if the input String is a valid URL */
    private boolean isValidCodebaseUrl(String resName, String str) {
        boolean codebaseValid = true;
        StringTokenizer st = new StringTokenizer(str);
        for (int i = 0; st.hasMoreTokens(); i++) {
            String url = st.nextToken();
            if (!url.endsWith("/") &&
                !url.endsWith(".jar") && !url.endsWith(".zip"))
            {
                if(resName == null) return false;
                print(resName+".codebase.slash", url);
                codebaseValid = false;
	    }
            try {
                new URL(url);
            } catch (MalformedURLException e) {
                if(resName == null) return false;
                print(resName+".codebase.syntax", url);
                codebaseValid = false;
            }
        }
        return codebaseValid;
    }//end isValidCodebaseUrl

    /** Determines if the input String corresponds to an existing file */
    private boolean isValidPolicyFile(String filename) {
        return isValidPolicyFile(resourceName,filename);
    }//end isValidPolicyFile

    /** Determines if the input String corresponds to an existing file */
    private boolean isValidPolicyFile(String resName, String filename) {
        File pf = new File(filename);
        if (!pf.exists()) {
            if(resName == null) return false;
            print(resName+".policy.exist", filename);
            return false;
        }
        return true;
    }//end isValidPolicyFile

    /** Determines if the input String can be used as a log directory */
    private boolean isValidLogDirectory(String dirPath) {
        return isValidLogDirectory(resourceName,dirPath);
    }//end isValidLogDirectory

    /** Determines if the input String can be used as a log directory */
    private boolean isValidLogDirectory(String resName, String dirPath) {
        File df = new File(dirPath);
        if (df.exists()) {
            if(resName == null) return false;
            print(resName+".logdir.exist", dirPath);
            return false;
        } else if(    (isValidCodebaseUrl(null,dirPath)) 
                   || (isValidPolicyFile(null,dirPath))
                   || (isClasspath(dirPath))
                   || (isProperty(dirPath))
                   || (isProgram(dirPath))
                   || (isCommaSeparatedList(dirPath)) )
        {
            if(resName == null) return false;
            print(resName+".logdir.invalid", dirPath);
            return false;
        }//endif
        return true;
    }//end isValidLogDirectory

    /** Determines if the input String represents a classpath identifier */
    private static boolean isClasspath(String arg) {
        if( (arg.equals("-classpath")) || (arg.equals("-cp")) ) return true;
        return false;
    }//end isClasspath

    /** Determines if the input String represents a valid java property */
    private static boolean isProperty(String arg) {
        if(arg.startsWith("-D")) return true;
        return false;
    }//end isProperty

    /** Determines if the input String can be used as a program */
    private static boolean isProgram(String arg) {
        /* If the argument doesn't begin with a 'dash', and it exists
         * as a file, not a directory, and the absolute path of the file
         * is given (that is, the argument begins with a file separator
         * such as '/' or 'drive:\') then assume it's a program
         */
        if (!arg.startsWith("-")) {
            File fd = new File(arg);
            if(     (fd.exists())
                && !(fd.isDirectory())
                &&  (fd.isAbsolute())  )
            {
                return true;
            }
        }
        return false;
    }//end isProgram

    /** Instantiates a LookupLocator using the input String */
    private static LookupLocator getLocator(String str) {
        try {
            return new LookupLocator(str);
        } catch(MalformedURLException e) {
            return null;
        }
    }//end getLocator

    /** Instantiates a LookupLocator using the input hostname and the
     *  default port number
     */
    private static LookupLocator hostToLocator(String hostname) {
        try {
            return new LookupLocator("jini://"+hostname+":4160/");
        } catch(MalformedURLException e) {
            return null;
        }
    }//end hostToLocator

    /** Determines if the input String is delimited by commas */
    private static boolean isCommaSeparatedList(String arg) {
        /* If arg contains 'comma(s)', return true */
        if(arg.indexOf(",") >= 0) return true;
        return false;
    }//end isCommaSeparatedList

    /** Parses a comma-separated String, extracting the sub-strings that
     *  represent group names (sub-strings NOT in the jini URL format)
     */
    private String[] getGroupsFromCommaSeparatedList(String list) {
        /* If the list contains "none" or "all", those override all other
         * groups. If the list contains both "none" and "all", "all"
         * takes precedence over "none". Additionally, if the group name
         * "public" is input, it will be returned as the 'un-named public
         * group' ("").
         */
        if(list.startsWith("-loc")) return DiscoveryGroupManagement.NO_GROUPS;
        String lst = list;
        if(list.startsWith("-group")) {//strip off -group,-groups
            int offset = list.indexOf(",");
            lst = list.substring(offset+1);
        }
        String publicGroup       = getString(resourceName+".groups.public");
        String noGroups          = getString(resourceName+".groups.none");
        String allGroups         = getString(resourceName+".groups.all");
        boolean containsNoGroups = false;
        StringTokenizer st = new StringTokenizer(lst, " \t\n\r\f,");
        int n = st.countTokens();
        if (n > 0) {
            ArrayList grpList = new ArrayList();
            for (int i=0; st.hasMoreTokens(); i++) {
                String group = st.nextToken();
                if (getLocator(group) != null) {//skip locator urls
                    continue;
                } else if (group.equals(publicGroup)) {
                    group = "";
                } else if(group.equals(noGroups)) {
                    containsNoGroups = true;
                } else if(group.equals(allGroups)) {
                    return DiscoveryGroupManagement.ALL_GROUPS;
                }
                grpList.add(group);
            }//end loop
            if(containsNoGroups)  return DiscoveryGroupManagement.NO_GROUPS;
            return ((String[])(grpList.toArray(new String[grpList.size()])));
        }//end if
        return defaultGroups;
    }//end getGroupsFromCommaSeparatedList

    /** Parses a comma-separated String, extracting the sub-strings that
     *  are given in the jini URL format (jini://hostname:[port])
     */
    private static LookupLocator[] getLocatorsFromCommaSeparatedList
                                                                 (String list)
    {
        if(list.startsWith("-group")) return new LookupLocator[0];
        String lst = list;
        if(list.startsWith("-loc")) {//strip off -loc,-locs,-locator,-locators
            int offset = list.indexOf(",");
            lst = list.substring(offset+1);
        }
        StringTokenizer st = new StringTokenizer(lst, " \t\n\r\f,");
        int n = st.countTokens();
        if (n > 0) {
            ArrayList locList = new ArrayList();
            for (int i=0; st.hasMoreTokens(); i++) {
                String locStr = st.nextToken();
                LookupLocator loc = getLocator(locStr);
                if(loc != null) {
                    locList.add(loc);
                } else {
                    loc = hostToLocator(locStr);
                    if(loc != null) {
                        locList.add(loc);
                    }
                }
            }//end loop
            return ((LookupLocator[])(locList.toArray
                                        (new LookupLocator[locList.size()])));
        }//end if
        return new LookupLocator[0];
    }//end getLocatorsFromCommaSeparatedList

    /** Examines a String array, extracting the elements that that
     *  represent group names (elements NOT in the jini URL format)
     */
    private static ParsedListResult parseGroupList(String[] subArgs) {
        ArrayList grpList = new ArrayList();
        int nParsed = 0;
        for(int i=0;i<subArgs.length;i++) {
            if(subArgs[i].startsWith("-"))      break;
            if(isClasspath(subArgs[i]))         break;
            if(isProperty(subArgs[i]))          break;
            if(isProgram(subArgs[i]))           break;
            LookupLocator loc = getLocator(subArgs[i]);
            if(loc == null) grpList.add(subArgs[i]);
            nParsed = nParsed+1;
        }
        String[] groups = (String[])(grpList.toArray
                                               (new String[grpList.size()]));
        return new ParsedListResult(nParsed,groups);
    }//end parseGroupList

    /** Examines a String array, extracting the elements that are given in
     *  the jini URL format (jini://hostname:[port])
     */
    private static ParsedListResult parseLocatorList(String[] subArgs) {
        ArrayList locList = new ArrayList();
        int nParsed = 0;
        for(int i=0;i<subArgs.length;i++) {
            if(subArgs[i].startsWith("-")) break;
            if(isClasspath(subArgs[i]))    break;
            if(isProperty(subArgs[i]))     break;
            if(isProgram(subArgs[i]))      break;
            LookupLocator loc = getLocator(subArgs[i]);
            if(loc != null) {
                locList.add(loc);
            } else {
                loc = hostToLocator(subArgs[i]);
                if(loc != null) {
                    locList.add(loc);
                }
            }
            nParsed = nParsed+1;
        }
        LookupLocator[] locators = (LookupLocator[])(locList.toArray
                                        (new LookupLocator[locList.size()]));
        return new ParsedListResult(nParsed,locators);
    }//end parseLocatorList

    /** Determines if the given String is a valid 'http', 'ftp', or 'file' 
     *  URL 
     */
    private static boolean isUrl(String str) {
        try {
            URL url = new URL(str);
            String protocol = url.getProtocol();
            if( protocol.equals("http") ) return true;
            if( protocol.equals("ftp") )  return true;
            if( protocol.equals("file") ) return true;
            return false;
        } catch(MalformedURLException e) {
            return false;
        }
    }//end isUrl

}//end class StartUtl
