/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.FiddlerImpl;
import com.sun.jini.fiddler.FiddlerProxy;
import com.sun.jini.fiddler.ParsedArgs;
import com.sun.jini.fiddler.StartUtl;

import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryService;

import net.jini.core.discovery.LookupLocator;

import java.io.IOException;
import java.rmi.server.RemoteObject;
import java.rmi.RMISecurityManager;
import java.util.Properties;

/**
 * Public method and command line interface for creating a non-activatable,
 * non-persistent (transient) instance of the Fiddler implementation of the
 * lookup discovery service.
 */
public class TransientFiddler {

    private static StartUtl startUtl = new StartUtl
                                     ("com.sun.jini.fiddler.resources.fiddler",
                                      "fiddler");
    /**
     * Command line interface for creating a non-activatable (transient)
     * instance of the lookup discovery service. The command line parameters
     * are as follows:
     * <p>
     * <pre>
     * logdir [-groups,group0,group1,...] [-locators,loc0,loc1,...]
     * </pre>
     *
     * @param logdir      required argument containing the name (with optional
     *                    path) of the directory in which the lookup discovery
     *                    service's state will be recorded for persistence
     * @param groups      optional argument containing the names of the
     *                    groups whose members are lookup service(s) the
     *                    lookup discovery service to be created should 
     *                    discover and join. A group name equal to "none" will
     *                    result in this lookup discovery service joining
     *                    no groups; a group name of "public" will result in
     *                    this lookup discovery service joining all lookup
     *                    service(s) it can find that belong to the un-named
     *                    public (""). The default is "public".
     * @param locator(s)  optional argument containing the names of the
     *                    hosts that are running the specific lookup service(s)
     *                    the lookup discovery service to be created should 
     *                    discover and join (these host names will eventually
     *                    be used to build corresponding instances of the
     *                    class net.jini.core.discovery.LookupLocator)
     */
    public static void main(String[] args) {
        ParsedArgs parsedArgs = startUtl.parseTransientArgs(args);
        if(parsedArgs == null) {
            startUtl.printTransientUsage();
            return;
        }
        String logDir            = parsedArgs.getLogDir();
        String[] groups          = parsedArgs.getGroups();
        LookupLocator[] locators = parsedArgs.getLocators();

        if(logDir != null) {
            /* ------------------------------------------------------------- */
            /* An RMISecurityManager must be set!
             *           
             * If a Security Manager is not set, exceptions can occur during
             * discovery (if needed classes are downloaded), but those
             * exceptions are "swallowed" by LookupDiscovery; so no discovery
             * will occur, and it will appear as if nothing is wrong.
             *           
             * Also if no SecurityManager is set, RMI will prevent the loading
             * of classes not found in the classpath; so a lookup will return
             * null instead of the expected references to service items or
             * attributes.
             */
            System.setSecurityManager(new RMISecurityManager());
            /* ------------------------------------------------------------- */

            /* Arg parsing and SecurityManager done. Now create the service */
            try {
                LookupDiscoveryService fiddler
                                           = create(logDir, groups, locators);
            } catch (Exception e) {
                e.printStackTrace(System.out);
            }
        }//endif(logDir != null)
    }//end main

    /**
     * This method creates a non-activatable and non-persistent (transient)
     * instance of the Fiddler implementation of the lookup discovery service
     * and returns a proxy that provides client-side access to that instance.
     * <p>
     * If no group names are entered on the command line, then the default
     * action that the lookup discovery service takes with respect to groups
     * is to join the un-named public group ("").
     * <p>
     * Although not persisted for recovery from system outages and crashes,
     * the transient version of this service records its state to files 
     * located in the directory entered on the command line. This feature
     * is particularly convenient for debugging.
     *
     * @param logDirname directory in which the state of the lookup discovery
     *                   service will be recorded
     * @param groups     <code>String</code> array containing the names of
     *                   the groups whose members are the lookup services that
     *                   should be discovered and joined
     * @param locators   array of instances of <code>LookupLocator</code> in
     *                   which each element is a locator of a specific lookup
     *                   service to discover and join
     * 
     * @return an instance of the LookupDiscoveryService interface which is
     *         a proxy that provides client-side access to the created
     *         lookup discovery service
     * 
     * @throws java.io.IOException 
     */
    public static LookupDiscoveryService create(String          logDirname,
                                                String[]        groups,
                                                LookupLocator[] locators)
                                                            throws IOException
    {
        FiddlerImpl fiddlerImpl = new FiddlerImpl(logDirname,groups,locators);
        return new FiddlerProxy((Fiddler)RemoteObject.toStub(fiddlerImpl),
                                  fiddlerImpl.getProxyID());
    }//end create

}//end class TransientFiddler
