/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.lease.landlord;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseDeniedException;

/** 
 * Provides an implementation of LeasePolicy that grants and renews
 * leases based on constant values for maximum and default lease
 * durations.
 *
 * When an instance of the class is created a
 * <code>LeaseManager</code> may be associated with it.  The manager
 * will be notified when a Lease is created or renewed.
 *
 * @see Landlord
 * @see LeasedResource
 * @see LeaseManager
 * @see net.jini.core.lease.Lease 
 */
public class LeaseDurationPolicy implements LeasePolicy {
    final private long maximum;       // Maximum lease or renewal we will grant
    final private long defaultLength; // Default lease or renewal length
    final private Landlord landlord;
    final private LeaseManager mgr;
    final private LandlordLeaseFactory factory;
    /**
     * Create a <code>LeaseDurationPolicy</code> that will handout the
     * leases with the specified maximum and default durations.
     * Leases with a requested duration less than maximum will be
     * granted exactly.
     *
     * @param maximum  The length in milliseconds of the longest lease
     *                 this  manager should handout.
     * @param defaultLength
     *                 The length in milliseconds of lease this
     *                 manager should handout if the client expresses
     *                 no preference.
     * @param landlord The owner of the leased resources being managed
     *                 by this manager.
     * @param mgr      <code>LeaseManager</code> that will be notified
     *                 of lease creations and renewals.  May be <code>
     *                 null</code> 
     * @param factory  Factory that will be used to generate new leases. 
     *                 If <code>null</code> a <code>LandlordLease</code> 
     *                 factory will be used.
     * @exception IllegalArgumentException thrown if
     *            <code>defaultLength</code> and <code>maximum</code>
     *            are not larger than 0 or if <code>landlord</code> is
     * <code>null</code>.  
     */
    public LeaseDurationPolicy(long maximum, long defaultLength, 
        Landlord landlord, LeaseManager mgr, LandlordLeaseFactory factory) 
    {
	if (!(maximum > 0))
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "maximum lease time must be larger than 0, passed:" + maximum);

	if (!(defaultLength > 0))
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "default lease lease time must be larger than 0 passed:" +
		 defaultLength);

	if (landlord == null)
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "must provide a Landlord");

	this.maximum = maximum;
	this.defaultLength = defaultLength;
	this.landlord = landlord;
	this.mgr = mgr;

	if (factory == null) 
	    factory = new LandlordLease.Factory();
	this.factory = factory;
    }

    /**
     * Factory method that returns lease objects for the specified
     * resource. The lease will call the <code>Landlord</code> object
     * passed to the manager's constructor to handle renewal and
     * cancel requests.  The cookie passed to these calls will be the
     * same as the cookie returned by
     * <code>resource.getCookie()</code>.  The duration of the lease
     * will be affected by the requested duration, and the default and
     * maximum parameters passed to the constructor.
     *
     * @param resource The resource the lease is for. 
     *                 <code>resource.setExpiration</code> will be called 
     *                 to initialize the expiration time of the resource.
     * @param requestedDuration
     *                 The length of lease the client wants.
     * @return A lease for the passed resource
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to Lease.ANYLENGTH or Lease.FOREVER. 
     */
    public Lease leaseFor(LeasedResource resource, long requestedDuration)
	throws LeaseDeniedException
    {
	final long now = currentTime();
	final long expiration =
	    calcExpiration(calculateDuration(resource, requestedDuration), 
			   now);	
	final long duration = expiration - now;

	final Lease rslt = 
	    factory.newLease(resource.getCookie(), landlord, expiration);

	resource.setExpiration(expiration);

	if (mgr != null)
	    mgr.register(resource, duration);

	return rslt;
    }

    /**
     * Renews the expiration of the passed resource.  The duration of
     * renewal is based on the requested duration
     * and the default and maximum parameters passed to the constructor.
     * @param resource The resource having its lease renewed.
     *                 <code>resource.setExpiration</code> will be called
     *                 to update the expiration time of the resource.
     * @param requestedDuration 
     *                 The new duration the client wants
     * @return The new duration the lease should have.
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to <code>Lease.ANYLENGTH</code> or
     * <code>Lease.FOREVER</code>.  
     */
    public long renew(LeasedResource resource, long requestedDuration) {
	final long now = currentTime();
	final long expiration =
	    calcExpiration(calculateDuration(resource, requestedDuration), 
			   now);	
	final long duration = expiration - now;

	final long oldExpiration = resource.getExpiration();
	resource.setExpiration(expiration);

	if (mgr != null)
	    mgr.renewed(resource, duration, oldExpiration);

	return duration;
    }

    /**
     * Returns <code>true</code> if the lease on the passed resource
     * has not expired yet.
     */
    public boolean ensureCurrent(LeasedResource resource) {
	return resource.getExpiration() > currentTime();
    }

    /**
     * Returns the duration this policy is willing to grant for the
     * passed resource at this time.  The duration actually granted
     * will be shorter if the duration extends pass the end of the
     * epoch.  Must return a positive number.
     * @param resource          The resource having a lease granted
     *                          or renewed
     * @param requestedDuration The duration the client wants 
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to <code>Lease.ANYLENGTH</code> or
     * <code>Lease.FOREVER</code>.  
     */
    protected long calculateDuration(LeasedResource resource, 
        long requestedDuration) 
    {
	if (requestedDuration == Lease.FOREVER)
	    requestedDuration = Long.MAX_VALUE;
	else if (requestedDuration == Lease.ANY)
	    requestedDuration = defaultLength;
	else if (requestedDuration < 0)
	    throw new 
		IllegalArgumentException("Negative lease duration " +
		    "requested");

	return Math.min(requestedDuration, maximum);
    }

    /**
     * Method that provides some notion of the current time in milliseconds
     * since the beginning of the epoch. Default implementation
     * calls System.currentTimeMillis()
     */
    protected long currentTime() {
	return System.currentTimeMillis();
    }

    /**
     * Calculates an expiration based on the passed time and a request
     * duration. Will clip the expiration to the end of the epoch
     * @param preferredDuration The duration the policy wants to grant.
     * @param now               The current time in milliseconds since
     *                          the beginning of the epoch
     * @return The duration the new expiration for the resource.
     */
    private long calcExpiration(long preferredDuration, long now)
    {
	long expiration = now + preferredDuration;
	// Any addition of two positive longs is guaranteed to be
	// negative if it overflowed
	if (expiration < 0) {
	    // Not enough time for the duration we want, set to end of
	    // epoch 
	    return Long.MAX_VALUE;
	}

	return expiration;
    }
}
