/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mahalo.log;

import java.util.Map;

import java.io.*;

/**
 * An implementation of a <code>Log</code>
 *
 * @author Gary Holness
 *
 * @see com.sun.jini.mahalo.log.Log
 */
public class SimpleLogFile implements Log {
    private long cookie;
    private ObjectOutputStream out;
    private FileOutputStream outfile;
    private String name;
    private Map catalogue;
    private Map files;
    private boolean valid;
    private boolean readonly;
    private static final boolean DEBUG = false;

    /**
     * Creates a read-only <code>SimpleLogFile</code>
     *
     * To be used for read-only access to a named <code>Log</code>.  This is
     * desired when recovering information from a <code>Log</code>.
     *
     * @param name names the file in which information is stored.
     *
     * @param catalogue <code>Map</code> which associates log files
     *                  to sequential numbers.  This is done so that
     *                  names for <code>LogFile</code>s can be
     *                  dissociated from the names chosen for the
     *                  physical files which store a <code>LogFile</code>'s
     *                  information.  When a <code>SimpleLogFile</code>
     *                  is invalidated, it needs to remove the association
     *                  between its identifier (cookie) and underlying          
     *                  file name.
     *
     * @param files <code>Map</code> which holds references to all
     *              available <code>LogFile</code>s.  When a 
     *              <code>SimpleLogFile</code> is invalidated, it
     *              needs to remove itself.  
     *
     * @see com.sun.jini.mahalo.log.Log
     * @see com.sun.jini.mahalo.log.LogManager
     * @see com.sun.jini.mahalo.log.MultiLogManager
     */
    public SimpleLogFile(String name, Map catalogue, Map files) {
	init(name, 0, catalogue, files);
	readonly = true;
    }

    /**
     * Creates a <code>SimpleLogFile</code>.
     *
     * @param name names the file in which information is stored.
     *
     * @param cookie identifier representing information being stored.
     *
     * @param catalogue <code>Map</code> which associates log files
     *                  to sequential numbers.  This is done so that
     *                  names for <code>LogFile</code>s can be
     *                  dissociated from the names chosen for the
     *                  physical files which store a <code>LogFile</code>'s
     *                  information.  When a <code>SimpleLogFile</code>
     *                  is invalidated, it needs to remove the association
     *                  between its identifier (cookie) and underlying
     *                  file name.
     *
     * @param files <code>Map</code> which holds references to all
     *              available <code>LogFile</code>s.  When a
     *              <code>SimpleLogFile</code> is invalidated, it
     *              needs to remove itself. 
     *
     * @see com.sun.jini.mahalo.log.Log
     * @see com.sun.jini.mahalo.log.LogManager
     * @see com.sun.jini.mahalo.log.MultiLogManager
     */
    public SimpleLogFile(String name, long cookie, Map catalogue, Map files) {
	init(name, cookie, catalogue, files);
    }

    private void init(String name, long cookie, Map catalogue, Map files) {
        if (name == null)
            throw new IllegalArgumentException("SimpleLogFile: null name");
 
	if (catalogue == null)
            throw new IllegalArgumentException("SimpleLogFile: null catalog");

	if (files == null)
            throw new IllegalArgumentException("SimpleLogFile: null files");

        this.name = name;
        this.cookie = cookie;
	this.catalogue = catalogue;
	this.files = files;
	valid = true;
    }


    /**
     * Returns the identifier associated with information in
     * this <code>Log</code>.
     *
     * @see com.sun.jini.mahalo.log.Log
     */
    public long cookie() {
	return cookie;
    }

    /**
     * Add a <code>LogRecord</code> to the <code>Log</code>.
     *
     * @param rec the record to be logged.
     *
     * @see com.sun.jini.mahalo.log.LogRecord
     */
    public synchronized void write(LogRecord rec) throws LogException {
	try {
	    if (!valid)
		throw new InvalidatedLogException("Cannot write to to " +
						"invalidated log");

	    if (readonly)
		throw new LogException("Unable to write to read only log");

	    if (out == null) {
                outfile = new FileOutputStream(name, true);
                out = new ObjectOutputStream(outfile);
	    }

	    out.writeLong(cookie);
	    out.writeObject(rec);
	    out.flush();

	    if (DEBUG)
		System.out.println("SimpleLogFile: write: " + rec);	
	} catch (InvalidClassException ice) {
	    System.out.println("SimpleLogFile: write: " + ice.getMessage());
	    ice.printStackTrace();
	} catch (NotSerializableException nse) {
	    System.out.println("SimpleLogFile: write: " + nse.getMessage());
	    nse.printStackTrace();
	} catch (IOException ioe) {
	    System.out.println("SimpleLogFile: write: " + ioe.getMessage());
	    ioe.printStackTrace();
        } catch (SecurityException se) {
            System.out.println("SimpleLogFile: Unable to create");
            se.printStackTrace();
        }

    }

    /**
     * Invalidate the log.
     */
    public synchronized void invalidate() throws LogException {
        Long key = new Long(cookie);

	try {
	valid = false;
	    if (out != null)
	        out.close();

	    if (outfile != null)
		outfile.close();

	    File fl = new File(name);
	    fl.delete();

	    if (DEBUG)
	       System.out.println("SimpleLogFile: invalidate: deleted " + name);
	} catch (SecurityException se) {
	    System.out.println("SimpleLogFile: invalidate: cannot invalidate");
	    se.printStackTrace();
	    throw new LogException("SimpleLogFile: invalidate: cannot " +
							    "invalidate");
	} catch (IOException ioe) {
	    System.out.println("SimpleLogFile: invalidate: cannot invalidate");
	    ioe.printStackTrace();
	    throw new LogException("SimpleLogFile: invalidate: cannot " +
							    "invalidate");
	}

	//Remove the catalogue entry first, since an unused
	//catalogue mapping means that the slot is empty
	//(no log file) and a new one can be used in the
	//slot.  We don't have to remove the files entry.
	//It will simply get overwritten, causing its
	//reference count to go to zero and it will be
	//garbage collected.
	//note: These are synchronized Maps

	catalogue.remove(key);
	files.remove(key);
    }

    /**
     * Recover information from the log.
     *
     * @param client who to inform with information from the log.
     *
     * @see com.sun.jini.mahalo.log.LogRecovery
     */
    public synchronized void recover(LogRecovery client) throws LogException {
	ObjectInputStream in = null;
	FileInputStream fin = null;
	LogRecord rec = null;
	long cookieval = 0;

	if (!valid)
	    throw new InvalidatedLogException("Cannot recover from " +
						"invalidated log");


	try {
	    fin = new FileInputStream(name);
	    in = new ObjectInputStream(fin);

	    cookieval = this.cookie = in.readLong();

	    rec = (LogRecord) in.readObject();

	    if (DEBUG)
	        System.out.println("SimpleLogFile: recover: recovering: " +
							    name);
	    while (true) {
		if (DEBUG)
		    System.out.println("SimpleLogFile: recover: cookie = " +
					cookie + " rec = " + rec);
		client.recover(cookieval, rec);
	        cookieval = in.readLong();
	        rec = (LogRecord) in.readObject();
	    }
	} catch (EOFException eofe) {
	    if (DEBUG)
		System.out.println("SimpleLogFile: recover: EOF");
	    ;	// nothing to do
	} catch (ClassNotFoundException cnfe) {
            System.out.println("SimpleLogFile: recover: unable to read");
	    cnfe.printStackTrace();
	} catch (InvalidClassException ice) {
            System.out.println("SimpleLogFile: recover: unable to read");
	    ice.printStackTrace();
	} catch (StreamCorruptedException sce) {
            System.out.println("SimpleLogFile: recover: unable to read");
	    sce.printStackTrace();
	} catch (OptionalDataException ode) {
            System.out.println("SimpleLogFile: recover: unable to read");
	    ode.printStackTrace();
	
	} catch (FileNotFoundException fnfe) {
	    //Swallow the exception on read only (recovery)
	    //access so as not to confuse the user
	    ;
        } catch (IOException ioe) {
            System.out.println("SimpleLogFile: recover: unable to read");
            ioe.printStackTrace();
        } catch (SecurityException se) {
            System.out.println("SimpleLogFile: recover: unable to read");
            se.printStackTrace();
        } finally {
	    try {
	    in.close();
	    fin.close();
	    } catch (IOException ioe) {
		System.out.println("SimpleLogFile: recover: unable to close");
		ioe.printStackTrace();
	    }
	    readonly = false;
	}

    }
}
