/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mahout.binder;

import java.rmi.*;
import java.rmi.registry.Registry;
import java.rmi.registry.LocateRegistry;

/**
 * Handles binding and unbinding a remote object into the RMI registry
 */
class RegistryHandler {
    /** The name to bind the object to */
    private String serviceName;

    /** The object to bind into the registry */
    private Remote service;

    /** The tread we are using for interaction with registry */
    private Thread registerThread;

    /** The port to find the registry on */
    private int    port;

    /** Array of retry backoff times */
    static final private long[] backoffs = {
	1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, // 1 second for the first 10 sec
	5000, 5000, 5000, 5000, 5000, // 5 seconds for the next 25 sec
	15000                         // 15 seconds after that
    };

    /**
     * Create a <code>RegistryHandler</code>
     * @param service     The object to bind into the registry
     * @param serviceName The name to bind the object under
     * @param port        The port to find the registry on
     */
    RegistryHandler(Remote service, String serviceName, int port) {
	this.serviceName = serviceName;
	this.service = service;
	this.port = port;
    }

    /**
     * Register the object passed in the constructor with the
     * registry.  This call only need be called once, though redundant
     * calls will not cause problems.  This call only kicks off the
     * binding process and will probably return before the object is
     * actually bound into the registry.
     */
    synchronized void register() {
	if (registerThread != null)
	    return;

	registerThread = new Thread(new RegisterThread(), "Registry Thread");
	registerThread.start();
    }

    /**
     * Unbind the object passed in the constructor from the registry and/or
     * stop the binding process kicked off by <code>register()</code>
     */
    synchronized void unregister() {
	if (registerThread.isAlive()) {
	    registerThread.interrupt();
	}

	try { 
	    final Registry reg = LocateRegistry.getRegistry(port);
	    reg.unbind(serviceName);
	    // Don't sweet exceptions, if we can't talk to it
	    // proably no else can ether (it is local, and it is
	    // not persistent )
	} catch (RemoteException e) {
	    System.err.println("destroy:Trouble unbinding from Registry");
	    e.printStackTrace();
	} catch (NotBoundException e) {
	    System.err.println("destroy:Trouble unbinding from Registry");
	    e.printStackTrace();
	}
    }

    /** 
     * <code>Runnable</code> that actually finds the registry and 
     * and binds service into the registry.
     */
    private class RegisterThread implements Runnable {
	/**
	 * Try to bind service into the rmi-registry, don't give up
	 * until we succeed (its does backoff on its re-tries
	 * however).  Periodically complains to
	 * <code>System.err</code> if it can't find the registry or
	 * can successfully rebind.  Will stop if interrupted.  
	 */
	public void run() {
	    long timeSinceLastPrint = 0;
	    int backoffIndex = -1;
	    Throwable last = null;
	    
	    while (true) {
		try {
		    final Registry reg = LocateRegistry.getRegistry(port);
		    reg.rebind(serviceName, service);
		    return;
		} catch (RemoteException e) {
		    // ignore and re-try
		    last = e;
		}

		// Must have failed
		backoffIndex++;
		if (backoffIndex >= backoffs.length)
		    backoffIndex = backoffs.length - 1;

		if (timeSinceLastPrint >= 15000) {
		    System.err.println("Having trouble registering with " +
				       "registry");
		    if (last != null) {
			System.err.println("Last error was:");
			last.printStackTrace();
		    }
		    timeSinceLastPrint = 0;
		    last = null;
		    System.err.println("...re-trying");
		}

		try {
		    Thread.sleep(backoffs[backoffIndex]);
		} catch (InterruptedException e) {
		    return;
		}
		
		timeSinceLastPrint += backoffs[backoffIndex];
	    }	    
	}
    }

}
