/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mahout.binder;

import java.io.IOException;

import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.server.RemoteServer;
import java.rmi.server.RemoteObject;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationException;
import java.rmi.activation.UnknownObjectException;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.entry.Entry;
import com.sun.jini.lookup.JoinManager;

/**
 * Remote and optionally activatable object that manages a service's
 * relationship with the directories it wants to bind into.
 * <p>
 * Implements remote versions of <code>JoinAdmin</code> and
 * <code>RegistryAdmin</code> so the service's admin proxy can delegate
 * to the <code>RemoteJoinAdminImpl</code> directly
 * <p>
 * Delegates to <code>JoinManager</code> to handle details of binding
 * into Jini lookups
 * <p>
 * Delegates to <code>RegistryHandler</code> to handle details of
 * binding into RMI registry.
 * <p>
 * Delegates persisting data to an object implementing
 * <code>JoinAdminState</code> so it can be used in ether a
 * non-activatable or activatable context, as well as to use different
 * persistence methods.
 * <p>
 * Implements <code>LocalActivatableObjectTable.Findable</code> so the
 * <code>LookupBinder</code> can find it when the service asks for a
 * <code>DirectoryAdmin</code> or to destroy itself
 * <p>
 * Implements <code>RefHolder</code> so it can bind non-remote objects
 * into RMI registry by binding itself into the registry.
 * <p>
 * @see net.jini.admin.JoinAdmin
 * @see com.sun.jini.mahout.RegistryAdmin
 * @see DirectoryAdmin
 * @see com.sun.jini.lookup.JoinManager
 * @see RegistryHandler
 * @see JoinAdminState
 * @see LocalActivatableObjectTable.Findable
 * @see RefHolder
 */
public class RemoteJoinAdminImpl implements DirectoryAdmin, RefHolder,
    LocalActivatableObjectTable.Findable
{
    /**
     * <code>JoinManager</code> that is handling the details of binding
     * into Jini lookups.
     *
     * @serial
     */
    private JoinManager  mgr;

    /**
     * The object that is coordinating our persistent state.
     * @serial
     */
    private JoinAdminState state;

    /**
     * Our activation ID.
     * @serial
     */
    private ActivationID   actID;

    /**
     * Object that is handing the details of binding into the RMI registry.
     * @serial
     */
    private RegistryHandler handler;

    /**
     * Constructor used for non-activatable services.
     * @param service The service that wants to be bound into Jini lookups
     *                and/or RMI registry
     * @param attributes The initial attribute list for the service.
     */
    public RemoteJoinAdminImpl(Object service, Entry[] attributes)
	throws RemoteException
    {
   	UnicastRemoteObject.exportObject(this);
	state = new TransientJoinAdminState(service, attributes);
    }

    //------------------------------------------------------
    // Methods/Constructors used in the activatable case

    /**
     * Activation constructor
     */
    public RemoteJoinAdminImpl(ActivationID id, MarshalledObject data)
	throws IOException, ClassNotFoundException
    {
	actID = id;
	Activatable.exportObject(this, id, 0);

	try {
	    restoreState(data);
	} catch (IOException e) {
	    corruptedState("Binder can't unpack persistent state", e);
	    throw e;
	} catch (ClassNotFoundException e) {
	    corruptedState("Binder can't unpack persistent state", e);
	    throw e;
	}

	LocalActivatableObjectTable.Findable key =
	    (LocalActivatableObjectTable.Findable)
	    RemoteObject.toStub(this);
	LocalActivatableObjectTable.register(key, this);
	
	advertise();
    }

    /** 
     * Utility method to report problems with corrupted state.
     * Attaches activation and activation group id to message and then
     * calls <code>reportError</code>.
     * @see reportError
     */
    protected void corruptedState(String msg, Exception e) {
	reportError(msg + " ActivationID:" + actID + " Group ID:" +
		    ActivationGroup.currentGroupID(), e);
    }

    /** 
     * Utility method to report errors. Dumps message and exception
     * stack trace to <code>System.err</code>. <code>e</code> may be
     * <code>null</code>.
     */
    private void reportError(String msg, Exception e) {
	System.err.println("binder:" + msg);
	if (e != null)
	    e.printStackTrace();
    }

    /**
     * Unpack our <code>JoinAdminState</code> object.
     */
    protected void restoreState(MarshalledObject data) 
	throws IOException, ClassNotFoundException
    {
	JoinAdminActivationState actState = 
	    (JoinAdminActivationState)data.get();
	state = actState.unpackJoinAdminState();
    }

    /**
     * Register an activatable RemoteJoinAdminImpl and return the
     * remote reference
     */
    static LocalActivatableObjectTable.Findable
	createActivatable(JoinAdminActivationState state, String codeSource,
	    ActivationGroupID groupID)
	throws RemoteException, ActivationException, IOException
    {
	ActivationDesc desc =
	    new ActivationDesc(groupID, RemoteJoinAdminImpl.class.getName(),
	        codeSource, new MarshalledObject(state), true);

	return (LocalActivatableObjectTable.Findable)
	    Activatable.register(desc);
    }

    /**
     * Register the service with the appropriate lookup services and/or
     * RMI registry 
     */
    protected void advertise() throws IOException {
	// Don't want two Join Managers
	if (mgr == null)
	    mgr = state.startNewJoinManager();

	if (state.isRegistryBindingEnabled()) {
	    handler = 
		new RegistryHandler(this, state.getName(), state.getPort());
	    handler.register();
	}
    }

    /**
     * Unregisters with RMID, the lookup services, and/or RMI registry
     */
    public void destroy() {
	// Unregister with lookup
	if (mgr != null)
	    mgr.terminate();

	// Remove ourselves from the registry
	if (handler != null)
	    handler.unregister();

	state.destroy();

	if (actID == null) {
	    // Must be unicast remote
	    try {
		UnicastRemoteObject.unexportObject(this, true);
	    } catch (NoSuchObjectException e) {
		// just being nice, we are probably exiting
	    }
	} // else we are activatable, destroying the group will have
	  // the same effect as unregistering
    }

    // Implementation of RemoteJoinAdmin, for get methods we just
    // delegate to JoinManger, for the set methods we call JoinMangager
    // to and then persist the change by calling the appropriate method
    // on our JoinAdminState.  If the call on our JoinAdminState we throw
    // a runtime exception since JoinAdmin methods don't let us throw
    // a IOException

    // Inherit doc comment from super interface
    public Entry[] getLookupAttributes() {
	return mgr.getAttributes();
    }


    // Inherit doc comment from super interface
    public void addLookupAttributes(Entry[] attrSets) {
	mgr.addAttributes(attrSets, true);
	try {
	    state.updateAttributes();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }


    // Inherit doc comment from super interface
    public void modifyLookupAttributes(Entry[] attrSetTemplates,
				       Entry[] attrSets) {
	mgr.modifyAttributes(attrSetTemplates, attrSets, true);

	try {
	    state.updateAttributes();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }


    // Inherit doc comment from super interface
    public String[] getLookupGroups() {
	return mgr.getGroups();
    }

    // Inherit doc comment from super interface
    public void addLookupGroups(String[] groups) {
 	try {
	    mgr.addGroups(groups);
	} catch (IOException e) {
	    throw new RuntimeException("Could not change groups");
	}

 	try {
	    state.updateGroups();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }

    // Inherit doc comment from super interface
    public void removeLookupGroups(String[] groups) {
	try {
	    mgr.removeGroups(groups);
	} catch (IOException e) {
	    throw new RuntimeException("Could not change groups");
	}

	try {
	    state.updateGroups();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }

    // Inherit doc comment from super interface
    public void setLookupGroups(String[] groups) {
	try {
	    mgr.setGroups(groups);
	} catch (IOException e) {
	    throw new RuntimeException("Could not change groups");
	}

	try {
	    state.updateGroups();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }
    
    // Inherit doc comment from super interface
    public LookupLocator[] getLookupLocators() {
	return mgr.getLocators();
    }

    // Inherit doc comment from super interface
    public void addLookupLocators(LookupLocator[] locators) {
	mgr.addLocators(locators);
	try {
	    state.updateLookupLocators();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }

    // Inherit doc comment from super interface
    public void removeLookupLocators(LookupLocator[] locators) {
	mgr.removeLocators(locators);
	try {
	    state.updateLookupLocators();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }

    // Inherit doc comment from super interface
    public void setLookupLocators(LookupLocator[] locators) {
	mgr.setLocators(locators);
	try {
	    state.updateLookupLocators();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change");
	}
    }

    // Required by the RegistryAdmin interface

    // Inherit doc comment from super interface
    public void enableRegistryBinding(String name, int port)
	throws IOException 
    {
	state.recordRegistryBindingEnable(name, port);
	handler = 
	    new RegistryHandler(this, state.getName(), state.getPort());
	handler.register();
    }

    // Required by refholder interface

    // Inherit doc comment from super interface
    public Object proxy() {
	return state.getService(); 
    }

    // Required LocalActivatableObjectTable.Findable

    // Inherit doc comment from super interface
    public void activate() {}
}


