/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mercury;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.rmi.MarshalledObject;

import net.jini.core.event.RemoteEvent;

/**
 * The <code>EventID</code> class is used to represent a unique event.
 * This class maintains the two pieces of information that make an event
 * unique: the event's source and id attributes.
 *
 * It's used in by mailbox code in order to maintain a list (on a per
 * registration basis) of events that caused an 
 * <code>UnknownEventException</code> to be received 
 * during an event notification attempt for the given event.
 */
class EventID implements Serializable {

    // TODO - generate serialversionid

    /** The event source */
    private transient Object source = null;

    /** 
     * The event id. 
     * @serial
     */
    private long id = 0;

    /** Simple constructor */
    public EventID(Object source, long id) {
        init(source, id);
    }

    /** 
     * Convenience constructor.
     * Initializes the object with attributes extracted from the
     * given <code>RemoteEvent</code> argument.
     */
    public EventID(RemoteEvent evt) {
        if (evt == null )
            throw new IllegalArgumentException(
		"RemoteEvent argument must be non-null");
        init(evt.getSource(), evt.getID());
    }

    /** 
     * Convenience initialization method.
     * Note that the private scoping is important here 
     * since it prevents a subclass from inadvertently overriding it.
     */
    private void init(Object source, long id) {
        if (source == null )
            throw new IllegalArgumentException(
		"Source argument must be non-null");
	this.source = source;
	this.id = id;
    }

    /** 
     * Return true if the given object is equal to <code>this</code> object
     * and false otherwise.  Two <code>EventID</code> objects are considered 
     * equal if their source and id attributes are equal.
     */
    public boolean equals (Object o) {
        if (o == null || (!(o instanceof EventID))) {
            return false;
        }
	EventID eid = (EventID)o;
        return (source.equals(eid.source) &&
                id == eid.id);
    }

    // inherit documentation from supertype
    public int hashCode() {
        // TODO - find a better hash algorithm?
        return (int)id; // just truncate for now
    }

    // inherit documentation from supertype
    public String toString() {
        return getClass().getName() + "[source=" + source + "]"
           + "[id=" + id + "]";
    }

    /**
     * @serialData Use default semantics for the id field and 
     * write the source field out as a MarshalledObject.
     * The source field is written out as a MarshalledObject
     * to preserve the codebase annotation as well as to 
     * guarantee that the object can be read from the stream
     * without corrupting it (e.g. because of a 
     * <code>ClassNotFoundException</code>). 
     */
    private void writeObject(ObjectOutputStream stream)
        throws IOException
    {
        stream.defaultWriteObject();
        stream.writeObject(new MarshalledObject(source));
    }

    /**
     * Initialize <code>id</code> field using default semantics and 
     * then unmarshall the value of <code>source</code> from the stream.
     */
    private void readObject(ObjectInputStream stream)
        throws IOException, ClassNotFoundException
    {
        stream.defaultReadObject();
        MarshalledObject mo = (MarshalledObject)stream.readObject();

        try {
            source = mo.get();
        } catch (Throwable e) {
            if (e instanceof Error &&
                !(e instanceof LinkageError ||
                  e instanceof OutOfMemoryError ||
                  e instanceof StackOverflowError)) {
                throw (Error)e;
	    }
        }
    }
}
