/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mercury;

import com.sun.jini.proxy.ThrowThis;
import com.sun.jini.proxy.UUID;

import java.io.Serializable;
import java.rmi.RemoteException;

import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.UnknownEventException;

/**
 * The ListenerProxy class implements the RemoteEventListener interface
 * and is the class used as the event "forwarding" target by clients of the 
 * mailbox service.  
 *
 * This class works in conjunction with the <code>ListenerProxyRep</code>
 * class in order to facilitate smart proxy functionality. 
 * Prior to JDK1.2.2, RMI semantics would attempt to substitute the associated
 * "stub" class for a Remote object being marshalled.  This behavior would
 * not allow the passing of state information (i.e. <code>registrationID</code>
 * in the <code>notify</code> invocations.  
 *
 * This class, instead, overrides the writeReplace method, which allows 
 * this class to substitute a class of its choice into the marshalling 
 * stream before RMI semantics are applied. In this case, a ListenerProxyRep 
 * (a non-Remote interface) is substituted for this class in the marshalling 
 * stream. The ListenerProxyRep, in turn, overrides its readResolve method in 
 * order to reconstitute a ListenerProxy object when it is unmarshalled.
 *
 * This "trick" essentially allows us to pass a smart proxy through
 * third party objects while maintaining the RemoteEventListener interface
 * contract (which could not occur using default RMI semantics).
 */

class ListenerProxy implements RemoteEventListener, Serializable {

    /**
     * The reference to the event mailbox service implementation
     *
     * @serial
     */
    private final MailboxBackEnd server;

    /**
     * The proxy's <code>UUID</code>
     *
     * @serial
     */
    private final UUID registrationID;

    /** Simple constructor */
    ListenerProxy(MailboxBackEnd ref, UUID regID) {
        server = ref;
        registrationID = regID;
    }

    /** 
     * Implementation of RemoteEventListener interface.
     * Simply forwards the event to the service implementation
     * along with the associated registration information.
     */
    public void notify(RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException 
    {
	try {
            server.notify(registrationID, theEvent);
	} catch (ThrowThis tt) {
	    tt.throwRemoteException();
	}

    }

    /* Override writeReplace to substitute ListenerProxyRep when marshalling */
    private Object writeReplace() {
	return new ListenerProxyRep(server, registrationID);
    }

    /** Returns the hash code generated by the hashCode method of the
     *  registration ID associated with an instance of this class.
     */
    public int hashCode() {
	return registrationID.hashCode();
    }

    /** Proxies for registrations with the same ID are considered equal. */
    public boolean equals(Object obj) {
        // TODO - check for null
	return (obj instanceof ListenerProxy &&
		registrationID.equals(((ListenerProxy)obj).registrationID));
    }

    /** Returns the unique registration ID associated with this listener */
    public UUID getID() {
	return registrationID;
    }
}


