/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mercury;


import com.sun.jini.constants.TimeConstants;
import com.sun.jini.debug.Debug;
import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.lease.landlord.LeaseDurationPolicy;
import com.sun.jini.lease.landlord.Landlord.RenewResults;
import com.sun.jini.lookup.entry.BasicServiceType;
import com.sun.jini.mahout.BasicServiceDesc; 
import com.sun.jini.mahout.ParseUtils; 
import com.sun.jini.proxy.ThrowThis;
import com.sun.jini.proxy.UUIDFactory;
import com.sun.jini.proxy.UUID;
import com.sun.jini.reliableLog.ReliableLog;
import com.sun.jini.reliableLog.LogException;
import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.thread.ReadersWriter;
import com.sun.jini.thread.ReadersWriter.ConcurrentLockException;
import com.sun.jini.thread.RetryTask;
import com.sun.jini.thread.TaskManager;

import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationSystem;
import java.rmi.activation.UnknownObjectException;
import java.rmi.MarshalledObject;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.ServerException;
import java.rmi.server.RemoteObject;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;
import java.util.TreeMap;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.UnknownEventException;
import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseException;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.LeaseMapException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.event.EventMailbox;
import net.jini.event.MailboxRegistration;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.discovery.LookupDiscovery;

/**
 * MailboxImpl implements the server-side of the event mailbox service.
 * <p>
 * Client-side proxies make the appropriate transformation from client
 * requests to the methods of this class.
 */
public class MailboxImpl implements MailboxBackEnd, TimeConstants {

    /* ServiceInfo values */
    private static final String PRODUCT = "EventMailbox";
    private static final String MANUFACTURER = "Sun Microsystems, Inc.";
    private static final String VENDOR = MANUFACTURER;
    private static final String VERSION = 
	com.sun.jini.constants.VersionConstants.SERVER_VERSION;

    /** The primary proxy to the current instance of this class */
    private MailboxProxy mailboxProxy;

    /** Concurrent object to control read and write access */
    private final ReadersWriter concurrentObj = new ReadersWriter();

    /** Map from <code>UUID</code> to <code>ServiceRegistration</code> */
    // HashMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private HashMap regByID = new HashMap(); 

    /**
     * Identity map of ServiceRegistration, ordered by lease expiration.
     * This is a parallel data structure to <code>regByID</code>.
     */
    // TreeMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private final TreeMap regByExpiration = new TreeMap();

    /** 
     * List of registrations that have event delivery enabled,
     * but don't have any event tasks currently scheduled.
     */
    // Using an ArrayList for random access performance. 
    // ArrayList is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private List pendingReg = new ArrayList();

    /**
     * Map of registrations that have event delivery enabled 
     * and have event tasks currently scheduled.
     */
    // HashMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private Map activeReg = new HashMap();

    /** Reliable log object to hold registration state records */
    private ReliableLog log;

    /** Log format version */
    private static final int LOG_VERSION = 1;

    /** Flag indicating whether system is in a state of recovery */
    private boolean inRecovery;

    /** Current number of records in the Log File since the last snapshot */
    private int logFileSize = 0;

    /** Log File must contain this many records before snapshot allowed */
    // TODO - allow this to be a configurable parameter
    private int logToSnapshotThresh = 50;

    /** Object on which the snapshot-taking thread will synchronize */
    private final Object snapshotNotifier = new Object();

    /** Snapshot-taking thread */
    private Thread snapshotter = null;

    /** Name of directory for persistence using ReliableLog */
    private String logDirname = null;

    /** ServiceID returned from the lookup registration process */
    private ServiceID serviceID = null;

    /** Our activation id, if we're activatable */
    private ActivationID activationID = null;

    /** <code>UUID</code> generator */
    private final UUIDFactory uuidFactory = new UUIDFactory();

    /** <code>EventLogIterator</code> generator */
    private final EventLogFactory eventLogFactory = new EventLogFactory();

    /** 
     * <code>UUID</code> that is shared among service proxies generated 
     * from this service instance. 
     * This value is used for determining equality among proxy objects.  
     * Service proxy objects with the same <code>UUID</code> are 
     * presumed to equal.
     */
    private UUID serviceProxyID;
  
    /** <code>LeaseDurationPolicy</code> for this service */
    private LeaseDurationPolicy leasePolicy = new LeaseDurationPolicy(
	    3 * HOURS,     // Maximum lease length
	    1 * HOURS,     // Default lease length 
	    this,          // Landlord reference (managing own leases) 
	    null,          // Lease manager reference (none used)
	    null);         // LandlordLeaseFactory reference (use default) 

    /** Manager for joining lookup services */
    private JoinManager joiner = null;

    /** 
     * DiscManager for joining lookup services. 
     * This can always be obtained from the JoinManager, so
     * this used just as a shortcut.
     */
    private LookupDiscoveryManager lookupDiscMgr = null;

    /** The attributes to use when joining lookup services */
    private Entry[] lookupAttrs = {};

    /** The lookup groups we should join. 
     * Default is to join no groups. 
     */
    private String[] lookupGroups = LookupDiscovery.NO_GROUPS;

    /** The lookup locators we should join 
     * Default is to join no locators. 
     */
    private LookupLocator[] lookupLocators = new LookupLocator[0];

    /** Event delivery thread */
    private Notifier notifier = null;

    /** Object for synchronizing with the expire thread */
    private final Object eventNotifier = new Object();

    /** Registration expirer thread */
    private ExpirationThread expirer = null;

    /** Earliest expiration time of a registration */
    private long minRegExpiration = Long.MAX_VALUE;

    /** Object for synchronizing with the expire thread */
    private final Object expirationNotifier = new Object();
 
    /** Debug utility class */
    private static final Debug debug = new Debug(MailboxBackEnd.DEBUG_PROP);

    /** PrintWriter handle for leasing related messages */
    private static final PrintWriter dbgLeases = debug.getWriter("leases");
    //private static final PrintWriter dbgLeases = new PrintWriter(System.err, true);

    /** PrintWriter handle for event delivery related messages */
    private static final PrintWriter dbgDelivery = debug.getWriter("delivery");
    //private static final PrintWriter dbgDelivery = new PrintWriter(System.err, true);

    /** PrintWriter handle for administrative related messages */
    private static final PrintWriter dbgAdmin = debug.getWriter("admin");
    //private static final PrintWriter dbgAdmin = new PrintWriter(System.err, true);

    /** PrintWriter handle for administrative related messages */
    //TODO - remove reference to System.out
    private static final PrintWriter dbgInit = debug.getWriter("init");
    //private static final PrintWriter dbgInit = new PrintWriter(System.err, true);

    /** PrintWriter handle for event related messages */
    private static final PrintWriter dbgReceive = debug.getWriter("receive");
    //private static final PrintWriter dbgReceive = new PrintWriter(System.err, true);

    /** PrintWriter handle for registration expiration related messages */
    private static final PrintWriter dbgExpiration = debug.getWriter("expiration");
    //private static final PrintWriter dbgExpiration = new PrintWriter(System.err, true);

    /** PrintWriter handle for registration expiration related messages */
    private static final PrintWriter dbgRecovery = debug.getWriter("recovery");
    //private static final PrintWriter dbgRecovery = new PrintWriter(System.err, true);

    /** 
     * Constructor for creating transient service instances. 
     * Note that transient only means non-activatable in this case.
     * State information is still logged to persistent storage.
     *
     * @param log directory where persistent state is maintained
     */
    public MailboxImpl(String log) throws IOException {
        UnicastRemoteObject.exportObject(this);
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::MailboxImpl(log=" + log + ")");
	}
        logDirname = log;
        init();
    }

    // inherit javadoc from supertype
    public UUID getProxyID() {
	return serviceProxyID;
    }

    /** Initialization common to both activatable and transient instances. */
    private void init() throws IOException {
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() ReliableLog --> " 
	        + logDirname);
	}
        log = new ReliableLog(logDirname, new LocalLogHandler());

	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() recovering persistent state");
	}
        inRecovery = true;
        log.recover();
        inRecovery = false;

				      
        // Would like to use serviceID instead, but don't want to 
        // delay construction in order to wait for the first 
        // lookup registration, which might not occur in a timely 
        // fashion.
	if (serviceProxyID == null) {
	    serviceProxyID = uuidFactory.newUUID();
        }

	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() taking snapshot");
	}
        log.snapshot();

        rebuildTransientState();

        // Create proxy that will be registered in the lookup service
        mailboxProxy = 
	    new MailboxProxy((MailboxBackEnd)RemoteObject.toStub(this),
                             serviceProxyID);

        // Create service attributes that will be registered in the 
        // lookup service
	lookupAttrs  = new Entry[] { 
	    new ServiceInfo(PRODUCT,MANUFACTURER,VENDOR,VERSION,"",""),
            new BasicServiceType("Event Mailbox Service")
        };
        
        /* Register this service with lookups for which it is configured */
	if (serviceID == null) { // First instance ... need service id
	    if (dbgInit != null) {
	        dbgInit.println("MailboxImpl::init() creating JoinMgr w/o serviceID");
	    }
	    joiner = new JoinManager(
		mailboxProxy,              // service object 
		lookupAttrs,               // service attributes
		new SrvcIDListener(),      // ServiceIDListener reference 
		null,			   // DiscoveryManagement ref - default
		null);                     // LeaseRenewalManager reference
		 		           // - using default manager
	} else { // Rejoin with (recovered) state information
	    if (dbgInit != null) {
	        dbgInit.println("MailboxImpl::init() creating JoinMgr w/ serviceID");
	    }
	    joiner = new JoinManager(
		mailboxProxy,              // service object
		lookupAttrs,               // service attributes
		serviceID,                 // Service id
		null,            	   // DiscoveryManagement ref - default
		null);                     // LeaseRenewalManager reference
		 		           // - using default manager
        }

        lookupDiscMgr = (LookupDiscoveryManager)joiner.getDiscoveryManager();

	notifier = new Notifier();
	expirer = new ExpirationThread();
	snapshotter = new SnapshotThread();

        if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() done");
        }
    } // End init()

    // Rebuilds associated data structures after a restart.
    private void rebuildTransientState() {

	// Reconstruct regByExpiration and pendingReg data structures,
	// if necessary.
	if (!regByID.isEmpty()) {
	    if (dbgInit != null) {
	        dbgInit.println("Rebuilding transient state ...");
	    }
	    Collection regs = regByID.values();
	    Iterator iter = regs.iterator();
	    ServiceRegistration reg = null;
	    UUID uuid = null;
	    while (iter.hasNext()) {
	        reg = (ServiceRegistration)iter.next();
	        uuid = (UUID)reg.getCookie();
	        if (dbgInit != null) {
	            dbgInit.println("Checking reg : " + reg);
		}
	        // Check if registration is still current
	        if (leasePolicy.ensureCurrent(reg)) {
	            if (dbgInit != null) {
	                dbgInit.println("Adding registration to "
	                    + "expiration watchlist");
		    }
	            regByExpiration.put(reg, reg);

	            // Check if registration needs to be added to the
	            // pending list. Note, we could have processed
	            // an "enabled" log record during recovery, so 
	            // only add it if it's not already there.
	            // We don't need to check activeReg since the
	            // the notifier hasn't kicked in yet. Aren't calling
	            // enableRegistration since it clears the "unknown
	            // events" list.
	            if (reg.getEventTarget() != null &&
	                !pendingReg.contains(uuid))
		    {
	                if (dbgInit != null) {
	                    dbgInit.println("Adding registration to "
	                        + "pending task list");
			}
        	        pendingReg.add(uuid);
	            }
	        } else {
	            // Registration has expired, so remove it.
	            regByID.remove(uuid);
	            if (dbgInit != null) {
	                dbgInit.println("Removing expired registration: ");
		    }
	        }
	    }
	}
        
    }

    ///////////////////////////
    // MailboxBackEnd Methods
    ///////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public MailboxRegistration register(long leaseDuration) 
	throws RemoteException
    {
	concurrentObj.writeLock();
	try {
	    return registerDo(leaseDuration);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void enableDelivery(UUID uuid, RemoteEventListener target) 
	throws RemoteException, ThrowThis 
    {
	concurrentObj.writeLock();
	try {
	    enableDeliveryDo(uuid, target);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }
	
    // This method's javadoc is inherited from an interface of this class
    public void disableDelivery(UUID uuid) 
	throws RemoteException, ThrowThis 
    {
	concurrentObj.writeLock();
	try {
	    disableDeliveryDo(uuid);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void notify(UUID registrationID, RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException, ThrowThis
    {
	concurrentObj.writeLock();
	try {
	    notifyDo(registrationID, theEvent);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }


    ////////////////////////
    // Administrable Methods
    ////////////////////////
    public Object getAdmin() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getAdmin called");
        }
	return (new MailboxAdminProxy(this, serviceProxyID));
    }


    ////////////////////////
    // Landlord Methods
    ////////////////////////

    // inherit javadoc from parent
    public long renew(Object cookie, long extension)
	    throws LeaseDeniedException, UnknownLeaseException, RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    return renewDo(cookie, extension);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public void cancel(Object cookie)
	    throws UnknownLeaseException, RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    cancelDo(cookie);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public RenewResults renewAll(Object[] cookie, long[] extension)
	    throws RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    return renewAllDo(cookie, extension);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public void cancelAll(Object[] cookie)
	    throws LeaseMapException, RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    cancelAllDo(cookie);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    //////////////////////////////////
    // KillVMAdmin methods
    //////////////////////////////////
    /**
     * Sleep time for <code>killVM</code>.
     *
     * @see killVM
     */
    private final static long KILL_DELAY = 500;

    // inherit doc comment
    public long killVM() {
        if (dbgAdmin != null) {
            dbgAdmin.println("killVM called");
        }
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server
	 * to which it is talking (this one) dies in the middle of
	 * the call.  So we start up a new thread that sleeps for 
	 * half a second and then shuts down the VM.  This lets
	 * the thread running shutDown return before the "end".
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		    ;	// oh, well
		}
		System.exit(0);
	    }
	}.start();
	return KILL_DELAY * 6;
    }

    //////////////////////////////////
    // DestroyAdmin methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public void destroy() {
        if (dbgAdmin != null) {
            dbgAdmin.println("destroy called");
        }
	(new DestroyThread()).start();
    }


    //////////////////////////////////
    // JoinAdmin methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public Entry[] getLookupAttributes() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupAttributes called");
        }
	concurrentObj.readLock();
	try {
	    /* no need to clone, never modified once created */
	    return lookupAttrs;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupAttributes(Entry[] attrSets) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupAttributes called");
        }
	concurrentObj.writeLock();
	try {
	    joiner.addAttributes(attrSets, true);
	    lookupAttrs = joiner.getAttributes();
	    /**
	    TODO - add persistence mechanism
	    EntryRep[] attrs = EntryRep.toEntryRep(attrSets, true);
	    try {
		addAttributesDo(myServiceID, 0, attrs);
	    } catch (UnknownLeaseException e) {
	    }
	    addLogRecord(new AttrsAddedLogObj(myServiceID, 0, attrs));
	    **/
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void modifyLookupAttributes(Entry[] attrSetTemplates,
				       Entry[] attrSets)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("modifyLookupAttributes called");
        }
	concurrentObj.writeLock();
	try {
	    joiner.modifyAttributes(attrSetTemplates, attrSets, true);
	    lookupAttrs = joiner.getAttributes();
	    /**
	    TODO - add persistence mechanism
	    EntryRep[] tmpls = EntryRep.toEntryRep(attrSetTemplates, false);
	    EntryRep[] attrs = EntryRep.toEntryRep(attrSets, false);
	    try {
		modifyAttributesDo(myServiceID, 0, tmpls, attrs);
	    } catch (UnknownLeaseException e) {
	    }
	    addLogRecord(new AttrsModifiedLogObj(myServiceID, 0,
						 tmpls, attrs));
	    **/
	} finally {
	    concurrentObj.writeUnlock();
	}
    }


    // This method's javadoc is inherited from an interface of this class
    public String[] getLookupGroups() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupGroups called");
	    dumpGroups(lookupGroups, dbgAdmin);
        }
	concurrentObj.readLock();
	try {
	    /* no need to clone, never modified once created */
	    return lookupGroups;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    try {
		lookupDiscMgr.addGroups(groups);
	    } catch (IOException e) {
		throw new RuntimeException(e.toString());
	    }
	    lookupGroups = lookupDiscMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void removeLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("removeLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    lookupDiscMgr.removeGroups(groups);
	    lookupGroups = lookupDiscMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("setLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    try {
		lookupDiscMgr.setGroups(groups);
	    } catch (IOException e) {
		throw new RuntimeException(e.toString());
	    }
	    lookupGroups = lookupDiscMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public LookupLocator[] getLookupLocators() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupLocators called");
        }
	concurrentObj.readLock();
	try {
	    /* no need to clone, never modified once created */
	    return lookupLocators;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    lookupDiscMgr.addLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void removeLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("removeLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    lookupDiscMgr.removeLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("setLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    lookupDiscMgr.setLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }


    //////////////////////////////////
    // StorageAdmin Methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public String getStorageLocation() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getStorageLocation called");
        }
	concurrentObj.readLock();
	try {
	    return logDirname;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setStorageLocation(String location)
	throws IOException, RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("setStorageLocation called with " + location);
        }

	if (location.equals(logDirname)) {
	    return;
	}

	concurrentObj.writeLock();
	try {
 	    ReliableLog newLog = new ReliableLog(location,
						 new LocalLogHandler());
	    newLog.snapshot();
	    ReliableLog oldLog = this.log;
	    this.log = newLog;
	    logDirname = location;
	    // Note: logDirname doesn't have to be persisted here
	    // since the activation system is doing it for us.
	    if (activationID != null) {
		try {
		    ActivationSystem sys = ActivationGroup.getSystem();
		    ActivationDesc desc = sys.getActivationDesc(activationID);
		    desc = new ActivationDesc(desc.getGroupID(),
					      desc.getClassName(),
					      desc.getLocation(),
					      new MarshalledObject(location),
					      desc.getRestartMode());
		    sys.setActivationDesc(activationID, desc);
		} catch (ActivationException e) {
		    throw new LogException("activation system problem", e);
		}
	    }
	    oldLog.deletePersistentStore();
	} finally {
	    concurrentObj.writeUnlock();
	}
    }


    //////////////////////////////////
    // Private Implementation Methods
    //////////////////////////////////

    // Utility method that adds the provided RemoteEvent
    // to the given registration's "unknown" event list.
    // If the mailbox receives an event for a registration
    // that is on this list, it will throw a UnknownEventException
    // back to the event sender.
    private void addUnknownEvent(UUID regID, EventID evid) {
	concurrentObj.writeLock();
	try {
	    ServiceRegistration reg = (ServiceRegistration)regByID.get(regID);
	    if(reg == null || !leasePolicy.ensureCurrent(reg)) {
		return; // nothing to do ... registration is gone
	    }
            if (dbgDelivery != null) {
                dbgDelivery.println("addUnknownEvent called for: " + reg);
            }

	    reg.getUnknownEvents().put(evid, evid);
	    addLogRecord(new UnknownEventExceptionLogObj(regID, evid));

            if (dbgDelivery != null) {
                dbgDelivery.println("UnknownEvents size: " 
                    + reg.getUnknownEvents().size());
            }
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // Utility method that tries to obtain the ServiceRegistration
    // object associated with the given service id.  If there is
    // no associated ServiceRegistration object, or it has expired,
    // this method will throw a "wrapped" NoSuchObjectException.
    // Note: If the NoSuchObject exception was not wrapped, then
    // the RMI system would wrap into a ServerException (since it's 
    // a RemoteEvent.  The proxies for this service know how to unwrap
    // the ThrowThis object and re-throw the NoSuchObjectException
    private ServiceRegistration getServiceRegistration(UUID regID) 
        throws ThrowThis
    {
        ServiceRegistration reg = (ServiceRegistration)regByID.get(regID); 
	if(reg == null) {
	    throw new ThrowThis( 
		new NoSuchObjectException("Not managing requested " +
	            "registration object"));
	} else if(!leasePolicy.ensureCurrent(reg)) { 
	    throw new ThrowThis( 
		new NoSuchObjectException("Requested registration object" +
		    "has expired"));
	}
	return reg;
    }

    private void removeRegistration(UUID regID, ServiceRegistration reg) {
	if (dbgLeases != null) {
	    dbgLeases.println("Removing --> " + reg);
	}
 
        regByID.remove(regID);
	regByExpiration.remove(reg);
	boolean exists = pendingReg.remove(regID);
	NotifyTask task = (NotifyTask)activeReg.remove(regID);
	if (task != null) { // cancel active task, if any
	    task.cancel();
	    if (dbgLeases != null) {
	        dbgLeases.println("Cancelling active notification task for "
	            + regID);
	    }
	}

	// Sanity check
	if (exists && task != null) {
	    if (dbgLeases != null) {
	        dbgLeases.println("ERROR: Registration was found "
	            + "on both the active and pending lists");
	    }
	    // TODO - throw internal exception
	}
    }

    private Registration registerDo(long duration) {
        UUID uuid = uuidFactory.newUUID();
        EventLogIterator iter = eventLogFactory.iterator(uuid);
        ServiceRegistration reg = new ServiceRegistration(uuid, iter);
        Lease l = null;
        try {
            l = leasePolicy.leaseFor(reg, duration);
	} catch (LeaseDeniedException lde) {
	    // Should never happen in our implementation.
	    // Re-throw as an internal exception.
	    throw new InternalMailboxException("Registration lease was denied", 
		                               lde);
	}

        // Add registration to data structures 
        addRegistration(reg);

        // log a record of this action
        addLogRecord(new RegistrationLogObj(reg));

        // Check expiration and check if expiration thread needs to 
        // wakeup earlier than scheduled.
	if (reg.getExpiration() < minRegExpiration) {
	    minRegExpiration = reg.getExpiration();
            if (dbgExpiration != null) {
                dbgExpiration.println("Notifying expiration thread");
            }
	    concurrentObj.waiterNotify(expirationNotifier);
	}

	if(dbgLeases != null) {
	    dbgLeases.println("Generated new lease for: ");
	    reg.dumpInfo(dbgLeases);
	}

	return new Registration(uuid, this, l);
    }

    /** 
     * Utility method used to add a registration to our state information. 
     * This method is invoked from <code>registerDo()</code> as well as
     * <code> RegistrationLogObj.apply()</code>.  
     */
    private void addRegistration(ServiceRegistration reg) {
	regByID.put(reg.getCookie(), reg);
	regByExpiration.put(reg, reg);
    }

    /** Performs the actual renewal logic */
    private long renewDo(Object cookie, long extension)
        throws UnknownLeaseException 
    {
        ServiceRegistration reg = null;
	long expiration = 0;
        
	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to renew " + cookie 
	        + "'s lease for " + extension/1000 + " sec"); 
	}
	// Get registration of interest
	try { 
            // Note: the following method will throw a ThrowThis exception
            // if the registration is invalid (i.e. expired or non-existent)
	    reg = getServiceRegistration((UUID)cookie);
	    expiration = leasePolicy.renew(reg, extension);
	    addLogRecord(new RegistrationRenewedLogObj((UUID)cookie, 
	                                               reg.getExpiration()));
	    // force a resort to occur
	    regByExpiration.remove(reg);
	    regByExpiration.put(reg, reg);
	} catch (ThrowThis tt) {
	    // Registration doesn't exist or has expired
	    if(dbgLeases != null) {
	        dbgLeases.println("Lease for " + cookie + " was NOT renewed");
	    }
	    throw new UnknownLeaseException("Not managing requested lease");
	}

	if(dbgLeases != null) {
	    dbgLeases.println("Lease for " + cookie + " was renewed");
	    reg.dumpInfo(dbgLeases);
	}

	return expiration;
    }
    
    
    private void cancelDo(Object cookie) throws UnknownLeaseException {
        ServiceRegistration reg = null;

    	if(dbgLeases != null) {
    	    dbgLeases.println("Attempting to cancel lease for: " + cookie); 
    	}

	try { 
            // Note: the following method will throw a ThrowThis exception
            // if the registration is invalid (i.e. expired or non-existent)
	    reg = getServiceRegistration((UUID)cookie);
            removeRegistration((UUID)cookie, reg);
            if (reg.getExpiration() == minRegExpiration) {
                if (dbgExpiration != null) {
                    dbgExpiration.println("Notifying expiration thread");
                }
	        concurrentObj.waiterNotify(expirationNotifier);
	    }
	} catch (ThrowThis tt) {
	    // Registration doesn't exist or has expired
	    throw new UnknownLeaseException("Not managing requested lease");
	}
    	if(dbgLeases != null) {
    	    dbgLeases.println("Lease cancelled for: " + cookie); 
    	}
    }
    
    private RenewResults renewAllDo(Object[] cookie, long[] extension)
	    throws RemoteException 
    {
	int count = cookie.length;
	long[] granted = new long[count];
	ArrayList exceptions = new ArrayList(count);

	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to renew a batch of " +
			      count + " leases"); 
	}

	for (int i = 0; i < count; i++) {
	    try {
		granted[i] = renewDo(cookie[i], extension[i]);
	    } catch (LeaseException le) {
		// Set flag for client-side handling
		granted[i] = -1;
		exceptions.add(le);
	    }
	}

	if(exceptions.size() == 0)  {
	    if(dbgLeases != null) {
	        dbgLeases.println("Batch renew totally successful");
	    }
	    return new RenewResults(granted);
	} else {
	    if(dbgLeases != null) {
	        dbgLeases.println("Batch renew contained exceptions");
	    }
	    // Note: Can't just cast exceptions.toArray() to Exception[]
	    Exception[] es = new Exception[exceptions.size()];
	    Iterator it = exceptions.iterator();
	    for (int i = 0; it.hasNext(); i++)
		es[i] = (Exception) it.next();
	    
	    return new RenewResults(granted, es);
	}
    }	     

    private void cancelAllDo(Object[] cookie) 
	throws LeaseMapException, RemoteException 
    {
	int count = cookie.length;
	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to cancel a batch of " +
		count + " leases"); 
	}

	Map map = null;
	for (int i = 0; i < count; i++) {
	    try {
		cancelDo(cookie[i]);
	    } catch (UnknownLeaseException e) {
		if (map == null)
		     map = new HashMap();
		map.put(cookie[i], e);
	    }
	}

	if (map != null)
	     throw new LeaseMapException("Trouble cancelling", map);
    }
	
    // This method's javadoc is inherited from an interface of this class
    private void enableDeliveryDo(UUID uuid, RemoteEventListener target) 
	throws ThrowThis 
    {
        // Proxy already checks for target == null, but just in case
        if (target == null) {
            //TODO - Throw internal exception
        }

	// Check for the resubmission of one of our own listeners
	if (target instanceof ListenerProxy) {
	    UUID id = ((ListenerProxy)target).getID(); 
	    if (regByID.get(id) != null) {
	        if (dbgDelivery != null) {
	            dbgDelivery.println("Disallowing resubmission of one of" 
		       +  " this service's own listeners: " + id); 
	        }
	        throw new IllegalArgumentException("Cannot submit a " +
		    "listener that was provided by the same EventMailbox " +
		    "service");
	    }
	}

	enableRegistration(uuid, target);
	addLogRecord(new RegistrationEnabledLogObj(uuid, target));

	// Notify the notifier of new "pending" entry
	concurrentObj.waiterNotify(eventNotifier);

	if (dbgDelivery != null) {
	    dbgDelivery.println("Enabling delivery for " + uuid +
		" to " + target);
	}
    }

    // 
    private void enableRegistration(UUID uuid, RemoteEventListener target) 
	throws ThrowThis 
    {
        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(uuid); 
	reg.setEventTarget(target); 

	// Clear unknownEvent collection regardless if the same target 
	// is provided. We'll assume that an active re-registration means 
	// that the client's listener has changed (for the better) somehow.
	reg.getUnknownEvents().clear();

        // Add this registration to set of enabled registrations
        // if it's not already there. Note that an enabled registration
        // can be in either the pendingReg or activeReg data structures
        // at any one time.
	if (!pendingReg.contains(uuid) && !activeReg.containsKey(uuid))
	    pendingReg.add(uuid);
    }

    private void disableDeliveryDo(UUID uuid) throws ThrowThis {
        disableRegistration(uuid);
        addLogRecord(new RegistrationDisabledLogObj(uuid)); 
    }
        
    private void disableRegistration(UUID uuid) throws ThrowThis {
	if (dbgDelivery != null) {
	    dbgDelivery.println("Attempting to disable delivery for " + uuid);
	}

        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(uuid); 

        // cache pre-disable value for debugging
        RemoteEventListener rel = reg.getEventTarget();

        // Set delivery target to null
	reg.setEventTarget(null); 

	// Remove registration from pending and active lists
	boolean inPending = pendingReg.remove(uuid);  
	NotifyTask task = (NotifyTask)activeReg.remove(uuid);
	if (task != null) { // cancel active task, if any
	    task.cancel();
	    if (dbgDelivery != null) {
	        dbgDelivery.println("Cancelling active notification task for " 
	            + uuid);
	    }
	}

        // If rel != null then the registration was previously enabled.
        // If so, then the its uuid should have appeared in either
        // the active or pending data structures. This can happen 
        // during recovery but otherwise there's a problem.
	if (rel != null && !inPending && task == null && !inRecovery) {
	    if (dbgDelivery != null) {
	        dbgDelivery.println("*** Registration was not found "
	            + "on the active or pending lists");
	    }
	    // TODO - throw internal exception
	}

	if (dbgDelivery != null) {
	    dbgDelivery.println("Disabled delivery for " + uuid);
	}
    }

    // This method's javadoc is inherited from an interface of this class
    private void notifyDo(UUID registrationID, RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException, ThrowThis
    {
        if (dbgReceive != null) {
            dbgReceive.println("Received event for " + registrationID);
	    dumpEvent(theEvent, dbgReceive);
	}

        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(registrationID); 

	// Check for event validity
	EventID evtid = new EventID(theEvent);
	if (reg.getUnknownEvents().containsKey(evtid)) {
            if (dbgReceive != null) {
                dbgReceive.println("Discarding event -- it's unknown");
	    }
	    throw new UnknownEventException();
	}

	try {
	    reg.iterator().add(theEvent);
	} catch (IOException ioe) {
	    // TODO - implement recovery code for persistent version
	    // ignore for now since we aren't persistent
	}

	if (reg.getEventTarget() != null) {
	    concurrentObj.waiterNotify(eventNotifier);
	}

        if (dbgReceive != null) {
	    reg.dumpInfo(dbgReceive);
	}
	
    }

    private void setServiceID(ServiceID id) {
	if (dbgInit != null) {
	    dbgInit.println("serviceID being set to " + id);
	}
        serviceID = id;
    }

    ///////////////////////////
    // Inner Classes
    //////////////////////////

    /** Class which listens for the announcement of a service ID assignment
     *  by the JoinManager
     */
    private class SrvcIDListener implements ServiceIDListener
    {
        public SrvcIDListener() {
            super();
        }

        /** The JoinManager will invoke this method when it receives a
         *  valid ServiceID from a lookup service; passing in that 
         *  ServiceID so that it may be stored for persistence.
         */
        public void serviceIDNotify(ServiceID id) {
            setServiceID(id);
            addLogRecord(new ServiceIDLogObj(id));
        }
    }//end class SrvcIDListener



    /**
     * The notifier thread.  This thread is responsible for notifying
     * client supplied notification targets when their associated registrations
     * have been enabled. 
     * <p>
     * The constructor calls <code>start()</code> so there is no need to explicitly
     * start this thread.
     */
    private class Notifier extends Thread implements TimeConstants {
        /** 
         * <code>TaskManager</code> that will be handling the notification tasks 
         */
        private TaskManager	taskManager;	
    
        /** Random number generator that will be used for implementing
         * a simple load balancing scheme. Seed it with the current time.
         */
        private Random rand = new Random(System.currentTimeMillis());
    
        // Time to wait between notification checks
        private final static long	PAUSE_TIME = 5000; // 5 seconds
    
        /**
         * Simple constructor
         */
        Notifier() {
    	    super("Notifier");
    	    taskManager = new TaskManager();
    	    start();
        }
    
        /**
         * Loop looking for new entries to check for, and notifications to
         * be sent.
         */
        public void run() {
	    try {
		concurrentObj.writeLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }

	    try {
	        while (!isInterrupted()) {
    	            int count = pendingReg.size();
    	            if (dbgDelivery != null) {
		        dbgDelivery.println(getName() + " checking " + count 
		            + " possible registrations");
		    }
    	            // don't need to check for size() > 0 (below) since
    	            // we hold the lock. 
    	            // TODO: optimize loop to ignore indices w/o available events
    	            // TODO: optimize loop to skip nextInt if only one choice
    	            while (count-- > 0) { 
    	                int index = rand.nextInt(pendingReg.size());
    	                UUID uuid = (UUID)pendingReg.get(index);
    	                ServiceRegistration reg = null;
        	        try { 
                            // Note: the following method will throw a 
                            // ThrowThis exception if the registration is 
                            // invalid (i.e. expired or non-existent)
                            reg = getServiceRegistration(uuid); 
    	                    if (reg.iterator().hasNext()) {
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println(getName() 
		                        + " Scheduling"
		                        + " delivery task for " + reg + "["
		                        + reg.iterator().size() 
		                        + " events left]");
				}
    	                        NotifyTask t = new NotifyTask(taskManager, uuid);
    	                        taskManager.add(t);
    	                        activeReg.put(uuid, t);
    	                        pendingReg.remove(index);
    		            }
			} catch (ThrowThis tt) {
			    // Invalid registration ... skip
			}
    		    }
		    try {
                        if (dbgDelivery != null ) {
                            dbgDelivery.println(getName()
                                + " delayed until " 
                                + new Date(PAUSE_TIME + 
                                           System.currentTimeMillis()));
			}
			// There is a potential for missing "notify" calls
			// so we can't just use a large timeout value.
			concurrentObj.writerWait(eventNotifier, PAUSE_TIME);
		    } catch (ConcurrentLockException e) {
			return;
		    }
		}

    	    } finally {
    	        concurrentObj.writeUnlock();
    	    }
    	                
    	    if (dbgDelivery != null) {
	        dbgDelivery.println(getName() + " finished ...");
	    }
        }
    }
    
    /*
     * Static stuff for NotifyTask. Unfortunately, we can't have 
     * static references in an inner class.
     */
    private static final long	MAX_TIME = 1 * HOURS;
    private final static int	MAX_ATTEMPTS = 5;	// max times to retry
    
    /**
     * A task that represent an event notification task for a particular
     * registration.
     */
    class NotifyTask extends RetryTask {
        private UUID	regID;
    	private long startTime = System.currentTimeMillis();
    
    	/**
    	 * Create an object to represent an event notification task.
    	 */
    	NotifyTask(TaskManager tm, UUID regID) {
    	    super(tm);
    	    this.regID = regID;
    	}

    	/**
    	 * This utility function tries to obtain the next event for the
    	 * given registration. If it finds one, the event is returned. 
    	 * Otherwise null is returned.
    	 */
    	private RemoteEvent getNextEvent(ServiceRegistration reg) {
    	    RemoteEvent evt = null;
    	    if (reg.iterator().hasNext()) {
		try {
		    evt =  (RemoteEvent)reg.iterator().next();
		} catch (IOException ioe) {
		    //TODO - implement recovery for persistent storage
		}
	    }
	    return evt;
    	}

    	/**
    	 * This utility function tries to obtain the next event for the
    	 * given registration. If it finds one, the event is returned. 
    	 * Otherwise null is returned.
    	 */
    	private void deleteNextEvent(ServiceRegistration reg) {
    	    if (reg.iterator().hasNext()) {
		try {
		    reg.iterator().remove();
		} catch (IOException ioe) {
		    //TODO - implement recovery for persistent storage
		}
	    } 
    	}    

    	/**
    	 * This utility function attempts to disable the given registration. 
    	 */
    	private boolean disableRegistration(UUID regID, RemoteEventListener l) {
    	    boolean disabled = true;
	    concurrentObj.writeLock();
	    try {
	        try {
                    // Note: the following method will throw a ThrowThis exception
                    // if the registration is invalid (i.e. expired or non-existent)
	            ServiceRegistration reg = getServiceRegistration(regID); 

	            // Check to see if the listener has changed since 
	            // the initial call to this method. If so, 
	            // assume that the new listener doesn't need to be
	            // disabled.  If not, continue with the disabling
	            // of the registration.
		    RemoteEventListener currentListener = reg.getEventTarget();
		    if (currentListener != null && currentListener.equals(l)) {
		        disableDeliveryDo(regID);
		    } else {
	                if (dbgDelivery != null) {
	                    dbgDelivery.println("Disabling registration for " 
	                        + regID + " skipped due to listener change.");
	                }
	                disabled = false;
		    }
		} catch (ThrowThis tt) {
		    // registration is gone ... ignore
		}
	    } finally {
	        concurrentObj.writeUnlock();
	    }
	    return disabled;
    	}  

    	/**
    	 * Try to notify the target.  Return <code>true</code> if the
    	 * notification was successful. Return <code>false</code> otherwise.
    	 * If we return false, this task will be rescheduled at some
    	 * point in the future.
    	 */
    	public boolean tryOnce() {
    
    	    boolean succeeded = false;           // attempt this task again?      
    	    boolean deleteEvent = false;         // delete the retrieved event? 
    	    boolean doNotify = true;             // attempt notification?
    	    ServiceRegistration reg = null;      // 
    	    RemoteEventListener listener = null; // event delivery target
	    RemoteEvent ev = null;               // event to deliver

    	    if (dbgDelivery != null) {
		dbgDelivery.println("Attempting event delivery for: " + regID
		    + " at " + new java.util.Date(System.currentTimeMillis()));
	    }
    
            // Put an upper time limit on how long to keep
            // this task on the active list.  This prevents
            // slow/bad event recipients from monopolizing
            // a thread in the thread pool. This will not 
            // prevent a "malicious" event recipient from 
            // hanging up this thread if it never returns
            // from the notify() call.
	    long curTime = System.currentTimeMillis();
    	    if (curTime - startTime() > MAX_TIME) {
		succeeded = true;
		deleteEvent = false;
    	        doNotify = false;
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Cancelling delivery due to time " +
		        "limit expiration.");
	        }
	    } else {
                // Get required delivery information
                concurrentObj.readLock();
        	try {
        	    try { 
                        // Note: the following method will throw a ThrowThis exception
                        // if the registration is invalid (i.e. expired or non-existent)
                        reg = getServiceRegistration(regID); 
        		listener = reg.getEventTarget();
        		ev = getNextEvent(reg);
    
                        // Check to see if either the listener or the event 
                        // is null. If so, skip the notify() call.
                        // If the event is not null, make an 
                        // additional check to see if the event is on the 
                        // unknown event list for this registration.
                        // If so, then skip the notify() call as well.
        		if (listener == null || ev == null) {
        		    succeeded = true;    // don't try again
        		    deleteEvent = false; // don't delete event
        		    doNotify = false;    // skip notify
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Cancelling delivery " +
		                   " because of null listener or event");
	                    }
			} else if (reg.getUnknownEvents().containsKey(
			               new EventID(ev))) {
        		    succeeded = true;    // don't try again
        		    deleteEvent = true;  // delete event
        		    doNotify = false;    // skip notify
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Cancelling delivery " +
		                   " because of unknown event");
	                    }
        		}
		    } catch (ThrowThis tt) { 
    		        succeeded = true;        // don't try again
    		        deleteEvent = false;     // don't remove event
        		doNotify = false;        // skip notify
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Cancelling delivery " +
		                " because of unknown registration");
	                }
    		    }
        	} finally {
                    concurrentObj.readUnlock();
        	}
	    } // end else
    	    
    	    // Important - don't hold any locks during a remote invocation
    	    if (doNotify) {
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Delivering evt: " + ev);
		    dbgDelivery.println("\tID#:  " + ev.getID());
		    dbgDelivery.println("\tSeq#: " + ev.getSequenceNumber());
	        }
		try {
		    listener.notify(ev);
		    succeeded = true;
    		    deleteEvent = true;
    	            if (dbgDelivery != null) {
		        dbgDelivery.println("Delivery was successful");
		    }
		} catch (UnknownEventException e) {
		    // they didn't want to know about this, so stop them 
		    // from getting future notifications, too.
    	            if (dbgDelivery != null) {
		        dbgDelivery.println("Caught UnknownEventException " +
		            " during notify");
	            }
		    addUnknownEvent(regID, new EventID(ev));
		    succeeded = true;
		    deleteEvent = true;
	        } catch (RemoteException re) {
	            if (((re instanceof ServerException) &&
	                 (re.detail != null && 
	                  re.detail instanceof NoSuchObjectException)) ||
	                re instanceof NoSuchObjectException) 
		    {
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Caught a definite " +
		                " RemoteException during notify");
	                }
		        disableRegistration(regID, listener);
		        succeeded = true;
		        deleteEvent = false;
		    } else {
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Caught an indefinite " +
		                " RemoteException during notify");
	                }
		        succeeded = false;
		        deleteEvent = false;
		    }
		} // end catch    
	    } // end if
    
            // If we still aren't successful after MAX_ATTEMPTS
            // then give up
    	    if (!succeeded && attempt() > MAX_ATTEMPTS) {
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Maximum delivery attempts reached");
		}
		succeeded = true;		
		deleteEvent = true;
	    }
    
            // Check if data structures need to be modified
            // If so, make the change and notify the 
            // notifier thread.
            if (succeeded || deleteEvent) {
                concurrentObj.writeLock();
        	try {
    		    try { 
                        // Note: the following method will throw a ThrowThis 
                        // exception if the registration is invalid 
                        // (i.e. expired or non-existent)
                        reg = getServiceRegistration(regID); 
                        if (succeeded) {
                            if (reg.getEventTarget() != null) {
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println("Putting task back "
		                        + "onto pending list");
		                }
        	                activeReg.remove(regID);
        	                pendingReg.add(regID);
			    } else { // reg was disabled
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println("Removing task ...");
				}
				// Note: disabling delivery should
				// have taken care of this for us.
        	                if (activeReg.remove(regID) != null ||
        	                    pendingReg.remove(regID) == true  ) {
    	                            if (dbgDelivery != null) {
		                        dbgDelivery.println("ERROR: Found "
		                            + "pending/active task for a "
		                            + "disabled registration"); 
				    }
				    // TODO - throw internal exception
				}
			    }
    		        }
        	        if (deleteEvent) {
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Deleting event ...");
			    }
        	            deleteNextEvent(reg); 
    		        }
    		    } catch (ThrowThis tt) { 
    		        // Registration is gone ... nothing to do
    		    }
        	} finally {
                    concurrentObj.writeUnlock();
        	}
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Waking up notifier");
		}
	        concurrentObj.waiterNotify(eventNotifier);
	    }

    	    return succeeded;
    	}
    
        /**
         * Return <code>true</code> if this task needs to runAfter any
         * of the tasks in the provided list and <code>false</code>
         * otherwise.  The notifier thread ensures that there should only
         * be one task per registration, so this method just returns
         * false.
         */
    	public boolean runAfter(java.util.List list, int max) {
    	    return false;
       	}
    }


    /**
     * Termination thread code.  We do this in a separate thread to
     * avoid deadlock, because Activatable.inactive will block until
     * in-progress RMI calls are finished.
     */
    private class DestroyThread extends Thread {

	/** Create a non-daemon thread */
	public DestroyThread() {
	    super("DestroyThread");
	    /* override inheritance from RMI daemon thread */
	    setDaemon(false);
	}

	public void run() {
	    if (dbgAdmin != null) {
		dbgAdmin.println("DestroyThread started ...");
	    }

	    /* must unregister before unexport */
	    if (activationID != null) {
		try {
		    Activatable.unregister(activationID);
		} catch (RemoteException e) {
		    /* give up until we can at least unregister */
		    return;
		} catch (ActivationException e) {
		}
	    }
	    try {
	        /* wait for any pending operations to complete */
		while (!UnicastRemoteObject.unexportObject(MailboxImpl.this,
							   false))
		{
		    Thread.yield();
		}
	    } catch (NoSuchObjectException e) {
	    }

	    /* all daemons must terminate before deleting persistent store */
	    if (dbgAdmin != null) {
		dbgAdmin.println("Terminating JoinManager ...");
	    }
	    joiner.terminate();

	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Notifier ...");
	    }
	    notifier.interrupt();
	    
	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Expirer ...");
	    }
	    expirer.interrupt();

	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Snapshotter ...");
	    }
	    snapshotter.interrupt();

	    try {
	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Notifier ...");
	        }
	        notifier.join();

	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Expirer ...");
	        }
	        expirer.join();	        

	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Snapshotter ...");
	        }
	        snapshotter.join();
	    } catch (InterruptedException e) {
	    }

	    log.deletePersistentStore();

	    if (activationID != null) {
		/* inactive will set current group ID to null */
		ActivationGroupID gid = ActivationGroup.currentGroupID();
		try {
		    while (!Activatable.inactive(activationID)) {
			Thread.yield();
		    }
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
		try {
		    ActivationGroup.getSystem().unregisterGroup(gid);
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
	    }
	    if (dbgAdmin != null) {
		dbgAdmin.println("DestroyThread finished ...");
		//dbgAdmin.println("DestroyThread sleeping 10s...");
		//try{ Thread.sleep(10000); } catch (Exception e) {}
	    }
	}
    }

    /** Registration expiration thread code */
    private class ExpirationThread extends Thread implements TimeConstants {

	/** Create a daemon thread */
	public ExpirationThread() {
	    super("ExpirationThread");
	    setDaemon(true);
            if (dbgExpiration != null ) {
                dbgExpiration.println(getName() + " started ...");
            }
            start();
	}

	/** 
	 * Loop over time-sorted collection and expire (remove) any entries
	 * that have expired.
	 */
	public void run() {
            ServiceRegistration reg = null;
            long delay = 0;

	    try {
		concurrentObj.writeLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }

	    try {
	        while (!isInterrupted()) {
		    long now = System.currentTimeMillis();
                    if (dbgExpiration != null ) {
                        dbgExpiration.println(getName() 
                                + " Checking regs @ " + new Date(now));
		    }
		    while (!regByExpiration.isEmpty()) {
                        reg = (ServiceRegistration)regByExpiration.firstKey(); 
			minRegExpiration = reg.getExpiration();
                        if (dbgExpiration != null ) {
                            dbgExpiration.println(getName() 
                                + " Checking " + reg 
                                + " which expires @ " 
                                + new Date(minRegExpiration));
                        }
			if (minRegExpiration > now) 
			    break;
                        removeRegistration((UUID)reg.getCookie(), reg);
		    }
		    try {
		        delay = minRegExpiration - now;
		        delay = (delay >= 0) ? delay : Long.MAX_VALUE;
                        if (dbgExpiration != null ) {
                            dbgExpiration.println(getName()
                                + " delayed until " + new Date(delay + now));
			}
			concurrentObj.writerWait(expirationNotifier, delay);
		    } catch (ConcurrentLockException e) {
			return;
		    }

		}
	    } finally {
	        concurrentObj.writeUnlock();
            }

            if (dbgExpiration != null ) {
                dbgExpiration.println(getName() + " finished ...");
	    }
	}
    } // End class


    ////////////////////////////////
    // Persistence Methods/Classes
    ////////////////////////////////

    /**
     * Interface defining the method(s) that must be implemented by each of
     * the concrete LogObj classes. This allows for the definition of
     * object-dependent invocations of the appropriate implementation of
     * the method(s) declared in this interface.
     */
    private static interface LogRecord extends Serializable {
	void apply(MailboxImpl mb);
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the serviceID notification from the lookup service occurs (which
     * should only happen once per instantiation).
     */
    private static class ServiceIDLogObj implements LogRecord {
	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The service ID.
	 *
	 * @serial
	 */
	private ServiceID id;

	/** Simple constructor */
        public ServiceIDLogObj(ServiceID id) {
            this.id = id;
        }

	/**
	 * Sets the mailbox's service ID to this object's stored value.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	    }
            mb.setServiceID(id);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a new registration is created.
     */
    private static class RegistrationLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The registration.
	 *
	 * @serial
	 */
	private ServiceRegistration reg;

	/** Simple constructor */
	public RegistrationLogObj(ServiceRegistration reg) {
	    this.reg = reg;
	}

	/**
	 * Adds the stored registration object to the mailbox's state.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        reg.dumpInfo(dbgRecovery);
	    }
	    mb.addRegistration(reg);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is enabled.
     */
    private static class RegistrationEnabledLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private transient RemoteEventListener target;

        /** Simple constructor */
        public RegistrationEnabledLogObj(UUID id, RemoteEventListener l) {
            regID = id;
            target = l;
        }

	/**
	 * @serialData RemoteEventListener as a MarshalledObject
	 */
	private void writeObject(ObjectOutputStream stream)
	    throws IOException
	{
	    stream.defaultWriteObject();
	    stream.writeObject(new MarshalledObject(target));
	}

	/**
	 * Unmarshals the event listener.
	 */
	private void readObject(ObjectInputStream stream)
	    throws IOException, ClassNotFoundException
	{
	    stream.defaultReadObject();
	    MarshalledObject mo = (MarshalledObject)stream.readObject();
	    try {
		target = (RemoteEventListener)mo.get();
	    } catch (Throwable e) {
		if (e instanceof Error &&
		    !(e instanceof LinkageError ||
		      e instanceof OutOfMemoryError ||
		      e instanceof StackOverflowError))
		    throw (Error)e;
	    }
	}

	/**
	 * Enables the stored RemoteEventListener object with the
	 * associated, stored registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("\t" + regID);
	        dbgRecovery.println("\t" + target);
	    }
	    try {
	        mb.enableRegistration(regID, target);
	    } catch (ThrowThis tt) {
	        // TODO - throw internal exception since this shouldn't happen
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is disabled.
     */
    private static class RegistrationDisabledLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private UUID regID;

	/** Simple constructor */
        public RegistrationDisabledLogObj(UUID id) {
            regID = id;
        }

	/**
	 * Disables the associated, stored registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("\tregID: " + regID);
	    }
	    try {
	        mb.disableRegistration(regID);
	    } catch (ThrowThis tt) {
	        // TODO - throw internal exception since this shouldn't happen
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of groups to join is changed.
     */
    private static class LookupGroupsChangedLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The new groups to join.
	 *
	 * @serial
	 */
	private String[] groups;

	/** Simple constructor */
	public LookupGroupsChangedLogObj(String[] groups) {
	    this.groups = groups;
	}

	/**
	 * Sets the value of <code>lookupGroups</code> to the provided
         * set.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        mb.dumpGroups(groups, dbgRecovery);
    	    }
	    mb.lookupGroups = groups;
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of locators of lookup services to join is changed.
     */
    private static class LookupLocatorsChangedLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The new locators to join.
	 *
	 * @serial
	 */
	private LookupLocator[] locators;

	/** Simple constructor */
	public LookupLocatorsChangedLogObj(LookupLocator[] locators) {
	    this.locators = locators;
	}

	/**
	 * Sets <code>lookupLocators</code> to value of the provided
	 * set.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        mb.dumpLocators(locators, dbgRecovery);
	    }
	    mb.lookupLocators = locators;
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is renewed.
     */
    private static class RegistrationRenewedLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The absolute time granted for this renewal.
	 *
	 * @serial
	 */
	private long expirationTime;

	/** Simple constructor */
	public RegistrationRenewedLogObj(UUID id, long time) {
	    regID = id;
	    expirationTime = time;
	}

	/**
	 * Sets the expiration time of the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Reg: " + regID + " will expire at " 
	            + new Date(expirationTime));
	    }

            // Note: The registration might be expired at this point
            // but we can't be sure until the recovery process is
            // complete.
            ServiceRegistration reg = 
                (ServiceRegistration)mb.regByID.get(regID);
            if (reg != null) { 
                reg.setExpiration(expirationTime);
	    } else {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("*** Registration not renewed "
	                + "- not found");
		}
		// TODO - throw internal exception
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is renewed.
     */
    private static class RegistrationCancelledLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/** Simple constructor */
	public RegistrationCancelledLogObj(UUID id) {
	    regID = id;
	}

	/**
	 * Sets the expiration time of the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Cancelling Reg: " + regID);
	    }

            // Note: The registration might be expired at this point
            // but we can't be sure until the recovery process is
            // complete.
            ServiceRegistration reg = 
                (ServiceRegistration)mb.regByID.get(regID);
            mb.removeRegistration(regID, reg);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * an UnknownEventException is during event notification. 
     * The exception is associated with the given registration.
     */
    private static class UnknownEventExceptionLogObj implements LogRecord {

	// TODO - generate serialVersionUID
	// private static final long serialVersionUID = 0; 

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The internal representation of the remote event which
	 * caused the UnknownEventException.
	 *
	 * @serial
	 */
	private EventID evtID;

        /** Simple constructor */
        public UnknownEventExceptionLogObj(UUID id, EventID eid) {
            regID = id;
            evtID = eid;
        }

	/**
	 * Adds the given <code>EventID</code> to the set of unknownEvents
	 * for the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Adding: " + evtID + " to " + regID);
	    }
	    ServiceRegistration reg = 
	        (ServiceRegistration)mb.regByID.get(regID);
	    reg.getUnknownEvents().put(evtID, evtID);
	}
	
    }
    

    /**
     * Add a state-change record to persistent storage.
     * <p>
     * Whenever a significant change occurs to the Mailbox's state, this
     * method is invoked to record that change in a file called a log file.
     * Each record written to the log file is an object reflecting both
     * the data used and the ACTIONS taken to make one change to the
     * Mailbox's state at a particular point in time. If the number of
     * records contained in the log file exceeds the pre-defined threshold,
     * a snapshot of the current state of the Mailbox will be recorded.
     * <p>
     * Whenever one of the following state changes occurs, this method
     * will be invoked with the appropriate implementation of the
     * LogRecord interface as the input argument.
     * <ul>
     * <li> registration creation
     * <li> registration renewal
     * <li> registration deletion/cancellation
     * <li> registration enabled
     * <li> registration disabled
     * </ul>
     * 
     * @see MailboxImpl.LocalLogHandler
     */
    private void addLogRecord(LogRecord rec) {
	try {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Adding log record: " + rec);
	    }
	    log.update(rec, true);
	    if (++logFileSize >= logToSnapshotThresh) {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("Notifying snapshot thread");
	        }
                concurrentObj.waiterNotify(snapshotNotifier);
	    }
	} catch (Exception e) {
	    /* if log updating fails, then one of the following must be done:
	     *   -- output the problem to a separate file and exit()
	     *   -- output the problem to a separate file and continue
	     *   -- set an "I have a problem" attribute & send notification
	     * XXX this issue will be addressed at a later time
	     */
	    e.printStackTrace();
	}
    }


    /**
     * Write the current state of the Mailbox to persistent storage.
     * <p>
     * A 'snapshot' of the Mailbox's current state is represented by
     * the data contained in certain fields of the Mailbox. That data
     * represents many changes -- over time -- to the Mailbox's state.
     * This method will record that data to a file referred to as the
     * snapshot file.
     * <p>
     * The data written by this method to the snapshot file -- as well as
     * the format of the file -- is shown below:
     * <ul>
     * <li> our class name
     * <li> log format version number
     * <li> our service ID
     * <li> current event ID
     * <li> current configuration parameters
     * <li> contents of the container holding the current registered services
     * <li> null (termination 'marker' for the set of registered services)
     * <li> contents of the container holding the current registered events
     * <li> null (termination 'marker' for the set of registered events)
     * </ul>
     * Each data item is written to the snapshot file in serialized form.
     * 
     * @see Mailbox.LocalLogHandler
     */

    private void takeSnapshot(OutputStream  out) throws IOException {
	if (dbgRecovery != null) {
	    dbgRecovery.println("Taking snapshot ...");
	}
	ObjectOutputStream stream = new ObjectOutputStream(out);

	stream.writeUTF(MailboxImpl.class.getName());
	stream.writeInt(LOG_VERSION);
	stream.writeObject(serviceID);
	stream.writeObject(serviceProxyID);
	stream.writeInt(logToSnapshotThresh);
	stream.writeObject(lookupGroups);
	stream.writeObject(lookupLocators);
	stream.writeObject(regByID);
	stream.flush();
	// TODO - persist LookupAttributes
    }

    /**
     * Retrieve the contents of the snapshot file and reconstitute the 'base'
     * state of the Mailbox from the retrieved data.
     * <p>
     * The data retrieved by this method from the snapshot file is shown
     * below:
     * <ul>
     * <li> our class name
     * <li> log format version number
     * <li> our service ID
     * <li> current configuration parameters
     * <li> contents of the container holding registration information
     * </ul>
     * During recovery, the state of the Mailbox at the time of a crash
     * or outage is re-constructed by first reconstituting the 'base state'
     * from the snapshot file; and then modifying that base state according
     * to the records retrieved from the log file. This method is invoked to
     * perform the first step in that reconstruction. As each registered
     * service or event is retrieved, it is resolved and then inserted into
     * its appropriate container object.
     * <p>
     * Because events can be generated before the next snapshot is taken,
     * the event sequence numbers must be incremented. This is because the
     * event specification requires that set of event sequence numbers
     * be monotonically increasing. Since the number of events that might
     * have been sent is arbitrary, each sequence number will be incremented
     * by a 'large' number so as to guarantee adherence to the specification.
     * 
     * @see Mailbox.LocalLogHandler
     */
    private void recoverSnapshot(InputStream in)
	throws IOException, ClassNotFoundException
    {
	if (dbgRecovery != null) {
	    dbgRecovery.println("Recovering snapshot ...");
	}
	int i = 0;
	ObjectInputStream stream = new ObjectInputStream(in);
	if (!MailboxImpl.class.getName().equals(stream.readUTF()))
	    throw new IOException("log from wrong implementation");
	if (stream.readInt() != LOG_VERSION)
	    throw new IOException("wrong log format version");
	serviceID = (ServiceID)stream.readObject();
	serviceProxyID = (UUID)stream.readObject();
	logToSnapshotThresh = stream.readInt();
	lookupGroups = (String[])stream.readObject();
	lookupLocators = (LookupLocator[])stream.readObject();
	regByID = (HashMap)stream.readObject();
	if (dbgRecovery != null) {
	    dbgRecovery.println("serviceID:           " + serviceID);
	    dbgRecovery.println("logToSnapshotThresh: " + logToSnapshotThresh);
	    dbgRecovery.println("lookupGroups:        ");
	    dumpGroups(lookupGroups, dbgRecovery);
	    dbgRecovery.println("lookupLocators:      ");
	    dumpLocators(lookupLocators, dbgRecovery);
	    Collection regs = regByID.values();
	    Iterator iter = regs.iterator();
	    while (iter.hasNext()) {
	        ServiceRegistration reg = (ServiceRegistration)iter.next();
	        reg.dumpInfo(dbgRecovery);
	    }
	}
	// TODO - recover LookupAttributes
    }


    /**
     * Handler class for the persistent storage facility.
     * <p>
     * At any point during processing in the mailbox, there will exist
     * both a 'snapshot' of the mailbox's state and a set of records
     * detailing each significant change that has occurred to the state
     * since the snapshot was taken. The snapshot information and the
     * incremental change information will be stored in separate files
     * called, respectively, the snapshot file and the log file. Together,
     * these files are used to recover the state of the mailbox after a
     * crash or a network outage (or if the mailbox or its ActivationGroup
     * is un-registered and then re-registered through the Activation Daemon).
     * <p>
     * This class contains the methods that are used to record and recover
     * the snapshot of the mailbox's state; as well as the method used to
     * apply the state changes that were recorded in the log file.
     * <p>
     * When the ReliableLog class is instantiated, a new instance of this
     * class is passed to its constructor so that the methods of this
     * class may be invoked by the methods defined in the ReliableLog.
     * Because this class extends the LogHandler class associated with
     * the ReliableLog class, this class must provide implementations of
     * the abstract methods declared in the LogHandler. Also, some of the
     * methods defined in this class override the methods of the LogHandler
     * in order to customize the handling of snapshot creation and
     * retrieval.
     * <p>
     * Each significant change to the mailbox's state is written to the
     * log file as an individual record (when addLogRecord() is invoked).
     * After the number of records logged exceeds a pre-defined threshold,
     * a snapshot of the state is recorded by invoking -- through the
     * ReliableLog and its LogHandler -- the snapshot() method defined in
     * this class. After the snapshot is taken, the log file is cleared
     * and the incremental log process starts over.
     * <p>
     * The contents of the snapshot file reflect the DATA contained in
     * the fields making up the current state of the mailbox. That data
     * represents many changes -- over time -- to the mailbox's state.
     * On the other hand, each record written to the log file is an object
     * that reflects both the data used and the ACTIONS taken to make one
     * change to the mailbox's state at a particular point in time.
     * <p>
     * The data written to the snapshot file is shown below:
     * <ul>
     * <li> our service ID
     * <li> current configuration parameters
     * <li> contents of the container holding the current registrations
     * <li> null (termination 'marker' for the set of registered services)
     * </ul>
     * The type of state changes that will generate a new record in the log
     * file are:
     * <ul>
     * <li> registration created
     * <li> registration renewed
     * <li> registration cancelled
     * <li> registration enabled
     * <li> registration disabled
     * </ul>
     * During recovery, the state of the mailbox at the time of a crash
     * or outage is re-constructed by first retrieving the 'base' state from
     * the snapshot file; and then modifying that base state according to
     * the records retrieved from the log file. The reconstruction of the
     * base state is achieved by invoking the recover() method defined in
     * this class. The modifications recorded in the log file are then
     * applied to the base state by invoking the applyUpdate() method
     * defined in this class. Both recover() and applyUpdate() are invoked
     * through the ReliableLog and its associated Loghandler.
     * <p>
     * NOTE: The following lines must be added to the mailbox's policy file
     * <pre>
     *     permission java.io.FilePermission "dirname",   "read,write,delete";
     *     permission java.io.FilePermission "dirname/-", "read,write,delete";
     * </pre>
     *     where 'dirname' is the name of the directory path (relative or
     *     absolute) where the snapshot and log file will be maintained.
     */
    private class LocalLogHandler extends LogHandler {
	
	/** Simple constructor */
	public LocalLogHandler() { }

	/* Overrides snapshot() defined in ReliableLog's LogHandler class. */
	public void snapshot(OutputStream out) throws IOException {
	    takeSnapshot(out);
	}

	/* Overrides recover() defined in ReliableLog's LogHandler class. */
	public void recover(InputStream in)
	    throws IOException, ClassNotFoundException
	{
	    recoverSnapshot(in);
	}

	/**
	 * Required method implementing the abstract applyUpdate()
	 * defined in ReliableLog's associated LogHandler class.
	 * <p>
	 * During state recovery, the recover() method defined in the
	 * ReliableLog class is invoked. That method invokes the method
	 * recoverUpdates() which invokes the method readUpdates(). Both
	 * of those methods are defined in ReliableLog. The method
	 * readUpdates() retrieves a record from the log file and then
	 * invokes this method.
	 * <p>
	 * This method invokes the version of the method apply() that
	 * corresponds to the particular type of 'log record' object
	 * that is input as the first argument. The log record object and its
	 * corresponding apply() method are defined in one of the so-called
	 * LogObj classes. Any instance of one the LogObj classes is an
	 * implementation of the LogRecord interface. The particular
	 * implementation that is input to this method is dependent on the
	 * type of record that was originally logged. The apply() method
	 * will then modify the state of the Mailbox in a way dictated
	 * by the type of record that was retrieved.
	 */
	public void applyUpdate(Object logRecObj) {
	    ((LogRecord)logRecObj).apply(MailboxImpl.this);
	}
    }

    /**
     * Snapshot-taking thread. 
     * <p>
     * A snapshot is taken when -- after writing a new record to the 
     * log file -- it is determined that the size of the log file has 
     * exceeded a certain threshold. The code which adds the new record 
     * to the log file and which, in turn, decides that a snapshot
     * must be taken is "wrapped" in a writer mutex. That is, synchronization
     * of processing is achieved in the Mailbox through a "reader/writer"
     * mutex construct. This construct allows only one writer at any one
     * time; but allows an unlimited number of simultaneous readers as
     * long as no writer has locked the mutex. During steady-state, it is
     * anticipated that there will be far more readers in use 
     * than writers. Since the process of
     * taking a snapshot can be time-consuming, if the whole snapshot-taking
     * process occupies that single writer mutex, then a significant number
     * of read requests will be un-necessarily blocked; possibly resulting
     * in an unacceptable degradation in response time. 
     * <p>
     * It is for the above reason that the process of taking a snapshot is
     * performed in a separate thread. The thread waits on the monitor
     * belonging to the snapshotNotifier instance until it is notified
     * (or "signaled") that a snapshot must be taken. The notification
     * is sent by another thread, created by the Mailbox, which
     * determines when the conditions are right for a snapshot. The
     * notification takes the form of an interrupt indicating that the
     * Snapshot monitor is available. Although the interrupt is sent 
     * while the writer mutex is locked, the act of sending the notification
     * is less time-consuming than the act of taking the snapshot itself.
     * When the thread receives a notification, it awakens and requests a
     * lock on the reader mutex (this is all done in the readerWait() method).
     * Because a reader -- not a writer -- mutex is locked, read-only
     * processes still have access to the system state, so lookups can be
     * performed; and the reader mutex prevents changes to the data while
     * the snapshot is in progress.  
     * <p>
     * Note that the current snapshot is guaranteed to complete before the
     * next snapshot request is received. This is because even though
     * the act of taking a snapshot can be viewed as a writer process, 
     * the fact that the next snapshot notification will be wrapped in a
     * writer mutex, combined with the fact that a writer mutex can not
     * be locked while a reader mutex is locked, allows the snapshot to
     * be treated as a reader process.
     */
    private class SnapshotThread extends Thread {

	/** Create a daemon thread */
	public SnapshotThread() {
	    super("SnapshotThread");
	    setDaemon(true);
	    start();
	}

	public void run() {
	    try {
		concurrentObj.readLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }
	    try {
		while (!isInterrupted()) {
		    try {
			concurrentObj.readerWait(snapshotNotifier,
						 Long.MAX_VALUE);
                        try {
                            log.snapshot();
		            logFileSize = 0;
	                } catch (InterruptedIOException e) {
			    return;
	                } catch (Exception e) {
			    if (e instanceof LogException &&
				((LogException)e).detail instanceof
				InterruptedIOException)
				return;
	                    /* XXX
	                     * if taking the snapshot fails for any reason,
                             * then one of the following must be done:
	                     *   -- output the problem to a file and exit
	                     *   -- output the problem to a file and continue
	                     *   -- set an "I have a problem" attribute and
                             *      then send a notification
	                     * this issue will be addressed at a later time
	                     */
			    e.printStackTrace();
		        }
		    } catch (ConcurrentLockException e) {
			return;
		    }
		}
	    } finally {
		concurrentObj.readUnlock();
	    }
	}
    }


    ///////////////////////
    // Activation Methods
    ///////////////////////

    /**
     * If for some reason we can't activate we dump a message to the log 
     * and unregister with activation.
     * @param m message to dump to <code>System.err</code>
     * @param e If non-<code>null</code> this exception is dumped to
     * <code>System.err</code> as well 
     */
    private void alertAndUnregister(String m, Throwable e) {
	System.err.println("Fatal error during activation:" + m);
	if (e != null)
	    e.printStackTrace();
	
	Exception failed = null;

	System.err.println("Attempting to unregister with activation");
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (ActivationException ee) {
	    failed = ee;
	} catch (RemoteException ee) {
	    failed = ee;
	}

	if (failed != null) {
	    System.err.println("Failed to unregister with activation");
	    failed.printStackTrace();
	} else {
	    System.err.println("Succeeded in unregistering");
	}
    }


    /**
     * Activation constructor
     *
     * @param activationID activation ID passed in by the activation daemon.
     * @param data state data needed to re-activate a Norm server
     */
    public MailboxImpl(ActivationID activationID, MarshalledObject data) 
	throws IOException, ClassNotFoundException
    {
	boolean runtimeThrown = false;
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl is being activated");
	}
	this.activationID = activationID;
	Remote ourStub = null;

	try {
	    try {
	        if (dbgInit != null) {
	            dbgInit.println("MailboxImpl is being exported");
	        }
		ourStub = Activatable.exportObject(this, activationID, 0);
	    } catch (RemoteException e) {
		 alertAndUnregister("Failure exporting object", e);
		 throw e;
	    }

	    // Get Log Dir Name
	    try {
                logDirname = (String)data.get();
	        if (dbgInit != null) {
	            dbgInit.println("Obtained persistence path: " + logDirname);
	        }
	    } catch (IOException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;
	    } catch (ClassNotFoundException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;	  
	    } catch (ClassCastException e) {
		alertAndUnregister("Failure recoving activation data", e);
		runtimeThrown = true;
		throw e;
	    }

	    init();

	} catch (RuntimeException e) {
	    if (runtimeThrown == false) { 
		alertAndUnregister("Unexpected exception during activation",
				   e);
	    }

	    throw e;
	} catch (Error e) {
	    alertAndUnregister("Unexpected exception during activation", e);
	    throw e;
	}	  
    }


    // Setup - code that registers a MailboxImpl with activation

    /**
     * Create an activation group suitable for a Mercury server.
     * @param desc Classpath, codebase, security policy file, VM, VM args,
     *		   and VM properties for activation group.
     */
    private static ActivationGroupID createGroup(BasicServiceDesc desc)
	throws ActivationException, RemoteException
    {
	Properties props = desc.getProperties();

	if (props == null)
	    props = new Properties();
	else
	    props = (Properties)props.clone();

	props.put("java.security.policy", desc.getPolicyPath());
	props.put("java.rmi.server.codebase", desc.getCodebase());

	String argv[] = desc.getJVMOptions();
	if (argv == null) {
	    argv = new String[]{"-cp", desc.getClassPath()};
	} else {
	    String[] nargv = new String[argv.length + 2];
	    System.arraycopy(argv, 0, nargv, 0, argv.length);
	    nargv[argv.length] = "-cp";
	    nargv[argv.length + 1] = desc.getClassPath();
	    argv = nargv;
	}

        if (dbgInit != null) {
            dbgInit.println("Creating activation group");
            dbgInit.print("Command: " + desc.getJVMPath());
            for (int i = 0; i < argv.length; i++) {
                dbgInit.print(" " + argv[i]);
            }
            dbgInit.println();
            dbgInit.println("Props:");
            props.list(dbgInit);
	}

	CommandEnvironment cmd = 
	    new CommandEnvironment(desc.getJVMPath(), argv);
	return ActivationGroup.getSystem().registerGroup(
					 new ActivationGroupDesc(props, cmd));
    }

    /**
     * Create an activatable and persistent EventMailbox service in
     * its own activation group.
     * <p>
     * The activation group will be configured according to the passed
     * <code>PersistentServiceDesc</code> object.  The service will
     * persist its data in the directory returned by
     * <code>desc.getPersistentsPath()</code> method. This directory
     * should not exist.
     * <p>
     * Note that when the service is later destroyed, it will destroy
     * its activation group, regardless of whether or not there are
     * other activatable objects still registered in the same group.
     * It will also delete its persistence directory.
     * <p>
     * The server will not bind itself into a directory service until
     * it has been configured through its administration interface.
     * The service's administration interface can be obtained by
     * casting it to a <code>net.jini.admin.Administrable</code> and
     * calling <code>getAdmin()</code>.	 The returned value will be a
     * reference to the service's admin interface.
     * <p>
     *
     * @see net.jini.admin.Administrable
     */
    public static EventMailbox create(BasicServiceDesc desc) 
	throws RemoteException, IOException, ActivationException
    {
	ActivationGroupID actID = createGroup(desc);

	try {
	    return create(actID, desc.getPersistentsPath());
	} catch (Exception e) {
	    try {
		ActivationGroup.getSystem().unregisterGroup(actID);
	    } catch (Exception ee) {
	    }
	    if (e instanceof ActivationException)
		throw (ActivationException)e;
	    else if (e instanceof IOException)
		throw (IOException)e;
	    throw (RuntimeException)e;
	}
    }


    /**
     * Create an activatable and persistent Mailbox service in
     * the passed activation group.  The service will persist its data
     * in the passed directory.	 This directory should not exist.
     * <p>
     * The passed activation group needs to have the necessary Mercury
     * class files in its classpath and the necessary Mercury client
     * classes files in its codebase.
     * <p>
     * Note that when the service is later destroyed, it will destroy
     * its activation group, regardless of whether or not there are
     * other activatable objects still registered in the same group.
     * It will also delete its persistence directory.
     * <p>
     * The server will not bind itself into a directory service until
     * configure though its administration interface.  The service's
     * administration interface can be obtained by casting it to a
     * <code>net.jini.admin.Administrable</code> and calling
     * <code>getAdmin()</code>.	 The returned value will be a
     * reference to the service's admin interface.
     * <p>
     *
     * @see net.jini.admin.Administrable
     */
    private static EventMailbox create(ActivationGroupID gid,
				       String logDirName) 
	throws IOException, ActivationException
    {
        if (dbgInit != null) {
            dbgInit.println("Creating Activatable object");
	}

	// Tell the activation system about us
	ActivationDesc actDesc =
	    new ActivationDesc(gid, MailboxImpl.class.getName(), null,
			       new MarshalledObject(logDirName), true);

	// Register with activation and finish setting up the binder
        MailboxBackEnd mb = (MailboxBackEnd)Activatable.register(actDesc);

	return new MailboxProxy(mb, mb.getProxyID());
    }



    /**
     * Start up a new event mailbox service. See the Mercury release notes for
     * more details on how system properties are interpreted.
     * The command line parameters are:
     * <pre>
     * codebase policyfile dbdir [group,...] [javaprog] [javaoptions]
     * </pre>
     *
     * @param codebase	  required argument containing the location with which
     *			  the mailbox service's remote classes will be
     *			  annotated; indicating the location from which those
     *			  classes should be downloaded. This argument is
     *			  passed to the activation system's group descriptor
     *			  as the java.rmi.server.codebase property so that this
     *			  property will be set in the VM created during
     *			  activation of the lookup service. This codebase
     *			  string must be a sequence of properly formatted URLs;
     *			  where each URL must end with ".jar", ".zip", or "/".
     *
     * @param policyfile  required argument containing the name (with optional
     *			  path) of the file containing the security policies
     *			  for the mailbox service. This argument is passed to
     *			  the activation system's group descriptor as the
     *			  java.security.policy property so that this
     *			  property will be set in the VM created during
     *			  activation of the lookup service.
     *
     * @param dbdir	  required argument containing the name (with optional
     *			  path) of the directory in which the mailbox service's
     *			  state will be recorded for persistence
     *
     * @param group(s)	  optional argument(s) containing the list of multicast
     *			  discovery groups that the mailbox should attempt
     *			  to join.
     *			  A group name equal to "none" will result in this
     *			  renewal service not using multicast discovery; a group
     *			  name of "public" will result in this renewal service 
     *			  registering with lookups belonging to the public 
     *			  group.  The default is "public".
     *
     * @param javaprog	  optional argument containing the name of a file
     *			  - including the full path - that will be used as
     *			  the 'cmdpath' argument in the creation of the
     *			  CommandEnvironment object; which specifies startup
     *			  options for the activation group. This argument can
     *			  be used to override default system properties and
     *			  to specify implementation-defined options for the
     *			  activation group. Upon activation of the renewal
     *			  service, the activation system will start the
     *			  renewal service by executing the file represented
     *			  by this argument. Although this file can take any
     *			  form - a script, a java executable, etc. - typically
     *			  it is simply a different version of the java
     *			  command than that used to execute the main()
     *			  method of NormServerImpl. In this way, one can
     *			  activate the renewal service with a different
     *			  version of java than the current version.
     *
     * @param javaoptions optional list of argument(s), each beginning with
     *			  a '-', that will be passed into the VM created
     *			  when the renewal service is activated. These
     *			  arguments can be any arguments understood by
     *			  the java command (or the javaprog if the java
     *			  command was overridden). These arguments can be
     *			  used to set system properties in the renewal service.
     *			  For example, a very useful property that can be set
     *			  using these arguments is the property that turns
     *			  off the just-in-time compiler (JIT). To turn off
     *			  the JIT, simply input the string "-Djava.compiler="
     *			  as one of these arguments.
     */
    public static void main(String[] args) {
        if (dbgInit != null) {
            dbgInit.println("Parsing command line arguments");
	}
	//TODO -- use resource strings for message output

	final BasicServiceDesc desc = ParseUtils.persistentParse(args);
	if (desc == null) {
            if (dbgInit != null) {
                dbgInit.println("Parsing command line arguments ... failed");
	    }
	    System.exit(-1);
	}

	System.setSecurityManager(new RMISecurityManager());
	try {
            if (dbgInit != null) {
                dbgInit.println("Creating activatable event mailbox service");
	    }
	    final EventMailbox srv = create(desc);

            if (dbgInit != null) {
                dbgInit.println("Configuring groups via JoinAdmin");
	    }
	    final JoinAdmin admin = (JoinAdmin)((Administrable)srv).getAdmin();
	    admin.setLookupGroups(desc.getGroups());
	} catch (Throwable e) {
	    System.err.println("MailboxImpl: main: unable to start " +
			       "EventMailbox service");
	    e.printStackTrace();
	    System.exit(-1);
	}
    }


    private void dumpGroups(String[] grps, PrintWriter pw) {
	if (grps == null)
	    pw.println("<ALL_GROUPS>");
	else if (grps.length == 0)
	    pw.println("<NO_GROUPS>");
	else {
	    for (int i = 0; i < grps.length; i++) {
		 pw.println("\t" + grps[i]);
	    }
	}
    }
    private void dumpLocators(LookupLocator[] locs, PrintWriter pw) {
	if (locs == null)
	    pw.println("<null>");
	else if (locs.length == 0)
	    pw.println("<EMPTY>");
	else {
	    for (int i = 0; i < locs.length; i++) {
		 pw.println("\t" + locs[i]);
	    }
	}
    }

    private void dumpEvent(RemoteEvent re, PrintWriter pw) {
	pw.println("RemoteEvent:");
	pw.println("\tID:   " + re.getID());
	pw.println("\tSeq#: " + re.getSequenceNumber());
    }


}
