/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mercury;

import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.proxy.UUID; 

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.rmi.MarshalledObject;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import net.jini.core.event.RemoteEventListener;

class ServiceRegistration implements LeasedResource, Comparable, Serializable {

    // TODO - generate serial version id

    /** Unique identifier object */
    private final UUID cookie;

    /** The current expiration for this registration */
    private long expiration = 0;

    /** The client provided notification target. */
    // This field is transient in order to allow readObject/writeObject
    // to un/wrap the the listener to/from a MarshalledObject. 
    // This is done to prevent the corruption of the input stream.
    // For example, if the associated codebase of the listener object
    // is no longer valid, then the stream is corrupted since the object
    // can't be reconstructed.  Wrapping the reference in a 
    // MarshalledObject always allows us to read from the stream.
    // We might not be able to reconstruct the listener field, but
    // the rest of the object will be fine.
    private transient RemoteEventListener eventTarget = null; 

    /** Event log iterator. */
    private EventLogIterator eventIterator = null;

    /** 
     * Map of collected eventID's that resulted in an 
     * UnknownEventException for the current <code>
     * eventTarget</code>. This <code>Map</code> is checked
     * upon each mailbox notification for this registration. 
     * If the provided event has the same event ID and source
     * as one in this <code>Map</code> then an UnknownEventException
     * is propagated back to the sender and the event is not logged.
     */
    private Map unknownEvents = new HashMap();

    /** Simple constructor */
    public ServiceRegistration(UUID cookie, EventLogIterator eventIterator) {
        this.cookie = cookie;
        this.eventIterator = eventIterator;
    }

    // inherit javadoc from parent
    public void setExpiration(long newExpiration) {
	 expiration = newExpiration;
    }

    // inherit javadoc from parent
    public long getExpiration() { 
	return expiration;
    }

    // inherit javadoc from parent
    public Object getCookie() {
	return cookie;
    }

    /** 
     * Return the identity map of eventsIDs that caused an 
     * UnknownEventException to be generated for the current
     * notification target. A Map is used instead of a generic
     * Collection since the space/time tradeoff for searches
     * seems worth it.
     */
    public Map getUnknownEvents() {
	return unknownEvents;
    }

    /** Return reference to the client supplied notification target */
    public RemoteEventListener getEventTarget() {
        return eventTarget;
    }

    /** Set reference to the client supplied notification target */
    public void setEventTarget(RemoteEventListener l) {
        eventTarget = l;
    }

    /** Set reference to the client supplied notification target */
    public EventLogIterator iterator() {
        return eventIterator; 
    }

    /**
     * Primary sort by leaseExpiration, secondary by leaseID.  The
     * secondary sort is immaterial, except to ensure a total order
     * (required by TreeMap).
     */
    public int compareTo(Object obj) {
        ServiceRegistration reg = (ServiceRegistration)obj;
        if (this == reg)
            return 0;
        if ( expiration < reg.expiration ||
             (expiration == reg.expiration &&
             (   (cookie.getMostSignificantBits() < 
		     reg.cookie.getMostSignificantBits())
              && (cookie.getLeastSignificantBits() < 
		     reg.cookie.getLeastSignificantBits()) ) ))
            return -1;
        return 1;
    }

    public String toString () {
        return getClass().getName() + ":" + cookie.toString();
    }

    void dumpInfo(PrintWriter pw) {
        pw.println("\t" + this.toString());
        pw.println("\tExpires at :" + new Date(expiration));
        pw.println("\tTarget is  :" + eventTarget);
        pw.println("\tEvents     :" + eventIterator.size());
        pw.println("\tUnknowns   :" + unknownEvents.size());
    }

    /**
     * @serialData RemoteEventListener as a MarshalledObject
     */
    private void writeObject(ObjectOutputStream stream)
        throws IOException
    {
        stream.defaultWriteObject();
        stream.writeObject(new MarshalledObject(eventTarget));
    }
    
    /**
     * Unmarshals the event listener.
     */
    private void readObject(ObjectInputStream stream)
        throws IOException, ClassNotFoundException
    {
        stream.defaultReadObject();
        MarshalledObject mo = (MarshalledObject)stream.readObject();
        try {
    	    eventTarget = (RemoteEventListener)mo.get();
        } catch (Throwable e) {
    	    if (e instanceof Error &&
    	        !(e instanceof LinkageError ||
    	          e instanceof OutOfMemoryError ||
    	          e instanceof StackOverflowError))
    	        throw (Error)e;
        }
    }
    
}
