/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.mercury;

import com.sun.jini.debug.Debug;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.rmi.server.RemoteObject;
import java.rmi.RMISecurityManager;
import java.text.MessageFormat;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;
import net.jini.event.EventMailbox;

/**
 * Public method and command line interface for creating a transient instance
 * of the mercury implementation of the event mailbox service.
 */
public class TransientMailbox {

    /** At lease this number of args must be input on the command line */
    private static final int MIN_N_ARGS = 1;
    /** No more than this number of args can be input on the command line */
    private static final int MAX_N_ARGS = 2;
    /** The set of resources used to enable localization */
    private static ResourceBundle resources = null;

    /** Debug utility class */
    private static final Debug debug = new Debug(MailboxBackEnd.DEBUG_PROP);

    /** PrintWriter handle for initialization related debugging messages */
    private static final PrintWriter dbgInit = debug.getWriter("init");

    /**
     * Command line interface for creating a non-activatable (transient)
     * instance of the event mailbox service. The command line parameters are:
     * <p>
     * <pre>
     * logdir [group,...] 
     * </pre>
     *
     * @param logdir      required argument containing the name (with optional
     *                    path) of the directory in which the event mailbox
     *                    service's state will be recorded for persistence
     * @param group       optional argument containing the names of the
     *                    groups that the event mailbox should discover 
     *                    and join. A group name equal to "none" will
     *                    result in this lookup discovery service joining
     *                    no groups; a group name of "public" will result in
     *                    this lookup discovery service joining all lookup
     *                    service(s) it can find that belong to the unamed
     *                    public (""). The default is "public".
     */
    public static void main(String[] args) {
	if (args.length < MIN_N_ARGS || args.length > MAX_N_ARGS) {
	    print("mercury.transient.usage", null);
	    print("mercury.groups.usage", null);
	    System.exit(1);
	}
        /* Verify that the log directory does not already exist */
	if (new File(args[0]).exists()) {
	    print("mercury.logdir.exist", args[0]);
	    System.exit(1);
	}

        /* Parse the remaining argument for groups. */
	String noneGroup = getString("mercury.groups.none");
	String publicGroup = getString("mercury.groups.public");
	String allGroup = getString("mercury.groups.all"); 

        /* Initialize the default value for groups */
	String[] groups = {""}; // public

        /* Initialize the counter index */
	int i = MIN_N_ARGS;

        /* Handle the groups argument. The code below expects the groups to
         * be input as a single, comma-separated string of names (no spaces).
         */

        if (i < args.length) {
            StringTokenizer st = new StringTokenizer(args[i],",");
            String group = null;

            int n = st.countTokens();
            if (n == 1) { 
                group = st.nextToken();
                if (group.equals(noneGroup)) {
                   groups = new String[0];
                }
                else if (group.equals(publicGroup)) {
                   // Do nothing ... already the default.
                }
                else if (group.equals(allGroup)) {
                   groups = null;
                }
                else {
                   groups[0] = group; 
	        }
            }
            else if (n > 1) {
                groups = new String[n];

                for (int k=0; st.hasMoreTokens(); k++) {
                    group = st.nextToken();
                    if (group.equals(publicGroup)) {
                        group = "";
                    }
                    else if (group.equals(noneGroup)) {
                        groups = new String[0];
	                print("mercury.groups.none.usage", args[0]);
	                System.exit(1);
		    }
                    else if (group.equals(allGroup)) {
                        groups = null;
	                print("mercury.groups.all.usage", args[0]);
	                System.exit(1);
		    }
                    groups[k] = group;
                }
            }
        }//end group parsing

        /* ----------------------------------------------------------------- */
        /* An RMISecurityManager must be set!
         *           
         * If a Security Manager is not set, exceptions can occur during
         * discovery (if needed classes are downloaded), but those exceptions
         * are "swallowed" by LookupDiscovery; so no discovery will occur,
         * and it will appear as if nothing is wrong.
         *           
         * Also if no SecurityManager is set, RMI will prevent the loading of
         * classes not found in the classpath; so a lookup will return null
         * instead of the expected references to service items or attributes.
         */
	System.setSecurityManager(new RMISecurityManager());
        /* ----------------------------------------------------------------- */

        // output debugging info if asked
        if (dbgInit != null) {
            System.out.println("Starting EventMailbox Service with:");
            System.out.println("\tLog dir: " + args[0]);
            System.out.print("\tGroups [" + groups.length + "]: ");
            for (int k = 0; k < groups.length; k++) {
                System.out.print(groups[k] + " ");
	    }
            System.out.println();
	}

        /* Argument parsing complete. Now create the service */
	try {
	    EventMailbox emb = create(args[0]);

	    // TODO - use reference to obtain group admin interface
	    // and configure the service to join the appropriate groups/
	    final JoinAdmin admin = (JoinAdmin)((Administrable)emb).getAdmin();
	    admin.setLookupGroups(groups);
	} catch (Exception e) {
	    e.printStackTrace(System.out);
	}
    }//end main

    /**
     * This method creates a non-activatable (transient) instance of the 
     * Mercury implementation of the event mailbox service and returns a 
     * proxy that provides client-side access to that instance.
     * <p>
     * The service is initially a member of no groups.
     * <p>
     * Persistent state is maintained even though the service
     * is transient. This is particularly convenient for debugging purposes.
     *
     * @param logDirname directory in which the state of the lookup discovery
     *                   service will be recorded for persistence
     * 
     * @return an instance of the EventMailboxService interface which is
     *         a proxy that provides client-side access to the created
     *         event mailbox service
     * 
     * @throws java.io.IOException 
     */
    public static EventMailbox create(String logDirname) throws IOException {
        MailboxImpl mb = new MailboxImpl(logDirname);
        return new MailboxProxy((MailboxBackEnd)RemoteObject.toStub(mb), 
				mb.getProxyID());
    }

    /**
     * Retrieves a string resource corresponding to a particular key from the
     * resource bundle. If the resource bundle does not yet exist, this
     * method will first load and instantiate the resource bundle.
     *
     * @param key a non-null case-sensitive string containing the name of 
     *            the resource whose string is to be retrieved.
     *
     * @return a non-null String containing the value of the desired resource
     */
    private static String getString(String key) {
	if (resources == null) {
	    try {
		resources = ResourceBundle.getBundle
                                    ("com.sun.jini.mercury.resources.mercury");
	    } catch (MissingResourceException e) {
		e.printStackTrace();
	    }
	}
	try {
	    return resources.getString(key);
	} catch (MissingResourceException e) {
	    return null;
	}
    }//end getString

    /**
     * Displays the resource string corresponding to a particular key, but 
     * with a given value substituted in the appropriate location in the
     * string.
     *
     * @param key a non-null case-sensitive string containing the name of 
     *            the resource whose string is to be printed.
     * @param val the value to substitute in the resource string before
     *            displaying 
     */
    private static void print(String key, String val) {
	String fmt = getString(key);
	if (fmt == null)
	    fmt = "no text found: \"" + key + "\" {0}";
	System.out.println(MessageFormat.format(fmt, new String[]{val}));
    }//end print
}
