/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.util.List;

import java.io.IOException;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import java.rmi.UnmarshalException;
import java.util.Map;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMap;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.lease.RenewalFailureEvent;


import net.jini.core.event.RemoteEvent;

import com.sun.jini.proxy.UUID;

import com.sun.jini.norm.event.EventFactory;

/**
 * Class that wraps client Leases.  Provides hooks for synchronization 
 * and data associated with each client lease while allowing us to 
 * use <code>LeaseRenewalManager</code>
 */
class ClientLeaseWrapper implements Lease, Serializable {
    /**
     * First exception thrown by the last contagious sequence of renewal
     * failures.  <code>null</code> if the renew has not been called yet or
     * the last renewal call seceded.
     * @serial
     */
    private Throwable firstFailure = null;

    /**
     * Client lease in marshaled form.
     * @serial
     */
    private MarshalledObject marshledClientLease;

    /**
     * Most current expiration time of client Lease that we know of	
     * @serial
     */
    private long clientLeaseExpiration;

    /**
     * Sequence number that uniquely identifies this wrapper
     * @serial
     */
    private long UID;

    /**
     * UUID of set we belong to
     * @serial
     */
    private UUID setUUID;

    /**
     * Transient copy of unpacked lease, <code>null</code> if we could not
     * unpack lease.
     */
    private transient Lease clientLease;

    /** 
     * Flag used to check if lease has been persisted since last
     * renewal, <code>true</code> means it has not been
     */
    private transient boolean renewed = false;

    /**
     * Reference to list of leases that have been renewed, but not
     * persisted
     */
    private transient List renewedList;

    /**
     * Simple constructor
     * @param clientLease Lease from client that is to be renewed.
     * @param UID ID number for this wrapper unique for all wrappers in a given
     *		  server
     * @param renewedList List that wrapper should go on after renewing their
     *			  client lease is renewed
     * @param setID	  <code>UUID</code> of renewal set this lease is in
     */
    ClientLeaseWrapper(Lease clientLease, long UID, List renewedList, 
		       UUID setID) throws IOException
    {
	this.renewedList = renewedList;
	this.UID = UID;
	this.clientLease = clientLease;

	setUUID = setID;

	clientLeaseExpiration = clientLease.getExpiration();
	clientLease.setSerialFormat(Lease.ABSOLUTE);
	marshledClientLease = new MarshalledObject(clientLease);
    }

    /**
     * Atomicly test and set the renewed flag.
     * @return the state of the renewed flag before it was set
     */
    private synchronized boolean testAndSetRenewed() {
	final boolean rslt = renewed;
	renewed = true;
	return renewed;
    }

    /**
     * Atomicly clear the renewed flag.
     */
    synchronized void clearRenewed() {
	renewed = false;
    }

    /**
     * Get a reference to the client lease unpacking it if necessary, if 
     * it can't be unpacked, return null.
     */
    Lease getClientLease() {
	if (clientLease != null)
	    return clientLease;

	try {
	    final Object upcl = marshledClientLease.get();
	    clientLease = (Lease)upcl;
	    return clientLease;
	} catch (IOException e) {
	    // fall through to return null
	} catch (ClassNotFoundException e) {
	    // fall through to return null
	}

	return null;
    }


    // Inherit java doc from super type
    public boolean equals(Object that) {
	if (that instanceof ClientLeaseWrapper)
	    return UID == ((ClientLeaseWrapper)that).UID;
	else
	    return false;
    }

    // Inherit java doc from super type
    public int hashCode() {
	return (int)UID;
    }

    // Inherit java doc from super type
    public long getExpiration() {
	final Lease cl = getClientLease();
	if (cl == null) {
	    synchronized (this) {		
		return clientLeaseExpiration;
	    }
	} else {
	    return cl.getExpiration();
	}
    }

    /**
     * Always throws <code>UnsupportedOperationException</code> since
     * a lease renewal service should never cancel a client lease.
     */
    public void cancel() {
	throw new UnsupportedOperationException("ClientLeaseWrapper.cancel:" + 
	     "LRS should not being canceling client leases ");
    }

    /**
     * Sets firstFailure if it is null, note this is only called from
     * <code>renew</code> so additional synchronization should not be necessary.
     */
    private void setFirstFailure(Throwable t) {
	if (firstFailure == null) { 
	    firstFailure = t;
		if (testAndSetRenewed()) 
		    renewedList.add(this);
	}	
    }

    // Inherit java doc from super type
    // This method assumes that renew is only called from a single thread
    public void renew(long duration)
	throws LeaseDeniedException, UnknownLeaseException, RemoteException 
    {
	if (clientLease == null) {
	    // try to unmarshall client lease
	    try {
		final Object upcl = marshledClientLease.get();
		clientLease = (Lease)upcl;
	    } catch (IOException e) {
		UnmarshalException t = new UnmarshalException(
		    "ClientLeaseWrapper.renew:Could not unmarshal client " +
		    "lease for renewal", e);
		setFirstFailure(t);
		throw t;
	    } catch (ClassNotFoundException e) {
		UnmarshalException t = new UnmarshalException(
		    "ClientLeaseWrapper.renew:Could not unmarshal client " +
		    "lease for renewal", e);
		setFirstFailure(t);
		throw t;
	    }
	}

	// If we get here clientLease must be non-null

	// $$$ do we want to ensure that the lease on the set is
	// current here or just rely on the expiration thread.
	try {
	    clientLease.renew(duration);
	} catch (LeaseDeniedException e) {
	    setFirstFailure(e);
	    throw (LeaseDeniedException)e;
	} catch (UnknownLeaseException e) {
	    setFirstFailure(e);
	    throw (UnknownLeaseException)e;
	} catch (RemoteException e) {
	    setFirstFailure(e);
	    throw (RemoteException)e;
	} catch (RuntimeException e) {
	    setFirstFailure(e);
	    throw (RuntimeException) e;
	} catch (Error e) {
	    setFirstFailure(e);
	    throw (Error)e;
	}

	// If we get here we must have successfully renewed the lease
	firstFailure = null;

	// Update the marshaled version of the lease 
	// $$$ We could only do this in writeObject but then we would have
	// to make sure it had changed since the last serialization, the 
	// renewed flag almost does this but not quite (maybe it should 
	// and we don't have it quite right...maybe writeObject should
	// clear the  flag instead of explicit clearing it after writing
	// the clw out?).  Not clear doing it here is significantly less
	// efficient.
	try {
	    marshledClientLease = new MarshalledObject(clientLease);
	} catch (IOException e) {
	    // $$$ Besides printing a message drop this on the floor,
	    // exception in some pathological cases we are extremely
	    // unlikely to get this exception, further more this is really
	    // just a logging error, which our current policy is to drop.
	    // In the abstract it would be better to do this in writeObject
	    // since this really is an error while trying to persist the 
	    // state of this object, however, that solution would require
	    // more synchronization.  Lastly this method may go away if
	    // we decide to stop using the LRM.
	    System.err.println("IOException while marshalling client lease " +
			       "after renewal");
	    e.printStackTrace();
	}

	// synchronize in case long assignment is not atomic
	synchronized (this) {
	    clientLeaseExpiration = clientLease.getExpiration();
	}

	// If necessary notify rest of system that this lease has to 
	// be persisted
	if (testAndSetRenewed()) {
	    synchronized (renewedList) {
		renewedList.add(this);
		renewedList.notifyAll();
	    }
	}	
    }

    /**
     * Always throws <code>UnsupportedOperationException</code>.  The
     * <code>LeaseRenewalManager</code> should never call this method and 
     * norm always serializes the wrapper with absolute times.
     */
    public void setSerialFormat(int format) {
	throw new UnsupportedOperationException(
	     "ClientLeaseWrapper.setSerialFormat:" + 
	     "LRS should not be setting serial format through the wrapper");
    }


    /**
     * Always throws <code>UnsupportedOperationException</code>.  The
     * <code>LeaseRenewalManager</code> should never call this method and 
     * norm always serializes the wrapper with absolute times.
     */
    public int getSerialFormat() {
	throw new UnsupportedOperationException(
	     "ClientLeaseWrapper.setSerialFormat:" + 
	     "LRS should not be setting serial format through the wrapper");
    }

    // Inherit java doc from super type
    public LeaseMap createLeaseMap(long duration) {
	// $$$ For now just return null (since canBatch always returns false
	return null;
    }

    // Inherit java doc from super type
    public boolean canBatch(Lease lease) {
	// $$$ For now just return false;
	return false;
    }

    /**
     * Override <code>readObject</code> to try an unpack client lease
     * @throws IOException if @see java.io.ObjectInputStream#defaultReadObject 
     * does
     * @throws ClassNotFoundException if 
     * @see java.io.ObjectInputStream#defaultReadObject does
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException
    {
	// fill in the object from the stream 
	in.defaultReadObject();

	// Try to get the client lease unpacked
	getClientLease();
    }

    /**
     * After recovering a lease wrapper call this method before using
     * any other of the wrappers methods and before allocating any new
     * wrappers can not recover itself.
     * @param renewedList List that wrapper should go on after renewing their
     *			  client lease is renewed
     * @param generator	  ID generator being used to generate IDs for 
     *			  client lease wrappers */
    void recoverTransient(List renewedList, UIDGenerator generator) {
	this.renewedList = renewedList;
	generator.inUse(UID);
    }

    /**
     * Return true if the unlying client lease has not yet been deserialized
     */
    boolean isDeformed() {
	return getClientLease() == null;
    }

    /**
     * Return true if this lease is in the specified set
     */
    boolean isInSet(UUID setUUID) {
	return setUUID == this.setUUID;
    }
    
    /**
     * Given a table mapping <code>UUID</code>s to
     * <code>LeaseSet</code> objects return the set associated with
     * this wrapper.
     */
    LeaseSet getLeaseSet(Map setTable) {
	return (LeaseSet)setTable.get(setUUID);
    }

    /**
     * Return an <code>EventFactory</code> that will create an
     * appropriate <code>RenewalFailureEvent</code> for the client
     * lease associated with this wrapper.  This method assumes
     * that no one else will be setting the serialization format
     * of the lease during the course of the call.
     * @param source Source object for the event
     * @throws IOException if the client lease could not be pre
     * marshalled.
     */
    EventFactory newFailureFactory(Object source) throws IOException {
	// If we are 
	Lease l = getClientLease();
	MarshalledObject ml = null;

	if (l == null) {
	    // We are still deformed, the best we can do is use the 
	    // pre-marshalled lease event though it has the wrong
	    // serial format
	    // $$$ is this the right thing to do?
	    ml = marshledClientLease;
	} else {
	    l.setSerialFormat(Lease.DURATION);
	    ml = new MarshalledObject(l);
	}

	return new FailureFactory(source, ml, firstFailure); 
    }

    /**
     * Nested top-level implementation of <code>EventFactory</code> that
     * generates  <code>RenewalFailureEvent</code> events
     */
    // $$$ We could have this object share state with wrapper, but weird
    // things could happen if the wrapper then changes -- right now
    // the wrapper won't change after this method is called but in the 
    // future who is to say.
    private class FailureFactory implements EventFactory {
	/** Source for event */
	private Object source;
	
	/** Client lease that could not be renewed in marshalled form */
	private MarshalledObject marshalledLease;

	/** 
	 * Throwable (if any) that was thrown when we tried to renew the lease
	 */
	private Throwable throwable;

	/**
	 * Simple constructor
	 * @param source Event source
	 * @param marshalledLease Client lease that could not be renewed
	 *        in marshalled form.
	 * @param throwable Exception (if any) that was thrown when 
	 *        the lease could not be renewed.
	 */
	private FailureFactory(Object source,
			       MarshalledObject marshalledLease,
			       Throwable throwable)
	{
	    this.source = source;
	    this.marshalledLease = marshalledLease;
	    this.throwable = throwable;
	}

	// Inherit java doc from super type
	public RemoteEvent createEvent(long             eventID, 
				       long             seqNum, 
				       MarshalledObject handback) 
	{
	    return new RenewalFailureEvent(source, eventID, seqNum,
	        handback, marshalledLease, throwable);
	}

    }
}

