/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import net.jini.core.lease.Lease;

/**
 * <code>DeformedLeaseList</code> provides a list for keeping track of
 * of client lease that could not be fully recovered during log
 * recovery.  Unless otherwise note all the methods in the class are
 * thread safe.
 */
class DeformedLeaseList {
    /** 
     * A list of all the leases we are tracking.  This is a list of weak
     * references to client lease wrapper objects
     */
    private Set leases = null;

    /**
     * Add a wrapped client lease to the list of deformed client leases.
     * This method is not thread safe.
     * @param clw A deformed client lease
     */
    void add(ClientLeaseWrapper clw) {
	if (leases == null)
	    leases =  new java.util.HashSet();
	leases.add(clw);
    }

    /**

     * Query the list to see if the pasted client lease is in this
     * list of deformed leases.  If it is returned the existing client
     * lease wrapper for this client lease.  During the query any
     * client leases wrapper (including the target if found) which are
     * discovered to no longer be deformed will be inserted into the
     * passed table using their associated client lease as a key and removed
     * from the list of deformed leases.
     * @param cl The client lease which may be referenced by a deformed 
     *           client lease
     * @param table A table mapping client leases to client lease wrappers.
     *           Any client lease wrappers encountered during the 
     *           query that are no longer deformed will be placed in this table.
     *           It is assumed that no other thread is trying to access
     *           this table
     * @return If it could be found the client lease wrapper that goes with
     *         <code>cl</code>.  <code>null</code> otherwise.
     */
    ClientLeaseWrapper query(Lease cl, Map table) {
	// Quick check so we are not always blocking on this
	if (leases == null) 
	    return null;

	synchronized (this) {
	    // Recheck now that we have the lock
	    if (leases == null) 
		return null;
	    
	    try {
		ClientLeaseWrapper clw;  // our result;

		for (Iterator i = leases.iterator(); i.hasNext(); ) {
		    clw = (ClientLeaseWrapper)i.next();
		    final Lease ccl = clw.getClientLease();
		    
		    if (ccl != null) {
			// We've got a live one!, remove it from the
			// deformed list and place it in the table
			table.put(ccl, clw);
			i.remove();
		    
			// Is it the one we are looking for?
			if (ccl.equals(cl)) {
			    return clw;
			}
		    } 
		}
	    } finally {
		// Can we GC the leases set?
		if (leases.isEmpty()) 
		    leases = null;
	    }

	    return null;
	}
    }
}

