/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.lang.reflect.Field;
import java.util.Map;
import java.util.Set;

import java.io.Serializable;
import java.io.IOException;
import java.rmi.MarshalledObject;

import net.jini.core.event.RemoteEvent;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;

import net.jini.lease.ExpirationWarningEvent;

import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.proxy.UUID;

import com.sun.jini.norm.event.EventType;
import com.sun.jini.norm.event.EventTypeGenerator;
import com.sun.jini.norm.event.EventFactory;
import com.sun.jini.norm.event.SendMonitor;

/**
 * Norm's internal representation of LeaseRenewalSets.  Unless otherwise
 * noted the methods of this class assume synchronization being handled
 * by the caller.
 */
class LeaseSet implements Serializable, LeasedResource {
    /**
     * Expiration time of the set
     * @serial
     */
    private long expiration;
  
    /**
     * ID that uniquely identifies this set
     * @serial
     */
    private UUID ID;

    /**
     * Map from wrapped client leases to membership Expiration
     * @serial
     */ 
    private Map leases = new java.util.HashMap();

    /**
     * <code>SetProxy</code> that corresponds to this set.
     * @serial
     */ 
    // $$$ Do we really want this to be serializable, persisting this
    // to the stream is convent, but a bit wasteful.  Might want to
    // revisit this as part of figuring out how to update set lease expiration
    private SetProxy setProxy;

    /**
     * Time before <code>expiration</code> to send expiration warning events
     */
    private long minWarning = NormServer.NO_LISTENER;

    /**
     * The event type for expiration warning events
     * @serial
     */
    private EventType warningEventType;

    /**
     * <code>SendMonitor</code> for expiration warning events
     */
    private transient SendMonitor warningMontior;

    /**
     * The event type for failure events
     * @serial
     */
    private EventType failureEventType;

    /**
     * <code>SendMonitor</code> for renewal failure events
     */
    private transient SendMonitor failureMontior;

    /**
     * <code>PersistentStore<code> that changes should be loged to.
     */
    private transient PersistentStore store;

    // Constructors and state restoration
    /**
     * Simple constructor.  Note expiration will be set when we allocate
     * a lease for this set.
     * @param ID for this set
     * @param generator object set can use to create new event type objects
     * @param store <code>PersistentStore<code> that changes should be loged
     *              to.
     */
    LeaseSet(UUID ID, EventTypeGenerator generator, PersistentStore store) {
	this.store = store;
	this.ID = ID;
	setMonitors();
	
	try {
	    warningEventType = 
		generator.newEventType(null, null, warningMontior);
	    failureEventType =
		generator.newEventType(null, null, failureMontior);
	} catch (IOException e) {
	    // Because we are passing null for the listener we will
	    // never get an exception
	}
    }

    /**
     * Utility method to set the <code>failureMontior</code> and
     * <code>warningMontior</code> fields
     */
    private void setMonitors() {
	try {
	    final Field warningField = 
		LeaseSet.class.getDeclaredField("warningEventType");
	    warningMontior = new SendMonitorImpl(warningField);

	    final Field failureField = 
		LeaseSet.class.getDeclaredField("failureEventType");
	    failureMontior = new SendMonitorImpl(failureField);
	} catch (NoSuchFieldException e) {
	    // This should never happen
	    throw new InternalNormException("Some one removed the " +
	        "\"warningEventType\" and/or \"failureEventType\" fields " +
		"from LeaseSet!", e);
	}
    }

    /**
     * We override <code>readObject</code> to so we can set the
     * <code>failureMontior</code> and <code>warningMontior</code>
     * fields.
     */
    private void readObject(java.io.ObjectInputStream in) 
	throws IOException, ClassNotFoundException
    {
	in.defaultReadObject();
	setMonitors();
    }

    /** 
     * Restore the transient state of the set that can't be restored 
     * automatically after a log recovery.
     * @param generator  Event type generator associated with this set's events
     * @param store      <code>PersistentStore<code> that changes should be 
     *                   loged to.
     * @return The Map of client lease wrappers to membership expirations
     */
    Map restoreTransientState(EventTypeGenerator  generator, 
			      PersistentStore     store)
    {
	this.store = store;
	warningEventType.restoreTransientState(generator, warningMontior);
	failureEventType.restoreTransientState(generator, failureMontior);
	return leases;
    }

    /**
     * Add the specified wrapped client lease to the set.
     * @param clw Lease to be added
     * @param membershipExpiration When should the lease be removed from the
     *				   set
     */
    void add(ClientLeaseWrapper clw, long membershipExpiration) {
	leases.put(clw, new Long(membershipExpiration));
	final Object u = new AddClientLease(this, clw, membershipExpiration);
	store.update(u);
    }

    /**
     * Remove the specified wrapped client lease from the set
     * @param cwl Lease to removed
     */
    void remove(ClientLeaseWrapper clw) {
	if (null == leases.remove(clw)) 
	    // if we are here the lease must have already been 
	    // removed by some other thread, don't bother logging
	    return;

	final Object u = new RemoveClientLease(this, clw);
	store.update(u);
    }

    /**
     * Destroy a lease set
     * @return a <code>Set</code> with all of the sets WrappedClientLeases
     */
    Set destroy() {
	setExpiration(-1);
	final Object u = new CancelLeaseSet(getUUID());
	store.update(u);
	return leases.keySet();
    }

    /** 
     * Set/update/clear the expiration warning listener
     * @param listener   The new listener
     * @param minWarning How long be for the lease on the set expires
     *                   should the event be sent. 
     * @param handback   The new handback
     * @return If <code>listener</code> is non-<code>null</code> return
     * an <code>EventRegistration</code> otherwise return <code>null</code>
     * @throws IOException if listener can not be serialized 
     */
    EventRegistration setExpirationWarningListener(
	RemoteEventListener listener, 
	long                minWarning,
	MarshalledObject    handback)
        throws IOException     					       
    {
	this.minWarning = minWarning;
	warningEventType.setListener(listener, handback);

	final Object u = new WarningEventRegistration(this);
	store.update(u);

	if (listener == null) 
	    return null;

	return new EventRegistration(warningEventType.getEventID(),
				     setProxy, 
				     // $$$ do we want to make sure
				     // this lease is as uptodate as posable
				     setProxy.getRenewalSetLease(), 
				     warningEventType.getLastSequenceNumber());
    }

    /**
     * Set/update/clear the renewal failure listener
     * @param listener The new listener
     * @param handback The new handback
     * @return If <code>listener</code> is non-<code>null</code> return
     * an <code>EventRegistration</code> otherwise return <code>null</code>
     * @throws IOException if listener can not be serialized 
     */
    EventRegistration setRenewalFailureListener(RemoteEventListener listener, 
						MarshalledObject    handback)
        throws IOException     					       
    {
	failureEventType.setListener(listener, handback);

	final Object u = new FailureEventRegistration(this);
	store.update(u);

	if (listener == null) 
	    return null;

	return new EventRegistration(failureEventType.getEventID(),
				     setProxy, 
				     // $$$ do we want to make sure
				     // this lease is as uptodate as posable
				     setProxy.getRenewalSetLease(), 
				     failureEventType.getLastSequenceNumber());
    }

    /**
     * Handle failures to renew a lease by removing the lease from the set
     * and if needed schedule sending an event.
     * @param clw The wrapped client lease for the lease that could not
     *            be removed.
     */
    void renewalFailure(ClientLeaseWrapper clw) {
	if (null == leases.remove(clw))
	    // if we are here the lease must have already been 
	    // removed by some other thread, don't bother logging
	    return;

	Object u;
	try {
	    final long seqNum = 
		failureEventType.sendEvent(clw.newFailureFactory(setProxy));
	    u = new RenewalFailure(this, clw, seqNum);
	} catch (IOException e) {
	    // Can't form event just drop it
	    // $$$ is this the right thing to do?
	    u = new RemoveClientLease(this, clw);
	}

	store.update(u);
    }

    /**
     * Send an expiration warning event for this set
     */
    void sendWarningEvent() {
	warningEventType.sendEvent(new WarningFactory());
    }

    /**
     * Nested class that implements <code>EventFactory</code> that
     * generates <code>ExpirationWarningEvent</code>s.
     */
    private class WarningFactory implements EventFactory {
	// Inherit java doc from super type
	public RemoteEvent createEvent(long             eventID, 
				       long             seqNum, 
				       MarshalledObject handback) 
	{
	    return new ExpirationWarningEvent(setProxy, eventID, seqNum,
	        handback, 
	        // $$$ This is a bug, the lease really needs to be updated
 	        // to the current expiration
	        setProxy.getRenewalSetLease());				      
	}
	
    }

    /**
     * Each <code>LeaseSet</code> keeps a reference to the corresponding 
     * <code>SetProxy</code>.  Use this method to set this reference.
     * <p>
     * Note: this method is not synchronized.
     */
    void setSetProxy(SetProxy proxy) {
	setProxy = proxy;
    }


    /**
     * Return true if there is a non-<code>null</code> listener registered
     * for the expiration warning event.
     * <p>
     * Note, this method assumes the current thread owns the set's lock
     */
    boolean haveWarningRegistration() {
	return warningEventType.haveListener();
    }

    /**
     * Return the absolute time when a expiration warning should be sent.
     */
    long getWarningTime() {
	return expiration - minWarning;
    }

    /**
     * Log the renewal of a client lease.
     * @param clw the wrapper for the client lease that was renewed
     */
    void logRenewal(ClientLeaseWrapper clw) {
	if (!leases.containsKey(clw))
	    // Some other thread must have removed this lease from
	    // the set after renewal, don't bother logging change
	    return;

	final Object u = new RenewClientLease(this, clw);
	store.update(u);
    }

    // Methods need to meet contract of LeasedResource	
    // Inherit java doc from super type
    public void setExpiration(long newExpiration) {
	expiration = newExpiration;
    }

    // Inherit java doc from super type
    public long getExpiration() {
	return expiration;
    }

    // Inherit java doc from super type
    public Object getCookie() {
	return ID;
    }

    /**
     * Return the <code>UUID</code> for this set.
     */
    UUID getUUID() {
	return ID;
    }

    private class SendMonitorImpl implements SendMonitor {
	/** field in parent with the <code>EventType</code> object */
	final private Field eventTypeField;

	/**
	 * Simple constructor
	 * @param eventTypeField Field in the parent that hold the event
	 * type this is the monitor for 
	 */
	private SendMonitorImpl(Field eventTypeField) {
	    this.eventTypeField = eventTypeField;
	}

	// Methods needed to meet contract of SendMonitor
	
	// Inherit java doc from super type
	public void definiteException(EventType           type,
				      RemoteEventListener listener,
				      RemoteEvent         ev,
				      Throwable           t)
	{
	    try {
		final EventType eventType = 
		    (EventType)eventTypeField.get(LeaseSet.this);
		final boolean changed = 
		    eventType.clearListenerIfEquals(listener, 
						    ev.getRegistrationObject());
		if (changed) {
		    // Need to log the change
		    Object u;
		    if (ev instanceof ExpirationWarningEvent)
			u = new WarningEventRegistration(LeaseSet.this);
		    else 
			u = new FailureEventRegistration(LeaseSet.this);
		    store.update(u);
		}
	    } catch (IllegalAccessException e) {
		// This will never happen
	    }
	}
	
	// Inherit java doc from super type
	public boolean isCurrent() {
	    synchronized (LeaseSet.this) {
		final long now = System.currentTimeMillis();
		return (now < LeaseSet.this.getExpiration());
	    }
	}
    }

    // All the inner classes we use to log changes to a lease set

    /**
     * Class used to log changes to the set expiration time.  
     */
    static class ChangeSetExpiration extends LeaseSetOperation {
	/**
	 * Updated expiration time
	 * @serial
	 */
	private long expiration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	ChangeSetExpiration(LeaseSet set) {
	    super(set.getUUID());
	    expiration = set.expiration;
	}
	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.setExpiration(expiration);
	}
    }

    /**
     * Class used to log adding a client lease to the set
     */
    private static class AddClientLease extends LeaseSetOperation {
	/**
	 * Wrapped version of client lease
	 * @serial
	 */
	private ClientLeaseWrapper clw;

	/**
	 * Membership expiration time of lease that was added
	 * @serial
	 */
	private long membershipExpiration;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 * @param long membership expiration time.
	 */
	private AddClientLease(LeaseSet set, ClientLeaseWrapper clw, 
			       long membershipExpiration) {
	    super(set.getUUID());
	    this.clw = clw;
	    this.membershipExpiration = membershipExpiration;
	}

	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.leases.put(clw, new Long(membershipExpiration));
	}
    }

    /**
     * Class used to log the removal of a client lease from the set
     */
    private static class RemoveClientLease extends LeaseSetOperation {
	/**
	 * Client lease to be removed
	 * @serial
	 */
	private ClientLeaseWrapper clw;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 */
	private RemoveClientLease(LeaseSet set, ClientLeaseWrapper clw) {
	    super(set.getUUID());
	    this.clw = clw;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.leases.remove(clw);
	}
    }

    /**
     * Class used to log a renewal failure
     */
    private static class RenewalFailure extends RemoveClientLease {
	/**
	 * Event ID of the corresponding renewal failure event (if any)
	 * @serial
	 */
	private long evID;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 * @param evID event ID of the renewal event that was sent
	 *             to mark this renewal failure
	 */
	private RenewalFailure(LeaseSet set, ClientLeaseWrapper clw,
			       long evID)
	{
	    super(set, clw);
	    this.evID = evID;
	}
	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    super.apply(set);
	    set.failureEventType.setLastSequenceNumber(evID);
	}
    }

    /**
     * Class used to log renewals of client leases
     */
    private static class RenewClientLease extends LeaseSetOperation {
	/**
	 * Wrapped version of client lease that was renewed.
	 * @serial
	 */
	private ClientLeaseWrapper clw;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 */
	private RenewClientLease(LeaseSet set, ClientLeaseWrapper clw) {
	    super(set.getUUID());
	    this.clw = clw;
	}
	
	// Inherit java doc from super type
	void apply(LeaseSet set) throws StoreException {
	    // Old version of the wrapped client lease will be equals()
	    // to the new one, so this remove should work
	    final Long membershipExpiration = (Long)set.leases.remove(clw);
	    if (membershipExpiration != null) 
		set.leases.put(clw, membershipExpiration);
	    else
		// $$$ What should we really do if we get back null?
		System.err.println("Log corrupted, missing client lease " +
				   clw + ", continuing");
	}
    }

    /**
     * Class used to log changes to warning event registrations
     */
    private static class WarningEventRegistration extends LeaseSetOperation {
	/**
	 * Warning time assocated with warning event registration
	 * @serial
	 */
	private long warningTime;

	/**
	 * EventType object that resulted from registration change
	 * @serial
	 */
	private EventType registration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	private WarningEventRegistration(LeaseSet set) {
	    super(set.getUUID());
	    warningTime = set.minWarning;
	    registration = set.warningEventType;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) throws StoreException {
	    set.minWarning = warningTime;
	    set.warningEventType = registration;
	}
    }

    /**
     * Class used to log changes to failure event registrations
     */
    private static class FailureEventRegistration extends LeaseSetOperation {
	/**
	 * EventType object that resulted from registration change
	 * @serial
	 */
	private EventType registration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	private FailureEventRegistration(LeaseSet set) {
	    super(set.getUUID());
	    registration = set.failureEventType;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) throws StoreException {
	    set.failureEventType = registration;
	}
    }
}
