/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;
import java.util.Collections;
import java.util.Properties;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.InterruptedIOException;

import java.rmi.MarshalledObject;
import java.rmi.Remote;
import java.rmi.MarshalledObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;

import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationSystem;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;

import net.jini.lookup.entry.ServiceInfo;

import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMapException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.LeaseDeniedException;

import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.lease.LeaseRenewalService;
import net.jini.lease.LeaseRenewalSet;
import net.jini.lease.LeaseRenewalManager;

import com.sun.jini.lease.landlord.LeasePolicy;
import com.sun.jini.lease.landlord.LeaseDurationPolicy;
import com.sun.jini.lease.landlord.Landlord.RenewResults;

import com.sun.jini.proxy.UUID;
import com.sun.jini.proxy.UUIDFactory;
import com.sun.jini.proxy.ThrowThis;

import com.sun.jini.lookup.entry.BasicServiceType;

import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.reliableLog.LogException;

import com.sun.jini.mahout.KillVMAdmin;
import com.sun.jini.mahout.ParseUtils;
import com.sun.jini.mahout.PersistentServiceDesc;
import com.sun.jini.mahout.BasicServiceDesc;

import com.sun.jini.constants.VersionConstants;

import com.sun.jini.norm.event.EventTypeGenerator;
import com.sun.jini.norm.lookup.JoinState;

/**
 * Implementation of NormServer.
 */
public class NormServerImpl implements NormServer {
    /** Current version of log format */
    private static final int CURRENT_LOG_VERSION = 1;

    /** Our activation ID */
    private ActivationID activationID;

    /** The name of our log dir */
    private String logDir;

    /** Our JoinManager */
    private JoinState joinState;

    /** Map of UUIDs to LeaseSets */
    private Map setTable = Collections.synchronizedMap(new java.util.HashMap());

    /** Map of client leases to wrapped equivalents  */
    private Map leaseTable = new java.util.HashMap();

    /** Lease renewal manager we that actually renews the leases */
    private LeaseRenewalManager lrm = new LeaseRenewalManager();

    /** Object the expires sets and generates expiration warning events */
    private LeaseExpirationMgr expMgr = new LeaseExpirationMgr(this);

    /** Policy we use for granting and renewing renewal set leases */
    private LeasePolicy setLeasePolicy;

    /** Our persistant store */
    private PersistentStore store;

    /** Factory we use to create UUIDs */
    private UUIDFactory setIDFactory = new UUIDFactory();

    /** Factory we use to create ClientLeaseWrapper IDs */
    private UIDGenerator idGen = new UIDGenerator() ;

    /** List of leases that have been renewed by not persisted */
    private List renewedList = new java.util.LinkedList();

    /** 
     * List of client leases that could not be full recovered during
     * log recovery
     */
    private DeformedLeaseList deformedLeases = new DeformedLeaseList();
	
    /**
     * Thread that pulls wrapped client leases off the renewedList and logs
     * them to disk
     */
    private RenewLogThread renewLogger;

    /** Object use to generate new event types */
    private EventTypeGenerator generator = new EventTypeGenerator();

    /**
     * Object that transfers events from the renewal manager to 
     * us we can remove dead lease and send events
     */
    private FailureNotifer failureNotifer;

    /** Log File must contain this many records before snapshot allowed */
    private int logToSnapshotThresh = 200;

    /** Weight factor applied to snapshotSize when deciding to take snapshot */
    private float snapshotWt = 10;

    /** Our remote stub */
    private NormServer ourStub = null;

    /** Thread the performs snapshots when signaled */
    private SnapshotThread snapshotter;

    ////////////////////////////////
    // Methods defined in NormServer
  
    // Inherit java doc from super type
    // $$$ This method is a bit long, unfortunately the best candidate
    // block for pulling out into a method (the manipulation of the leaseTable)
    // needs to return two values....
    public void renewFor(UUID id, Lease leaseToRenew,
			 long membershipDuration) 
	throws ThrowThis
    {
	final LeaseSet set = (LeaseSet)setTable.get(id);
	if (set == null)
	    throw new ThrowThis(new NoSuchObjectException("Can't find set"));

	// $$$ Should ensure that the lease was not granted by this
	// server

	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // The set's lease has expired, don't add anything
		    // to the set!
		    throw new ThrowThis(
		        new NoSuchObjectException("Set has expired"));
		}

		ClientLeaseWrapper clw;
		boolean alreadyInSet;

		// Make sure the lease is not owned by some other set
		synchronized (leaseTable) {
		    clw = (ClientLeaseWrapper)leaseTable.get(leaseToRenew);
		    if (clw == null) {
			// Could it be that this was a lease that
			// could not be recovered? Check the deformed list
			clw = deformedLeases.query(leaseToRenew, leaseTable);
		    }

		    if (clw == null) {
			// We don't know about this lease, create a new
			// wrapper and place it in the table

			alreadyInSet = false;
			try {
			    // Until the lease gets removed we always
			    // what it serialized using the absolute
			    // format 
			    leaseToRenew.setSerialFormat(Lease.ABSOLUTE);
			    clw = new ClientLeaseWrapper(leaseToRenew,
							 idGen.newID(),
							 renewedList,
							 id);
			} catch (IOException e) {
			    throw new IllegalArgumentException(
			       "NormServerImpl.renewFor:Handed lease that " +
			       "can't be marshalled");
			}
			leaseTable.put(leaseToRenew, clw);
		    } else {
			// Two possibilities, ether it is already in this
			// which case we just want to update, or it is in
			// another set in which case we throw ILAE
			if (clw.isInSet(id)) 
			    alreadyInSet = true;
			else
			    throw new IllegalArgumentException(
				"NormServerImpl.renewFor:Lease already in " +
				"another set created by this renewal service");
		    }
		}
		 
		long membershipExpiration;

		if (membershipDuration == Lease.ANY) {
		    membershipExpiration = Lease.ANY;
		} else {
		    membershipExpiration = membershipDuration +
			System.currentTimeMillis();

		    // Check for overflow
		    if (membershipExpiration < 0)
			membershipExpiration = Long.MAX_VALUE;
		}

		set.add(clw, membershipExpiration);
		
		if (alreadyInSet)
		    try {
			lrm.setExpiration(clw, membershipExpiration);
		    } catch (UnknownLeaseException e) {
			// $$$ This can happen if there was some problem
			// renewing the lease, or its LRM expiration just
			// ran out.  Not clear if just shoving it back
			// in is really the right answer since quite
			// likely the next thing to happen will be
			// then generation of a removal event...not
			// clear if that should undo THIS add
			// Maybe we don't want to LRM to remove leases
			// at all?
			lrm.renewUntil(clw, membershipExpiration, 
				       failureNotifer);
		    }
		else
		    lrm.renewUntil(clw, membershipExpiration, failureNotifer);
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    // Inherit java doc from super type
    public Lease remove(UUID id, Lease leaseToRemove) throws ThrowThis {
	final LeaseSet set = (LeaseSet)setTable.get(id);
	if (set == null)
	    throw new ThrowThis(new NoSuchObjectException("Can't find set"));

	// The most uptodate ref to the lease we have
	Lease leaseRemoved = null;
	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Lease has expired add anything to it!
		  throw new ThrowThis(
		      new NoSuchObjectException("Set has expired"));
		}
		
		ClientLeaseWrapper clw;
		synchronized (leaseTable) {
		    clw = (ClientLeaseWrapper)leaseTable.get(leaseToRemove);

		    // Down to set.remove() could be outside of of the
		    // sync block...just seems like a waste to grab
		    // the leaseTable lock, let it go, do a little bit
		    // of work, and grab it again...the removal from
		    // the table does have to be after the lease is
		    // remove from the set/LRM [Otherwise another
		    // thread could add the lease to another set and
		    // add to the LRM but then have it removed by this
		    // thread.]  [Of course it may be more than just a
		    // little bit of work, set.remove() does need to
		    // flush an object to disk...if we do acquire 
		    // leaseTable's lock twice we need to do a 2nd get
		    // and make sure that the clw is in set.

		    if (clw == null) 
			// Make sure we are not dealing with a deformed lease
			clw = deformedLeases.query(leaseToRemove, leaseTable);

		    if (clw == null)
			// Lease must have been removed already
			return null;

		    try {
			lrm.remove(clw);
		    } catch (UnknownLeaseException e) {
			// This can happen if there was some problem
			// renewing the lease or its LRM expiration just
			// ran out. Since we are removing the lease anyway
			// ignore. 
		    }

		    set.remove(clw);		   
		    leaseTable.remove(leaseToRemove);
		    // At this point we can assume clw is no longer deformed
		    leaseRemoved = clw.getClientLease();
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    

	if (leaseRemoved == null) 
	    return null;

	// Whenever we serialize a lease we have to make sure that
	// its serial form will stay the same during the
	// serialization.  Since we have removed it from the set we
	// don't have to worry about this lease being serialized to disk
	// any more so changing the serial format here should be safe.
	leaseRemoved.setSerialFormat(Lease.DURATION);
	return leaseRemoved;
    }

    // Inherit java doc from super type
    public EventRegistration setExpirationWarningListener(
				 UUID		     id,
				 RemoteEventListener listener, 
				 long		     minWarning, 
				 MarshalledObject    handback)
	throws ThrowThis 
    {
	final LeaseSet set = (LeaseSet)setTable.get(id);
	if (set == null)
	    throw new ThrowThis(new NoSuchObjectException("Can't find set"));

	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Lease has expired, don't do anything to it
		    throw new ThrowThis(
		        new NoSuchObjectException("Set has expired"));
		}

		try {
		    final boolean haveBefore = set.haveWarningRegistration();
		    final EventRegistration rslt =
			set.setExpirationWarningListener(listener,
							 minWarning, handback);

		    if (haveBefore != set.haveWarningRegistration()) {
			// Ether we had a registration before and we
			// don't now or do now and not before, ether
			// way wack the expiration manager so it can
			// schedule the right task.
			expMgr.reschedule(set);
		    }

		    return rslt;
		} catch (IOException e) {
		    // This means the listener could not be serialized,
		    // re-throw as a IllegalArgumentException 
		    throw new IllegalArgumentException("Passed a listener " +
			      "that could not be serialized");
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    /**
     * Remote a set if its expiration time has been reached.
     */
    void expireIfTime(LeaseSet set) {
	try {
	    store.acquireMuatorLock();

	    synchronized (set) {
		if (setLeasePolicy.ensureCurrent(set)) {
		    // Someone must have renewed the lease...don't expire
		    return;
		}

		removeSet(set);
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    /**
     * Schedule the sending of an expiration warning event.
     * This could be a method on the set it self but this keeps all
     * of the high level synchronization logic in one file.     
     */
    void sendWarningEvent(LeaseSet set) {
	// We don't need to acquire the store lock because we 
	// won't be mutating any persistant state.
	// We will be reading state of the set so we do need to
	// sync on it to ensure we get a constant view
	synchronized (set) {
	    if (!setLeasePolicy.ensureCurrent(set)) {
		// Must have just been canceled or expired, don't send event
		return;
	    }

	    set.sendWarningEvent();
	}
    }

    // Inherit java doc from super type
    public EventRegistration setRenewalFailureListener(
				 UUID		     id,
				 RemoteEventListener listener, 
				 MarshalledObject    handback)
	throws ThrowThis
    {
	final LeaseSet set = (LeaseSet)setTable.get(id);
	if (set == null)
	    throw new ThrowThis(new NoSuchObjectException("Can't find set"));

	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Lease has expired, don't do anything to it
		    throw new ThrowThis(
		        new NoSuchObjectException("Set has expired"));
		}

		try {
		    return set.setRenewalFailureListener(listener,
							 handback);
		} catch (IOException e) {
		    // This means the listener could not be serlized,
		    // re-throw as a IllegalArgumentException 
		    throw new IllegalArgumentException("Passed a listener " +
			      "that could not be serialized");
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    /**
     * Handle failures to renew a lease by removing the lease from its set
     * and if needed schedule sending an event.
     * @param clw The wrapped client lease for the lease that could not
     *            be removed.  <code>clw.isDeformed</code> must be
     *            <code>false</code>
     */
    void renewalFailure(ClientLeaseWrapper clw) {
	final LeaseSet set = clw.getLeaseSet(setTable);
	if (set == null) 
	    // set must have just been removed, no state to update, no
	    // events to send, just return
	    return;

	try {
	    store.acquireMuatorLock();	    

	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // expired, no state to update, no
		    // events to send, just return
		    return;
		}

		set.renewalFailure(clw);

		final Lease clientLease = clw.getClientLease();
		if (clientLease != null) {
		    // Not deformed, we need to remove both the client
		    // lease and the wrapper from the leaseTable
		    synchronized (leaseTable) {
			leaseTable.remove(clientLease);
		    }
		}

		// Note, we don't need to remove the lease from the
		// LRM because it has already been removed.
	    }
	} finally {
	    store.releaseMuatorLock();
	}    	
    }

    /////////////////////////////////////////
    // Methods defined in LeaseRenewalService

    // Inherit java doc from super type
    public LeaseRenewalSet createLeaseRenewalSet(long leaseDuration) {
	final UUID newID      = setIDFactory.newUUID();
	final LeaseSet newSet = new LeaseSet(newID, generator, store);
	
	Lease newLease;

	try {
	    newLease = setLeasePolicy.leaseFor(newSet, leaseDuration);
	} catch (LeaseDeniedException e) {
	    // This will never happen because we never use a policy that
	    // denies granting (or renewing) a lease complain bitterly
	    // and throw a runtime exception
	    System.err.println("NormServerImpl.createLeaseRenewalSet:" +
		"Got LeaseDeniedException creating lease - this should not " +
		"happen!");
	    throw new InternalNormException("Error creating lease", e);
	}

	final SetProxy rslt = new SetProxy(ourStub, newID, newLease);
	newSet.setSetProxy(rslt);

	try {
	    store.acquireMuatorLock();
	    setTable.put(newID, newSet);
	    final Object u = new CreateLeaseSet(newSet);
	    store.update(u);
	} finally {
	    store.releaseMuatorLock();
	}

	return rslt;
    }

    //////////////////////////////////////////
    // Thread to persist client lease renewals
    private class RenewLogThread extends Thread {
	/** Create a daemon thread */
	private RenewLogThread() {
	    super("log renewals thread");
	    setDaemon(true);
	}

	public void run() {
	    while (true) {
		try {
		    ClientLeaseWrapper clw;
		    synchronized (renewedList) {
			// If there is an item on the list pull it off for
			// processing, otherwise wait and try again
			if (renewedList.isEmpty()) {
			    try {
				renewedList.wait();
				continue; // go back to top of loop
			    } catch (InterruptedException e) {
				// someone wants us dead
				return;
			    }
			} else {
			    clw = (ClientLeaseWrapper)renewedList.remove(0);
			}			
		    }
		    
		    // A lease was renewed, log the new state
		    final LeaseSet set = clw.getLeaseSet(setTable);
		    if (set == null) 
			// set must have just been removed, no state
			// to update, go to next item in list
			continue;
		    
		    try {	
			store.acquireMuatorLock();  
			synchronized (set) {
			    if (!setLeasePolicy.ensureCurrent(set)) {
				// expired, no state to update, go to next item
				// in list
				continue;
			    }
			    
			    clw.clearRenewed();
			    // Small window here where the lease can
			    // be renewed, it's renewed flag is re-set
			    // and because it was cleared the clw ends
			    // up on renewed list, we then come back
			    // to this thread and get log the new
			    // state, and come back later and log that
			    // state again.  Since this is just
			    // slightly wasteful, not incorrect this
			    // is ok. [Reversing these two lines of
			    // course would be wrong...]
			    set.logRenewal(clw);	
			}
		    } finally {
			store.releaseMuatorLock();
		    }

		    // Give other threads a chance to run
		    Thread.yield();
		} catch (RuntimeException e) {
		    System.err.println("Runtime exception in RenewLogThread" +
				       "...restarting");
		    e.printStackTrace();
		}
	    }
	}
    }


    //////////////////////////////
    // Methods defined in Landlord

    // Inherit java doc from super type
    public long renew(Object cookie, long extension) 
	throws UnknownLeaseException, LeaseDeniedException
    {
	final LeaseSet set = (LeaseSet)setTable.get(cookie);
	if (set == null) 
	    throw new UnknownLeaseException("No lease for cookie:" + cookie);
	try {
	    store.acquireMuatorLock();	    

	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Lease has expired, don't renew
		    throw new 
			UnknownLeaseException("Lease has already expired");
		}
		// No one can expire the lease, so it is safe to update.
		// $$$ Might be better to make an extra call to currentTime
		// and calculate the new duration just be for returning
		final long granted = setLeasePolicy.renew(set, extension);

		// Log update 
		final Object u = new LeaseSet.ChangeSetExpiration(set);
		store.update(u);
		return granted;
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    // Inherit java doc from super type
    public void cancel(Object cookie) throws UnknownLeaseException {
	final LeaseSet set = (LeaseSet)setTable.get(cookie);
	if (set == null) 
	    throw new UnknownLeaseException("No lease for cookie:" + cookie);

	try {
	    store.acquireMuatorLock();
	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    //Someone else beat us to it, just return
		    return;
		}

		removeSet(set);
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    /**
     * Do the heavy lifting on removing a set, assumes the locks on the
     * set and store have been aquired.
     */
    private void removeSet(LeaseSet set) {
	setTable.remove(set.getCookie());	
	// set.destroy will persist the removal of the set, and 
	// change the set's expiration to -1 which will cause any
	// other operations on the set to throw NoSuchObjectException
	// or UnknownLeaseException (they will all lock on the set
	// and ensure current before doing anything substave)
	final Set leases = set.destroy(); 
	synchronized (leaseTable) {
	    for (Iterator i = leases.iterator(); i.hasNext(); ) {
		ClientLeaseWrapper clw = (ClientLeaseWrapper)i.next();
		Lease cl = clw.getClientLease();
		
		if (cl == null)
		    // Must be deformed, not in LRM or leaseTable, ignore
		    continue;

		try {
		    lrm.remove(clw);
		} catch (UnknownLeaseException e) {
		    // This can happen if there was some problem
		    // renewing the lease, or its LRM expiration just
		    // ran out.	 Since we are remove the lease any
		    // way ignore.
		}

		leaseTable.remove(cl);
	    }
	}
    }

    // Inherit java doc from super type
    public RenewResults renewAll(Object[] cookies, long[] extensions) {
	long[] granted	= new long[cookies.length];
	List exceptions = new java.util.ArrayList();

	for (int i = 0; i < cookies.length; i++) {
	    Exception failure = null;

	    try {
		granted[i] = renew(cookies[i], extensions[i]);
	    } catch (LeaseDeniedException e) {
		failure = e;
	    } catch (UnknownLeaseException e) {
		failure = e;
	    }

	    if (failure != null) {
		exceptions.add(failure);
		granted[i] = -1;
	    }
	}

	if (exceptions.size() == 0)
	    return new RenewResults(granted);
	else {
	    Exception[] es = new Exception[exceptions.size()];
	    Iterator it = exceptions.iterator();
	    for (int i = 0; it.hasNext(); i++)
		es[i] = (Exception) it.next();
	    return new RenewResults(granted, es);
	}
    }

    // Inherit java doc from super type
    public void cancelAll(Object[] cookies) throws LeaseMapException {
	Map map = null;
	for (int i = 0; i < cookies.length; i++) {
	    try {
		cancel(cookies[i]);
	    } catch (UnknownLeaseException e) {
		if (map == null)
		    map = new java.util.HashMap();
		map.put(cookies[i], e);
	    }
	}

	if (map != null)
	    throw new LeaseMapException("cancelling", map);

    }

    ////////////////////////////////////////////////
    // Methods and classes needed by PersistantStore

    /**
     * Called by <code>PersistentStore</code> after every update to give
     * server a chance to trigger a snapshot
     * @param updateCount Number of updates since last snapshot
     */
    void updatePerformed(int updateCount) {
	// First check to see if the size of the log is larger than a 
	// minimum threshold, this keeps snapshots from happening when 
	// the state of the server is very small (like initially when
	// the snapshot size is zero)
	if (updateCount >= logToSnapshotThresh) {
	    // Compare the size of log to what next snapshot would be
	    final int snapshotSize = setTable.size() + leaseTable.size();
	    if ((float)updateCount >= snapshotWt*((float)snapshotSize)) {
		// Both conditions meet, trigger snapshot
		snapshotter.takeSnapshot();
	    }
	}
    }

    /**
     * Perform the 3rd stage of log recovery, restoring the various pieces of 
     * transient state (populating the LRM, leaseTable, restoring various 
     * transient fields).
     */
    void restoreTransientState() {
	for (Iterator i = setTable.values().iterator(); i.hasNext(); ) {
	    // $$$ might want to dump sets that have expired
	    
	    final LeaseSet set = (LeaseSet)i.next();
	    final Map leases = set.restoreTransientState(generator, store);
	
	    // Go through all the leases in the set and add them to 
	    // the right tables.
	    final Set contents = leases.entrySet();
	    for (Iterator j = contents.iterator(); j.hasNext(); ) {

		// We sync here so all the updates will be flushed to
		// memory before the the LRM's threads start working on
		// this lease
		ClientLeaseWrapper clw;
		long membershipExpiration; 
		synchronized (this) {
		    final Map.Entry entry = (Map.Entry)j.next();
		    clw = (ClientLeaseWrapper)entry.getKey();
		    membershipExpiration = ((Long)entry.getValue()).longValue();

		    // Let the clw recover its state	    
		    clw.recoverTransient(renewedList, idGen);

		    // Do we put this lease in the table or the deformed list?
		    final Lease cl = clw.getClientLease();
		    if (cl == null) 
			deformedLeases.add(clw);
		    else
			leaseTable.put(cl, clw);
		}

		// Note: there is no race condition here because
		// failureNotifer and renewedList will buffer any
		// renewals/renewal failure until the rest of the server
		// is ready to handle them
		lrm.renewUntil(clw, membershipExpiration, failureNotifer);
	    }
	}
    }

    /**
     * Implementation of <code>LogHandler</code> used by NormServerImpl
     */
    private class OurLogHandler extends LogHandler {
	// Inherit java doc from super type
	// Snapshot format
	//     Generator we use to create EventType objects
	//     The number of LeaseSet objects in the setTable
	//     All of the LeaseSet objects
	public void snapshot(OutputStream out) throws Exception {
	    final ObjectOutputStream oostream = new ObjectOutputStream(out);

	    oostream.writeInt(CURRENT_LOG_VERSION);
	    oostream.writeObject(generator);
	    oostream.writeInt(setTable.size());
	    final Collection sets = setTable.values();
	    for (Iterator i = sets.iterator(); i.hasNext();) {
		oostream.writeObject(i.next());
	    }
	    oostream.flush();

	    // $$$ We are missing an optimization here, as write each
	    // set we could check to see if any of the client leases
	    // in that set have been renewed, but not logged and clear
	    // them from the renewedList (this assumes that the
	    // renewal changes the persisted state of the wrapper, not
	    // the processing by logRenewal()).  This is not a bug
	    // since any clw we pull off the renewedList and persist
	    // after this point will have state at lease as uptodate
	    // as this snapshot.
	}

	// Inherit java doc from super type
	public void recover(InputStream in) throws Exception {
	    final ObjectInputStream oistream = new ObjectInputStream(in);

	    int version;
	    version = oistream.readInt();

	    if (version != CURRENT_LOG_VERSION) {
		    throw new CorruptedStoreException("Incompatible version " +
		        "ID in log, looking for " + CURRENT_LOG_VERSION +
			", got " + version);

	    }

	    generator = (EventTypeGenerator)oistream.readObject();
	    final int size = oistream.readInt();
	    setTable = Collections.synchronizedMap(new java.util.HashMap(size));
	    for (int i = 0; i < size; i++) {
		final LeaseSet set = (LeaseSet)oistream.readObject();
		setTable.put(set.getUUID(), set);
	    }
	}

	// Inherit java doc from super type
	public void applyUpdate(Object update) throws Exception {
	    final LoggedOperation op = (LoggedOperation)update;
	    op.apply(setTable);
	}
    }

    /**
     * Thread that performs the actual snapshots, done in a separate thread
     * so it will not hang up in progress remote calls
     */
    private class SnapshotThread extends Thread {
	/** Create a daemon thread */
	private SnapshotThread() {
	    super("snapshot thread");
	    setDaemon(true);
	}

	/** Signal this thread that is should take a snaps */
	private synchronized void takeSnapshot() {
	    notifyAll();
	}

	public void run() {
	    while (!isInterrupted()) {
		synchronized (this) {
		    try {
			wait();
		    } catch (InterruptedException e) {
			return;
		    }
		}

		try {
		    System.err.println("Taking snapshot");
		    store.snapshot();
		    System.err.println("Snapshot done");
		} catch (InterruptedIOException e) {
		    return;
		} catch (Exception e) {
		    if (e instanceof LogException &&
			((LogException)e).detail instanceof
			InterruptedIOException)
			return;
		    /* $$$
		     * if taking the snapshot fails for any reason,
		     * then one of the following must be done:
		     *   -- output the problem to a file and exit
		     *   -- output the problem to a file and continue
		     *   -- set an "I have a problem" attribute and
		     *      then send a notification
		     * this issue will be addressed at a later time
		     */
		    e.printStackTrace();
		}
	    }
	}
    }

    ///////////////////////////////////
    // Methods defined in Administrable

    // Inherit java doc from super type
    public Object getAdmin() {
	return this;
    }

    ///////////////////////////////
    // Methods defined in JoinAdmin

    // Inherit java doc from super type
    public Entry[] getLookupAttributes() {
	return joinState.getAttributes();
    }

    // Inherit java doc from super type
    public void addLookupAttributes(Entry[] attrSets) {
	joinState.addAttributes(attrSets, true);
    }

    // Inherit java doc from super type
    public void modifyLookupAttributes(Entry[] attrSetTemplates, 
				       Entry[] attrSets) 
    {
	joinState.modifyAttributes(attrSetTemplates, attrSets, true);
    }
  
    // Inherit java doc from super type
    public String[] getLookupGroups() {
	return joinState.getGroups();
    }

    // Inherit java doc from super type
    public void addLookupGroups(String[] groups) {
	joinState.addGroups(groups);
    }

    // Inherit java doc from super type
    public void removeLookupGroups(String[] groups) {
	joinState.removeGroups(groups);
    }

    // Inherit java doc from super type
    public void setLookupGroups(String[] groups) {
	joinState.setGroups(groups);
    }

    // Inherit java doc from super type
    public LookupLocator[] getLookupLocators() {
	return joinState.getLocators();
    }

    // Inherit java doc from super type
    public void addLookupLocators(LookupLocator[] locators) {
	joinState.addLocators(locators);
    }

    // Inherit java doc from super type
    public void removeLookupLocators(LookupLocator[] locators) {
	joinState.removeLocators(locators);
    }

    // Inherit java doc from super type
    public void setLookupLocators(LookupLocator[] locators) {
	joinState.setLocators(locators);
    }

    //////////////////////////////////
    // Methods defined in DestroyAdmin

    // Inherit java doc from super type
    public void destroy() {
	Exception failed = null;

	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (Exception t) {
	    System.err.println("While unexporting and/or unregistering " +
			       "with RMID :");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

	try {
	    store.destroy();
	} catch (Exception t) {
	    System.err.println("While destroying persistent store:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

	killVM();	// go away
    }

    /**
     * Sleep time for <code>killVM</code>.
     *
     * @see killVM
     */
    private final static long KILL_DELAY = 10;

    // inherit doc comment
    public long killVM() {
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server
	 * to which it is talking (this one) dies in the middle of
	 * the call.  So we start up a new thread that sleeps for a
	 * hundredth of a second and then shuts down the VM.  This lets
	 * the thread running shutDown return before the denouement.
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		    ;	// oh, well
		}
		System.exit(0);
	    }
	}.start();
	return KILL_DELAY * 100;
    }

    //////////////////////////////////
    // Methods defined in DestroyAdmin

    // Inherit java doc from super type
    public synchronized void setSnapshotWeight(float weight) {
	snapshotWt = weight;
    }

    // Inherit java doc from super type
    public synchronized float getSnapshotWeight() {
	return snapshotWt;
    }

    // Inherit java doc from super type
    public synchronized void setLogToSnapshotThreshold(int threshold) {
	logToSnapshotThresh = threshold;
    }

    // Inherit java doc from super type
    public synchronized int getLogToSnapshotThreshold() {
	return logToSnapshotThresh;
    }


    ////////////////////////////////////
    // Server setup and activation code

    // Activation

    /**
     * If for some reason we can't activate we dump a message to the log 
     * and unregister with activation.
     * @param m message to dump to <code>System.err</code>
     * @param e If non-<code>null</code> this exception is dumped to
     * <code>System.err</code> as well 
     */
    private void alertAndUnregister(String m, Throwable e) {
	System.err.println("Fatal error during activation:" + m);
	if (e != null)
	    e.printStackTrace();
	
	Exception failed = null;

	System.err.println("Attempting to unregister with activation");
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (ActivationException ee) {
	    failed = ee;
	} catch (RemoteException ee) {
	    failed = ee;
	}

	if (failed != null) {
	    System.err.println("Failed in unregister attempt");
	    failed.printStackTrace();
	}
	
	System.err.println("Succeeded in unregistering");
    }

    /**
     * Activation constructor
     *
     * @param activationID activation ID passed in by the activation daemon.
     * @param data state data needed to re-activate a Norm server
     */
    public NormServerImpl(ActivationID activationID, MarshalledObject data) 
	throws IOException, ClassNotFoundException
    {
	boolean runtimeThrown = false;
	System.err.println("Norm is activating");
	this.activationID = activationID;

	try {
	    try {
		ourStub = 
		    (NormServer)Activatable.exportObject(this, activationID, 0);
	    } catch (RemoteException e) {
		 alertAndUnregister("Failure exporting object", e);
		 throw e;
	    }

	    // Get Log Dir Name
	    try {
		logDir = (String)data.get();
	    } catch (IOException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;
	    } catch (ClassNotFoundException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;	  
	    } catch (ClassCastException e) {
		alertAndUnregister("Failure recoving activation data", e);
		runtimeThrown = true;
		throw e;
	    }

	    // We use some of these during the recovery process
	    failureNotifer = new FailureNotifer(this);
	    renewLogger = new RenewLogThread();
	    snapshotter = new SnapshotThread();

	    try {
		store = new PersistentStore(logDir, ourStub, 
					    new OurLogHandler(), this);
		// Creating the store completes the first two stages of 
		// log recovery (reading the snapshot and the updates)
		// Perform the last stage here of restoring transient state
		restoreTransientState();
	    } catch (StoreConflictException e) {
		alertAndUnregister("Another Norm Server is using this " +
		    "log (" + logDir + ")", e);
		runtimeThrown = true;
		throw new IllegalStateException("Trying to start a Norm " +
		    "server with a log owned by another server");
	    } catch (CorruptedStoreException e) {
		alertAndUnregister("Log corrupted, can't recover ", e);
		runtimeThrown = true;
		throw new IllegalStateException("Corrupted log");
	    } catch (StoreException e) {
		alertAndUnregister("Can't recover log", e);
		runtimeThrown = true;
		throw new IllegalStateException("Corrupted log");
	    }

	    // Create new baseline snapshot
	    try {
		store.snapshot();
	    } catch (IOException e) {
		alertAndUnregister("Can't create new baseline snapshot", e);
		throw e;
	    }

	    try {
		joinState = new JoinState(this, lrm);
		store.addSubStore(joinState);
	    } catch (StoreException e) {
		alertAndUnregister("Can't create JoinState", e);
		runtimeThrown = true;
		throw new IllegalStateException("Can't create JoinState");
	    }

	    // Make sure we have a ServiceInfo and BasicServiceType
	    // attribute
	    try {
		final Entry[] currentAttrs = joinState.getAttributes();
		boolean haveServiceInfo = false;
		boolean haveBasicServiceType = false;

		for (int i=0; i<currentAttrs.length; i++) {
		    if (currentAttrs[i] instanceof ServiceInfo)
			haveServiceInfo = true;
		    else if (currentAttrs[i] instanceof BasicServiceType)
			haveBasicServiceType = true;
		}

		Entry[] toAdd = null;
		if (!haveServiceInfo && !haveBasicServiceType)
		    toAdd = new Entry[2];
		else if (!haveServiceInfo || !haveBasicServiceType)
		    toAdd = new Entry[1];
		
		if (toAdd != null) {
		    int i = 0;
		    if (!haveServiceInfo) {
			toAdd[i] = new ServiceInfo("Lease Renewal Service", 
		            "Sun Microsystems, Inc.",  
			    "Sun Microsystems, Inc.", 
			    VersionConstants.SERVER_VERSION,
			    "", "");
			i++;
		    } 

		    if (!haveBasicServiceType) {
			toAdd[i] = 
			    new BasicServiceType("Lease Renewal Service");
			i++;
		    }
			
		    joinState.addAttributes(toAdd, false);
		}
	    } catch (Exception e) {
 		System.err.println("Problem setting up service controlled " +
				   "attributes");
		e.printStackTrace();
		System.err.println("Ignoring and continuing");
	    }

	} catch (RuntimeException e) {
	    if (runtimeThrown == false) { 
		alertAndUnregister("Unexpected exception during activation",
				   e);
	    }

	    throw e;
	} catch (Error e) {
	    alertAndUnregister("Unexpected exception during activation", e);
	    throw e;
	}	  

	setLeasePolicy = 
	    new LeaseDurationPolicy(
	        1000 * 3600 * 2, // At most grant 2 hour leases
		1000 * 3600 * 1, // default lease is one hour
		ourStub, expMgr, null);

	// $$$ By rights this code should be in
	// restoreTransientState(), however we can't have an independent
	// thread running around changing persistant state util we get
	// to this point (I think the only real issue is the baseline
	// snapshot) and once we place a set in the expMgr it the
	// underlying wakeup queue will start running which can cause
	// calls to expireIfTime() (and sendWarningEvent() though those
	// should not be a problem since they don't log anything).
	//
	// I would prefer to ether modify WakeupQueue() to have a "start now"
	// method (eqv to how we create failureNotifer above but call 
	// start() bellow) or be able to hold an exclusive snapshot
	// lock until the initial snapshot is done.  Ether should allow
	// this code to be moved into restoreTransientState

	for (Iterator i = setTable.values().iterator(); i.hasNext(); ) {
	    final LeaseSet set = (LeaseSet)i.next();
	    synchronized (set) {
		expMgr.schedule(set);
	    }
	}

	failureNotifer.start();
	renewLogger.start();
	snapshotter.start();

	System.err.println("Norm is activated");
    }


    // Setup - code that registers a NormServerImpl with activation

    /**
     * Create an activation group suitable for a Norm server.
     * @param desc Classpath, codebase, security policy file, VM, VM args,
     *		   and VM properties for activation group.
     */
    private static ActivationGroupID createGroup(PersistentServiceDesc desc)
	throws ActivationException, RemoteException
    {
	Properties props = desc.getProperties();

	if (props == null)
	    props = new Properties();
	else
	    props = (Properties)props.clone();

	props.put("java.security.policy", desc.getPolicyPath());
	props.put("java.rmi.server.codebase", desc.getCodebase());

	String argv[] = desc.getJVMOptions();
	if (argv == null) {
	    argv = new String[]{"-cp", desc.getClassPath()};
	} else {
	    String[] nargv = new String[argv.length + 2];
	    System.arraycopy(argv, 0, nargv, 0, argv.length);
	    nargv[argv.length] = "-cp";
	    nargv[argv.length + 1] = desc.getClassPath();
	    argv = nargv;
	}

	CommandEnvironment cmd = 
	    new CommandEnvironment(desc.getJVMPath(), argv);
	return ActivationGroup.getSystem().registerGroup(
					 new ActivationGroupDesc(props, cmd));
    }

    /**
     * Create an activatable and persistent, LeaseRenewalService in
     * its own activation group.
     * <p>
     * The activation group will be configured according to the passed
     * <code>PersistentServiceDesc</code> object.  The service will
     * persist its data in the directory returned by
     * <code>desc.getPersistentsPath()</code> method. This directory
     * should not exist.
     * <p>
     * Note that when the service is later destroyed, it will destroy
     * its activation group, regardless of whether or not there are
     * other activatable objects still registered in the same group.
     * It will also delete its persistence directory.
     * <p>
     * The server will not bind itself into a directory service until
     * it has been configured through its administration interface.
     * The service's administration interface can be obtained by
     * casting it to a <code>net.jini.admin.Administrable</code> and
     * calling <code>getAdmin()</code>.	 The returned value will be a
     * reference to the service's admin interface.
     * <p>
     *
     * @see net.jini.admin.Administrable
     */
    public static LeaseRenewalService create(PersistentServiceDesc desc) 
	throws RemoteException, IOException, ActivationException
    {
	ActivationGroupID actID = createGroup(desc);

	try {
	    return create(actID, desc.getPersistentsPath());
	} catch (Exception e) {
	    try {
		ActivationGroup.getSystem().unregisterGroup(actID);
	    } catch (Exception ee) {
	    }
	    if (e instanceof ActivationException)
		throw (ActivationException)e;
	    else if (e instanceof IOException)
		throw (IOException)e;
	    throw (RuntimeException)e;
	}
    }


    /**
     * Create an activatable and persistent, NormServer in
     * the passed activation group.  The service will persist its data
     * in the passed directory.	 This directory should not exist.
     * <p>
     * The passed activation group needs to have the necessary Mahalo
     * class files in its classpath and the necessary Mahalo client
     * classes files in its codebase.
     * <p>
     * Note that when the servicer is later destroyed, it will destroy
     * its activation group, regardless of whether or not there are
     * other activatable objects still registered in the same group.
     * It will also delete its persistence directory.
     * <p>
     * The server will not bind itself into a directory service until
     * configure though its administration interface.  The service's
     * administration interface can be obtained by casting it to a
     * <code>net.jini.admin.Administrable</code> and calling
     * <code>getAdmin()</code>.	 The returned value will be a
     * reference to the service's admin interface.
     * <p>
     *
     * @see net.jini.admin.Administrable
     */
    private static LeaseRenewalService create(ActivationGroupID gid,
					      String logDirName) 
	throws IOException, ActivationException
    {
	// We are activatable
	// Check path for validity, and create directory
	com.sun.jini.system.FileSystem.makeNewDir(logDirName);

	// Tell the activation system about us
	ActivationDesc actDesc =
	    new ActivationDesc(gid, NormServerImpl.class.getName(), null,
			       new MarshalledObject(logDirName), true);

	// Register with activation and finish setting up the binder
	NormServer ourRef = (NormServer)Activatable.register(actDesc);

	PersistentStore.setupStore(logDirName, ourRef);
	return ourRef;
    }

    /**
     * Start up a new lease renewal service. See Norm release notes for
     * more details on how system properties are interpreted.
     * The command line parameters are:
     * <pre>
     * codebase policyfile dbdir [group,...] [javaprog] [javaoptions]
     * </pre>
     *
     * @param codebase	  required argument containing the location with which
     *			  the renewal service's remote classes will be
     *			  annotated; indicating the location from which those
     *			  classes should be downloaded. This argument is
     *			  passed to the activation system's group descriptor
     *			  as the java.rmi.server.codebase property so that this
     *			  property will be set in the VM created during
     *			  activation of the lookup service. This codebase
     *			  string must be a sequence of properly formatted URLs;
     *			  where each URL must end with ".jar", ".zip", or "/".
     *
     * @param policyfile  required argument containing the name (with optional
     *			  path) of the file containing the security policies
     *			  for the renewal service. This argument is passed to
     *			  the activation system's group descriptor as the
     *			  java.security.policy property so that this
     *			  property will be set in the VM created during
     *			  activation of the lookup service.
     *
     * @param dbdir	  required argument containing the name (with optional
     *			  path) of the directory in which the renewal service's
     *			  state will be recorded for persistence
     *
     * @param group(s)	  optional argument(s) containing the list of multicast
     *			  discovery groups that the renewal should look in for
     *			  lookup servers in for lookup servers to register with.
     *			  A group name equal to "none" will result in this
     *			  renewal service not using multicast discovery; a group
     *			  name of "public" will result in this renewal service 
     *			  registering with lookups belonging to the public 
     *			  group.  The default is "public".
     *
     * @param javaprog	  optional argument containing the name of a file
     *			  - including the full path - that will be used as
     *			  the 'cmdpath' argument in the creation of the
     *			  CommandEnvironment object; which specifies startup
     *			  options for the activation group. This argument can
     *			  be used to override default system properties and
     *			  to specify implementation-defined options for the
     *			  activation group. Upon activation of the renewal
     *			  service, the activation system will start the
     *			  renewal service by executing the file represented
     *			  by this argument. Although this file can take any
     *			  form - a script, a java executable, etc. - typically
     *			  it is simply a different version of the java
     *			  command than that used to execute the main()
     *			  method of NormServerImpl. In this way, one can
     *			  activate the renewal service with a different
     *			  version of java than the current version.
     *
     * @param javaoptions optional list of argument(s), each beginning with
     *			  a '-', that will be passed into the VM created
     *			  when the renewal service is activated. These
     *			  arguments can be any arguments understood by
     *			  the java command (or the javaprog if the java
     *			  command was overridden). These arguments can be
     *			  used to set system properties in the renewal service.
     *			  For example, a very useful property that can be set
     *			  using these arguments is the property that turns
     *			  off the just-in-time compiler (JIT). To turn off
     *			  the JIT, simply input the string "-Djava.compiler="
     *			  as one of these arguments.
     */
    public static void main(String[] args) {
	final BasicServiceDesc desc = ParseUtils.persistentParse(args);
	if (desc == null) System.exit(-1);

	System.setSecurityManager(new RMISecurityManager());
	try {
	    final LeaseRenewalService srv = create(desc);

	    // Config
	    final JoinAdmin admin = (JoinAdmin)((Administrable)srv).getAdmin();
	    admin.setLookupGroups(desc.getGroups());
	} catch (Throwable e) {
	    System.err.println("NormServerImpl: main: unable to start " +
			       "LeaseRenewalService");
	    e.printStackTrace();
	    System.exit(-1);
	}
    }
}
