/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.OutputStream;
import java.io.ObjectOutput;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.ObjectInputStream;

import java.util.List;
import java.util.Iterator;

import com.sun.jini.thread.ReadersWriter;
import com.sun.jini.reliableLog.ReliableLog;
import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.system.FileSystem;

import com.sun.jini.norm.lookup.SubStore;

/**
 * Class that actually stores a Norm server's state to disk.  Basicly
 * a wrapper around ReliableLog with the addition of lock management.
 */
class PersistentStore {
    /** file name we write the proxy into the log dir under */
    private final static String PROXY_FILE_NAME = "NormServerStub";

    /** Object we use to reliable and penitently log updates to our state */
    private ReliableLog log;

    /**
     * No mutation of persistent state can occur during a snapshot,
     * however, we can have multiple mutators, use
     * <code>ReadersWriter</code> object to manage this invariant.
     * Note as far as the lock is concerned the mutators are the
     * readers and snapshot thread (the reader) since for us mutation
     * is the non-exclusive operation
     */
    final private ReadersWriter mutatorLock = new ReadersWriter();

    /**
     * Thread local that tracks if (and how many times) the current thread
     * has acquired a non-exclusive mutator lock.
     */
    final static private ThreadLocal lockState = new ThreadLocal();

    /** Cache a <code>Long</code> object with a zero value */
    final static private Long zero = new Long(0);

    /** Location of the persistent store */
    final private File storeLocation;

    /** Object that handles the recovery of logs  */
    final private LogHandler logHandler;

    /** The NormServer we are part of */
    final private NormServerImpl server;

    /** Number of updates since last snapshot */
    private int updateCount;

    /** A list of all of the substores */
    private List subStores = new java.util.LinkedList();

    /**
     * Construct a store that will persist its data to the specified
     * directory.
     * @param logDir Directory where the store should persist its data.
     *               must exist.
     * @param svrRef Service's remote stub, used to check if 
     *               this is our log, store does not hold on this reference
     * @param logHandler Object that will process the log and last snapshot
     *               to recover the server's state
     * @param server the server is called back after an update so it can
     *               decide whether or not to do a snapshot.
     * @throws StoreException if there is a problem setting up the store
     */
    PersistentStore(String logDir, Object srvRef, LogHandler logHandler,
		    NormServerImpl server)
	throws StoreException
    {
	this.logHandler = logHandler;
	this.server = server;
	storeLocation = new File(logDir);

	// Ensure that we are the only one using this log
	Object lrs = null;
	final File stubFile = new File(storeLocation, PROXY_FILE_NAME);

	FileInputStream istream = null;
	ObjectInputStream oistream = null;

	try {
	    istream = new FileInputStream(stubFile);
	    oistream = new ObjectInputStream(istream);
	    lrs = oistream.readObject();

	    // $$$ should we try to close first up here so we don't
	    // miss any exceptions on close?
	} catch (IOException e) {
	    throw new CorruptedStoreException("Persistent store corrupted:" +
	        "Can't read remote stub out of " + stubFile, e);
	} catch (ClassNotFoundException e) {
	    throw new CorruptedStoreException("Persistent store corrupted" +
		"object in " + stubFile + " is of an unexpected class", e);
	} catch (ClassCastException e) {
	    throw new CorruptedStoreException("Persistent store corrupted" +
		"object in " + stubFile + " is of an unexpected class", e);
	} finally {
	    safeClose(istream, oistream);
	}

	if (!lrs.equals(srvRef)) 
	    throw new StoreConflictException("Another server is using this " +
	        "log (" + logDir + ")");

	try {
	    log = new ReliableLog(storeLocation.getCanonicalPath(), logHandler);
	} catch (IOException e) {
	    throw new CorruptedStoreException("Failure creating reliable log",
					      e);
	}

	try {
	    log.recover();
	} catch (IOException e) {
	    throw new CorruptedStoreException("Failure recoving reliable log",
					      e);	    
	}
    }

    /**
     * Method used by server setup code to initialize the store on disk
     * @param logDir Name of directory to place the persistent store in
     * @param ref    Stub to server, used as interlock to prevent multiple
     *               servers from using the same log directory.
     * @throws IOException if the store can not be initialized.
     */
    static void setupStore(String logDir, Object ref) throws IOException {
	// Write proxy into log dir for use as interlock	
	FileOutputStream ostream = null;
	ObjectOutputStream oostream = null;

	try {
	    ostream = new FileOutputStream(new File(logDir, PROXY_FILE_NAME));
	    oostream = new ObjectOutputStream(ostream);
	    oostream.writeObject(ref);
	    oostream.flush();

	    // $$$ should we try to close first up here so we don't
	    // miss any exceptions on close?
	} finally {
	    safeClose(ostream, oostream);
	}
    }

    /**
     * Destroy the store
     * @throws IOException if it has difficulty removing the log files 
     */
    void destroy() throws IOException {
	// Prep all the sub-stores to be destroyed
	for (Iterator i = subStores.iterator(); i.hasNext(); ) {
	    SubStore subStore = (SubStore)i.next();
	    subStore.prepareDestroy();
	}

	log.deletePersistentStore();
	FileSystem.destroy(storeLocation, true);
    }

    /**
     * Inform the store of a sub-store
     */
    void addSubStore(SubStore subStore) throws StoreException {
	try {
	    final String subDir = subStore.subDirectory();

	    if (subDir == null) {
		subStore.setDirectory(storeLocation);
	    } else {
		subStore.setDirectory(new File(storeLocation, subDir));
	    }
	    
	    subStores.add(subStore);
	} catch (IOException e) {
	    throw new StoreException("Failure adding substore " + subStore,
				     e);

	}
    }

    ///////////////////////////////////////////////////////////////
    // Utility methods for closing files after an IOException

    /**
     * Utility method to close a file after an
     * <code>IOException</code>, note we swallow any
     * <code>IOException<code> because it is assumed the exception
     * that triggered the call is more important.  
     * @param s stream to close, may be <code>null</code>
     * @param os an optional object stream that wraps <code>s</code>
     *           <code>null</code>
     */
    static private void safeClose(InputStream s, ObjectInput os) {
	try {
	    if (os != null) {
		s.close();
		return;
	    }
	} catch (IOException e) {
	    safeClose(s);
	}

	safeClose(s);
    }

    /**
     * Utility method to close a file after an
     * <code>IOException</code>, note we swallow any
     * <code>IOException<code> because it is assumed the exception
     * that triggered the call is more important.  
     * @param s stream to close, may be <code>null</code>
     */
    static private void safeClose(InputStream s) {
	try {
	    if (s != null)
		s.close();
	} catch (IOException ie) {
	    // ignore, not catastrophic, if we are in this method
	    // then we are already cleaning up from an IOException
	}
    }

    /**
     * Utility method to close a file after an
     * <code>IOException</code>, note we swallow any
     * <code>IOException<code> because it is assumed the exception
     * that triggered the call is more important.  
     * @param s stream to close, may be <code>null</code>
     * @param os an optional object stream that wraps <code>s</code>
     *           <code>null</code>
     */
    static private void safeClose(OutputStream s, ObjectOutput os) {
	try {
	    if (os != null) {
		s.close();
		return;
	    }
	} catch (IOException e) {
	    safeClose(s);
	}

	safeClose(s);
    }

    /**
     * Utility method to close a file after an
     * <code>IOException</code>, note we swallow any
     * <code>IOException<code> because it is assumed the exception
     * that triggered the call is more important.  
     * @param s stream to close, may be <code>null</code>
     */
    static private void safeClose(OutputStream s) {
	try {
	    if (s != null)
		s.close();
	} catch (IOException ie) {
	    // ignore, not catastrophic, if we are in this method
	    // then we are already cleaning up from an IOException
	}
    }

    /////////////////////////////////////////////////////////////////
    // Methods for obtaining and releasing the locks on the store

    /**
     * Block until we can acquire a non-exclusive mutator lock on the
     * servers 's persistent state.  This lock should be acquired in a
     * <code>try</code> block and a <code>releaseMuatorLock</code>
     * call should be placed in a <code>finally</code> block.
     */
    void acquireMuatorLock() {
	// Using read lock because we want a non-exclusive lock
	mutatorLock.readLock();

	Long lockStateVal = (Long)lockState.get();
	if (lockStateVal == null) 
	    lockStateVal = zero;

	lockState.set(new Long(lockStateVal.longValue() + 1));
    }

    /**
     * Release one level of  mutator locks if this thread holds at lease one.
     */
    void releaseMuatorLock() {
	Long lockStateVal = (Long)lockState.get();
	if (lockStateVal == null) 
	    lockStateVal = zero;

	final long longVal = lockStateVal.longValue();
	if (longVal == 0)
	    // No lock to release, return
	    return;

	if (longVal == 1) {
	    // Last one on stack release lock
	    // Using read lock because we want a non-exclusive lock
	    mutatorLock.readUnlock();
	    lockStateVal = zero;
	} else {
	    lockStateVal = new Long(longVal - 1);
	}

	lockState.set(lockStateVal);
    }

    //////////////////////////////////////////////////////////////////
    // Methods for writing records to the log and taking and
    // coordinating snapshots

    /**
     * Log an update. Will flush to disk before returning.
     * @throws IllegalStateException if the current thread does not hold
     * a non-exclusive mutator lock
     * @throws IOException see @see com.sun.jini.reliableLog.ReliableLog#update
     */
    void update(Object o) {
	final Long lockStateVal = (Long)lockState.get();
	if (lockStateVal == null || lockStateVal.longValue() == 0) 
	    throw new IllegalStateException("PersistentStrore.update:" +
	        "Must acquire mutator lock before calling update()");

	synchronized (this) { 
	    try {
		log.update(o, true);
		updateCount++;
		server.updatePerformed(updateCount);
	    } catch (IOException e) {
		// $$$ should proably be propagating this exception
		System.err.println("IOException while updating log");
		e.printStackTrace();
	    }
	}
    }

    /**
     * Generate a snapshot, will perform the necessary locking to ensure no
     * threads are mutating the state of the server before creating the 
     * snapshot.
     * @throws IOException see 
     * @see com.sun.jini.reliableLog.ReliableLog#snapshot
     */
    void snapshot() throws IOException {
	try {
	    // Using write lock because we want an exclusive lock
	    mutatorLock.writeLock();
	    updateCount = 0;

	    // Don't need to sync on this because
	    // mutatorLock.writeLock() gives us an exclusive lock
	    log.snapshot();
	} finally {
	    // Using write lock because we want an exclusive lock
	    mutatorLock.writeUnlock();
	}
    }
}

