/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm;

import java.io.Serializable;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;

import net.jini.core.lease.Lease;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.lease.LeaseRenewalSet;

import com.sun.jini.proxy.UUID;
import com.sun.jini.proxy.ThrowThis;

/**
 * Client side proxy for Norm's lease renwal sets.  Used an object of
 * type NormServer to communicate back to server.
 */
class SetProxy implements LeaseRenewalSet, Serializable {
    /** 
     * ID of the set this object is a proxy for
     * @serial
     */
    final private UUID setID;

    /** 
     * Lease for this set
     * @serial
     */
    final private Lease ourLease;
     
    /**
     * Remote ref to server we belong to 
     * @serial
     */
    final private NormServer server;


    /**
     * Simple constructor
     * @param server object proxy should call back to
     * @param id ID of set this proxy is for
     * @param lease on for set this is proxy is for
     */
    SetProxy(NormServer server, UUID id, Lease lease) {
	this.server = server;
	setID = id;
	ourLease = lease;
    }

    // Inherit java doc from super type
    public void renewFor(Lease leaseToRenew, long membershipDuration)
	throws RemoteException 
    {
	// Do some client side error checking
	if (leaseToRenew == null) 
	    throw new NullPointerException("LeaseRenewalSet.renewFor:Must " +
					   "pass a non-null lease");

	if (!(membershipDuration == Lease.ANY ||
	      membershipDuration == Lease.FOREVER ||
	      membershipDuration > 0))
	    throw new IllegalArgumentException(
	        "LeaseRenewalSet.renewFor:Requested membership " +
		"duration must be Lease.ANY, Lease.FOREVER, or positive");

	// It would be nice to do the check here on leaseToRenew being
	// granted by this set, however, because we are trying to keep
	// LandlordLease out of the norm.jar we would then also have
	// to keep SetProxy out of norm.jar which could be a real
	// hassle, so for now we will do the check on the server side.

	try {
	    server.renewFor(setID, leaseToRenew, membershipDuration);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public Lease remove(Lease leaseToRemove) throws RemoteException {
	// Do some client side error checking
	if (leaseToRemove == null) 
	    throw new NullPointerException("LeaseRenewalSet.remove:Must " +
	        "pass a non-null lease");

	try {
	    return server.remove(setID, leaseToRemove);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }


    // Inherit java doc from super type
    public EventRegistration setExpirationWarningListener(
	                         RemoteEventListener listener, 
				 long                minWarning, 
				 MarshalledObject    handback)
	throws RemoteException    
    {
	// Do some client side error checking
	if (listener == null) 
	    throw new NullPointerException(
	        "LeaseRenewalSet.setExpirationWarningListener:Must " +
		"pass a non-null listener");

	if (minWarning < 0)
	    throw new IllegalArgumentException(
	        "LeaseRenewalSet.setExpirationWarningListener:minWarning " +
		"must be positive");

	try {
	    return server.setExpirationWarningListener(setID,
	        listener, minWarning, handback);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }

    // Inherit java doc from super type
    public void clearExpirationWarningListener() throws RemoteException {
	try {
	    server.setExpirationWarningListener(setID, null, 
						NormServer.NO_LISTENER, null);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public EventRegistration setRenewalFailureListener(
				 RemoteEventListener listener, 
				 MarshalledObject    handback)
	throws RemoteException
    {
	// Do some client side error checking
	if (listener == null) 
	    throw new NullPointerException(
	        "LeaseRenewalSet.setRenewalFailureListener:Must " +
		"pass a non-null listener");

	try {
	    return server.setRenewalFailureListener(setID, listener, handback);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }


    // Inherit java doc from super type
    public void clearRenewalFailureListener() throws RemoteException {
	try {
	    server.setRenewalFailureListener(setID, null, null);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public Lease getRenewalSetLease() {
	return ourLease;
    }

    // Inherit java doc from super type
    public int hashCode() {
	return setID.hashCode();
    }

    // Inherit java doc from super type
    public boolean equals(Object obj) {
	if (!(obj instanceof SetProxy))
	    return false;

	final SetProxy that = (SetProxy)obj;

	return (setID.equals(that.setID) && server.equals(that.server));
    }
}
