/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm.event;

import java.io.IOException;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.ServerError;
import java.util.List;

import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.UnknownEventException;

import com.sun.jini.thread.TaskManager;
import com.sun.jini.thread.RetryTask;


/**
 * Representation of an event type the supports a single registrant.
 * The registrant, event ID, and event sequence number information
 * portions of this object are preserved when serialized.  The send
 * monitor, and currently scheduled event sends are not.
 */
public class EventType implements java.io.Serializable {
    /**
     * Listener registered for events of this type.  Stored in 
     * marshaled form so object the <code>EventType</code> object can
     * be recovered even if the listener's codebase is not available.
     * If this field is <code>null</code>, <code>listener</code> and 
     * <code>handback</code> will be also.  
     * @serial
     */
    private MarshalledObject marshaledListener;

    /** Transient cache of listener in unmarshaled form */
    private transient RemoteEventListener listener;

    /**
     * Handback object associated with current listener.
     * @serial
     */
    private MarshalledObject handback;

    /**
     * Last event sequence number we used
     * @serial
     */
    private long lastSeqNum;

    /**
     * Our event ID
     * @serial
     */
    private long evID;

    /** 
     * Object we check with to ensure leases have not expired and notify
     * when we get a definite exception during an event send attempt	
     * If this field is <code>null</code> generator will be also
     */
    private transient SendMonitor monitor;

    /** 
     * Event type generator that created us. If this field is
     * <code>null</code> monitor will be also.
     */
    private transient EventTypeGenerator generator;

    /**
     * Simple constructor.  Initially the last sequence number is set to 0.
     * @param generator EventTypeGenerator that is creating this event type
     * @param monitor   Object which is to monitor the sending of events
     *                  of this type.
     * @param evID      Event ID of this event type
     * @param listener  The listener events of this type should go to
     * @param handback  The object that should be passed to listener
     *                  as part of the event.     
     * @throws IOException if listener can not be serialized 
     */
    EventType(EventTypeGenerator generator, SendMonitor monitor, long evID,
	      RemoteEventListener listener, MarshalledObject handback)
        throws IOException
    {
	if (generator == null)
	    throw new NullPointerException("EventType(): Must create event " +
                "type objects with a non-null generator");

	if (monitor == null)
	    throw new NullPointerException("EventType(): Must create event " +
                "type objects with a non-null monitor");

	this.generator = generator;
	this.monitor = monitor;
	this.evID = evID;
	setLastSequenceNumber(0);
	setListener(listener, handback);
    }    

    /** Utility method to null out listener */
    private void clearListener() {
	this.listener = null;
	this.handback = null;
	marshaledListener = null;	    
    }
       

    /**
     * (Re)set the listener for this type of event.  Any pending
     * events that have not yet been sent will be sent to the new
     * listener, passing the new handback.  Setting the listener to
     * <code>null</code> will cancel the sending of all pending
     * events.  
     * @param listener The listener events of this type should go to
     * @param handback The object that should be passed to listener
     *                 as part of the event.     
     * @throws IOException if listener can not be serialized */
    public synchronized void setListener(RemoteEventListener listener, 
					 MarshalledObject    handback)
        throws IOException
    {
	if (listener == null) {
	    clearListener();
	} else {	    
	    marshaledListener = new MarshalledObject(listener);
	    this.listener = listener;
	    this.handback = handback;
	}
    }

    /**
     * Returns <code>true</code> if there is a listener registered for this 
     * event type.
     */
    public boolean haveListener() {
	return marshaledListener != null;
    }

    /**
     * Convince method to get the listener.
     * @return The listener or <code>null</code> if the listener can't be
     * unpacked or there is no listener.
     */
    private RemoteEventListener getListener() {
	if (!haveListener()) 
	    return null;

	if (listener != null) 
	    return listener;

	// There is a listener, but it is not unpacked yet, try to unpack
	try {
	    listener = (RemoteEventListener)marshaledListener.get();
	} catch (IOException e) {
	    // swallow and return null bellow

	    // $$$ is this really the right thing to do?
	    // we proably really have a corrupted marshaledListener here
	} catch (ClassNotFoundException e) {
	    // swallow and return null bellow
	}

	// If get() threw an exception listener is still null
	return listener;
    }

    /**
     * Atomically replace both the listener and handback registered to
     * receive events of this type with <code>null</code> iff they are
     * both equal to the past listener and handback objects
     * respectively.  If the replacement occurs it will have the
     * same effect as calling <code>setListener(null, null)</code>.
     * <p>
     * Can be used by code that needs to remove event registrations in
     * response to exceptions thrown during event delivery without
     * risking clobbering new registrations.
     * @param oldListener That should be replaced by <code>null</code>
     * @param oldHandback That should be replaced by <code>null</code>
     * @return <code>true</code> if the state of the object was
     * changed and <code>false</code> otherwise.
     * @see EventType#setListener */
    public synchronized boolean clearListenerIfEquals(
        RemoteEventListener oldListener, MarshalledObject oldHandback) 
    {
	if (!haveListener()) {
	    // no one is registered so listener is already cleared
	    // just return false (no change);
	    return false;
	}

	final RemoteEventListener l = getListener();

	if (l == null) {
	    // If l is null then we must have a listener but can't unpack it.
	    // If it was equal to oldListener we would be able to unpack it,
	    // therefor l != oldListener, no change, return false;
	    return false;
	}	    

	if (l.equals(oldListener) && handback.equals(oldHandback)) {
	    // both match, replace both with null
	    clearListener();
	}

	// One or both did not match, no change
	return false;
    }

    /**
     * Set the object's notion of the last sequence number.  The next event
     * scheduled to be sent will have a sequence number one greater than
     * the value past to this call.
     * <p>
     * Note: this method is not synchronized.
     * @param seqNum value for the last sequence number
     */
    public void setLastSequenceNumber(long seqNum) {
	lastSeqNum = seqNum;
    }

    /**
     * Return the sequence number of the last event that was scheduled
     * to be sent.  Intended primarily for creating
     * <code>EventRegistration</code> objects and the like.
     */
    public synchronized long getLastSequenceNumber () {
	return lastSeqNum;
    }
    
    /**
     * Return the <code>long</code> that was uniquely associated with this
     * object when it was created.
     */
    public long getEventID() {
	return evID;
    }

    /**
     * Schedule the sending of an event.  This event will be sent to
     * the currently registered listener.  If the listener changes
     * before the the event is successfully sent the event will be sent
     * to the new listener.  If the currently listener is
     * <code>null</code> this call will have no affect.
     * @param factory An object that will be used to create the
     * <code>Event</code> object when necessary.
     * @return The sequence number assigned to the event, if the
     * current listener is <code>null</code> this will be the same
     * value as returned by the last call to this method.
     * @throws IllegalStateException if this method is called
     *         after the object has be deserialized and before
     *         <code>restoreTransientState</code> has been called
     * @see EventType#restoreTransientState
     */
    public synchronized long sendEvent(EventFactory factory) {
	if (generator == null) {
	    // Have not had our state restored, complain
	    throw new IllegalStateException("EventType.sendEvent:" +
	        "called before state was fully restored");
	}

	// If we don't have a listener we don't do anything
	if (!haveListener())
	    return lastSeqNum;
	
	final TaskManager mgr = generator.getTaskManager();
	mgr.add(new SendTask(mgr, factory, ++lastSeqNum));
	
	return lastSeqNum;
    }

    /**
     * Restore the transient state of this object after recovering it
     * from a serialization stream.  Nether argument can be
     * <code>null</code>.
     * <p>
     * Note: this method is not synchronized.  
     * @param generator The <code>EventTypeGenerator</code> that was used
     *                  to create this EventType object originally.
     * @param monitor   The object that monitors the progress of events
     *                  set by this object.  */
    public void restoreTransientState(EventTypeGenerator generator,  
                                      SendMonitor        monitor)
    {
	if (generator == null)
	    throw new NullPointerException("EventType.restoreTransientState:" +
	        "Must called with a non-null generator");

	if (monitor == null)
	    throw new NullPointerException("EventType.restoreTransientState:" +
	        "Must called with a non-null monitor");

	this.generator = generator;
	this.monitor   = monitor;
	generator.recoverEventID(evID);
    }

    /**
     * Sub-class of <code>RetryTask</code> used by <code>EventType</code>
     * to send events
     */
    private class SendTask extends RetryTask {
	/** Max time we are willing to let a send attempt to go on for */
	final static private long MAX_TIME = 1000 * 60 * 60 * 24; //~ 1 Day
	
	/** Factory used to create the <code>RemoteEvent</code> to be sent */
	final private EventFactory eventFactory;

	/** Sequence number the event should have */
	final private long seqNum;

	/**
	 * Simple constructor
	 * @param taskManager <code>TaskManager</code> this task is to be
	 *                    put into.
	 * @param eventFactory <code>EventFactory</code> that will be used
	 *                     to create the event to be sent.
	 * @param seqNum      The sequence number of the event
	 */
	private SendTask(TaskManager taskManager, EventFactory eventFactory,
			 long seqNum)
	{
	    super(taskManager);
	    this.eventFactory = eventFactory;
	    this.seqNum = seqNum;
	}

	// Inherit java doc from super type
	public boolean tryOnce() {
	    final long now = System.currentTimeMillis();
	    if (now - startTime() > MAX_TIME)
		return true;	// we have been trying too long, stop here

	    if (!EventType.this.monitor.isCurrent())
		return true;	// lease gone, stop here

	    // Local copies of listener and handback so they won't
	    // be clobbered by setListener calls
	    RemoteEventListener listener;
	    MarshalledObject handback;
	    
	    synchronized (EventType.this) {
		if (!EventType.this.haveListener())
		    return true; // No currently registered listener, stop here

		listener = EventType.this.getListener();
		if (listener == null)
		    return false; // There is a listener, but we can't unpack it
		                  // schedule a re-try latter

		handback = EventType.this.handback;
	    }

	    // $$$ We could have EventType keep a version # on the 
	    // handback and only create a new event if the version number
	    // on the hand back has changed.
	    final RemoteEvent event =
		eventFactory.createEvent(EventType.this.evID, seqNum, handback);
	    
	    // Try sending 
	    try {
		listener.notify(event);
		return true;
	    } catch (NoSuchObjectException e) {
		// This is a definite exception tell the monitor and stop
		EventType.this.monitor.definiteException(EventType.this,
							 listener, event, e);
		return true;
	    } catch (ServerError e) {
		// This is a definite exception tell the monitor and stop
		EventType.this.monitor.definiteException(EventType.this,
							 listener, event, e);
		return true;
	    } catch (RemoteException e) {
		// This is not a definite exception, try again latter
		return false;
	    } catch (UnknownEventException e) {
		// This is a definite exception tell the monitor and stop
		EventType.this.monitor.definiteException(EventType.this,
							 listener, event, e);
		return true;
	    } catch (RuntimeException e) {
		// This is a definite exception tell the monitor and stop
		EventType.this.monitor.definiteException(EventType.this,
							 listener, event, e);
		return true;
	    }
	}

	// Inherit java doc from super type
	public boolean runAfter(List tasks, int size) {
	    // We don't need to run these tasks in any particular order
	    return false;
	}
    }
}
