/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.norm.event;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.rmi.MarshalledObject;
import net.jini.core.event.RemoteEventListener;
import com.sun.jini.thread.TaskManager;

/**
 * Factory class for <code>EventType</code> objects.  All
 * <code>EventType</code> objects created by the same generator (or
 * associated with the same generator by a
 * <code>EventType.restoreTransientState<code> call) will use the same
 * thread pool to manage their event send threads.
 *
 * @see EventType 
 * @see EventType.restoreTransientState
 */
public class EventTypeGenerator implements java.io.Serializable {
    /**
     * Next event ID
     * @serial
     */
    private long nextEvID = 1;

    /**
     * Task manager used to send events
     */
    private transient TaskManager taskManager = new TaskManager();

    /**
     * Create a new <code>EventType</code> object.
     * @param listener the listener initially associated with the
     *                 event type. May be <code>null</code>.
     * @param handback The handback object initially associated with
     *                 the event type. May be <code>null</code>.
     * @param monitor  Object to callback when an event sending
     *        attempt fails with a definite exception and to 
     *        ensure that the lease on the event is still current.
     *        May not be <code>null</code>
     * @return The new <code>EventType</code> object.
     * @throws IOException if listener can not be serialized 
     */
    public synchronized EventType newEventType(RemoteEventListener listener, 
					       MarshalledObject    handback,
					       SendMonitor         monitor)
	throws IOException
    {
	EventType rslt = new EventType(this, monitor, nextEvID, 
				       listener, handback);
	// Do this afterward so we don't use up an ID if 
	// we can't create the new object
	nextEvID++;
	return rslt;
    }

    /**
     * Called by event types during transient state recovery to ensure
     * the generator knows about there event ID
     * <p>
     * Note: this method is not synchronized.  
     * @param evID Event ID of recovered <code>EventType</code> object
     */
    void recoverEventID(long evID) {
	if (evID >= nextEvID)
	    nextEvID = evID + 1;
    }

    /**
     * Return the task manager that <code>EventType</code> objects created
     * by this generator should use to send their events.
     */
    TaskManager getTaskManager() {
	return taskManager;
    }

    /**
     * Override <code>readObject</code> to create a <code>TaskManager</code>
     * @see java.io.ObjectInputStream#defaultReadObject docs
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException
    {
	// fill in the object from the stream 
	in.defaultReadObject();

	taskManager = new TaskManager();
    }
}
