/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.discovery;

import net.jini.discovery.*;
import net.jini.core.discovery.*;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.IOException;
import java.util.*;

/** 
 * This class is a helper utility class that organizes and manages all
 * discovery-related activities on behalf of a client or service. Rather
 * than providing its own facility for coordinating and maintaining
 * all of the necessary state information related to group names,
 * <code>LookupLocator</code> objects, and <code>DiscoveryListener</code>
 * objects, clients and services can employ this class to provide those
 * facilities on their behalf.
 *
 * @see net.jini.DiscoveryManagement
 * @see net.jini.DiscoveryGroupManagement
 * @see net.jini.DiscoveryLocatorManagement
 * 
 * @author Zane Pan
 */
public class LookupDiscoveryManager implements DiscoveryManagement,
                                               DiscoveryGroupManagement,
                                               DiscoveryLocatorManagement
{
    final public static int FROM_GROUP = 1;
    final public static int FROM_LOCATOR = 2;
    
    private final ArrayList listeners = new ArrayList(1);
    private final ArrayList discoveredSet = new ArrayList();
    private LookupLocatorDiscovery locatorDisc;
    private LookupDiscovery lookupDisc;
    private DiscMgrListener locatorlistener = 
                            new DiscMgrListener(FROM_LOCATOR);
    private DiscMgrListener grouplistener = 
                            new DiscMgrListener(FROM_GROUP);

    private final class ProxyReg  {
	public ServiceRegistrar proxy;
	private int from;
	private boolean bDiscarded = false;
	public ProxyReg(ServiceRegistrar proxy, int from) {
	    if(proxy == null)
		throw new IllegalArgumentException(
					 "proxy cannot be null");
	    this.proxy = proxy;
	    this.from = from;
	}	    

	public boolean equals(Object obj) {
	    if (obj instanceof ProxyReg){
		return proxy.equals(((ProxyReg)obj).proxy);
	    } else return false;
	}

	public int hashCode() {
	    return proxy.hashCode();
	}
	
	public synchronized void addFrom(int from) {
	    this.from |= from;
	}

	public synchronized boolean removeFrom(int from) {
	    if( from == 0 )
		    throw new RuntimeException
			("call removeFrom with bad argument");
	    this.from &= ~from;
	    return (this.from == 0);
	}

	public synchronized int getFrom() {
	    return this.from;
	}

	public void discard() {
	    //System.out.println("failed, call discard");
	    synchronized(this) {
		bDiscarded = true;
	    }
	    if((from & FROM_GROUP) != 0)
		lookupDisc.discard(proxy);

	    if((from & FROM_LOCATOR) != 0)
		locatorDisc.discard(proxy);
	}

	public synchronized boolean isDiscarded() {
	    return bDiscarded;
	}
    }

    private final class DiscMgrListener implements DiscoveryListener {
	private int from;
	public DiscMgrListener(int from) {
	    this.from = from;
	}

	public void discovered(DiscoveryEvent e) {
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    ArrayList newProxys = new ArrayList(1);
	    for(int i=0; i<proxys.length; i++) {
		 ProxyReg reg = findReg(proxys[i]);
		 if(reg == null) {
		     reg = new ProxyReg(proxys[i], from);
		     addDiscoveredSet(reg);
		     newProxys.add(proxys[i]);
		 } else
		     reg.addFrom(from);
	    }
	    notifyListener(newProxys, true);
	}

	public void discarded(DiscoveryEvent e) {
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    ArrayList drops = new ArrayList(1);
	    for(int i=0; i<proxys.length; i++) {
		 ProxyReg reg = findReg(proxys[i]);
		 if(reg != null) {
		     if(removeDiscoveredSet(reg))
			 drops.add(reg.proxy);
		 }
	    }
	    notifyListener(drops, false);
	}
	
	private void addDiscoveredSet(ProxyReg reg) {
	    synchronized(discoveredSet) {
		discoveredSet.add(reg);
	    }
	}

	private boolean removeDiscoveredSet(ProxyReg reg) {
	    boolean bret = reg.removeFrom(from);
	    if(bret) {
		synchronized(discoveredSet) {
		    discoveredSet.remove(reg);
		}
	    }
	    return bret;
	}
    }

    /** 
     * Constructs an instance of this class that will organize and manage 
     * all discovery-related activities on behalf of the client or service
     * that instantiates this class.
     * <p>
     * If <code>null</code> (<code>DiscoveryGroupManagement.ALL_GROUPS</code>)
     * is input to the <code>groups</code> parameter, then attempts will be
     * made via group discovery to discover all lookup services located within
     * range of the entity that constructs this class. If the empty array
     * (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input to that
     * parameter, no group discovery will be performed until the set of
     * groups to discover is populated.
     * <p>
     * If an empty array or a <code>null</code> reference is input to the
     * <code>locators</code> parameter, no locator discovery will be performed
     * until the set of locators to discover is populated.
     *
     * @param groups   <code>String</code> array, none of whose elements may
     *                 be <code>null</code>, consisting of the names of the
     *                 groups whose members are lookup services the client
     *                 or service wishes to discover.
     * @param locators array of instances of <code>LookupLocator</code>, none
     *                 of whose elements may be <code>null</code>, and in
     *                 which each element corresponds to a specific lookup
     *                 service the client or service wishes to discover via
     *                 locator discovery.
     * @param listener a reference to <code>DiscoveryListener</code> object
     *                 that will be notified when a targeted lookup service
     *                 is discovered or discarded.
     *
     * @throws java.io.IOException because construction of this class may
     *         initiate the discovery process, which can throw an
     *         <code>IOException</code> when socket allocation occurs.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either one or more of the elements of the <code>groups</code>
     *         parameter is <code>null</code>, or one or more elements of
     *         the <code>locators</code> parameter is null.
     *
     * @see net.jini.core.discovery.LookupLocator
     * @see net.jini.discovery.DiscoveryListener
     */
    public LookupDiscoveryManager(String[] groups,  
				  LookupLocator[] locators,
				  DiscoveryListener listener)
                                                           throws IOException
    {
        /* Configure for group discovery */
 	lookupDisc = new LookupDiscovery(groups);
	lookupDisc.addDiscoveryListener(grouplistener);
        /* Configure for locator discovery */
	locatorDisc = new LookupLocatorDiscovery(locators);
	locatorDisc.addDiscoveryListener(locatorlistener);
        /* Initiate the discovery process */
	if(listener != null)
	    listeners.add(listener);
    }

    /**
     * Returns an array consisting of the elements of the managed set
     * of locators; that is, instances of <code>LookupLocator</code> in
     * which each instance corresponds to a specific lookup service to
     * discover. The returned set will include both the set of 
     * <code>LookupLocator</code>s corresponding to lookup services 
     * that have already been discovered as well as the set of those
     * that have not yet been discovered. If the managed set of locators
     * is empty, this method will return the empty array. This method
     * returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         of the managed set of locators
     *
     * @see net.jini.DiscoveryLocatorManagement#getLocators
     */
    public LookupLocator[] getLocators(){
	return locatorDisc.getLocators();
    }

    /**
     * Adds a set of locators to the managed set of locators. Elements in the
     * input set that duplicate (using the <code>LookupLocator.equals</code>
     * method) elements already in the managed set will be ignored. If the
     * empty array is input, the managed set of locators will not change.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators to add to the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators to augment.
     *         That is, the current managed set of locators is
     *         <code>null</code>.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#addLocators
     */
    public void addLocators(LookupLocator[] locators){
	locatorDisc.addLocators(locators);
    }

    /**
     * Deletes a set of locators from the managed set of locators, and discards
     * any already-discovered lookup service that corresponds to a deleted
     * locator. For any lookup service that is discarded as a result of an
     * invocation of this method, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, this method takes no action.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators that will be removed from the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators from which
     *         remove elements.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#removeLocators
     */
    public void removeLocators(LookupLocator[] locators) {
	locatorDisc.removeLocators(locators);
    }

    /**
     * Replaces all of the locators in the managed set with locators from
     * a new set, and discards any already-discovered lookup service that
     * corresponds to a locator that is removed from the managed set
     * as a result of an invocation of this method. For any such lookup
     * service that is discarded, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, locator discovery will cease until this
     * method is invoked with an input parameter that is non-<code>null</code>
     * and non-empty.
     *
     * @param locators <code>LookupLocator</code> array consisting of the 
     *                 locators that will replace the current locators in the
     *                 managed set.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#setLocators
     */
    public void setLocators(LookupLocator[] locators) {
	locatorDisc.setLocators(locators);

    }
    
    /** 
     * Returns an array consisting of the elements of the managed set
     * of groups; that is, the names of the groups whose members are the
     * lookup services to discover. If the managed set of groups is empty,
     * this method will return the empty array. If there is no managed set
     * of groups, then null is returned; indicating that all groups are to
     * be discovered. This method returns a new array upon each invocation.
     *
     * @return <code>String</code> array consisting of the elements of the
     *         managed set of groups
     *
     * @see net.jini.DiscoveryGroupManagement#getGroups
     */
    public String[] getGroups() {
	return lookupDisc.getGroups();
    }

    /**   
     * Adds a set of group names to the managed set of groups. Elements in
     * the input set that duplicate elements already in the managed set
     * will be ignored. Once a new name is added to the managed set,
     * attempts will be made to discover all (as yet) undiscovered lookup
     * services that are members of the group having that name. If the empty
     * array (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input, the
     * managed set of groups will not change.
     *
     * Note that any entity that invokes this method must have
     * <code>DiscoveryPermission</code> on each of the groups in the
     * new set, otherwise a <code>SecurityException</code> will be
     * propagated through this method.
     *
     * @param groups <code>String</code> array consisting of the group names
     *               to add to the managed set.
     *
     * @throws java.io.IOException because an invocation of this method may
     *         result in the re-initiation of the discovery process, which can
     *         throw an <code>IOException</code> when socket allocation occurs.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of groups to augement.
     *         That is, the current managed set of groups is <code>null</code>.
     *         If the managed set of groups is <code>null</code>, all groups
     *         are being discovered; thus, requesting that a set of groups be
     *         added to the set of all groups makes no sense.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>groups</code>
     *         parameter, or one or more of the elements of the
     *         <code>groups</code> parameter is <code>null</code>. If a
     *         <code>null</code> is input, then the entity is effectively
     *         requesting that "all groups" be added to the current managed
     *         set of groups; which is not allowed. (Note that if the entity
     *         wishes to change the managed set of groups from a finite set
     *         of names to "all groups", the <code>setGroups</code> method
     *         should be invoked with <code>null</code> input.)
     *
     * @see net.jini.DiscoveryGroupManagement#addGroups
     */
    public void addGroups(String[] groups) throws IOException {
	lookupDisc.addGroups(groups);
    }

    /**   
     * Deletes a set of group names from the managed set of groups. If the
     * empty array (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input,
     * this method takes no action.
     *
     * @param groups <code>String</code> array consisting of the group names
     *               that will be removed from the managed set.
     *
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of groups from which
     *         remove elements.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>groups</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryGroupManagement#removeGroups
     */
    public void removeGroups(String[] groups) {
	lookupDisc.removeGroups(groups);
    }

    /**   
     * Replaces all of the group names in the managed set with names from
     * a new set. Once a new group name has been placed in the managed
     * set, if there are lookup services belonging to that group that have
     * already been discovered, no event will be sent to the entity's
     * listener for those particular lookup services. Attempts to discover
     * all (as yet) undiscovered lookup services belonging to that group
     * will continue to be made.
     * <p>
     * If null (<code>DiscoveryGroupManagement.ALL_GROUPS</code>) is input
     * to this method, then attempts will be made to discover all (as yet)
     * undiscovered lookup services that are within range, and which
     * are members of any group. If the empty array
     * (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input, then
     * group discovery will cease until this method is invoked with an
     * input parameter that is non-<code>null</code> and non-empty.
     *
     * Note that any entity that invokes this method must have
     * <code>DiscoveryPermission</code> on each of the groups in the
     * new set, otherwise a <code>SecurityException</code> will be
     * propagated through this method.
     *
     * @param groups <code>String</code> array consisting of the group
     *               names that will replace the current names in the
     *               managed set.
     *
     * @throws java.io.IOException because an invocation of this method may
     *         result in the re-initiation of the discovery process, which can
     *         throw an <code>IOException</code> when socket allocation occurs.
     *
     * @see net.jini.DiscoveryGroupManagement#setGroups
     */
    public void setGroups(String[] groups) throws IOException {
	lookupDisc.setGroups(groups);
    }

    /**
     * Adds an instance of <code>DiscoveryListener</code> to the set of
     * objects listening for discovery events. Once the listener is
     * registered, it will be notified of all lookup services discovered
     * to date, and will then be notified as new lookup services are
     * discovered or existing lookup services are discarded.
     * <p>
     * If <code>null</code> is input, this method takes no action. If the
     * listener input to this method duplicates (using the <code>equals</code>
     * method) another element in the current set of listeners, no action
     * is taken.
     *
     * @param l an instance of <code>DiscoveryListener</code> corresponding
     *          to the listener to add to the set of listeners.
     *
     * @see net.jini.DiscoveryListener
     * @see net.jini.DiscoveryManagement#addDiscoveryListener
     */
    public void addDiscoveryListener(DiscoveryListener l) {
	ArrayList set;
	synchronized(listeners) {
	    if(!listeners.contains(l)) {
		listeners.add(l);
	    }
	    if(discoveredSet.isEmpty())
		return;
	    set = (ArrayList)discoveredSet.clone();
	}
	notifyListener(l, set, true);
    }

    /**
     * Removes a listener from the set of objects listening for discovery
     * events. If the listener object input to this method does not exist
     * in the set of listeners, then this method will take no action.
     *
     * @param l an instance of <code>DiscoveryListener</code> corresponding
     *          to the listener to remove from the set of listeners.
     *
     * @see net.jini.DiscoveryListener
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public void removeDiscoveryListener(DiscoveryListener l) {
	synchronized(listeners) {
	    listeners.remove(l);
	}
    }

    /**
     * Returns an array of instances of <code>ServiceRegistrar</code>, each
     * corresponding to a proxy to one of the currently discovered lookup
     * services. For each invocation of this method, a new array is returned.
     *
     * @return array of instances of <code>ServiceRegistrar</code>, each
     *         corresponding to a proxy to one of the currently discovered
     *         lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public ServiceRegistrar[] getRegistrars() {
	//checkTerminated();
	ArrayList proxySet = new ArrayList(1);;
	synchronized(discoveredSet) {
	    int k = 0;
	    Iterator iter = discoveredSet.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg = (ProxyReg)iter.next();
		if(!reg.isDiscarded())
		    proxySet.add(reg.proxy);
	    }
	}
	ServiceRegistrar[] ret = 
		new ServiceRegistrar[proxySet.size()];
	
	proxySet.toArray(ret);
	return ret;
    }

    /**
     * Removes an instance of <code>ServiceRegistrar</code> from the
     * managed set of lookup services, making the corresponding lookup
     * service eligible for re-discovery. This method takes no action if
     * the parameter input to this method is <code>null</code>, or if it
     * does not match (using <code>equals</code>) any of the elements in
     * the managed set.
     *
     * @param proxy the instance of <code>ServiceRegistrar</code> to discard
     *              from the managed set of lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#discard
     */
    public void discard(ServiceRegistrar proxy) {
	ProxyReg reg = findReg(proxy);
	if(reg != null)
	    reg.discard();
    }

    /**
     * Terminates all threads, ending all discovery processing being
     * performed by the current instance of this class.
     * <p>
     * After this method has been invoked, no new lookup services will
     * be discovered, and the effect of any new operations performed
     * on the current instance of this class are undefined.
     *
     * @see net.jini.DiscoveryManagement#terminate
     */
    public void terminate() {
	synchronized(listeners) {
	    listeners.clear();
	}
	lookupDisc.terminate();
	locatorDisc.terminate();
    }

    /* return where the proxy come from. */
    public int getFrom(ServiceRegistrar proxy) {
	ProxyReg reg = findReg(proxy);
	if(reg != null)
	    return reg.getFrom();
	return 0;
    }

    private ProxyReg findReg(ServiceRegistrar proxy) {
	synchronized(discoveredSet) {
	    Iterator iter = discoveredSet.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg =(ProxyReg)iter.next();
		if(reg.proxy.equals(proxy))
		    return reg;
	    }
	}
	return null;
    }

    private void notifyListener(ArrayList proxySet, boolean bDiscoverd) {
	if(proxySet.isEmpty())
	    return;
	ArrayList notifies;
	synchronized(listeners) {
	    notifies = (ArrayList)listeners.clone();
	}
	    
	Iterator iter = notifies.iterator();
	while(iter.hasNext()) {
	    DiscoveryListener l = (DiscoveryListener)iter.next();
	    notifyListener(l, proxySet, bDiscoverd);
	}
    }
    
    private void notifyListener(DiscoveryListener l, 
				ArrayList proxySet,
				boolean bDiscoverd) {
	ServiceRegistrar[] proxys = new ServiceRegistrar[proxySet.size()];
	proxySet.toArray(proxys);
	DiscoveryEvent evt =  
	    new DiscoveryEvent(LookupDiscoveryManager.this,proxys);
	if(bDiscoverd)
	    l.discovered(evt);
	else
	    l.discarded(evt);
    }

}
