/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */
package net.jini.discovery;


import net.jini.discovery.DiscoveryListener;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.lease.*;
import net.jini.core.entry.Entry;
import net.jini.discovery.*;
import net.jini.core.lookup.*;

import java.io.IOException;
import java.lang.reflect.*;
import java.util.*;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;

/**
 * This class encapsulates the functionality required of an entity that
 * wishes to employ the unicast discovery protocol to discover a lookup
 * service. This utility provides an implementation that makes the process
 * of finding specific lookup services much simpler for both services and
 * clients.
 * <p>
 * Because this class participates in only the unicast discovery protocol,
 * and because the unicast discovery protocol imposes no restriction on the
 * physical location of the entity relative to a lookup service, this utility
 * can be used to discover lookup services running on hosts that are located
 * far from, or near to, the host on which the entity is running. This lack
 * of a restriction on location brings with it a requirement that the
 * discovering entity supply this class with specific information about the
 * desired lookup services; namely, the location of the device(s) hosting
 * each lookup service. This information is supplied through an instance
 * of the <code>LookupLocator</code> class.
 *
 * @author Zane Pan
 */
public class LookupLocatorDiscovery implements DiscoveryManagement,
                                               DiscoveryLocatorManagement
{
    private static class LocatorReg implements Comparable {
	public ServiceRegistrar proxy = null;
	public final LookupLocator l;
	/** 
	 * Since cnt only gets modified in the constructor, we therefore
	 * do not need to synchronize on it.
	 */
	private static final long MINI_RETRY = 15000;
	private static final long[] sleepTime = {0,5*1000, 10*1000, 20*1000,
	                                         30*1000, 60*1000, 
						 2*60*1000, 4*60*1000,
						 8*60*1000, 15*60*1000};
	private static int cnt = 0;
	public long nextTryTime;
	private int numFails = 0;
	private final int id;
	private long time = 0;
	public LocatorReg(LookupLocator l) {
	    id = cnt++;
	    this.l = l;
	    calcNextTryTime();
	}
	
	public void calcNextTryTime() {
	    nextTryTime =  System.currentTimeMillis() + sleepTime[numFails];
	    if(numFails < sleepTime.length -1)
		numFails++;
	}
	
	public boolean equals(Object obj) {
	    return l.equals(((LocatorReg)obj).l);
	}

	
	public int hashCode() {
	    return l.hashCode();
	}
       
	public int compareTo(Object obj) {
	    if (equals(obj)) {
		return 0;
	    }
	    LocatorReg e = (LocatorReg)obj;
	    if(nextTryTime == e.nextTryTime ) {
		if(id < e.id )
		    return 1;
		else
		    return -1;
	    }
	    if (nextTryTime < e.nextTryTime)
		return 1;
	    return -1;
	}

	public boolean tryGetProxy() {
	    if (proxy != null ) {
		throw new IllegalArgumentException
		    ("LookupLocator has been discovered already");

	    }
	    try {
		proxy = l.getRegistrar();
		time =  System.currentTimeMillis();
		return true;
	    } catch (Exception e) {
		//System.out.println("fail************");
		calcNextTryTime();
		return false;
	    }
	}

       /**
	* This method only gets called from discard(). It tries to 
	* prevent the discard method from being called too frequently. It
	* measures the time from when discovered() gets called to when
	* discard() gets called. If the time is less than MINI_RETRY,
	* we delay our next discover.
	*/
	public void fixupNextTryTime()  {
	    if( (System.currentTimeMillis() - time) > MINI_RETRY ) {
		numFails = 0;
	    }
	    
	    //System.out.println("numFails:" + numFails + "   time:" +
	    //		      (System.currentTimeMillis() - time) );
	} 
	

    }
 
    /**
     * Stores LookupLocators that have not been discovered yet. It is
     * sorted on the nextTryTime. The last key has the least value.
     */ 
    private final TreeMap locatorsMap = new TreeMap();
    /** Stores LookupLocators that have been discovered */
    private final ArrayList discoveredLocators = new ArrayList(1);
    /**
     * Thread to keep trying to find proxies from LookupLocators in
     * the TreeMap. The thread terminates if the TreeMap is empty,
     * and gets created if the TreeMap is not empty.
     */
    private DiscoveryThread discoveryThread = null;
    /** Thread that handles pending notifications. */
    private Notifier notifierThread;
    /** Notifications to be sent to listeners. */
    private LinkedList pendingNotifies = new LinkedList();
    /** Stores DiscoveryListeners **/
    private final ArrayList listeners = new ArrayList(1);

    /** Store the LocatorReg during RMI call */
    private LocatorReg regInRMICall = null;

    /**
     * Create a LookupLocatorDiscovery, with an initial array of
     * LookupLocators to be managed.  For each managed LookupLocator,
     * unicast discovery is performed to obtain a ServiceRegistrar
     * proxy for that lookup service.
     * 
     * @param locators the locators to discover
     *
     */
    public LookupLocatorDiscovery(LookupLocator[] locators) {
	synchronized(this) {
	    discoverLocators(locators);
	}
    }

    /**
     * Add a DiscoveryListener to the listener set. The listener's
     * discovered method gets called right way with an array of 
     * ServiceRegistrars that have already been discoved, and will
     * be called in the future whenever additional lookup services
     * are discovered.
     *
     * @param l the new DiscoveryListener to add
     *
     */
    public void addDiscoveryListener(DiscoveryListener l) {
	synchronized(this) {
	    if(listeners.contains(l))
		return;
	    listeners.add(l);
	    if(!discoveredLocators.isEmpty()) {
		ArrayList list = new ArrayList(1);
		list.add(l);
		addNotify(list, buildServiceRegistrar(), false);
	    }
	}
    }


    /**
     * Remove a DiscoveryListener from the listener set. It does
     * nothing if the DiscoveryListener does not exist in the 
     * the listener set.
     *
     * @param l the existing DiscoveryListener to remove
     *
     */
    public synchronized void removeDiscoveryListener(DiscoveryListener l) {
	int index = listeners.indexOf(l);
	if(index != -1)
	    listeners.remove(index);
    }

    /**
     * Removes an instance of <code>ServiceRegistrar</code> from the
     * managed set of lookup services, making the corresponding lookup
     * service eligible for re-discovery. This method takes no action if
     * the parameter input to this method is <code>null</code>, or if it
     * does not match (using <code>equals</code>) any of the elements in
     * the managed set.
     *
     * @param proxy the instance of <code>ServiceRegistrar</code> to discard
     *              from the managed set of lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#discard
     */
    public void discard(ServiceRegistrar proxy) {
	synchronized(this) {
	    LookupLocator lct = findRegFromProxy(proxy);
	    if(lct == null)
		return;
	    LocatorReg reg = removeDiscoveredLocator(lct);
	    reg.proxy = null;
	    reg.fixupNextTryTime();
	    reg.calcNextTryTime();
	    addToMap(reg);
	    if(!listeners.isEmpty()) {
		ArrayList drops = new ArrayList(1);
		drops.add(proxy);
		notifyDropped(drops);
	    }
	}
    }

    /**
     * Returns an array consisting of the elements of the managed set
     * of locators; that is, instances of <code>LookupLocator</code> in
     * which each instance corresponds to a specific lookup service to
     * discover. The returned set will include both the set of 
     * <code>LookupLocator</code>s corresponding to lookup services 
     * that have already been discovered as well as the set of those
     * that have not yet been discovered. If the managed set of locators
     * is empty, this method will return the empty array. This method
     * returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         of the managed set of locators
     *
     * @see net.jini.DiscoveryLocatorManagement#getLocators
     */
    public synchronized LookupLocator[] getLocators() {
	int size = discoveredLocators.size() + locatorsMap.size();
	if(regInRMICall != null )
	    size +=1;

	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) 
	    ret[k++] = ((LocatorReg)iter.next()).l;

	iter = locatorsMap.values().iterator();
	while(iter.hasNext())
	    ret[k++] = ((LocatorReg)iter.next()).l;

	if(regInRMICall != null )
	    ret[k] = regInRMICall.l;

	return ret;
    }

    /**
     * Adds a set of locators to the managed set of locators. Elements in the
     * input set that duplicate (using the <code>LookupLocator.equals</code>
     * method) elements already in the managed set will be ignored. If the
     * empty array is input, the managed set of locators will not change.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators to add to the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators to augment.
     *         That is, the current managed set of locators is
     *         <code>null</code>.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#addLocators
     */
    public synchronized void addLocators(LookupLocator[] locators) {
	discoverLocators(locators);
    }

    /**
     * Deletes a set of locators from the managed set of locators, and discards
     * any already-discovered lookup service that corresponds to a deleted
     * locator. For any lookup service that is discarded as a result of an
     * invocation of this method, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, this method takes no action.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators that will be removed from the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators from which
     *         remove elements.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#removeLocators
     */
    public void removeLocators(LookupLocator[] locators) {
	synchronized(this) {
	    ArrayList drops = new ArrayList(1);
	    for(int i=0; i<locators.length; i++) {
		LocatorReg reg = removeDiscoveredLocator(locators[i]);
		if(reg != null) {
		    drops.add(reg.proxy);
		    continue;
		}
		reg = findReg(locators[i]);
		if(reg != null) {
		    removeLocatorReg(reg);
		    notifyAll();
		}
	    }
	    if(!drops.isEmpty() && !listeners.isEmpty())
		notifyDropped(drops);
	}
    }

    private void removeLocatorReg(LocatorReg reg) {
	if(reg == regInRMICall )
	    regInRMICall = null;
	else
	    locatorsMap.remove(reg);
    }

     /**
     * Replaces all of the locators in the managed set with locators from
     * a new set, and discards any already-discovered lookup service that
     * corresponds to a locator that is removed from the managed set
     * as a result of an invocation of this method. For any such lookup
     * service that is discarded, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, locator discovery will cease until this
     * method is invoked with an input parameter that is non-<code>null</code>
     * and non-empty.
     *
     * @param locators <code>LookupLocator</code> array consisting of the 
     *                 locators that will replace the current locators in the
     *                 managed set.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#setLocators
     */
   public void setLocators(LookupLocator[] locators) {
	synchronized(this) {
	    ArrayList drops = new ArrayList(1);
	    Iterator iter = discoveredLocators.iterator();
	    while(iter.hasNext()) {
		LocatorReg reg = (LocatorReg)iter.next();
		if(!isArrayContains(locators, reg.l)) {
		    iter.remove();
		    drops.add(reg.proxy);
		}
	    }
	    iter = locatorsMap.values().iterator();
	    while(iter.hasNext()) {
		LocatorReg reg = (LocatorReg)iter.next();
		if(!isArrayContains(locators, reg.l)) {
		    iter.remove();
		}
	    }
	    if(regInRMICall != null) {
	       if(!isArrayContains(locators, regInRMICall.l))
		   regInRMICall = null;
	    }
	    discoverLocators(locators);
	    if(!drops.isEmpty() && !listeners.isEmpty())
		notifyDropped(drops);
	}
    }

    /**
     * Terminates all threads, ending all locator discovery processing being
     * performed by the current instance of this class.
     * <p>
     * After this method has been invoked, no new lookup services will
     * be discovered, and the effect of any new operations performed
     * on the current instance of this class are undefined.
     *
     * @see net.jini.DiscoveryManagement#terminate
     */
    public synchronized void terminate() {
	if(discoveryThread != null )
	    discoveryThread.interrupt();
	regInRMICall = null;
    }

    private ServiceRegistrar[] buildServiceRegistrar() {
	int k = 0;
	ServiceRegistrar[] proxys =
	    new ServiceRegistrar[discoveredLocators.size()];
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    proxys[k++] = reg.proxy;
	}
	return proxys;
    }

    private void discoverLocators(LookupLocator[] lcts) {
	if ( lcts == null)
	    return;
	LookupLocator lct;
	for(int i=0; i<lcts.length; i++) {
	    if(isDiscovered(lcts[i]))
		continue;
	    LocatorReg reg = findReg(lcts[i]);
	    if(reg == null) {
		reg = new LocatorReg(lcts[i]);
		addToMap(reg);
	    }
	}
    }

    private LocatorReg findReg(LookupLocator lct) {
	if(checkRegInRMICall(lct))
	    return regInRMICall;

	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if (reg.l.equals(lct))
		return reg;
	}
	return null;
    }

    private boolean checkRegInRMICall(LookupLocator lct) {
	if (regInRMICall == null )
	    return false;
	return regInRMICall.l.equals(lct);
    }

    private class DiscoveryThread extends Thread {

	public DiscoveryThread() {
	    super("DiscoveryThread");
	    setDaemon(true);
	}

        public void run() {
	    try {
		while (!isInterrupted()) {
		    synchronized (LookupLocatorDiscovery.this) {
			if(locatorsMap.isEmpty()) {
			    discoveryThread = null;
			    return;
			}
			LocatorReg reg = (LocatorReg)locatorsMap.lastKey();
			long tryTime = reg.nextTryTime; 
			long now = System.currentTimeMillis();
			if(tryTime > now ) {
			    //System.out.println("#sleep:." + (tryTime-now));
			    LookupLocatorDiscovery.this.wait(tryTime-now);
			    continue;
			}
			regInRMICall = reg;
			locatorsMap.remove(reg);
		    }
		    regTryGetProxy(regInRMICall);
		}
	    } catch (InterruptedException ex) {
	    }
	}
    }

    private void regTryGetProxy(LocatorReg reg) {
	//System.out.println("regTryGetProxy:" + reg.l.getHost() + 
	//		   "  id:" + reg.id);
	boolean b = reg.tryGetProxy();
	synchronized (this) {
	    if(regInRMICall == null )
		return;
	    regInRMICall = null;
	    if(!b) {
		addToMap(reg);
		return;
	    }
	    discoveredLocators.add(reg);
	    if(!listeners.isEmpty())
		addNotify((ArrayList)listeners.clone(),
			  new ServiceRegistrar[]{reg.proxy}, false);
	}
    }

    private void addToMap(LocatorReg reg) {
	locatorsMap.put(reg, reg);
	if(discoveryThread == null ) {
	   discoveryThread =  new DiscoveryThread();
	   discoveryThread.start();
	} else
	    notifyAll();
    }

    private boolean isDiscovered(LookupLocator lct) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
 	    LocatorReg reg = (LocatorReg)iter.next();
	    if(reg.l.equals(lct))
		return true;
	}
	return false;
    }

    private LocatorReg removeDiscoveredLocator(LookupLocator lct) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if(reg.l.equals(lct)) {
		iter.remove();
		return reg;
	    }
	}
	return null;
    }

    private void notifyDropped(ArrayList drops) {
	ServiceRegistrar[] proxys = new ServiceRegistrar[drops.size()];
	drops.toArray(proxys);
	addNotify((ArrayList)listeners.clone(), proxys, true);
    }

    /**
     * Add a notification task to the pending queue, and start a
     * notifierThread if one isn't running.
     */
    private void addNotify(ArrayList notifies,
			   ServiceRegistrar[] regs,
			   boolean discard)
    {
	synchronized (pendingNotifies) {
	    pendingNotifies.addLast(new NotifyTask(notifies, regs, discard));
	    if (notifierThread == null) {
		notifierThread = new Notifier();
		notifierThread.start();
	    }
	}
    }

    /** Simple holder for notify task data */
    private static class NotifyTask {
	/** The listeners to notify */
	public final ArrayList listeners;
	/** The proxies that have been discovered or discarded */
	public final ServiceRegistrar[] regs;
	/** True if discarded, else discovered */
	public final boolean discard;

	public NotifyTask(ArrayList listeners,
			  ServiceRegistrar[] regs,
			  boolean discard)
	{
	    this.listeners = listeners;
	    this.regs = regs;
	    this.discard = discard;
	}
    }

    /** Task thread for calling listeners with notifications */
    private class Notifier extends Thread {

	/** Create a daemon thread */
	public Notifier() {
	    super("event notifier");
	    setDaemon(true);
	}

	public void run() {
	    while (true) {
		NotifyTask task;
		synchronized (pendingNotifies) {
		    if (pendingNotifies.isEmpty()) {
			notifierThread = null;
			return;
		    }
		    task = (NotifyTask)pendingNotifies.removeFirst();
		}
		for (Iterator iter = task.listeners.iterator();
		     iter.hasNext(); )
		{
		    DiscoveryEvent e =
			new DiscoveryEvent(LookupLocatorDiscovery.this,
					   (ServiceRegistrar[])task.regs.clone());
		    DiscoveryListener l = (DiscoveryListener)iter.next();
		    if (task.discard)
			l.discarded(e);
		    else
			l.discovered(e);
		}
	    }
	}
    }

    private boolean isArrayContains(Object[] a, Object obj) {
	for(int i=0; i<a.length; i++ ) {
	    if(a[i].equals(obj))
		return true;
	}
	return false;
    }

    private LookupLocator findRegFromProxy(ServiceRegistrar proxy) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if((reg.proxy).equals(proxy ))
		return reg.l;
	}
	return null;    
    }
    
    /* debug function */
    private void printMap () {
	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();	    
	    System.out.println("printMap reg:" + reg.id);
	}
    }

    /**
     * Returns the set of <code>LookupLocator</code> objects representing the
     * desired lookup services that are currently discovered. If no lookup
     * services are currently discovered, this method returns the empty array.
     * This method returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         from the managed set of locators that correspond to lookup
     *         services that have already been discovered.
     */
    public synchronized LookupLocator[] getDiscoveredLocators() {
	int size = discoveredLocators.size();
	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) 
	    ret[k++] = ((LocatorReg)iter.next()).l;
	return ret;
    }

    /**
     * Returns the set of <code>LookupLocator</code> objects representing the
     * desired lookup services that have not yet been discovered. If all
     * of the desired lookup services are currently discovered, this method
     * returns the empty array. This method returns a new array upon each
     * invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         from the managed set of locators that correspond to lookup
     *         services that have not yet been discovered.
     */
    public synchronized LookupLocator[] getUndiscoveredLocators() {
	int size = locatorsMap.size();
	if(regInRMICall != null )
	    size +=1;
	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext())
	    ret[k++] = ((LocatorReg)iter.next()).l;
	if(regInRMICall != null )
	    ret[k] = regInRMICall.l;
	return ret;
    }

    /**
     * Returns an array of instances of <code>ServiceRegistrar</code>, each
     * corresponding to a proxy to one of the currently discovered lookup
     * services. For each invocation of this method, a new array is returned.
     *
     * @return array of instances of <code>ServiceRegistrar</code>, each
     *         corresponding to a proxy to one of the currently discovered
     *         lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public ServiceRegistrar[] getRegistrars() {
        synchronized(this) {
            if((discoveredLocators == null) || (discoveredLocators.isEmpty())){
                return new ServiceRegistrar[0];
            }
            return buildServiceRegistrar();
        }
    }
}
