/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.discovery;

import net.jini.core.lookup.ServiceRegistrar;

import java.rmi.MarshalledObject;

/**
 * When unmarshalling an instance of MarshalledObject, one of the following 
 * checked exceptions is possible: an IOException can occur while 
 * deserializing the object from its internal representation; and a 
 * ClassNotFoundException can occur if, while deserializing the object 
 * from its internal representation, either the class file of the object 
 * cannot be found, or the class file of either an interface or a class 
 * referenced by the object being deserialized cannot be found. 
 * Typically, a ClassNotFoundException occurs when the codebase from 
 * which to retrieve the needed class file is not currently available.
 * <p>
 * This class provides a mechanism that clients of the lookup discovery 
 * service may use for efficient handling of the exceptions that may 
 * occur when unmarshalling elements of a set of marshalled instances 
 * of the ServiceRegistrar interface. When elements in such a set are 
 * unmarshalled, the LookupUnmarshalException class may be used to 
 * collect and report pertinent information generated when failure 
 * occurs while unmarshalling the elements of the set.
 * <p>
 * The information that may be of interest to entities that receive this
 * exception class is contained in the following fields: a set of
 * ServiceRegistrar instances in which each element is the result of 
 * a successful unmarshalling attempt, a set of marshalled instances of 
 * ServiceRegistrar in which each element could not be successfully 
 * unmarshalled, and a set of exceptions (IOException, ClassNotFoundException,
 * or some unchecked exception) in which each element corresponds to one
 * of the unmarshalling failures.
 * <p>
 * Thus, when exceptional conditions occur while unmarshalling a set of 
 * marshalled instances of ServiceRegistrar, this class can be used not 
 * only to indicate that an exceptional condition has occurred, but also 
 * to provide information that can be used to perform error handling 
 * activities such as: determining if it is feasible to continue with 
 * processing, reporting errors, attempting recovery, and debugging.
 * <p>
 * Note that this exception class should be used only to report exceptional
 * conditions occurring when unmarshalling a set of marshalled instances 
 * of the ServiceRegistrar interface.
 *
 * @see net.jini.core.lookup.ServiceRegistrar
 */
public class LookupUnmarshalException extends Exception {

    private static final long serialVersionUID = 2956893184719950537L;

    /**
     * Array containing the set of instances of ServiceRegistrar that were
     * successfully unmarshalled during the process in which at least one
     * failure occurred.
     *
     * @serial
     */
    private ServiceRegistrar[] unmarshalledRegs = null;

    /**
     * Array containing the set of ServiceRegistrar instances that could not
     * be unmarshalled. This set should contain at least one element.
     *
     * @serial
     */
    private MarshalledObject[] stillMarshalledRegs = null;

    /**
     * Array containing the set of exceptions that occurred during the 
     * unmarshalling process. Each element in this set should be
     * an instance of IOException, ClassNotFoundException, or some
     * unchecked exception. Furthermore, there should be a one-to-one
     * correspondence between each element in this set and each element
     * in the set of still-to-be-unmarshalled ServiceRegistrar instances.
     * That is, the element of this set corresponding to index i should
     * be an instance of the exception that occurred while attempting to 
     * unmarshal the element at index i of stillMarshalledRegs.
     *
     * @serial
     */
    private Throwable[] exceptions = null;

    /** String describing the nature of the exception.
     *  <p>
     *  Since this exception should be thrown only as a result of a failure
     *  while unmarshalling a set of marshalled instances of ServiceRegistrar,
     *  the description String is fixed to one value.
     *
     * @serial
     */
    private static String descStr
                        = "failed to unmarshal at least one ServiceRegistrar";
    /**
     * Constructs a new instance of LookupUnmarshalException.
     *
     * @param unmarshalledRegs    Array containing the set of instances of
     *                            ServiceRegistrar that were successfully
     *                            unmarshalled.
     *                            
     * @param stillMarshalledRegs Array containing the set of marshalled
     *                            ServiceRegistrar instances that could not
     *                            be unmarshalled.
     *                   
     * @param exceptions          Array containing the set of exceptions that
     *                            occurred during the unmarshalling process.
     *                            Each element in this set should be an
     *                            instance of IOException,
     *                            ClassNotFoundException, or some unchecked
     *                            exception. Furthermore, there should be a
     *                            one-to-one correspondence between each
     *                            element in this set and each element in
     *                            the stillMarshalledRegs parameter.
     * <p>
     *                            That is, the element of this set
     *                            corresponding to index i should be an
     *                            instance of the exception that occurred
     *                            while attempting to unmarshal the element
     *                            at index i of the stillMarshalledRegs 
     *                            parameter.
     *
     * @throws java.lang.IllegalArgumentException this exception occurs
     *         when the number of elements in the exceptions parameter is
     *         not equal to the number of elements in the stillMarshalledRegs
     *         parameter.
     */
    public LookupUnmarshalException(ServiceRegistrar[] unmarshalledRegs,
                                    MarshalledObject[] stillMarshalledRegs,
                                    Throwable[] exceptions) 
    {
        super(descStr);
        /* If the number of exceptions does not equal number of registrars
         * that could not be unmarshalled, throw exception
         */
        if(exceptions.length != stillMarshalledRegs.length) {
            throw new IllegalArgumentException
                           ("exceptions.length ("+exceptions.length
                            +") is not equal to stillMarshalledRegs.length"
                            +stillMarshalledRegs.length+")");
        }
        this.unmarshalledRegs = unmarshalledRegs;
        this.stillMarshalledRegs = stillMarshalledRegs;
        this.exceptions = exceptions;
    }

    /**
     * Accessor method that returns an array consisting of instances of 
     * ServiceRegistrar, where each element of the array corresponds to a 
     * successfully unmarshalled object. Note that the same array is
     * returned on each invocation of this method; that is, a copy is
     * not made.
     *
     * @return array of instances of ServiceRegistrar, where each element
     *         corresponds to a successfully unmarshalled object.
     */
    public ServiceRegistrar[] getUnmarshalledRegs() {
        return unmarshalledRegs;
    }

    /**
     * Accessor method that returns an array consisting of instances of 
     * MarshalledObject, where each element of the array is a marshalled 
     * instance of the ServiceRegistrar interface, and corresponds to an 
     * object that could not be successfully unmarshalled. Note that the 
     * same array is returned on each invocation of this method; that is,
     * a copy is not made.
     *
     * @return array of marshalled instances of ServiceRegistrar, where
     *         each element corresponds to an object in which failure
     *         occurred while attempting to unmarshal the object.
     */
    public MarshalledObject[] getStillMarshalledRegs() {
        return stillMarshalledRegs;
    }

    /**
     * Accessor method that returns an array consisting of instances of 
     * Throwable, where each element of the array corresponds to one of 
     * the exceptions that occurred during the unmarshalling process.
     * Note that the same array is returned on each invocation of this
     * method; that is, a copy is not made.
     * <p>
     * Each element in the return set should be an instance of IOException, 
     * ClassNotFoundException, or some unchecked exception. Additionally, 
     * there should be a one-to-one correspondence between each element in 
     * the array returned by this method and the array returned by the 
     * getStillMarshalledRegs method. That is, the i-th element of the set 
     * returned by this method should be an instance of the exception that 
     * occurred while attempting to unmarshal the i-th element of the set 
     * returned by the getStillMarshalledRegs method.
     *
     * @return array of instances of Throwable, where each element
     *         corresponds to one of the exceptions that occurred during
     *         the unmarshalling process.
     */
    public Throwable[] getExceptions() {
        return exceptions;
    }

    /**
     * Accessor method that returns the String describing the nature of
     * the exception. Note that the same String is returned on each
     * invocation of this method; that is, a copy is not made.
     *
     * @return String object describing the nature of the exception.
     */
    public String getMessage() {
        return super.getMessage();
    }
}
