/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.lease;

import net.jini.core.lease.*;

/**
 * With respect to an entity that uses the <code>LeaseRenewalManager</code>
 * to manage leases granted to the entity, this class defines the event
 * that is sent by the <code>LeaseRenewalManager</code> to the entity's
 * registered listener when the <code>LeaseRenewalManager</code> fails to
 * renew one of the leases it is managing for the entity. Such failures
 * typically occur because: the expiration time of the lease has been
 * reached (and the lease has already expired) before the renewal request
 * can be made; the renewal request has been denied by the lease grantor,
 * even even though the desired absolute expiration time of the lease has
 * not yet been reached; or the renewal request could not be completed
 * before the expiration of the lease.
 * <p>
 * This class encapsulates information about both the lease on which the
 * failure occurred, as well as information about the condition that 
 * caused the renewal attempt to fail.
 * <p>
 * This class is a subclass of the class <code>EventObject</code>, adding
 * the following additional items of abstract state: a reference to the
 * affected lease; a long value representing the desired absolute expiration
 * of the affected lease; and the exception (if any) that caused the event
 * to be sent. In addition to the methods of the <code>EventObject</code>
 * class, this class defines methods through which this additional state
 * may be retrieved.
 * 
 * @see net.jini.core.Lease
 * @see net.jini.lease.LeaseRenewalManager
 * @see net.jini.lease.LeaseListener
 * @see java.util.EventObject
 */
public class LeaseRenewalEvent extends java.util.EventObject {
    /**
     * The failed lease.
     *
     * @serial
     */
    private Lease lease;
    /**
     * The desired expiration of the failed lease.
     *
     * @serial
     */
    private long expiration;
    /**
     * The exception that caused the failure, if any.
     *
     * @serial
     */
    private Exception ex;

    /** 
     * Constructs an instance of this class that encapsulates the information
     * the <code>LeaseRenewalManager</code> wishes to associate with the
     * affected lease.
     *
     * @param source     reference to the instance of the
     *                   <code>LeaseRenewalManager</code> that generated the
     *                   event
     * @param lease      the lease to which the event pertains
     * @param expiration the desired (absolute) time of expiration for the
     *                   affected lease
     * @param ex         an instance of the exception (if any) that caused
     *                   the event to be sent
     *
     * @see net.jini.core.lease.Lease
     * @see net.jini.lease.LeaseRenewalManager
     */
    public LeaseRenewalEvent(LeaseRenewalManager source,
			     Lease lease,
			     long expiration,
			     Exception ex) 
    {
	super(source);
	this.lease = lease;
	this.expiration = expiration;
	this.ex = ex;
    }

    private static final long serialVersionUID = -7536467680016825016L;

    /** 
     * Returns a reference to the lease to which the event pertains.
     *
     * @return the <code>Lease</code> object corresponding to the lease
     *         on which the renewal attempt failed
     *
     * @see net.jini.core.lease.Lease
     */
    public Lease getLease() {
	return lease;

    }

    /** 
     * Returns the desired absolute expiration time of the lease to which
     * event pertains.
     *
     * @return a <code>long</code> value that represents the desired absolute
     *         expiration time of the lease on which the renewal attempt
     *         failed
     */
    public long getExpiration() {
	return expiration;
    }

    /** 
     * Returns the exception (if any) that caused the event to be sent.
     * The conditions under which the event may be sent, and the related
     * values returned by this method, are as follows:
     * <ul>
     * <li> if the lease's expiration time is reached before the
     *      <code>LeaseRenewalManager</code> can request renewal, a
     *      <code>LeaseRenewalEvent</code> will be sent with no associated
     *      exception. In this case, invoking this method will return
     *      <code>null</code>.
     * <li> if the <code>LeaseRenewalManager</code> sends a
     *      <code>LeaseRenewalEvent</code> because a renewal request was
     *      denied by the lease grantor even though the lease's expiration
     *      time has not yet been reached, this method will return a
     *      <code>LeaseDeniedException</code>.
     * <li> if the <code>LeaseRenewalManager</code> sends a 
     *      <code>LeaseRenewalEvent</code> because a renewal request could
     *      not be completed before the expiration of the lease, this
     *      method will return an <code>UnknownLeaseException</code>.
     * </ul>
     *
     * @return an instance of <code>Exception</code> or <code>null</code>,
     *         indicating the condition that caused the 
     *         <code>LeaseRenewalManager</code> to fail to renew the 
     *         affected lease
     *         
     * @see net.jini.core.lease.LeaseDeniedException
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public Exception getException() {
	return ex;
    }
}
