/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.lease;

import net.jini.core.lease.*;
import java.util.*;
import java.rmi.RemoteException;

/**
 * This class is a simple mechanism that provides for the systematic renewal
 * and overall management of leases placed on resources that are advertised
 * by services, for which an entity has registered interest.
 * <p>
 * Providing a leasing mechanism in time-based distributed systems can help
 * to prevent the accumulation of outdated and unwanted resources. Such
 * leasing mechanisms generally require renewed proof of interest to continue
 * the existence of a leased resource. Thus, any client or service that
 * requests the use of the leased resources provided by another service
 * may be granted access to those resources for a given period of time; and
 * must continue to request renewal of the lease on each resource for as
 * long as the client or service wishes to have access to the resource.
 * <p>
 * When entities are faced with the task of managing the leases granted
 * by the various services with which the entities interact, such entities
 * may find the use of this class minimizes the work resulting from having
 * to repeatedly implement this required lease management functionality.
 * This class is a simple mechanism that provides for the systematic renewal
 * and overall management of leases placed on resources that are advertised
 * by services, for which an entity has registered interest.
 * <p>
 * Note that this class is not remote. Clients and services wishing to use
 * this class will create an instance of this class in their own address
 * space to locally manage the leases granted to them.
 * <p>
 * This class distinguishes between two time values associated with lease
 * expiration: the desired time of expiration for the lease; and the actual
 * time of expiration granted when the lease is created or renewed. Both time
 * values are absolute times, not relative time durations.
 * <p>
 * Methods of this class that accept <code>Lease.FOREVER</code> or
 * <code>Lease.ANY</code> as the desired expiration time will treat
 * <code>Lease.FOREVER</code> and <code>Lease.ANY</code> as semantically
 * identical. That is, if either <code>Lease.FOREVER</code> or
 * <code>Lease.ANY</code> is requested, renewal requests will be made
 * indefinitely until a cancellation request is made. It is important to
 * note that although these values are not distinguished in the semantics
 * of the <code>LeaseRenewalManager</code> class itself, they may ultimately
 * be distinguished in the semantics of the entity that granted the lease.
 * Whenever this class requests that a lease-granting entity renew a lease
 * to expire at some absolute time, the value of the desired lease expiration
 * time (whatever that value may be - <code>Lease.FOREVER</code>, 
 * <code>Lease.ANY</code>, or any other value) will be passed on to that
 * lease-granting entity, leaving it to the lease-granting entity to define
 * the semantics of the requested lease expiration time.
 * <p>
 * The operations in this class are appropriately synchronized for concurrent
 * operation. Additionally, this class makes certain guarantees with respect
 * to concurrency. When this class makes a remote call (for example, when
 * requesting the renewal of a lease), any invocations made on the methods
 * of this class will not be blocked. Similarly, this class makes a
 * re-entrancy guarantee with respect to the listener objects registered 
 * with this class. Should this class invoke a method on a registered
 * listener (a local call), calls from that method to any other method of
 * this class are guaranteed not to result in a deadlock condition.
 * 
 * @see net.jini.core.Lease
 *
 * @author Zane Pan 
 */

public class LeaseRenewalManager {
    private static final long renewbatchTimeGap =5*60*1000;
    private static final long RENEW_RTT = 1000 * 10;
    private static class Entry implements Comparable {
	// Since the cnt only gets modified in the constructor, it does
	// not need to be synchronized.
	private static int cnt = 0;
	private long endTime;
	public Entry(Lease lease, long expiration, LeaseListener listener) {
	    this.endTime =  lease.getExpiration();
	    this.lease = lease;
	    this.expiration = expiration;
	    this.listener = listener;
	    calcRenew();
	    id = cnt++;
	}
	
	/** create a fake entry for tailMap */
	public Entry(long renew) {
	    this.renew = renew + renewbatchTimeGap;
	    id = Integer.MAX_VALUE;
	}

	public int id;
	public Lease lease;
	public long expiration;
	public LeaseListener listener;
	public long renew;
	private RemoteException ex = null;
	public long getRenewDuration(long now) {
	    if(expiration == Lease.FOREVER || expiration ==Lease.ANY )
		return expiration;
	    else
		return (expiration - now);
	}
	
	/** calculate the renew time for the lease entry */
	public void calcRenew() {
	    calcRenew(System.currentTimeMillis());
	}
	public void calcRenew(long now) {
	    this.ex = null;
	    this.endTime = lease.getExpiration();	    
	    if (expiration != Lease.ANY && expiration <= endTime) {
		renew = Long.MAX_VALUE; // no need for renew anymore
		return;
	    } 

	    long delta = endTime - now;
	    if(delta <= RENEW_RTT ) {
		// we may not have enough time to do the renew, but 
		// try to renew it as soon as possible.
		renew = now;
		return;
	    } else if(delta <= RENEW_RTT*2)
		delta = RENEW_RTT;
	    else if(delta <= RENEW_RTT*8)
		delta /=2;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta /= 8;
	    else if (delta <= 1000 * 60 * 60 * 24 * 14)
		delta = 1000 * 60 * 60 * 24;
	    else
		delta = 1000 * 60 * 60 * 24 * 3;
	    renew = endTime - delta;
	
	}
	
	/** call the method if renew fails, it re-calculates the renew time */
	public void delayRenew(RemoteException ex) {
	    this.ex = ex;
	    long delta = endTime - renew;
	    if(delta <= RENEW_RTT )
		return;
	    else if (delta <= RENEW_RTT*3)
		 delta = RENEW_RTT;
	    else if (delta <= 1000 * 60 * 60)
		delta /= 3;
	    else if (delta <= 1000 * 60 * 60 * 24)
		delta = 1000 * 60 * 30;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta = 1000 * 60 * 60 * 3;
	    else
		delta = 1000 * 60 * 60 * 8;
	    renew += delta;
	}
	
	/** 
	 *  Used by the treeMap to sort the entry by renew. The first key 
	 *  has the longest renew 
         */
	public int compareTo(Object obj) {
	    if (this == obj)
		return 0;

	    Entry e = (Entry)obj;
	    if(renew == e.renew ) {
		if(id < e.id )
		    return 1;
		else
		    return -1;
	    }
	    if (renew < e.renew)
		return 1;
	    return -1;
	}

	public boolean canBatch(Entry e) {
	    return this.lease.canBatch(e.lease);
	} 

	public LeaseMap createLeaseMap(long now) {
	    return this.lease.createLeaseMap(getRenewDuration(now));
	}
    }
    
    /** a TreeMap to store Entry */
    private final TreeMap leases = new TreeMap();
    /** the renew thread */
    private RenewThread renewer = null;
    /** the time for the renewThread to wake up */
    private long minRenew = Long.MAX_VALUE;

    /** 
     * No-argument constructor that creates an instance of this class that
     * initially manages no leases.
     */
    public LeaseRenewalManager() {
    }

    /** 
     * Constructs an instance of this class that will initially manage
     * a single lease. Employing this form of the constructor is 
     * equivalent to invoking the no-argument form of the constructor
     * followed by an invocation of the <code>renewUntil<code> method.
     *
     * @param lease       reference to the initial lease to manage
     * @param expiration  the desired (absolute) time of expiration for the
     *                    <code>lease</code> parameter
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> is input, no notifications will
     *                    be sent.
     *
     * @see net.jini.lease.LeaseListener
     * @see net.jini.lease.LeaseRenewalManager#renewUntil
     */
    public LeaseRenewalManager(Lease lease,
                               long expiration,
			       LeaseListener listener) 
    {
	addLease(lease, expiration, listener);
    }

    /** 
     * Adds a lease to the set of leases being managed by this class (the 
     * set of "managed leases").
     * <p>
     * If the lease input to this method is already in the set of managed
     * leases, the listener object and the desired expiration associated
     * with that lease will be replaced with the new listener and expiration.
     * The lease will remain in the set of managed leases until one of the
     * following occurs:
     * <ul>
     * <li> the lease expires
     * <li> the lease is cancelled
     * <li> an explicit removal of the lease from the set is requested
     * <li> an <code>UnknownLeaseException</code>, a
     *      <code>LeaseDeniedException</code>, or any other non-remote
     *      exception is received during a lease renewal attempt
     * <li> a (remote) <code>NoSuchObjectException</code> is received during
     *      a lease renewal attempt
     * </ul>
     * This method will interpret the value of the <code>expiration</code>
     * parameter as the desired absolute system time after which the lease
     * is no longer valid. That argument provides the ability to indicate an
     * expiration time that extends beyond the actual expiration of the lease.
     * If the value input to that argument does indeed extend beyond the
     * actual lease expiration time, then the lease will be systematically
     * renewed at appropriate times until one of the conditions listed above
     * occurs. If the value input is less than or equal to the actual
     * expiration time, nothing will be done to modify the time when the
     * lease actually expires. That is, the lease will not be renewed with
     * an expiration time that is less than the originally granted expiration.
     * In particular, if either <code>Lease.FOREVER</code> or
     * <code>Lease.ANY</code> is requested, renewal requests will be made
     * (using <code>Lease.ANY</code>) indefinitely until a cancellation
     * request is made.
     * <p>
     * If a non-<code>null</code> object reference is passed in as the
     * <code>listener</code> parameter, the object passed to that parameter
     * will receive notification of exceptional conditions occurring upon
     * a renewal attempt. In particular, exceptional conditions may include
     * the following:
     * <ul>
     * <li> <code>UnknownLeaseException</code>
     * <li> <code>LeaseDeniedException</code>
     * <li> <code>RemoteException</code>
     * <li> any other non-<code>RemoteException</code>
     * </ul>
     * If either an <code>UnknownLeaseException</code> or a 
     * <code>LeaseDeniedException</code> occurs during a lease renewal
     * request, the event will be wrapped in an instance of the
     * <code>LeaseRenewalEvent</code> class and then sent to the listener.
     * <p>
     * If a <code>RemoteException</code> occurs during a renewal request
     * for a particular lease, renewal requests will continue to be made
     * for that lease until either the lease is renewed, or the lease
     * expiration time has been exceeded. In the latter case, the first
     * <code>RemoteException</code> received will be sent to the listener.
     * <p>
     * If <code>null</code> is passed in as the <code>listener</code>
     * parameter, the entity will receive no notification of exceptional
     * conditions occurring as a result of a lease renewal attempt.
     *
     * @param lease       reference to the new lease to manage
     * @param expiration  the desired (absolute) time of expiration for the
     *                    <code>lease</code> parameter
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> is input, no notifications will
     *                    be sent.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>lease</code> parameter.
     *
     * @see net.jini.lease.LeaseListener
     * @see net.jini.lease.LeaseRenewalEvent
     * @see net.jini.core.lease.UnknownLeaseException
     * @see net.jini.core.lease.LeaseDeniedException
     */
    public void renewUntil(Lease lease, 
			   long expiration, 
			   LeaseListener listener) 
    {
	addLease(lease, expiration, listener);

    }
 
    /** 
     * Adds a lease to the set of leases being managed by this class (the 
     * set of "managed leases").
     * <p>
     * The semantics of this method are identical to those of the 
     * <code>renewUntil</code> method, with the <code>expiration</code>
     * parameter of <code>renewUntil</code> set to a value of 
     * <code>duration + current time (in milliseconds)</code>.
     * <p>
     * This method tests for arithmetic overflow in the expiration time
     * computed from the value of <code>duration</code> parameter 
     * (current time + duration). Should such overflow be present, a value
     * of <code>Lease.FOREVER</code> is used to represent the lease's
     * expiration time.
     *
     * @param lease       reference to the new lease to manage
     * @param expiration  the desired duration (relative time) of the
     *                    <code>lease</code> parameter
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> is input, no notifications will
     *                    be sent.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>lease</code> parameter.
     *
     * @see net.jini.lease.LeaseRenewalManager#renewUntil
     * @see net.jini.lease.LeaseListener
     * @see net.jini.lease.LeaseRenewalEvent
     * @see net.jini.core.lease.UnknownLeaseException
     * @see net.jini.core.lease.LeaseDeniedException
     */
    public void renewFor(Lease lease, long duration, LeaseListener listener) {
	long curTime = System.currentTimeMillis();
	if( duration < Lease.FOREVER - curTime  )// check overflow.
	   duration += curTime;
	else
	   duration = Lease.FOREVER;  
	addLease(lease, duration, listener);
    }

    /** 
     * add a lease for the LeaseManager to manager.
     *
     * @param lease the lease object 
     * @param expiration the absolute expiration time of the lease
     * @param LeaseListener When lease expires before the duration, the  
     * leaseExpired method of the listener gets called. If lease renew
     * throws the LeaseDenied exception, the leaseDenied method gets
     * called. If listener is null, no notification method gets called.
     */
    private synchronized void addLease(Lease lease,long expiration,
				      LeaseListener listener) 
    {
	Entry e = findEntryDo(lease);
	if(e != null ) 
	    leases.remove(e);
	addEntry (new Entry(lease, expiration, listener));
    }
        
    /**
     * Returns the current desired time of expiration requested for a
     * particular lease, (not the actual expiration that was granted when
     * the lease was created or renewed). 
     *
     * @param lease       the lease from which the current desired time of
     *                    expiration is being requested
     *
     * @return a <code>long</code> value corresponding to the current 
     *         desired time of expiration of the <code>lease</code> parameter
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized long getExpiration(Lease lease)
	throws UnknownLeaseException 
    {
	Entry e = findEntry(lease);
	return e.expiration;
    }

    /**
     * Replaces the current desired expiration of a given lease from the
     * managed set with a new desired expiration time.
     * <p>
     * Note that an invocation of this method with a lease that is currently
     * a member of the managed set is equivalent to an invocation of the
     * <code>renewUntil</code> method with the lease's current listener
     * input to that method's <code>listener</code> parameter. Specifically,
     * if the value of the <code>expiration</code> parameter is less than
     * or equal to the lease's current desired expiration, this method takes
     * no action.
     *
     * @param lease       the lease whose desired time of expiration should
     *                    be replaced
     *
     * @param expiration  <code>long</code> value representing the new
     *                    desired time of expiration for the
     *                    <code>lease</code> parameter
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.lease.LeaseRenewalManager#renewUntil
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized void setExpiration(Lease lease, long expiration)
	throws UnknownLeaseException 
    {
	Entry e = findEntry(lease);
	long oldExpiration = e.expiration;
	e.expiration = expiration;
	if( oldExpiration >= expiration)
	    return;

	leases.remove(e);
	e.calcRenew();
	addEntry(e);
    }

    /**
     * Both removes a given lease from the managed set, and cancels the
     * given lease.
     * <p>
     * Note that even if an exception is thrown as a result of the cancel
     * operation, the lease will still have been removed from the set of
     * leases managed by this class. Additionally, any exception thrown by
     * the <code>cancel</code> method of the lease object itself may also
     * be thrown by this method.
     *
     * @param lease the lease to remove and cancel
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been successfully cancelled, (but the lease is 
     *         guaranteed to have been removed from the managed set).
     *
     * @see net.jini.core.lease.Lease#cancel
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public void cancel(Lease lease)
	throws UnknownLeaseException, java.rmi.RemoteException
    {
	Entry e;
	synchronized(this) {
	    e = findEntry(lease);
	    leases.remove(e);
	    calcMinRenew();
	}
	e.lease.cancel();
    }

    /** add an lease Entry to the treeMap */
    private void addEntry(Entry ne) {
	leases.put(ne, ne);
	calcMinRenew();
    }

    /**
     * Removes a given lease from the managed set of leases; but does not
     * cancel the given lease.
     *
     * @param lease the lease to remove from the managed set
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized void remove(Lease lease) throws UnknownLeaseException {
	Entry e = findEntry(lease);
	leases.remove(e);
	calcMinRenew();

    }

    /**
     * Removes all leases from the managed set of leases. This method does
     * not request the cancellation of the removed leases.
     */
    public synchronized void clear() {
	leases.clear();
	calcMinRenew();

    }

    /** calculate the time for the renewThread to wake up */
    private void calcMinRenew() {
 	long renew = Long.MAX_VALUE;
	long now = System.currentTimeMillis();
	Iterator iter = leases.values().iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if(e.renew >= e.endTime ) {
		if(now >=e.endTime)
		     iter.remove();
		continue;
	    }
	    if (e.endTime <= now ) {
		//lease has expired and no listener, just remove it.
		if(e.listener == null ) {
		    iter.remove();
		    continue;
		}
	    }
	    
	    if (!canBatch(e)) {
		if (e.renew < renew - RENEW_RTT ) {
		     renew = e.renew;
		} else { 
		    renew -= RENEW_RTT;
		}
	    }
	}

	if(renewer == null && renew < Long.MAX_VALUE )
	    createRenewThread();
	else 
	    if(minRenew > renew || renew ==  Long.MAX_VALUE  ) 
		this.notifyAll();	    
	minRenew = renew;

    }
    
    /**
     * Return true if e can be batched with another entry that expires
     * between e.renew - renewbatchTimeGap and e.renew.
     */
    private boolean canBatch(Entry e) {
	Iterator iter = leases.tailMap(e).values().iterator();
	iter.next(); // skip e itself
	while (iter.hasNext()) {
	    Entry be = (Entry)iter.next();
	    if (e.renew - be.renew > renewbatchTimeGap)
		break;
	    if (e.canBatch(be))
		return true;
	}
	return false;
    }

    /** find a lease entry in the treeMap */
    private Entry findEntry(Lease lease) throws UnknownLeaseException {
	Entry e = findEntryDo(lease);
	if(e != null )
	    return e;
	else
	    throw new UnknownLeaseException();
    }

    private Entry findEntryDo(Lease lease) {
	Iterator iter = leases.values().iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if (e.lease.equals(lease))
		return e;
	}
	return null;
    }
    
    /** renew a lease in the entry */
    private void renewEntry(Entry e) {
	try {
	    e.lease.renew(e.getRenewDuration(System.currentTimeMillis()));
	    synchronized(this) {
		if(leases.remove(e) != null ){
		    e.calcRenew();
		    addEntry(e);
		}
	    }
	} catch (java.rmi.RemoteException ex) {
		synchronized(this) {
		    if(leases.remove(e) != null) {
			e.delayRenew(ex);
			addEntry(e);
		    }
		}
	} catch (Exception ex) {
	    boolean b;
	    synchronized(this) {
		b = leases.remove(e) != null && (e.listener != null);
		calcMinRenew();
	    }
	    if(b)
		e.listener.notify(
		    new LeaseRenewalEvent(this, e.lease, e.expiration, ex));
	}
    }

    private synchronized void batchDelayRenew(ArrayList bList,
                                              RemoteException ex)
    {
	Iterator iter = bList.iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if(leases.remove(e) != null) {
		e.delayRenew(ex);
		leases.put(e, e);
	    }
	}	
	calcMinRenew();	
    }
    
    private void updateAfterRenew(LeaseMap batchLeaseMap,
					       Map exMap, List bList)
    {
	ArrayList al = new ArrayList(1);
	synchronized(this) {
	    Iterator iter = bList.iterator();
	    while(iter.hasNext()) {
		Entry e = (Entry)iter.next();
		if( leases.remove(e) == null)
		    continue;

		if( exMap != null && exMap.containsKey(e.lease)) {
		    al.add(e);

		} else {
		    e.calcRenew();
		    leases.put(e, e);
		}
	    }
	    calcMinRenew();
	}
	for(int i=0; i<al.size(); i++) {
	    Entry e = (Entry)al.get(i);
	    if(e.listener != null ) 
		e.listener.notify( new LeaseRenewalEvent(this, e.lease, 
							 e.expiration,
				   (LeaseException)exMap.get(e.lease)));
	    
	}
    }

    private void batchRenew(ArrayList bList) {
	LeaseMap batchLeaseMap = createBatchLeaseMap(bList);
	try {
	    batchLeaseMap.renewAll();
	    updateAfterRenew(batchLeaseMap, null, bList);
	} catch (LeaseMapException ex) {
	    updateAfterRenew(batchLeaseMap, ex.exceptionMap, bList);
	} catch (java.rmi.RemoteException ex) {
	    batchDelayRenew(bList, ex);
	}	
    }

    /** create a batch set for batch renew */
    private ArrayList createBatchList(Entry e) {
	Entry fakeEntry = new Entry(e.renew);
	SortedMap lMap = leases.tailMap(fakeEntry);
	Iterator iter = lMap.values().iterator();
	ArrayList batchList = new ArrayList(1);
	while(iter.hasNext()) {
	    Entry be = (Entry)iter.next();
	    if (e.canBatch(be)) {
		batchList.add(be);
	    }
	}
	return batchList;
    }

    /** create a batch leaseMap for batch renew */
    private LeaseMap createBatchLeaseMap(ArrayList bList) {
	Iterator iter = bList.iterator();
	long now = System.currentTimeMillis();
	LeaseMap batchLeaseMap = ((Entry)iter.next()).createLeaseMap(now);
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    batchLeaseMap.put(e.lease, new Long(e.getRenewDuration(now)));
	}
	return batchLeaseMap;
    }

    /** create the renewThread */
    private void createRenewThread() {
	renewer = new RenewThread();
	renewer.start();
    }

    private class RenewThread extends Thread {

	public RenewThread() {
	    super("Renew lease");
	    setDaemon(true);
	}

        public void run() {
	    try {
		while (true) {
		    Entry e;
		    ArrayList bList=null;
		    boolean bExpired = false;
		    synchronized (LeaseRenewalManager.this) {
			/**
			 * minRenew < now can happen when we try to renew
                         * a lease that is going to expire soon. An examples
			 * of this case is that a lease fails to renew the 
			 * several times, then the delayRenew calculates
			 * the new renew time to current time. 
			 */
			long now = System.currentTimeMillis();
			while (minRenew > now ) {
			    //System.out.println("sleep:." + (minRenew-now));
			    if (minRenew == Long.MAX_VALUE) {
				renewer = null;
				//System.out.println("thread exit");
				return;
			    }
			    LeaseRenewalManager.this.wait(minRenew - now);
			    now = System.currentTimeMillis();
			}
			e = (Entry)leases.lastKey();
			bExpired = e.endTime <= now;
			if(bExpired) {
			    leases.remove(e);
			    calcMinRenew();
			} else
			    bList = createBatchList(e);
		    }
		    if(bExpired) {
			if(e.listener != null ) 
			    e.listener.notify
                                         (new LeaseRenewalEvent
                                                 (LeaseRenewalManager.this, 
					          e.lease,e.expiration,e.ex));
		    } else {
			if(bList.size() > 1)
			    batchRenew(bList);
			else
			    renewEntry(e);
		    }
		}
	    } catch (InterruptedException ex) {
	    }
	}
    }
    /** a debug function to display the lease info in the treeMap */
    /*************************************************************
    public  void printRenewTime() {
    	Iterator iter = leases.values().iterator();
	System.out.println("++++++++++++++++++++++++++++++++++++++");
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    System.out.println("renew:" + ( e.renew - 5000000) + 
		     "  expiration:" + (e.expiration - 5000000)+
		     " id:" + e.id);
	}
	System.out.println("--------------------------------------");
    }
    ***************************************************************/
}
