/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */
package net.jini.lease;

import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import net.jini.core.lease.Lease;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.EventRegistration;


/**
 * Defines the interface to lease renewal sets.  The interface is not
 * a remote interface; each implementation of the renewal service
 * exports proxy objects that implement the LeaseRenewalSet interface
 * that use an implementation-specific protocol to communicate with
 * the actual remote server.  All of the proxy methods obey normal RMI
 * remote interface semantics except where explicitly noted.  Two
 * proxy objects are equal if they are proxies for the same set
 * created by the same renewal service.  Every method invocation (on
 * both LeaseRenewalService and LeaseRenewalSet) is atomic with
 * respect to other invocations.
 *
 * A number of the methods in this class throw
 * <code>RemoteException</code>, each of these may throw the
 * <code>java.rmi.NoSuchObjectException</code> subclass.  If a client
 * receives a <code>NoSuchObjectException</code> when calling a method
 * on a renewal set the client can infer that the set has been
 * destroyed; however, it should not infer that the renewal service
 * has been destroyed.
 *
 * @see LeaseRenewalSet 
 */
public interface LeaseRenewalSet {
    /**
     * Add a lease to the set.  The renewal service will renew the
     * lease until the membership duration expires or a renewal
     * attempt results in a definite exception.  Each renewal will use
     * the time reaming in the membership duration for the renewal
     * duration.  If <code>Lease.ANY</code> is passed for membership
     * duration the set will continue to renew the lease until it
     * receives a definite exception, each time it will use
     * <code>Lease.ANY</code> for the duration.  If
     * <code>Lease.FOREVER</code> is passed for the membership
     * duration the set will continue to renew the lease until it
     * receives a definite exception each time it will use
     * <code>Lease.FOREVER</code> for the duration.
     * <p>
     * If the passed lease is already in the set this call will have
     * the effect of associating the old lease with the new membership
     * duration.
     *
     * @param leaseToRenew       The lease to be added to the renewal set
     * @param membershipDuration Value for membership duration in milliseconds
     *
     * @throws IllegalArgumentException if
     * <code>membershipDuration</code> is not positive,
     * <code>Lease.ANY</code>, or <code>Lease.FOREVER</code>.  Will
     * also throw <code>IllegalArgumentException</code> if
     * <code>leaseToRenew</code> is in another set created by the same
     * renewal server.
     *
     * @throws NullPointerException if <code>leaseToRenew</code> is
     * <code>null</code>.
     *
     * @throws java.rmi.RemoteException 
     */
    public void renewFor(Lease leaseToRenew, long membershipDuration)
	throws RemoteException;

    /**
     * Removes the specified lease from set.  If the lease is currently
     * in the set it will be returned, otherwise <code>null</code> will
     * be returned.  
     *
     * @param leaseToRemove Lease to be removed from the set.
     * @return The removed lease if it was in the set or
     * <code>null</code> if it was not     
     * @throws NullPointerException if <code>leaseToRemove</code> is
     * <code>null</code>.
     *
     * @throws java.rmi.RemoteException 
     */
    public Lease remove(Lease leaseToRemove) 
	throws RemoteException;
    

    /** 
     * Register for the expiration warning event associated with this
     * set.  The expiration warning event is sent at a client
     * specified time before the lease on the set is to expire.
     * <p>
     * If there is already a listener registered for this event, the
     * new listener will replace it.  Events sent the new listener
     * will be in the same sequence and have the same event type id as
     * those sent to the old listener.
     *
     * @param listener   The listener to be notified when this event occurs.
     * @param minWarning How long be for the lease on the expires
     *                   should the event be sent.
     * @param handback   An object to be handed back to the listener
     *                   when the warning event occurs.
     * @return A <code>EventRegistration</code> describing the event
     * registration.
     *
     * @throws IllegalArgumentException if <code>minWarning</code> negative.
     *
     * @throws NullPointerException if <code>listener</code> is
     * <code>null</code>.
     *
     * @throws java.rmi.RemoteException 
     */
    public EventRegistration setExpirationWarningListener(
	                         RemoteEventListener listener, 
				 long                minWarning, 
				 MarshalledObject    handback)
	throws RemoteException;

    /**       
     * Remove the listener currently registered for the expiration
     * warning.  It is safe to call this method even if no listener is
     * currently registered.
     * @throws java.rmi.RemoteException 
     */
    public void clearExpirationWarningListener()
	    throws RemoteException;

    /** 
     * Register for the renewal failure event associated with this
     * set.  The renewal failure event is sent when the renewal
     * service attempts to renew a lease in a given set and receives a
     * definite exception or if a lease in a given set expires before
     * its membership duration elapses.
     * <p>
     * If there is already a listener registered for this event, the
     * new listener will replace it.  Events sent the new listener
     * will be in the same sequence and have the same event type id as
     * those sent to the old listener.
     *
     * @param listener   The listener to be notified when this event occurs.
     * @param handback   An object to be handed back to the listener
     *                   when the warning event occurs.
     * @return A <code>EventRegistration</code> describing the event
     * registration.
     *
     * @throws NullPointerException if <code>listener</code> is
     * <code>null</code>.
     *
     * @throws java.rmi.RemoteException
     */
    public EventRegistration setRenewalFailureListener(
				 RemoteEventListener listener, 
				 MarshalledObject    handback)
	throws RemoteException;

    /**       
     * Remove the listener currently registered for the renewal
     * failure event.  It is safe to call this method even if no
     * listener is currently registered.  
     *
     * @throws java.rmi.RemoteException
     */
    public void clearRenewalFailureListener()
	throws RemoteException;

    /**
     * Returns the lease that controls the lifetime of this set.  Can
     * be used to extend or end the sets lifetime.  Note that this
     * does not make a remote call.
     *
     * @return the lease that controls lifetime of this set
     */
    public Lease getRenewalSetLease();
}


