/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */
package net.jini.lease;

import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import net.jini.core.lease.Lease;
import net.jini.core.event.RemoteEvent;

/**
 * Type of event generated by a lease renewal set when it can't renew a lease
 * it is responsible for.
 * 
 * @see LeaseRenewalSet 
 */
public class RenewalFailureEvent extends RemoteEvent {
    private static final long serialVersionUID = 8891457049195932943L;

    /**
     * Throwable returned by <code>getThrowable</code> method.
     * @serial
     */
    private Throwable throwable;

    /**
     * Lease, in marshaled form, returned by <code>getLease</code>
     * method.  
     * @serial 
     */
    private MarshalledObject marshaledLease;

    /**
     * Transient cache of lease returned by <code>getLease</code> method.  
     */
    private transient Lease lease = null;

    /**
     * Simple constructor.
     *
     * @param source the ServiceRegistrar that generated the event
     * @param eventID the registration eventID
     * @param seqNum the sequence number of this event
     * @param handback the client handback
     * @param marshaledLease the lease which could not be renewed in
     *        marshaled form.
     * @param throwable the first throwable that was thrown in the last
     *        chain of renewal failures.
     */
    public RenewalFailureEvent(Object source, 
			       long eventID,
			       long seqNum,
			       MarshalledObject handback,
			       MarshalledObject marshaledLease,
			       Throwable throwable) 
    {
	super(source, eventID, seqNum, handback);
	this.throwable = throwable;
	this.marshaledLease = marshaledLease;
    }

    /**
     * Returns the lease that could not be renewed.  When deserialized
     * the lease is left in marshaled form.  It is only unmarshalled
     * when this call is made.  If a call to this method fails future
     * calls will attempt to re-unmarshal the lease.  Once the lease is
     * successfully unmarshalled it is cached in a transient field so
     * future calls will not result in <code>IOException</code> or
     * <code>ClassNotFoundException</code>.
     *
     * @throws IOException can be thrown if there are difficulties
     * unmarshalling the lease usually this will be some sort class
     * miss-match.
     *
     * @throws ClassNotFoundException can be thrown if there are
     * difficulties unmarshalling lease, usually this will indicate one
     * of the classes associated with the lease's implementation could
     * not be loaded.
     */
    public Lease getLease() throws IOException, ClassNotFoundException {
	if (lease == null) 
	    lease = (Lease)marshaledLease.get();

	return lease;
    };

    /**
     * Returns the <code>Throwable</code> that was thrown by the first
     * failed renewal attempt in the last series of consecutive
     * failures.
     */
    public Throwable getThrowable() {
	return throwable;
    }
}

