/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */
package net.jini.lookup;

import java.rmi.server.UnicastRemoteObject;
import net.jini.core.lease.*;
import net.jini.core.entry.Entry;
import net.jini.core.discovery.*;
import net.jini.discovery.*;
import net.jini.core.lookup.*;

import net.jini.lease.*;
import net.jini.discovery.LookupLocatorDiscovery;

import java.util.*;
import java.io.IOException;
import java.rmi.RemoteException;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.EventRegistration;

/**
 * The <code>ClientLookupManager</code> class is a helper utility class
 * that any client-like entity can use to "discover" services registered 
 * with any number of lookup services of interest. On behalf of such
 * entities, this class maintains - as much as possible - up-to-date
 * state information about both the lookup services the entity wishes
 * to query, and the services the entity wishes to acquire and use.
 * By maintaining current service state information, the entity can 
 * implement efficient mechanisms for service access and usage.
 * <p>
 * There are three basic usage patterns for this class. In order of
 * importance and typical usage, those patterns are:
 * <p>
 * <ul>
 *   <li> the entity requests that the <code>ClientLookupManager</code>
 *        create a cache (an instance of <code>LookupCache</code>) which 
 *        will asynchronously "discover", and locally store, references to
 *        services that match criteria defined by the entity; services
 *        which are registered with one or more lookup services managed
 *        by the <code>ClientLookupManager</code> on behalf of the entity.
 *        The cache can be viewed as a set of service references that the 
 *        entity can access locally as needed through one of the public, 
 *        non-remote methods provided in the cache's interface. Thus, rather
 *        than making costly remote queries of multiple lookup services at
 *        the point in time when the entity needs the service, the entity
 *        can simply make local queries on the cache for the services that
 *        the cache acquired and stored at a prior time. An entity should
 *        employ this pattern when the entity must make <i>frequent</i>
 *        queries for multiple services. By populating the cache with
 *        multiple instances of the desired services, redundancy in the
 *        availability of those services can be provided. Thus, if an
 *        instance of a service is found to be unavailable when needed,
 *        the entity can execute a local query on the cache rather than
 *        one or more remote queries on the lookup services to acquire
 *        an instance that is available. To employ this pattern, the entity
 *        invokes the method <code>createLookupCache</code>.
 *   <li> the entity can register with the event mechanism provided by the
 *        <code>ClientLookupManager</code>. This event mechanism allows the
 *        entity to request that it be notified when a service of interest
 *        is discovered for the first time, or has encountered a state change
 *        such as removal from all lookup services, or attribute set changes.
 *        Although interacting with a local cache of services in the way
 *        described in the first pattern can be very useful to entities that
 *        need frequent access to multiple services, some client-like 
 *        entities may wish to interact with the cache in a reactive manner.
 *        For example, an entity such as a service browser typically wishes
 *        to be notified of the arrival of new services of interest as well
 *        as any changes in the state of the current services in the cache.
 *        In these situations, polling for such changes is usually viewed as
 *        undesirable. If the cache were to also provide an event mechanism
 *        with notification semantics, the needs of entities that employ
 *        either pattern can be satisfied. To employ this pattern, the entity
 *        must create a cache and supply it with an instance of the
 *        <code>ServiceDiscoveryListener</code> interface that will receive
 *        instances of <code>ServiceDiscoveryEvent</code> when events of
 *        interest, related to the services in the cache, occur.
 *   <li> the entity, through the public API of the 
 *        <code>ClientLookupManager</code>, can directly query the lookup 
 *        services managed by the <code>ClientLookupManager</code> for 
 *        services of interest; employing semantics similar to the semantics
 *        employed in a typical lookup service query made through the
 *        <code>ServiceRegistrar</code> interface. Such queries will result
 *        in a remote call being made at the same time the service is needed
 *        (unlike the first pattern in which remote calls typically occur
 *        prior to the time the service is needed). This pattern may be useful
 *        to entities needing to find services on an infrequent basis, or 
 *        when the cost of making a remote call is outweighed by the 
 *        overhead of maintaining a local cache (for example, due to limited
 *        resources). Although an entity that needs to query lookup service(s)
 *        can certainly make such queries through the 
 *        <code>ServiceRegistrar</code> interface, the 
 *        <code>ClientLookupManager</code> provides a broad API with 
 *        semantics that are richer than the semantics of the 
 *        <code>lookup</code> methods provided by the
 *        <code>ServiceRegistrar</code>. This API encapsulates functionality
 *        that many client-like entities may find more useful when managing
 *        both the set of desired lookup services, and the service queries
 *        made on those lookup services. To employ this pattern, the entity
 *        simply instantiates this class with the desired parameters, and
 *        then invokes the appropriate version of the <code>lookup</code> 
 *        method when the entity wishes to acquire a service that matches
 *        given criteria.
 * </ul>
 * <p>
 * All three mechanisms just described - local queries on the cache, service 
 * discovery notification, and remote lookups - employ the same 
 * template-matching scheme as that employed in the 
 * <code>ServiceRegistrar</code> interface. Additionally, each mechanism
 * allows the entity to supply an object referred to as a <i>filter</i>. 
 * Such an object is a non-remote object that defines additional matching
 * criteria that the <code>ClientLookupManager</code> applies when searching
 * for the entity's services of interest. This filtering facility is
 * particularly useful to entities that wish to extend the capabilities of
 * the standard template-matching scheme.
 * <p>
 * Note that this class is not remote. Clients and services that wish to use
 * this class will create an instance of this class in their own address
 * space to manage the state of discovered services and their associated
 * lookup services locally.
 * 
 * @see net.jini.discovery.DiscoveryManagement
 * @see net.jini.lookup.LookupCache
 * @see net.jini.lookup.ServiceDiscoveryListener
 * @see net.jini.lookup.ServiceDiscoveryEvent
 * @see net.jini.core.lookup.ServiceRegistrar
 * @author Zane Pan
 */
public class ClientLookupManager {
    /* 
     * The task class for implementing register, lookup, notify and
     * dropProxy task.
     */
    private static abstract class  Task {
	protected ProxyReg reg;
	public Task(ProxyReg reg) {
	    this.reg = reg;
	}
	/* check if the task is on a specific ProxyReg, return true if it is */
	public boolean isFromProxy(ProxyReg reg) {
	    return this.reg.equals(reg);
	}
	public abstract void exec();
    }
    
    /* Used as the callback in the lookup(tmpl, wait) to create
     * createLookupCache.
     */
    private final static class ServiceDiscoveryListenerImpl 
    implements ServiceDiscoveryListener
    {
	ArrayList items = new ArrayList();
	public synchronized void serviceAdded(ServiceDiscoveryEvent event) {
	    items.add(event.getPostEventServiceItem());
	    this.notifyAll();
	}

	public void serviceRemoved(ServiceDiscoveryEvent event){
	}
	public void serviceChanged(ServiceDiscoveryEvent event){
	}
	
	public synchronized ServiceItem[] getServiceItem() {
	    ServiceItem[] r = new ServiceItem[items.size()];
	    items.toArray(r);
	    items.clear();
	    return r;
	}
    }

    /* 
     * Used to group together the lease and event sequence number. For each
     * LookupCache, there is a HashMap that maps a ProxyReg to an EventReg.
     */
    private final static class EventReg  {
	/* The Event notification lease */
	public Lease lease;
	/* The current event sequence number for the Service template */
	public long seqNo;
	public EventReg(Lease lease, long seqNo) {
	    this.lease = lease;
	    this.seqNo = seqNo;
	}
    }
    
    /*
     * Used in the LookupCache.For each LookupCache, there is a HashMap that
     * maps ServiceId to a ServiceItemReg. The ServiceItemReg class keep
     * tracking where the ServiceItem come from.
     */
    private final static class ServiceItemReg  {
	/* to store ServiceRegistrars that has the ServiceItem registered. */
	private final ArrayList proxys = new ArrayList();
	/* flag to indicator that ServiceItem has been discarded */
	private boolean bDiscarded = false;
	public ServiceItem item;
	public ServiceItemReg(ServiceRegistrar proxy, ServiceItem item) {
	    addProxy(proxy);
	    this.item = item;
	}

	/* only get called from the taskThread. no sync is needed */
	public void addProxy(ServiceRegistrar proxy) {
	    if(!proxys.contains(proxy))
		proxys.add(proxy);
	}

	/* only get called from the taskThread. no synch is needed */
	public void removeProxy(ServiceRegistrar proxy) {
	    int index = proxys.indexOf(proxy);
	    if(index != -1)
		proxys.remove(index);
	}

	/* only get called from the taskThread. no lynch is needed */
	public boolean hasNoProxys() {
	    return proxys.isEmpty();
	}
	
	/* mark the ServiceItem discarded */ 
	public void setDiscarded(boolean b) {
	    bDiscarded = b;
	}

	/* mark the ServiceItem discarded */ 
	public boolean isDiscarded() {
	   return  bDiscarded;
	}

	/* a debug function that can be removed when we ship */
	public void printProxys() {
	    Iterator iter = proxys.iterator();
	    System.out.println("----------- proxys --------------");
	    while(iter.hasNext()) {
		try {
		    ServiceRegistrar proxy = (ServiceRegistrar)iter.next();
		    System.out.println("proxy:" + proxy.getLocator());
		} catch(Exception e){}
	    }
	    System.out.println("------------------ --------------");
	}
    }

    /*
     * A class to wrapper for a ServiceRegistrar, we may add some information
     * such as if the proxy is come from a group, a locator, or both in the
     * future.
     */
    private final static class ProxyReg  {
	public ServiceRegistrar proxy;
	public ProxyReg(ServiceRegistrar proxy) {
	    if(proxy == null)
		throw new IllegalArgumentException(
					 "proxy cannot be null");
	    this.proxy = proxy;
	}	    

	public boolean equals(Object obj) {
	    if (obj instanceof ProxyReg){
		return proxy.equals(((ProxyReg)obj).proxy);
	    } else return false;
	}

	public int hashCode() {
	    return proxy.hashCode();
	}
    }
    
    /* 
     * The task thread for executing the task in the taskQue. We only have one
     * instance, the taskThread in the ClientLookupManager
     */
    private final class TaskThread extends Thread {
	public TaskThread() {
	    super("taskThread");
	    setDaemon(true);
	}

	public void run(){
	    while(!isInterrupted()) {
		LookupCacheImpl cache;
		synchronized(taskQue) {
		    if(taskQue.isEmpty()) {
			taskThread = null;
			return;
		    }
		    cache = (LookupCacheImpl)taskQue.remove(0);
		}
		cache.execTasks();
	    }
	}    
    }

    /* The Listener class for the LeaseRenewalManager */
    private final class LeaseListenerImpl implements LeaseListener {
	private ServiceRegistrar proxy;
	public LeaseListenerImpl(ServiceRegistrar proxy) {
	    this.proxy = proxy;
	}
	/* When lease renewal fails, we discard the proxy  */
	public void notify(LeaseRenewalEvent e) {
	    discard(proxy);
	}
    }

    /* The Class for cache the matched objects */
    private final class LookupCacheImpl implements LookupCache {
	/* 
	 *  a RemoveEventListener class register with the proxy to notify any
	 * ServiceItem changes (NOMATCH_MATCH, MATCH_NOMATCH, MATCH_MATCH)
	 */
      	private final class LookupListener extends UnicastRemoteObject 
	implements RemoteEventListener,  java.io.Serializable  
	{	
	    public LookupListener() throws RemoteException {
		super();
	    }

	    public void notify(RemoteEvent evt) {
		ServiceEvent theEvent = (ServiceEvent)evt;
		notifyServiceMap((ServiceRegistrar)theEvent.getSource(),
				 theEvent.getServiceID(),
				 theEvent.getServiceItem(),
				 theEvent.getTransition(),
				 theEvent.getSequenceNumber());
	    }
	}  
	
	/* This class is for register the lookupListener to a
         * ServiceRegistrar
         */
	private final class RegisterListenerTask  extends Task {	    
	    public RegisterListenerTask(ProxyReg reg) {
		super(reg);
	    }
	    public void exec() {
		long duration = getLeaseDuration();
		if(duration < 0)
		    return;
		EventRegistration e = registerListener(reg.proxy,
						       tmpl, 
						       lookupListener,
						       duration);
		if(e != null) {
		    Lease lease = e.getLease();
		    synchronized(eventRegMap) {
			// cancel the lease if it has terminated.
			if(bCacheTerminated)
			    cancelLease(lease); 
			else 
			    eventRegMap.put
                                   (reg, new EventReg(lease, 
					              e.getSequenceNumber()));
		    }
		}
	    }
	}
	
	/* Use this class to do a Lookup task on a ServiceRegistrar. */
	private final class LookupTask extends Task {
	    public LookupTask(ProxyReg reg) {
		super(reg);
	    }
	    public void exec() {
		ServiceRegistrar proxy = reg.proxy;
		ServiceMatches matches;
		try {
		    matches = proxy.lookup(tmpl, MAX_MATCHES);
		} catch (RemoteException e) {
		    discard(proxy);
		    return;
		}

		updateServiceMap(proxy, matches.totalMatches>0 ?
				 matches.items : null);
		for(int i=0; i<matches.totalMatches; i++) {
		    if(checkItem(matches.items[i]))
			addService(proxy, matches.items[i]);
		}
	    }
	}

	/* Use this class to do the task for proxy discarded */
	private final class ProxyRegDropTask extends Task {	    
	    public ProxyRegDropTask(ProxyReg reg) {
		super(reg);
	    }
	    public void exec() {
		synchronized(eventRegMap) {
		    // lease has already been canceled by removeProxyReg 
		    if(eventRegMap.containsKey(reg))
		       eventRegMap.remove(reg);
		}
		updateServiceMap(reg.proxy, null);
	    }
	}

	/* Use this class to do the task for discard a service */
	private final class discardServiceTask extends Task {
	    private Object service;
	    public discardServiceTask(Object service) {
		super(null);
		this.service = service;
	    }

	    public void exec() {
		Iterator iter = getServiceIdMapEntrySetIterator();
		while(iter.hasNext()) {
		    Map.Entry e = (Map.Entry)iter.next();
		    ServiceItemReg itemReg = (ServiceItemReg)e.getValue();
		    //check on item.service so we don't need to sync on itemReg
		    if(itemReg.item.service == service) {
			ServiceID sid = (ServiceID)e.getKey();
			removeServiceIdMap(sid,itemReg.item);
			return;
		    }
		}
	    }
	}
	
	/* Use this class to do the event notify task */
	private final class NotifyEventTask extends Task{
	    private ServiceID sid;
	    private ServiceItem item;
	    private int transition;
	    public NotifyEventTask(ProxyReg reg,
				   ServiceID sid,
				   ServiceItem item,
				   int transition )
        {
		super(reg);
		this.sid = sid;
		this.item = item;
		this.transition = transition;
	    }
	    public void exec() {
		if(transition ==  ServiceRegistrar.TRANSITION_MATCH_NOMATCH ||
		   !checkItem(item)) {
		    //ServiceItem does not matched anymore.
		    ServiceItemReg itemReg;
		    synchronized(serviceIdMap) {
			itemReg = (ServiceItemReg)serviceIdMap.get(sid);
		    }
		    if(itemReg != null) {
			itemReg.removeProxy(reg.proxy);
			if(item != null || itemReg.hasNoProxys()) {
			    //item != null means that the item has changed to
			    //NOMATCH, but still registered with the proxy.
			    removeServiceIdMap(sid, itemReg.item);
			}
		    }
		} else // ServiceItem matches.
		    addService(reg.proxy, item);
	    }

	}
	private final static int ITEM_ADDED = 0;
	private final static int ITEM_REMOVED = 2;
	private final static int ITEM_CHANGED = 3;

	/* stores the listener for the event notifications */
	private LookupListener lookupListener;
	/* store the tasks */
	private final ArrayList tasks = new ArrayList();
	/* flag to indicate if the LookupCache has been terminated. */
	private boolean bCacheTerminated = false;
	/* stores the listener to get new item or remove item notifications */
	private final ArrayList sItemListeners = new ArrayList();
	/* map from ServiceID to ServiceItemReg */
	private final HashMap serviceIdMap = new HashMap();
	/* map from ProxyReg to EventReg */
	private final HashMap eventRegMap = new HashMap();
	/* serviceTemplate for the cache */
	private ServiceTemplate tmpl;
	/* the filter for cache */
	private ServiceItemFilter filter = null;
	/* it used in Event registration */
	private long leaseDuration;
	/* 
	 * log the time when the cache gets created. It used to calculate when
	 * the cache should expire.
	 */
	private final long startTime = System.currentTimeMillis();
	public LookupCacheImpl(ServiceTemplate tmpl, 
			       ServiceItemFilter filter,
			       ServiceDiscoveryListener sListener,
			       long leaseDuration) 
	    	    throws RemoteException
        {
	    this.tmpl = copyServiceTemplate(tmpl);
	    this.leaseDuration = leaseDuration;
	    this.filter = filter;
	    lookupListener = new LookupListener();
	    ArrayList set;
	    synchronized(proxyRegSet) {
		set = (ArrayList)proxyRegSet.clone();
	    }

	    synchronized(sItemListeners) {
		if(sListener != null )
		    sItemListeners.add(sListener);
	    }

	    for(int i=0; i<set.size(); i++) {
		ProxyReg reg = (ProxyReg)set.get(i);
		addProxyReg(reg);
	    }
	}
	
	/* Check if the cache has expired */
	public boolean hasExpired() {
	    return getLeaseDuration() <= 0;
	}
	
	// This method's javadoc is inherited from an interface of this class
	public void terminate() {
	    synchronized(caches) {
		int index = caches.indexOf(this);
		if(index != -1)
		    caches.remove(index);

	    }
	    removeTaskQue();
	    synchronized(eventRegMap) {
		bCacheTerminated = true;
		Set set = eventRegMap.entrySet();
		Iterator iter = set.iterator();
		while(iter.hasNext()) {
		    Map.Entry e = (Map.Entry)iter.next();
		    EventReg eReg = (EventReg)e.getValue();
		    cancelLease(eReg.lease);
		}
	    }
	    try {
		UnicastRemoteObject.unexportObject(lookupListener, true);
	    } catch(Exception e) {
	    }
	}

	// This method's javadoc is inherited from an interface of this class
	public ServiceItem lookup(ServiceItemFilter myFilter) {
	    checkCacheTerminated();
	    //printServiceIdMap();
	    ServiceItem[] ret = getServicesItems(myFilter);
	    if (ret.length == 0 )
		return null;
	    int rand = Math.abs(random.nextInt()) % ret.length;
	    return ret[rand];

	}

	// This method's javadoc is inherited from an interface of this class
	public ServiceItem[] lookup(ServiceItemFilter myFilter,int maxMatches){
	    checkCacheTerminated();
	    if (maxMatches < 1)
		throw new IllegalArgumentException("maxMatches must be > 0");

	    ArrayList items = new ArrayList(1);
	    ServiceItem[] sa = getServicesItems(myFilter);
	    int len = sa.length;
	    if (len == 0 )
		return new ServiceItem[0];
	    int rand = Math.abs(random.nextInt()) % len;
	    for(int i=0; i<len; i++) {
		items.add(sa[(i+rand) % len ]);
		if(items.size() == maxMatches)
		    break;
	    }
	    ServiceItem[] ret = new ServiceItem[items.size()];
	    items.toArray(ret);
	    return ret;
	}

	// This method's javadoc is inherited from an interface of this class
	public void discard(Object serviceReference) {
	    Iterator iter = getServiceIdMapEntrySetIterator();
	    while(iter.hasNext()) {
		Map.Entry e = (Map.Entry)iter.next();
		ServiceItemReg itemReg = (ServiceItemReg)e.getValue();
		synchronized(itemReg) {	
		    if(itemReg.item.service == serviceReference) 
			itemReg.setDiscarded(true);	
		}
	    }
	    synchronized(tasks) {
		Task t = new discardServiceTask(serviceReference);
		tasks.add(t);
	    }
	    addTaskQue();
	}

	/* returns the iterator of entry set in the copy of the ServiceIdMap */
	private Iterator getServiceIdMapEntrySetIterator() {
	    HashMap serviceIdMapCopy;
	    synchronized(serviceIdMap) {
		serviceIdMapCopy = (HashMap)serviceIdMap.clone();
	    }
	    Set set = serviceIdMapCopy.entrySet();
	    return set.iterator();
	}

	private ServiceItem[] getServicesItems(ServiceItemFilter myFilter) {
	    ArrayList items = new ArrayList(1);
	    Iterator iter = getServiceIdMapEntrySetIterator();
	    while(iter.hasNext()) {
		Map.Entry e = (Map.Entry)iter.next();
		ServiceItemReg itemReg = (ServiceItemReg)e.getValue();
		ServiceItem s;
		synchronized(itemReg) {
		    // need sync here since the task thread may update the 
		    // itemReg.item
		    if(itemReg.isDiscarded()) 
			continue;
		    s = itemReg.item;
		}
		if(myFilter == null || myFilter.check(s))
		    items.add(s);
	    }
	    ServiceItem[] ret = new ServiceItem[items.size()];
	    items.toArray(ret);
	    return ret;
	}

	// This method's javadoc is inherited from an interface of this class
	public void addListener(ServiceDiscoveryListener l){
	    checkCacheTerminated();
	    if(l == null ) 
		return;
	    synchronized(sItemListeners) {
		sItemListeners.add(l);
	    }
	    ServiceItem[] items = getServicesItems(null);
	    for(int i=0; i<items.length; i++) 
		addServiceNotify(items[i]);
	}
	
	// This method's javadoc is inherited from an interface of this class
	public void removeListener(ServiceDiscoveryListener l){
	    checkCacheTerminated();
	    if( l== null)
		return;
	    synchronized(sItemListeners) {
		int index = sItemListeners.indexOf(l);
		if(index != -1)
		    sItemListeners.remove(l);
	    }
	}

	/*
	 * Add a new ProxyReg to the lookupCache. Called by the constructor and
	 * the DiscMgrListener's discovered().
	 * @param reg a ProxyReg to add.
	 */
	public void addProxyReg(ProxyReg reg) {
	    RegisterListenerTask treg = new RegisterListenerTask(reg);
	    LookupTask tlookup = new LookupTask(reg);
	    synchronized(tasks) {
		//removeUselessTask(reg); since removeProxyReg already have done
		tasks.add(treg);
		tasks.add(tlookup);
	    }
	    addTaskQue();
	}

	/*
	 * Remove a ProxyReg from the lookupCache. Called by  DiscMgrListener's
	 * discarded().
	 * @param reg a ProxyReg to remove.
	 */
	public void removeProxyReg(ProxyReg reg) {
	    EventReg eReg;
	    synchronized(eventRegMap) {
		// let the taskThread to do the (modify)eventRegMap.remove
		eReg = (EventReg)eventRegMap.get(reg);
	    }
	    if(eReg != null) 
		cancelLease(eReg.lease);

	    ProxyRegDropTask t = new ProxyRegDropTask(reg);
	    synchronized(tasks) {
		removeUselessTask(reg);
		tasks.add(t);
	    }
	    addTaskQue();
	}
	
	/* throws RuntimeException if the lookupCache has been terminated */
	private void checkCacheTerminated() throws RuntimeException {
	    checkTerminated();
	    synchronized(eventRegMap) {
		if(bCacheTerminated)
		    throw new RuntimeException("Terminated Error");
	    }
	}	
	
	/* 
	 * Called by the lookupListener's notify(). check the event sequence
	 * number to create LookupTask or NotifyTask. add the task to the 
	 * task queue.
	 */
	private void notifyServiceMap(ServiceRegistrar proxy, 
				      ServiceID sid,
				      ServiceItem item,
				      int transition,
				      long seqNo)
        {
	    synchronized(tasks) {
		ProxyReg reg;
		synchronized(proxyRegSet) {
		    reg = findReg(proxy);
		    if (reg == null )
			return;
		}
		Task t;
		if( updateSeqNumber(reg, seqNo)) {
		    t = new NotifyEventTask(reg, sid, item, transition );
		}else {
		    // at lease one event has been lost, do the lookup task. 
		    // removeUselessTask(reg); but not the drop task
		    t = new LookupTask(reg);
		}
		tasks.add(t);
	    }
	    addTaskQue();
	}
	
	/* 
	 * return true if there is no gap between event sequence number.
	 * set the current event sequence number.
	 */
	private boolean updateSeqNumber(ProxyReg reg, long seqNo) {
	    long prevSeqNo;
	    synchronized(eventRegMap) {
		EventReg eReg = (EventReg)eventRegMap.get(reg);
		// eReg cannot be null here
		prevSeqNo = eReg.seqNo ;
		eReg.seqNo = seqNo;
	    }
	    return  (seqNo == ++prevSeqNo);
	}
	
	/* add the lookupCache to the taskQue, create the taskThread
         * if necessary
         */
	private void addTaskQue() {
	    synchronized(taskQue) {
		if(!taskQue.contains(this))
		    taskQue.add(this);
		if(taskThread == null) {
		    taskThread =  new TaskThread();
		    taskThread.start();
		}
	    }
	}
	
	/* remove the lookupCache from the taskQue, called only by
         * terminate()
         */
	private void removeTaskQue() {
	    synchronized(taskQue) {
		int index = taskQue.indexOf(this);
		if(index != -1)
		    taskQue.remove(index);
	    }
	}

	/* 
	 * remove all the tasks from this ProxyReg from the tasks. It gets 
	 * called when the ProxyReg has been discarded.
	 */
	private void removeUselessTask(ProxyReg reg) {
	    Iterator iter = tasks.iterator();
	    while(iter.hasNext()) {
		Task t = (Task)iter.next();
		if(t.isFromProxy(reg))
		    iter.remove();
	    }
	}
	
	/* execute all tasks in tasks ArrayList. Only called by the
         * taskThread
         */ 
	public void execTasks() {
	    ArrayList tasksClone;
	    synchronized(tasks) {
		if(tasks.isEmpty())
		    return;
		tasksClone = (ArrayList)tasks.clone();
		tasks.clear();
	    }
	    while(!tasksClone.isEmpty()) {
		Task t  = (Task)tasksClone.remove(0);
		t.exec();
	    }

	}
	
	/*
	 * Update the ServiceIdMap for a proxy. Only called by the taskThread.
	 * @param proxy a proxy to update.
	 * @param items the items that matches the template, but not necessary
	 * satisfy the filter.
	 */
  	private void updateServiceMap(ServiceRegistrar proxy,
				      ServiceItem[] items) 
        {
	    ArrayList notifyItems = new ArrayList(1);
	    Iterator iter = getServiceIdMapEntrySetIterator();
	    while(iter.hasNext()) {
		Map.Entry e = (Map.Entry)iter.next();
		ServiceItemReg itemReg = (ServiceItemReg)e.getValue();
		ServiceID sid = (ServiceID)e.getKey();
		ServiceItem newItem = findItem(sid, items);
		if(newItem != null ) {
		    if(checkItem(newItem)) // changed from match to match.
			itemMatchMatchChange(proxy, newItem, itemReg);		
		    else // changed from match to no_match.
			removeServiceIdMap(sid, itemReg.item);
		} else {
		    //service has been removed from the proxy, or service does
		    // not matched the template anymore.
		    itemReg.removeProxy(proxy);
		    if(itemReg.hasNoProxys())
			removeServiceIdMap(sid,itemReg.item);
		}
	    }
        }

	/* remove a entry in the serviceIdMap, send notification message */
	private void removeServiceIdMap(ServiceID sid, ServiceItem item) {
	    synchronized(serviceIdMap) {
		serviceIdMap.remove(sid);
	    }
	    removeServiceNotify(item);
	}
	
	/* return a ServiceItem that has same ServiceID from items */
	private ServiceItem findItem(ServiceID sid, ServiceItem[] items) {
	    if(items != null) {
		for(int i=0; i<items.length; i++) {
		    if(items[i].serviceID.equals(sid) )
			return items[i];
		}
	    }
	    return null;
	}

	/* Add a ServiceItem to  ServiceIdMap. Only called by the taskThread */
	private void addService(ServiceRegistrar proxy, ServiceItem sItem) {
	    ServiceItemReg itemReg;
	    synchronized(serviceIdMap) {
		itemReg = (ServiceItemReg)serviceIdMap.get(sItem.serviceID);
	    }
	    if( itemReg == null ) {
		itemReg = new ServiceItemReg(proxy, sItem);
		synchronized(serviceIdMap) {
		    serviceIdMap.put(sItem.serviceID, itemReg);
		}
		addServiceNotify(sItem);
	    }else 
		itemMatchMatchChange(proxy, sItem, itemReg);
	}
	
	private void itemMatchMatchChange(ServiceRegistrar proxy,
					  ServiceItem newItem,
					  ServiceItemReg itemReg ) {
	   
	    ServiceItem oldItem = itemReg.item ;
	    synchronized(itemReg) {
		if(itemReg.isDiscarded())
		    return;
		itemReg.item = newItem;
	    }			
	    itemReg.addProxy(proxy);
	    if(newItem.service.equals(oldItem.service)) {
		if(!LookupAttributes.equal(newItem.attributeSets,
					   oldItem.attributeSets)) {
		    //send serviceChanged notification.
		    changeServiceNotify(newItem, oldItem);
		}
	    } else {// the service has changed, 
		//send serviceRemove and serviceAdded notification.
		removeServiceNotify(oldItem);
		addServiceNotify(newItem);
	    }
	}
	
	/* returns true if the ServiceItem satisfies the filter. */
	private boolean checkItem(ServiceItem s) {
	    if(s == null || s.service == null )
		return false;
	    if(filter == null )
		return true;
	    return filter.check(s);
	}

	/* get expiration time of the lease */
	public long getLeaseDuration() {
	    if(leaseDuration == Long.MAX_VALUE)
		return Long.MAX_VALUE;
	    return leaseDuration  + startTime - System.currentTimeMillis();
	}

	/* notify a ServiceItem has been added. */
	private void addServiceNotify(ServiceItem item) {
	    serviceNotifyDo(null, item, ITEM_ADDED);
	} 

	/* notify a ServiceItem has been removed. */
	private void removeServiceNotify(ServiceItem item) {
	    serviceNotifyDo(item, null, ITEM_REMOVED);
	} 

	/* notify a ServiceItem has been changed, but still matches. */
	private void changeServiceNotify(ServiceItem newItem,
                                         ServiceItem oldItem )
        {
	    serviceNotifyDo(oldItem, newItem, ITEM_CHANGED);
	} 

	/* common code does service notification */
	private void serviceNotifyDo(ServiceItem oldItem,
				     ServiceItem item,
				     int action)
	{
	    ArrayList notifies;
	    synchronized(sItemListeners) {
		if(sItemListeners.isEmpty())
		    return;
	        notifies = (ArrayList)sItemListeners.clone();
	    }
	    Iterator iter = notifies.iterator();
	    ServiceDiscoveryEvent event = 
		new ServiceDiscoveryEvent(this, oldItem, item);
	    while (iter.hasNext()) {
		ServiceDiscoveryListener sl
                                      = (ServiceDiscoveryListener)iter.next();
		switch(action) {
		case ITEM_ADDED:sl.serviceAdded(event);break;
		case ITEM_REMOVED:sl.serviceRemoved(event);break;
		case ITEM_CHANGED:sl.serviceChanged(event);break;
		}
	    }
	}

	/* a debug function to print out the serviceIdMap */
	private void printServiceIdMap() {
	    if(serviceIdMap.isEmpty()) {
		System.out.println("************************** sid:");
		System.out.println("serviceIdMap is empty");
	    }
	    Iterator iter = getServiceIdMapEntrySetIterator();
	    while(iter.hasNext()) {
		Map.Entry e = (Map.Entry)iter.next();
		ServiceItemReg itemReg = (ServiceItemReg)e.getValue();
		ServiceID sid = (ServiceID)e.getKey();
		System.out.println("************************** sid:"+ sid);
		itemReg.printProxys();
	    }
	}
    }
    
    /* max matches for lookup or LookupCache */
    final private static int MAX_MATCHES = 100;
    /* stores LookupDiscoveryManager.*/
    private DiscoveryManagement discMgr;
    /* the leaseRenewalManager can been passed in, or it creates its own */
    private LeaseRenewalManager leaseRenewalMgr;
    /* Stores ServiceRegistrar that has been discovered */
    private final ArrayList proxyRegSet = new ArrayList();
    /* Stores DiscoveryListener **/
    private final ArrayList listeners = new ArrayList();
    /* Random number generator for use in lookup */
    private final Random random = new Random();
    /* to store LookupCache */
    private final ArrayList caches = new ArrayList();
    /* to store LookupCacheImpl that contains tasks */
    private final ArrayList taskQue = new ArrayList();
    /* taskThread get created if necessary */
    private TaskThread taskThread = null;
    /* flag to indicate if the ClientLookupManager has been terminated. */
    private boolean bTerminated = false;
    /* Listener class for the discovery notification */
    private class DiscMgrListener implements DiscoveryListener {
	/* new proxy has been discovered */
	public void discovered(DiscoveryEvent e) {
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    ArrayList newProxys = new ArrayList(1);
	    ArrayList notifies = null;
	    for(int i=0; i<proxys.length; i++) {
		//try {
		//   System.out.println("new proxy:" +
		//		       proxys[i].getLocator().getHost());
		//}catch(Exception ex){
		//}
		ProxyReg reg = new ProxyReg(proxys[i]);
		synchronized(proxyRegSet) {
		    proxyRegSet.add(reg);
		    newProxys.add(reg);
		}		
	    }

	    synchronized(listeners) {
		if(!listeners.isEmpty())
		    notifies = (ArrayList)listeners.clone();
	    }

	    Iterator iter = newProxys.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg = (ProxyReg)iter.next();
		cacheAddProxy(reg);
		if(notifies != null)
		    listenerDiscovered(reg.proxy, notifies);
	    }
	}
	/* proxy has been discarded */
	public void discarded(DiscoveryEvent e) {
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
	    ArrayList notifies;
	    ArrayList drops = new ArrayList(1);
	    synchronized(proxyRegSet) {
		for(int i=0; i<proxys.length; i++) {
		    ProxyReg reg = findReg(proxys[i]);
		    if(reg != null ) { // this check can be removed.
			proxyRegSet.remove(proxyRegSet.indexOf(reg));
			drops.add(reg);
		    } else
			throw new RuntimeException("discard error");

		}
	    }
	    Iterator iter = drops.iterator();
	    while(iter.hasNext())
		dropProxy((ProxyReg)iter.next());

	    synchronized(listeners) {
		if(listeners.isEmpty())
		    return;
		notifies = (ArrayList)listeners.clone();
	    }
	    listenerDropped(drops, notifies);
	}
    }

    /* add the proxy to all the LookupCache */
    private void cacheAddProxy(ProxyReg reg) {
	synchronized(caches) {
	    Iterator iter = caches.iterator();
	    while (iter.hasNext()) {
		LookupCacheImpl cache= (LookupCacheImpl)iter.next();
		cache.addProxyReg(reg);
	    }
	}
    }
    
    /* remove the proxy from all the lookupCache */
    private void dropProxy(ProxyReg reg ) {
	synchronized(caches) {
	    Iterator iter = caches.iterator();
	    while (iter.hasNext()) {
		LookupCacheImpl cache= (LookupCacheImpl)iter.next();
		cache.removeProxyReg(reg);
	    }
	}
    }

    /**
     * Constructs an instance of <code>ClientLookupManager</code> which will,
     * on behalf of the entity that constructs this class, discover and
     * manage a set of lookup services, as well as discover and manage sets
     * of services registered with those lookup services. The entity indicates
     * which lookup services to discover and manage through the parameters
     * input to this constructor.
     * <p>
     * As stated in the class description, this class has three usage patterns:
     * <p>
     * <ul>
     *   <li> the entity uses a <code>LookupCache</code> to locally store and
     *        manage discovered services so that those services can be accessed
     *        quickly
     *   <li> the entity registers with the event mechanism provided by a 
     *        <code>LookupCache</code> to be notified when services of interest
     *        are discovered
     *   <li> the entity uses the <code>ClientLookupManager</code> to perform
     *        remote queries of the lookup services, employing richer 
     *        semantics than that provided through the standard
     *        <code>ServiceRegistrar</code> interface
     * </ul>
     * <p>
     * Although the first two usage patterns emphasize the use of a cache 
     * object, that cache is acquired only through an instance of the
     * <code>ClientLookupManager</code> class.
     * <p>
     * It is important to note that some of the methods of this class 
     * (<code>createLookupCache</code> and the <i>blocking</i> versions of
     * <code>lookup</code> to be exact) can throw a
     * <code>RemoteException</code> when invoked. This is because each of
     * these methods may attempt to register with the event mechanism 
     * of at least one lookup service, a process that requires a remote object
     * (a listener) to be exported to the lookup service(s). Both the process
     * of registering with a lookup service's event mechanism and the process
     * of exporting a remote object are processes that can result in a 
     * <code>RemoteException</code>. 
     * <p>
     * In order to facilitate the exportation of the remote listener just
     * described, the <code>ClientLookupManager</code> class instantiates an
     * inner class that implements the <code>RemoteEventListener</code>
     * interface. Although this class defines, instantiates, and exports this
     * remote listener, <i>it is the entity's responsibility</i> to provide a
     * mechanism for any lookup service to acquire the stub of the exported
     * listener. The common mechanism employed for such a task is to 
     * store the necessary class files in a JAR file, advertise the location
     * of that JAR file by setting the <code>java.rmi.server.codebase</code> 
     * property of the entity to "point" at the JAR file, and run an HTTP
     * server to serve up the JAR file to any requesting lookup service.
     * <p>
     * If the entity does not provide a mechanism for lookup services to access
     * the remote listener exported by the <code>ClientLookupManager</code>,
     * the remote methods of the <code>ClientLookupManager</code> - the
     * methods involved in the two most important usage patterns of the
     * <code>ClientLookupManager</code> - will be of no use to the entity.
     * Because providing such a mechanism is of such importance, an example
     * of the key steps necessary to provide the mechanism are provided below.
     * Note that this is just one example of how to provide such a mechanism.
     * There may be other mechanisms that are more appropriate for some
     * entities and configurations.
     * <p>
     * Suppose the entity that intends to use the 
     * <code>ClientLookupManager</code> is named <i>myClient</i> and will run
     * on a host named <i>myHost</i>. Additionally, suppose that the 
     * <i>down-loadable</i> JAR file that myClient will make available 
     * to the lookup services that myClient's instance of the
     * <code>ClientLookupManager</code> will interact with is named 
     * <i>myClient-dl.jar</i>. Finally, suppose that myClient-dl.jar is
     * located in the directory <i>/files/clients/jars</i>, and will be
     * served by an HTTP server listening on port <i>8082</i>.
     * <p>
     * Then do the following:
     * <p>
     * <ul>
     *   <li> place in myClient-dl.jar, the class
 <code>net.jini.lookup.ClientLookupManager$LookupCacheImpl$LookupListener_Stub.class</code> (found in the Jini distribution)
     *   <li> place the class 
              <code>net.jini.core.event.RemoteEventListener.class</code>
              in myClient-dl.jar
     *   <li> place any other classes (unrelated to the remote listener
     *        exportation process described above) that other entities may
     *        need to download from myClient
     *   <li> start an HTTP server, listening on port 8082, that will serve
     *        JAR files from directory /files/clients/jars
     *   <li> start myClient with a codebase property set in the following way:
     *        <code>-Djava.rmi.server.codebase="http://myHost:8082/files/clients/jars/myClient-dl.jar"</code>
     *   <li> 
     * </ul>
     * <p>
     * The lookup services should now be able to access the remote listener
     * exported by the <code>ClientLookupManager</code> on behalf of the
     * <i>myClient</i> entity. Note that although it is not actually necessary
     * to place <code>net.jini.core.event.RemoteEventListener.class</code> in
     * myClient-dl.jar (because the lookup services should already have access
     * to that class), it is recommended practice to include in the JAR
     * file, all classes on which the primary class to be downloaded depends.
     * <p>
     * This constructor takes two arguments: an object that implements the
     * <code>DiscoveryManagement</code> interface and a reference to a
     * <code>LeaseRenewalManager</code> object. The constructor throws an
     * IOException because construction of a
     * <code>ClientLookupManager</code> may initiate the multicast
     * discovery process, a process that can throw an
     * <code>IOException</code>.
     *
     * @param discoveryMgr the <code>DiscoveryManagement</code> 
     *  		implementation through which notifications 
     *  		that indicate a lookup service has been 
     *  		discovered or discarded will be received.
     *  		If the value of the argument is <code>null</code>,
     *                  then an instance of the
     *  		<code>LookupDiscoveryManager</code> utility
     *  		class will be constructed to listen for events
     *  		announcing the discovery of only those lookup
     *  		services that are members of the public group.
     * 
     * @param leaseMgr the <code>LeaseRenewalManager</code> to use. A 
     *  		value of <code>null</code> may be passed as the
     *  		<code>LeaseRenewalManager</code> argument. If
     *  		the value of the argument is <code>null</code>,
     *  		an instance of the
     *  		<code>LeaseRenewalManager</code> class will be
     *  		created, initially managing no
     *  		<code>Lease</code> objects.
     * 
     * @throws IOException because construction of a  
     *  		<code>ClientLookupManager</code> may initiate
     *  		the multicast discovery process which can throw
     *  		an <code>IOException</code>.
     *
     * @see net.jini.discovery.DiscoveryManagement
     * @see net.jini.core.event.RemoteEventListener
     * @see net.jini.core.lookup.ServiceRegistrar
     */
    public ClientLookupManager(DiscoveryManagement discoveryMgr,
			       LeaseRenewalManager leaseMgr)
    throws IOException
    {
	DiscMgrListener l =  new DiscMgrListener();
	leaseRenewalMgr = ( leaseMgr != null) ? leaseMgr :
	                  new LeaseRenewalManager();
	if(discoveryMgr == null)
	    discMgr = new LookupDiscoveryManager(null, null, null);
	else 
	    discMgr = discoveryMgr; 
	discMgr.addDiscoveryListener(l); 
    }

    /* send discovered notify message */
    private void discoveredNotifyListener(DiscoveryListener l) {
	ServiceRegistrar[] proxys;
	synchronized(proxyRegSet) {
	    if(proxyRegSet.isEmpty())
		return;
	    proxys = buildServiceRegistrar();
	}
	DiscoveryEvent evt =  
	    new DiscoveryEvent(this, proxys);
	l.discovered(evt);

    }
    
    /* notify a list of ServiceRegistrar has been discarded */
    private void listenerDropped(ArrayList drops, ArrayList notifies) {
	ServiceRegistrar[] proxys = new ServiceRegistrar[drops.size()];
	drops.toArray(proxys);
	listenerDropped(proxys, notifies);
    }

    /* notify a array of ServiceRegistrar has been discarded */
    private void listenerDropped(ServiceRegistrar[] proxys, ArrayList notifies)
    {
	Iterator iter = notifies.iterator();
	while (iter.hasNext()) {
	    DiscoveryEvent evt = 
		new DiscoveryEvent(this, (ServiceRegistrar[])proxys.clone());
	    ((DiscoveryListener)iter.next()).discarded(evt);
	}
    }

    /* notify a new ServiceRegistrar has been discovered */
    private void listenerDiscovered(ServiceRegistrar proxy, ArrayList notifies)
    {
	Iterator iter = notifies.iterator();
	while (iter.hasNext()) {
	    DiscoveryEvent evt = 
		new DiscoveryEvent(this, new ServiceRegistrar[]{proxy});
	    ((DiscoveryListener)iter.next()).discovered(evt);
	}
    }
   
    /* returns array of ServiceRegistrar from our proxyRegSet */
    private ServiceRegistrar[] buildServiceRegistrar() {
	int k = 0;
	ServiceRegistrar[] proxys =
	    new ServiceRegistrar[proxyRegSet.size()];
	Iterator iter = proxyRegSet.iterator();
	while(iter.hasNext()) {
	    ProxyReg reg = (ProxyReg)iter.next();
	    proxys[k++] = reg.proxy;
	}
	return proxys;
    }

    /**
     * returns an Array of ServiceRegistrar that has been discovered.
     */
    public ServiceRegistrar[] getLookupServices() {
	checkTerminated();
	synchronized(proxyRegSet) {
	    return buildServiceRegistrar(); 
	}
    }

    /**
     * Queries each available lookup service in the set of lookup services
     * managed by the <code>ClientLookupManager</code> (the <i>managed set</i>)
     * for a service reference that matches criteria defined by the entity
     * that invokes this method. The semantics of this method are similar to
     * the semantics of the <code>lookup</code> method provided by the
     * <code>ServiceRegistrar</code> interface; employing the same 
     * template-matching scheme. Additionally, this method allows any entity
     * to supply an object referred to as a <i>filter</i>. Such an object is
     * a non-remote object that defines additional matching criteria that the
     * <code>ClientLookupManager</code> applies when searching for the
     * entity's services of interest. This filtering facility is particularly
     * useful to entities that wish to extend the capabilities of standard
     * template-matching.
     * <p>
     * Entities typically employ this method when they need infrequent access
     * to services, and when the cost of making remote queries is outweighed
     * by the overhead of maintaining a local cache (for example, because of
     * resource limitations).
     * <p>
     * This version of <code>lookup</code> returns a <i>single</i> instance
     * of <code>ServiceItem</code> corresponding to one of possibly many
     * service references that satisfy the matching criteria. If multiple
     * services matching the input criteria happen to exist, it is arbitrary
     * as to which reference is actually returned. It is for this reason that
     * entities that invoke this method typically care only that <i>a</i>
     * service is returned, not <i>which</i> service.
     * <p>
     * Note that, unlike other versions of <code>lookup</code> provided
     * by the <code>ClientLookupManager</code>, this version does not 
     * <i>block</i>. That is, this version will return immediately upon
     * failure (or success) to find a service matching the input criteria.
     *
     * @param tmpl   an instance of <code>ServiceTemplate</code> corresponding
     *               to the object to use for template-matching when searching
     *               for desired services. If <code>null</code> is input to
     *               this parameter, this method will use a <i>wildcarded</i>
     *               template (will match all services) when performing
     *               template-matching. Note that the effects of modifying
     *               contents of this parameter before this method returns
     *               are unpredictable and undefined.
     * @param filter an instance of <code>ServiceItemFilter</code> containing
     *               matching criteria that should be applied in addition to
     *               the template-matching employed when searching for desired
     *               services. If <code>null</code> is input to this parameter,
     *               then only template-matching will be employed to find the 
     *               desired services.
     *
     * @return a single instance of <code>ServiceItem</code> corresponding to
     *         a reference to a service that matches the criteria represented
     *         in the input parameters; or <code>null</code> if no matching
     *         service can be found. Note that if multiple services matching
     *         the input criteria exist, it is arbitrary as to which reference
     *         is returned.
     *
     * @see net.jini.core.lookup.ServiceRegistrar#lookup
     * @see net.jini.core.lookup.ServiceTemplate
     * @see net.jini.lookup.ServiceItemFilter
     */
    public ServiceItem lookup(ServiceTemplate tmpl, ServiceItemFilter filter)
    { 
	checkTerminated();
	ServiceRegistrar[] proxys;
	synchronized(proxyRegSet) {
	    proxys =  buildServiceRegistrar();
	}
	int len = proxys.length;
	if(len == 0 ) return null;

	int rand = Math.abs(random.nextInt()) % len;
	for(int i=0; i<len; i++) {
	    ServiceRegistrar proxy = proxys[(i + rand) % len];
	    ServiceItem sItem = null;
	    try {
		ServiceMatches sm = proxy.lookup(tmpl, MAX_MATCHES);
		sItem = getMatchedServiceItem(sm, filter);
	    } catch(RemoteException e) {
		discard(proxy);
	    } catch(Exception e) {
	    }
	    if(sItem != null) {
		return sItem;
	    }
	}
	return null;
    }

    /**
     * Queries each available lookup service in the managed set for a service
     * that matches the input criteria. The semantics of this method are
     * similar to the semantics of the <code>lookup</code> method provided by
     * the <code>ServiceRegistrar</code> interface; employing the same 
     * template-matching scheme. Additionally, this method allows any entity
     * to supply an object referred to as a <i>filter</i>. Such an object is
     * a non-remote object that defines additional matching criteria that the
     * <code>ClientLookupManager</code> applies when searching for the
     * entity's services of interest. This filtering facility is particularly
     * useful to entities that wish to extend the capabilities of standard
     * template-matching.
     * <p>
     * This version of <code>lookup</code> returns a <i>single</i> instance
     * of <code>ServiceItem</code> corresponding to one of possibly many
     * service references that satisfy the matching criteria. If multiple
     * services matching the input criteria happen to exist, it is arbitrary
     * as to which reference is actually returned. It is for this reason that
     * entities that invoke this method typically care only that <i>a</i>
     * service is returned, not <i>which</i> service.
     * <p>
     * Note that this version of <code>lookup</code> provides a 
     * <i>blocking</i> feature that is controlled through the
     * <code>wait</code> parameter. That is, this version will not return
     * until either a service that matches the input criteria has been
     * found, or the amount of time contained in the <code>wait</code>
     * parameter has passed. If, while waiting for the service of interest
     * to be found, the entity decides that it no longer wishes to wait the
     * entire period for this method to return, the entity may interrupt this
     * method by invoking the interrupt method from the <code>Thread</code>
     * class. The intent of this mechanism is to allow the entity to interrupt
     * this method in the same way it would a sleeping thread.
     * <p>
     * Entities typically employ this method when they need infrequent access
     * to services, are willing (or forced) to wait for those services to be
     * found, and consider the cost of making remote queries for those
     * services is outweighed by the overhead of maintaining a local cache
     * (for example, because of resource limitations).
     *
     * @param tmpl   an instance of <code>ServiceTemplate</code> corresponding
     *               to the object to use for template-matching when searching
     *               for desired services. If <code>null</code> is input to
     *               this parameter, this method will use a <i>wildcarded</i>
     *               template (will match all services) when performing
     *               template-matching. Note that the effects of modifying
     *               contents of this parameter before this method returns
     *               are unpredictable and undefined.
     * @param filter an instance of <code>ServiceItemFilter</code> containing
     *               matching criteria that should be applied in addition to
     *               the template-matching employed when searching for desired
     *               services. If <code>null</code> is input to this parameter,
     *               then only template-matching will be employed to find the 
     *               desired services.
     * @param wait   the amount of time (in milliseconds) to wait before
     *               ending the "search" and returning <code>null</code>.
     *               If a non-positive value is input to this parameter,
     *               then this method will not wait; it will simply query
     *               the available lookup services and return a matching 
     *               service reference or <code>null</code>.
     *
     * @return a single instance of <code>ServiceItem</code> corresponding to
     *         a reference to a service that matches the criteria represented
     *         in the input parameters; or <code>null</code> if no matching
     *         service can be found. Note that if multiple services matching
     *         the input criteria exist, it is arbitrary as to which reference
     *         is returned.
     * 
     * @throws java.lang.InterruptedException this exception occurs when the
     *         entity interrupts this method by invoking the interrupt method
     *         from the <code>Thread</code> class.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         a RemoteException occurs either as a result of an attempt
     *         to export a remote listener, or an attempt to register with the
     *         event mechanism of a lookup service.
     *
     * @see net.jini.core.lookup.ServiceRegistrar#lookup
     * @see net.jini.core.lookup.ServiceTemplate
     * @see net.jini.lookup.ServiceItemFilter
     * @see java.lang.Thread
     */
    public ServiceItem lookup(ServiceTemplate tmpl,
                              ServiceItemFilter filter,
			      long wait)
	throws InterruptedException, RemoteException
    {
	checkTerminated();
	ServiceItem sm = null;
	//System.out.println("*****sm:" + sm);
	if(sm != null)
	    return sm;
	ServiceDiscoveryListener cacheListener
                                        = new ServiceDiscoveryListenerImpl();
	LookupCacheImpl cache
                        = createLookupCache(tmpl, filter, cacheListener, wait);
	long duration = cache.getLeaseDuration();
	synchronized(cacheListener) {
	    while ( duration > 0 ) {
		cacheListener.wait(duration);
		sm = cache.lookup(null);
		if(sm != null )
		    return sm;
		duration = cache.getLeaseDuration();
	    }
	}
	cache.terminate();
	return sm;
    }

    /**
     * The <code>createLookupCache</code> method allows the client-like
     * entity to request that the <code>ClientLookupManager</code>
     * create a new managed set (or cache) and populate it with
     * services, which match criteria defined by the entity, and whose
     * references are registered with one or more of the lookup
     * services the entity has targeted for discovery.
     * <p>
     * This method returns an object of type <code>LookupCache</code>.
     * Through this return value, the entity can query the cache for
     * services of interest, manage the cache's event mechanism for
     * service discoveries, or terminate the cache.
     * <p>
     * An entity typically uses the object returned by this method to
     * provide local storage of, and access to, references to services
     * that it is interested in using. Entities needing frequent access
     * to numerous services will find the object returned by this
     * method quite useful because acquisition of those service
     * references is provided through local method invocations.
     * Additionally, because the object returned by this method provides
     * an event mechanism, it is also useful to entities wishing to
     * simply monitor, in an event-driven manner, the state changes that
     * occur in the services of interest.
     * <p>
     * Although not required, a common usage pattern for entities that
     * wish to use the <code>LookupCache</code> class to store and manage
     * "discovered" services is to create a separate cache for each service
     * type of interest.
     * 
     * @param tmpl template to match. It uses template-matching
     *        semantics to identify the service(s) to acquire from 
     *        lookup services in the managed set. If this value is 
     *        <code>null</code>, it is the equivalent of passing a 
     *        <code>ServiceTemplate</code> constructed with all 
     *        <code>null</code> arguments (all wildcards).
     * @param filter used to apply additional matching criteria to any 
     *        <code>ServiceItem</code> found through template-matching.
     *        If this value is <code>null</code>, no additional filtering
     *        will be applied beyond the template-matching. 
     * @param listener object that will receive notifications when 
     *        services matching the input criteria are discovered for 
     *        the first time, or have encountered a state change such as 
     *        removal from all lookup services or attribute set changes. 
     *        If this value is <code>null</code>, the cache resulting from 
     *        that invocation will send no such notifications.
     *
     * @return LookupCache used to query the cache for services of
     * 	      interest, manage the cache's event mechanism for service
     *        discoveries, or terminate the cache.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         a RemoteException occurs either as a result of an attempt
     *         to export a remote listener, or an attempt to register with the
     *         event mechanism of a lookup service.
     *
     * @see net.jini.lookup.ServiceItemFilter
     */
    public LookupCache createLookupCache(ServiceTemplate tmpl, 
					 ServiceItemFilter filter,
					 ServiceDiscoveryListener listener)
	throws RemoteException
    {
	checkTerminated();
	return createLookupCache(tmpl, filter, 
				 listener, Long.MAX_VALUE);
    }

    /** The <code>getDiscoveryManager</code> method will return an
     * object that implements the <code>DiscoveryManagement</code>
     * interface. The object returned by this method provides the
     * <code>ClientLookupManager</code> with the ability to set
     * discovery listeners and to discard previously discovered lookup
     * services when they are found to be unavailable.
     *
     * @return DiscoveryManagement implementation
     * @see net.jini.discovery.DiscoveryManagement
     */
    public DiscoveryManagement getDiscoveryManager() {
	return discMgr;
    }

    /** 
     * The <code>getLeaseRenewalManager</code> method will return an
     * instance of the <code>LeaseRenewalManager</code> class. The
     * object returned by this method manages the leases requested and
     * held by the <code>ClientLookupManager</code>. In general, these 
     * leases correspond to the registrations made by the 
     * <code>ClientLookupManager</code> with the event mechanism of
     * each lookup service in the managed set. 
     *
     * @return LeaseRenewalManager for this <code>ClientLookupManager</code>.
     * @see net.jini.lease.LeaseRenewalManager
     */
    public LeaseRenewalManager getLeaseRenewalManager() {
	return leaseRenewalMgr;
    }
    
    /** 
     * The <code>terminate</code> method performs cleanup duties
     * related to the termination of the event mechanism for lookup
     * service discovery, the event mechanism for service discovery,
     * and the cache management duties of the
     * <code>ClientLookupManager</code>.  
     * <p> 
     * For each instance of <code>LookupCache</code> created and 
     * managed by the <code>ClientLookupManager</code>, the
     * <code>terminate</code> method will do the following:
     * <ul>
     * <li>Either remove all listener objects registered for receipt
     * of <code>DiscoveryEvent</code> objects or, if the discovery
     * manager employed by the <code>ClientLookupManager</code> was
     * created by the <code>ClientLookupManager</code> itself,
     * terminate all discovery processing being performed by that
     * manager object on behalf of the entity.
     * <p> 
     * <li>Cancel all event leases granted by each lookup service in
     * the managed set of lookup services.
     * <p>
     * <li>Un-export all remote listener objects registered with each 
     * lookup service in the managed set.  
     * <p> 
     * <li>Terminate all threads involved in the process of retrieving
     * and storing references to discovered services of interest.
     * </ul> 
     * Calling any method after the termination will result in a
     * <code>RuntimeException</code>.
     *
     * @see net.jini.lookup.LookupCache
     * @see net.jini.discovery.DiscoveryEvent
     */
    public void terminate() {
	synchronized(this) {
	    bTerminated = true;
	}
	synchronized(taskQue) {
	    if(taskThread != null)
		taskThread.interrupt();
	}
	
	ArrayList cachesClone;
	synchronized(caches) {
	    if(caches.isEmpty())
		return;
	    cachesClone = (ArrayList)caches.clone();
	}
	Iterator iter = cachesClone.iterator();
	while (iter.hasNext()) {
	    LookupCacheImpl cache= (LookupCacheImpl)iter.next();
	    cache.terminate();
	}
    }

    /**
     * Queries each available lookup service in the managed set for service(s)
     * that match the input criteria. The semantics of this method are
     * similar to the semantics of the <code>lookup</code> method provided by
     * the <code>ServiceRegistrar</code> interface; employing the same 
     * template-matching scheme. Additionally, this method allows any entity
     * to supply an object referred to as a <i>filter</i>. Such an object is
     * a non-remote object that defines additional matching criteria that the
     * <code>ClientLookupManager</code> applies when searching for the
     * entity's services of interest. This filtering facility is particularly
     * useful to entities that wish to extend the capabilities of standard
     * template-matching.
     * <p>
     * Entities typically employ this method when they need infrequent access
     * to multiple instances of services, and when the cost of making remote
     * queries is outweighed by the overhead of maintaining a local cache
     * (for example, because of resource limitations).
     * <p>
     * This version of <code>lookup</code> returns an <i>array</i> of instances
     * of <code>ServiceItem</code> in which each element corresponds to a
     * service reference that satisfies the matching criteria. The number
     * of elements in the returned set will be no greater than the value of
     * the <code>maxMatches</code> parameter, but may be less.
     * <p>
     * Note that this version of <code>lookup</code> does not provide a 
     * <i>blocking</i> feature. That is, this version will return immediately
     * with whatever number of service references it can find, up to
     * the number indicated in the <code>maxMatches</code> parameter. If
     * no services matching the input criteria can be found on the first
     * attempt, an empty array is returned.
     *
     * @param tmpl       an instance of <code>ServiceTemplate</code>
     *                   corresponding to the object to use for 
     *                   template-matching when searching for desired services.
     *                   If <code>null</code> is input to this parameter, 
     *                   this method will use a <i>wildcarded</i> template 
     *                   (will match all services) when performing
     *                   template-matching. Note that the effects of modifying
     *                   contents of this parameter before this method returns
     *                   are unpredictable and undefined.
     * @param maxMatches this method will return no more than this number of
     *                   service references
     * @param filter     an instance of <code>ServiceItemFilter</code> 
     *                   containing matching criteria that should be applied 
     *                   in addion to the template-matching employed when 
     *                   searching for desired services. If <code>null</code> 
     *                   is input to this parameter, then only 
     *                   template-matching will be employed to find the 
     *                   desired services.
     *
     * @return an array of instances of <code>ServiceItem</code> where each
     *         element corresponds to a reference to a service that matches
     *         the criteria represented in the input parameters; or an
     *         empty array if no matching service can be found.
     *
     * @see net.jini.core.lookup.ServiceRegistrar#lookup
     * @see net.jini.core.lookup.ServiceTemplate
     * @see net.jini.lookup.ServiceItemFilter
     */
    public ServiceItem[] lookup(ServiceTemplate tmpl,
                                int maxMatches,
			        ServiceItemFilter filter )
    { 
	checkTerminated();
	if (maxMatches < 1)
	    throw new IllegalArgumentException("maxMatches must be > 0");

	ServiceRegistrar[] proxys;
	synchronized(proxyRegSet) {
	    proxys =  buildServiceRegistrar();
	}
	int len = proxys.length;
	ArrayList sItemSet = new ArrayList(len);
	if(len > 0) {
	    int rand = Math.abs(random.nextInt()) % len;
	    for(int i=0; i<len; i++) {
		ServiceRegistrar proxy = proxys[(i + rand) % len];
		try {
		    ServiceMatches sm = proxy.lookup(tmpl, MAX_MATCHES);
		    int totalMatches = sm.totalMatches ;		
		    if(totalMatches == 0)
			continue;
		    int r = Math.abs(random.nextInt()) % totalMatches;
		    for(int j=0; j<totalMatches; j++) {
			ServiceItem sItem = sm.items[(j+r) % totalMatches];
			if(sItem == null)
			    continue;
			if(filter != null && (!filter.check(sItem)))
			    continue;
			if(!sItemSet.contains(sItem))
			    sItemSet.add(sItem);
			if(sItemSet.size() >= maxMatches) {
			    break;
			}
		    }
		} catch(RemoteException e) {
		    discard(proxy);
		} catch(Exception e) {
		    System.out.println(e);
		}
	    }
	}
	ServiceItem [] r = new ServiceItem[sItemSet.size()];
	sItemSet.toArray(r);
	return r; 
    }

    /**
     * Queries each available lookup service in the managed set for service(s)
     * that match the input criteria. The semantics of this method are
     * similar to the semantics of the <code>lookup</code> method provided by
     * the <code>ServiceRegistrar</code> interface; employing the same 
     * template-matching scheme. Additionally, this method allows any entity
     * to supply an object referred to as a <i>filter</i>. Such an object is
     * a non-remote object that defines additional matching criteria that the
     * <code>ClientLookupManager</code> applies when searching for the
     * entity's services of interest. This filtering facility is particularly
     * useful to entities that wish to extend the capabilities of standard
     * template-matching.
     * <p>
     * This version of <code>lookup</code> returns an <i>array</i> of instances
     * of <code>ServiceItem</code> in which each element corresponds to a
     * service reference that satisfies the matching criteria. The number
     * of elements in the returned set will be no greater than the value of
     * the <code>maxMatches</code> parameter, but may be less.
     * <p>
     * Note that this version of <code>lookup</code> provides a 
     * <i>blocking</i> feature that is controlled through the
     * <code>wait</code> parameter in conjunction with the
     * <code>minMaxMatch</code> and the <code>maxMatches</code> parameters.
     * This method will not return until one of the following occurs:
     * <p>
     *  <ul>
     *    <li> the number of matching services found on the first attempt is
     *         greater than or equal to the value of the
     *         <code>minMaxMatch</code> parameter, in which case this method
     *         returns each of the services found up to the value of 
     *         the <code>maxMatches</code> parameter
     *    <li> the number of matching services found <i>after</i> the first
     *         attempt (that is, after the method enters the "wait state")
     *         is at least as great as the value of the
     *         <code>minMaxMatch</code> parameter in which case this method
     *         returns each of the services found up to the value of 
     *         the <code>maxMatches</code> parameter
     *    <li> the amount of time that has passed since this method entered
     *         the wait state exceeds the value of the <code>wait</code>
     *         parameter, in which case this method returns all of the
     *         currently discovered services
     *  </ul>
     * <p>
     * The purpose of the <code>minMaxMatch</code> parameter is to allow the
     * entity to balance its need for multiple matching service references
     * with its need to minimize the time spent in the wait state; time that
     * most would consider wasted if an acceptable number of matching service
     * references were found, but this method continued to wait until the end
     * of the designated time period.
     * <p>
     * If, while waiting for the minimum number of desired services to 
     * be discovered, the entity decides that it no longer wishes to wait the
     * entire period for this method to return, the entity may interrupt this
     * method by invoking the interrupt method from the <code>Thread</code>
     * class. The intent of this mechanism is to allow the entity to interrupt
     * this method in the same way it would a sleeping thread.
     * <p>
     * Entities typically employ this method when they need infrequent access
     * to multiple instances of services, are willing (or forced) to
     * wait for those services to be found, and consider the cost of making
     * remote queries for those services is outweighed by the overhead
     * of maintaining a local cache (for example, because of resource
     * limitations).
     *
     * @param tmpl        an instance of <code>ServiceTemplate</code>
     *                    corresponding to the object to use for 
     *                    template-matching when searching for desired
     *                    services. If <code>null</code> is input to this
     *                    parameter, this method will use a
     *                    <i>wildcarded</i> template  (will match all 
     *                    services) when performing template-matching. Note 
     *                    that the effects of modifying contents of this 
     *                    parameter before this method returns are 
     *                    unpredictable and undefined.
     * @param minMaxMatch this method will immediately exit the wait state
     *                    and return once this number of service references
     *                    is found
     * @param maxMatches  this method will return no more than this number of
     *                    service references
     * @param filter      an instance of <code>ServiceItemFilter</code> 
     *                    containing matching criteria that should be applied 
     *                    in addion to the template-matching employed when 
     *                    searching for desired services. If <code>null</code> 
     *                    is input to this parameter, then only 
     *                    template-matching will be employed to find the 
     *                    desired services.
     * @param wait        the amount of time (in milliseconds) to wait before
     *                    ending the "search" and returning an empty array.
     *                    If a non-positive value is input to this parameter,
     *                    then this method will not wait; it will simply query
     *                    the available lookup services and return whatever
     *                    matching service reference(s) it could find, up
     *                    to <code>maxMatches</code>.
     *
     * @return an array of instances of <code>ServiceItem</code> where each
     *         element corresponds to a reference to a service that matches
     *         the criteria represented in the input parameters; or an
     *         empty array if no matching service can be found within the
     *         time allowed.
     * 
     * @throws java.lang.InterruptedException this exception occurs when the
     *         entity interrupts this method by invoking the interrupt method
     *         from the <code>Thread</code> class.
     * 
     * @throws java.lang.IllegalArgumentException this exception occurs when
     *         one of the following conditions is satisfied:
     * <p><ul> <li>the <code>minMaxMatch</code> parameter is non-positive
     *         <li>the <code>maxMatches</code> parameter is non-positive
     *         <li>the value of <code>maxMatches</code> is <i>less than</i>
     *             the value of <code>minMaxMatch</code>
     *    </ul>
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         a RemoteException occurs either as a result of an attempt
     *         to export a remote listener, or an attempt to register with the
     *         event mechanism of a lookup service.
     *
     * @see net.jini.core.lookup.ServiceRegistrar#lookup
     * @see net.jini.core.lookup.ServiceTemplate
     * @see net.jini.lookup.ServiceItemFilter
     * @see java.lang.Thread
     */
    public ServiceItem[] lookup(ServiceTemplate tmpl,
                                int minMaxMatch,
                                int maxMatches,
			        ServiceItemFilter filter,
                                long wait )
	throws InterruptedException, RemoteException
    { 

	checkTerminated();

	if (minMaxMatch < 1)
	    throw new IllegalArgumentException("minMaxMatch must be > 0");

	if (maxMatches < minMaxMatch)
	    throw new IllegalArgumentException
                                     ("maxMatches must be > minMaxMatch");

	ServiceItem [] sItems = lookup(tmpl, maxMatches, filter);
	if(sItems.length >= minMaxMatch)
	    return sItems;
	
	ArrayList sItemSet = new ArrayList(sItems.length);
	for(int i=0; i<sItems.length; i++) {
	    //if(!sItemSet.contains(sItems[i])
	    sItemSet.add(sItems[i]);
	}
	ServiceDiscoveryListenerImpl cacheListener =
	    new ServiceDiscoveryListenerImpl();
	LookupCacheImpl cache
                       = createLookupCache(tmpl, filter, cacheListener, wait);
	long duration = cache.getLeaseDuration();
	synchronized(cacheListener) {
	    while ( duration > 0 ) {
		cacheListener.wait(duration);
		ServiceItem items[] = cacheListener.getServiceItem();
		for(int i=0; i<items.length; i++) {
		    if(!sItemSet.contains(items[i]))
			sItemSet.add(items[i]);
		}
		if(sItemSet.size() == minMaxMatch)
		    break;
		duration = cache.getLeaseDuration();
	    }
	}
	cache.terminate();
	ServiceItem [] r = new ServiceItem[sItemSet.size()];
	sItemSet.toArray(r);
	return r; 
    }
    
    private ServiceItem getMatchedServiceItem(ServiceMatches sm, 
					      ServiceItemFilter filter)
    {
	int len = sm.totalMatches;
	if(len > 0) {
	    int rand = Math.abs(random.nextInt()) % len;
	    for(int i=0; i<len; i++) {
		ServiceItem sItem = sm.items[(i+rand) % len];
		if(sItem == null)
		    continue;
		if(filter != null && (!filter.check(sItem)))
		    continue;
		return sItem;
	    }
	}
	return null;
    }

    /* create a LookupCache with specific lease duration */
    private LookupCacheImpl createLookupCache(ServiceTemplate tmpl, 
					 ServiceItemFilter filter,
					 ServiceDiscoveryListener listener,
					 long leaseDuration)
	throws RemoteException
    {
	if(tmpl == null)
	    tmpl = new ServiceTemplate(null, null, null);

	LookupCacheImpl cache = 
	    new LookupCacheImpl(tmpl,filter,listener, leaseDuration);
	synchronized(caches) {
	    caches.add(cache);
	}
	return cache;
    }

    /* find the ProxyReg that has the proxy equals to the given proxy */
    private ProxyReg findReg(ServiceRegistrar proxy) {
	Iterator iter = proxyRegSet.iterator();
	while(iter.hasNext()) {
	    ProxyReg reg =(ProxyReg)iter.next();
	    if(reg.proxy.equals(proxy))
		return reg;
	}
    	return null;
    }

    /* discard a ServiceRegistrar.*/
    private void discard(ServiceRegistrar proxy) {
	discMgr.discard(proxy);
    }

    /* cancel a event lease */
    private void cancelLease(Lease lease ) {
	try {
	    leaseRenewalMgr.cancel(lease);
	} catch (Exception e) {
	}
    }

    /*
     * This only get called from the taskThread.
     * register an eventLease. If caught RemoteException when try to register
     * the Event, discard the proxy and returns null.
     */
    private EventRegistration registerListener(ServiceRegistrar proxy,
					       ServiceTemplate tmpl,
					       RemoteEventListener l,
					       long duration)
    {
	EventRegistration e = null;
	try {
	    e = proxy.notify(tmpl,
			     ServiceRegistrar.TRANSITION_NOMATCH_MATCH |
			     ServiceRegistrar.TRANSITION_MATCH_NOMATCH |
			     ServiceRegistrar.TRANSITION_MATCH_MATCH,
			     l, 
			     null,
			     duration); 
	    leaseRenewalMgr.renewUntil(e.getLease(),
				       duration,
				       new LeaseListenerImpl(proxy));
	} catch (RemoteException ex) {
	    discard(proxy);
	}
	return e;
    }

    /* throws RuntimeException if it has been terminated. */
    private void checkTerminated() throws RuntimeException {
	synchronized(this) {
	    if(bTerminated)
		throw new RuntimeException("Terminated Error");
	}

    }

    static private ServiceTemplate copyServiceTemplate(ServiceTemplate tmpl) {
	Class[] serviceTypes = null;
	Entry[] attributeSetTemplates = null;
	if(tmpl.serviceTypes != null) {
	    int len = tmpl.serviceTypes.length;
	    serviceTypes = new Class[len];
	    System.arraycopy(tmpl.serviceTypes, 0, serviceTypes, 0, len );
	}
	    
 	if(tmpl.attributeSetTemplates != null) {
	    int len =  tmpl.attributeSetTemplates.length;
	    attributeSetTemplates = new Entry[len];
	    System.arraycopy(tmpl.attributeSetTemplates, 0,
                             attributeSetTemplates, 0, len); 
	}
	return new ServiceTemplate(tmpl.serviceID,
                                   serviceTypes,
                                   attributeSetTemplates);
    }
}
