/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.lookup;
  
import net.jini.core.lookup.ServiceItem;

/**
 * The LookupCache interface defines the methods provided by the object
 * created and returned by the ClientLookupManager when a client-like
 * entity invokes the createLookupCache method. It is within that object
 * that discovered service references, which match criteria defined by
 * the entity, are stored. Through this interface the entity may retrieve
 * one or more of the stored service references, register and un-register
 * with the cache's event mechanism, and terminate all of the cache's
 * processing. This interface is used by ClientLookupManager
 *
 * @see ClientLookupManager
 * @author Zane Pan
 */
public interface LookupCache {
   /** 
     * Finds a <code>ServiceItem</code> object that satisfies the
     * filter. 
     * <p>
     * The service item returned must have been previously discovered to 
     * be both registered with one or more of the lookup services in the 
     * managed set, and to match criteria defined by the entity.
     * <p>
     * The semantics of the <code>filter</code> argument are
     * identical to those of the <code>filter</code> argument specified
     * for a number of the methods defined in the interface of the
     * <code>ClientLookupManager</code> utility class. This argument is
     * intended to allow an entity to separate its filtering into two
     * steps: an initial filter applied during the discovery phase, and
     * a finer resolution filter applied upon retrieval from the cache.
     * As with the methods of the <code>ClientLookupManager</code>, if
     * <code>null</code> is the value of this argument, then no additional
     * filtering will be performed.
     *
     * @param filter used for matching <code>ServiceItem</code>s. A null 
     * 		     value means no additional filter should be applied.
     * 		    
     * @return ServiceItem that satisfies the filter, and that was 
     * 			previously discovered to be registered with one
     * 			or more lookup services in the managed set. A 
     * 			<code>null</code> value will be returned if no
     * 			<code>ServiceItem</code> is found that matches the 
     * 			criteria or if the cache is empty.
     * 			
     */
    public ServiceItem lookup(ServiceItemFilter filter);

   /** 
     * Finds an array of <code>ServiceItem</code>s that satisfy the
     * filter.
     * <p>
     * The service item returned must have been previously discovered to 
     * be both registered with one or more of the lookup services in the 
     * managed set, and to match criteria defined by the entity.
     * <p>
     * The semantics of the <code>filter</code> argument are
     * identical to those of the <code>filter</code> argument specified
     * for a number of the methods defined in the interface of the
     * <code>ClientLookupManager</code> utility class. This argument is
     * intended to allow an entity to separate its filtering into two
     * steps: an initial filter applied during the discovery phase, and
     * a finer resolution filter applied upon retrieval from the cache.
     * As with the methods of the <code>ClientLookupManager</code>, if
     * <code>null</code> is the value of this argument, then no
     * additional filtering will be performed.
     * 
     * @param filter used for matching <code>ServiceItem</code>s. A null 
     * 		     value means no additional filter should be applied.
     * 		    
     * @param maxMatches maximum number of matches to return. If this 
     *			value is set to <code>Integer.MAX_VALUE</code>
     *			then all elements in the cache that match the
     *			criteria will be returned.
     *
     * @return ServiceItem[] that satisfy the filter, and that were 
     * 			previously discovered to be registered with one
     * 			or more lookup services in the managed set. An 
     * 			empty array will be returned if no
     *                  <code>ServiceItem</code> is found that matches
     * 			the criteria or if the cache is empty.
     *
     * @throws IllegalArgumentException if <code>maxMatches</code> is a
     *			negative number.
     */
    public ServiceItem[] lookup(ServiceItemFilter filter, int maxMatches);

   /** 
     * Registers a <code>ServiceDiscoveryListener</code> object with
     * the event mechanism of a <code>LookupCache</code>. The listener 
     * object will receive a <code>ServiceDiscoveryEvent</code> upon the 
     * discovery, removal, or modification of one of the cache's
     * services. Once a listener is registered, it will be notified of
     * all service references discovered to date, and will be notified as
     * new services are discovered and existing services are modified or
     * discarded. If the parameter value is <code>null</code>, no action 
     * will be taken.
     *
     * @param l the <code>ServiceDiscoveryListener</code> object to
     * 			register.
     */
    public void addListener(ServiceDiscoveryListener l);


   /** 
     * Removes a <code>ServiceDiscoveryListener</code> object from the set  
     * of listeners currently registered with the <code>LookupCache</code>.
     * Once all listeners are removed from the cache's set of listeners, 
     * the cache will send no more <code>ServiceDiscoveryEvent</code> 
     * notifications.
     *
     * @param l the <code>ServiceDiscoveryListener</code> object to remove.
     */
    public void removeListener(ServiceDiscoveryListener l);

   /** 
     * Deletes a service reference from the cache and causes a notification 
     * to be sent to all registered listeners indicating that the service 
     * has been discarded . Note that after the service has been discarded, 
     * there is no guarantee that the service will again become eligible
     * for discovery.
     *
     * @param serviceReference the service reference to
     * 		discard.
     */
    public void discard(Object serviceReference);

   /** 
     * Performs cleanup duties related to the termination of the processing
     * being performed by a particular instance of
     * <code>LookupCache</code>. For that instance, this method cancels
     * all event leases granted by the lookup services that supplied the
     * contents of the cache, and un-exports all remote listener objects
     * registered with those lookup services. The terminate method is
     * typically called when the entity is no longer interested in the
     * contents of the <code>LookupCache</code>.
     */
    void terminate();
}
