/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JXP
 */

package net.jini.lookup;

import net.jini.core.lookup.ServiceItem;
/**
 * The <code>ServiceDiscoveryListener</code> interface defines the
 * methods used by objects such as a <code>LookupCache</code> to notify
 * an entity that events of interest related to the elements of the cache
 * have occurred. It is the responsibility of the entity wishing to be
 * notified of the occurrence of such events to construct an object
 * that implements the <code>ServiceDiscoveryListener</code> interface
 * and then register that object with the cache's event mechanism. Any
 * implementation of this interface must define the actions to take upon
 * receipt of an event notification. The action taken is dependent on
 * both the application and the particular event that has occurred.
 * <p>
 * Should an exception occur during an invocation of any of the methods
 * defined by this interface, the semantics of how that exception is
 * handled are undefined.
 * <p>
 * Each method defined by this interface must not modify the contents
 * of the <code>ServiceDiscoveryEvent</code> parameter; doing so can
 * result in unpredictable and undesirable effects on future processing
 * by the <code>ClientLookupManager</code>. It is for this reason that if
 * one of these methods modifies the contents of the parameter, the
 * effects are undefined.
 * <p>
 * This interface makes the following concurrency guarantee. For any
 * given listener object that implements this interface, no two methods
 * (either the same two methods or different methods) defined by the
 * interface can be invoked at the same time by the same cache. For
 * example, the serviceRemoved method must not be invoked while the
 * invocation of another listener's <code>serviceAdded</code> method is
 * in progress.
 * <p>
 * Finally, it should be noted that the intent of the methods of this
 * interface is to allow the recipient of the 
 * <code>ServiceDiscoveryEvent</code> to be informed that a service has been
 * added to, removed from, or modified in the cache. Calls to these methods
 * are synchronous to allow the entity that makes the call (for example, a
 * thread that interacts with the various lookup services of interest) to
 * determine whether or not the call succeeded. However, it is not part of
 * the semantics of the call that the notification return can be delayed
 * while the recipient of the call reacts to the occurrence of the event.
 * Thus, it is highly recommended that implementations of this interface
 * avoid time consuming operations, and return from the method as quickly
 * as possible. For example, one strategy might be to simply note the
 * occurrence of the <code>ServiceDiscoveryEvent</code>, and perform any
 * time consuming event handling asynchronously. 
 *
 * @see LookupCache
 * @see ServiceDiscoveryEvent
 * @author Zane Pan
 */
public interface ServiceDiscoveryListener {
    /** 
     * When the cache receives from one of the managed lookup services,
     * an event signaling the <i>registration</i> of a service of
     * interest for the <i>first time</i> (or for the first time since
     * the service has been discarded), the cache invokes the
     * <code>serviceAdded</code> method on all instances of
     * <code>ServiceDiscoveryListener</code> that are registered with the
     * cache; doing so notifies the entity that a service of interest has
     * been discovered.
     *
     * @param event an instance of <code>ServiceDiscoveryEvent</code> 
     * 			containing references to the service item 
     *			corresponding to the event, including 
     * 			representations of the service's state both 
     *			before and after the event.
     */
    void serviceAdded(ServiceDiscoveryEvent event);

    /** 
     * When the cache receives, from a managed lookup service, an event
     * signaling the removal of a service of interest from the last such
     * lookup service with which it was registered, the cache invokes
     * the <code>serviceRemoved</code> method on all instances of
     * <code>ServiceDiscoveryListener</code> that are registered with
     * the cache; doing so notifies the entity that a service of interest 
     * has been discarded.
     * 
     * @param event a <code>ServiceDiscoveryEvent</code> object 
     * 			containing references to the service item 
     *			corresponding to the event, including 
     *			representations of the service's state both 
     *			before and after the event.
     */
    void serviceRemoved(ServiceDiscoveryEvent event);

    /** 
     * When the cache receives, from a managed lookup service, an event
     * signaling the unique modification of the attributes of a service
     * of interest (across the attribute sets of all references to the
     * service), the cache invokes the <code>serviceChanged</code>
     * method on all instances of <code>ServiceDiscoveryListener</code>
     * that are registered with the cache; doing so notifies the entity
     * that the state of a service of interest has changed.
     * 
     * @param event a <code>ServiceDiscoveryEvent</code> object
     * 			containing references to the service item 
     * 			corresponding to the event, including 
     * 			representations of the service's state both
     * 			before and after the event.
     */
    void serviceChanged(ServiceDiscoveryEvent event);
}
