/*
 * Copyright 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun(TM) Community Source
 * License, Jini(TM) Technology Core Platform, v. 1.0 (the "License"); 
 * you may not use this file except in compliance with the License. You 
 * may obtain a copy of the License at http://www.sun.com/jini. Software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Jini Technology Core Platform code, version 1.0.
 * The Developer of the Original Code is Sun Microsystems, Inc. Portions
 * created by Sun Microsystems, Inc. are Copyright (C) 1999 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology
 * Core Platform Compatibility Kit Version 1.0
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s): Java(TM) 2 SDK v1.2.2 Solaris(TM) Reference
 *                   Implementation (Final) on Solaris 2.6
 * 
 *                   Java 2 SDK v1.2.2 Microsoft Windows 95/98/NT 
 *                   Production Release (Final) on Microsoft Windows 
 *                   NT 4.0 SP4
 * 
 * CopyrightVersion 1.1Alpha_JSK
 */
package com.sun.jini.fiddler;

import com.sun.jini.debug.Debug;

import com.sun.jini.fiddler.Fiddler;
import com.sun.jini.fiddler.FiddlerAdminProxy;
import com.sun.jini.fiddler.FiddlerProxy;
import com.sun.jini.fiddler.FiddlerRegistration;
import com.sun.jini.fiddler.FiddlerRenewResults;
import com.sun.jini.fiddler.FiddlerStartParams;

import com.sun.jini.constants.VersionConstants;

import com.sun.jini.lookup.entry.BasicServiceType;
import com.sun.jini.lookup.entry.LookupAttributes;

import com.sun.jini.proxy.UUID;
import com.sun.jini.proxy.UUIDFactory;

import com.sun.jini.reliableLog.ReliableLog;
import com.sun.jini.reliableLog.LogException;
import com.sun.jini.reliableLog.LogHandler;

import com.sun.jini.thread.ReadersWriter;
import com.sun.jini.thread.ReadersWriter.ConcurrentLockException;
import com.sun.jini.thread.TaskManager;

import net.jini.discovery.DiscoveryEvent;
import net.jini.discovery.DiscoveryListener;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.discovery.LookupDiscoveryRegistration;
import net.jini.discovery.LookupUnmarshalException;
import net.jini.discovery.RemoteDiscoveryEvent;

import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.lookup.entry.Comment;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.Status;
import net.jini.lookup.entry.StatusType;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.lease.Lease;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lookup.ServiceRegistrar;

import java.lang.reflect.Array;

import java.net.InetAddress;
import java.net.MalformedURLException;

import java.io.InterruptedIOException;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Serializable;

import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationSystem;
import java.rmi.MarshalledObject;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.RemoteObject;
import java.rmi.server.UnicastRemoteObject;

import java.security.SecureRandom;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;

/**
 * This class is the server side of an implementation of the lookup
 * discovery service. Multiple client-side proxy classes are used to
 * interact and communicate with this back-end server. Those proxy
 * classes are: <code>FiddlerProxy</code> (the proxy for the 
 * <code>LookupDiscoveryService</code> interface which defines how clients
 * register with the lookup discovery service), <code>FiddlerAdminProxy</code>
 * (the proxy for the <code>FiddlerAdmin</code> interface which specifies 
 * the methods through which administration duties such as joining, 
 * storage location, persistent state logging policy, and shutting
 * down the lookup discovery service can be performed), and
 * <code>FiddlerRegistration</code> (the proxy for the 
 * <code>LookupDiscoveryRegistration</code> interface which defines the
 * methods through which clients can perform duties such as group and
 * locator management, state retrieval, and discarding discovered but
 * unavailable lookup services so they are eligible for re-discovery).
 * <p>
 * It is through the proxies that communicate with this class that clients
 * interact with the lookup discovery service. When a client makes a method
 * invocation on one of the proxies, the proxy makes a corresponding call
 * on the methods specified in the <code>Fiddler</code> interface which
 * are implemented in this class, ultimately executing on the back-end server.
 */
public class FiddlerImpl implements Fiddler {

    /* ServiceInfo values */
    private static final String PRODUCT      = "Lookup Discovery Service";
    private static final String MANUFACTURER = "Sun Microsystems, Inc.";
    private static final String VENDOR       = MANUFACTURER;
    private static final String VERSION      = VersionConstants.SERVER_VERSION;

    /** When re-setting the bound on lease durations, that bound cannot be
     *  set to a value larger than this value */
    private static final long MAX_LEASE = 1000L * 60 * 60 * 24 * 365 * 1000;
    /** Log format version */
    private static final int LOG_VERSION = 1;

    /** The primary proxy to the current instance of this class */
    private FiddlerProxy fiddlerProxy;
    /** The service ID assigned by the lookup service to the instance
     *  of the lookup discovery service for which the current instance
     *  of this class serves as implementation
     */
    private ServiceID serviceID;
    /** The activation id of the current instance of the lookup discovery
     *  service, if it happens to be and activatable instance
     */
    private ActivationID activationID;
    /** The unique identifier assigned to each registration object created
     *  by the current instance of the lookup discovery service. This ID
     *  is used as an indicator of equality between registration objects, 
     *  as well as key into the various maps associated with each such
     *  registration and managed by the lookup discovery service.
     */
    private UUID proxyID = null;

    /**
     * Map from registrationID to registrationInfo.  Every registration is in
     * this map under its registrationID.
     */
    private final HashMap registrationByID = new HashMap(11);
    /**
     * Map from registrationInfo to registrationInfo (that is, to itself),
     * where the elements of the map are ordered by lease expiration time.
     */
    private final TreeMap registrationByTime = new TreeMap();

    /** Performs all group and locator discovery */
    private LookupDiscoveryManager discoveryMgr = null; 
    /**
     * The listener registered for both group discovery events and locator
     * discovery events.
     */
    private final LookupDiscoveryListener discoveryListener 
                                               = new LookupDiscoveryListener();

    /** For each registration created by the lookup discovery service, an
     *  event identifier that uniquely maps the registration to the
     *  registration's listener and managed sets will be generated and
     *  associated with the registration through the EventRegistration
     *  field of the registrationInfo. This event ID is unique across
     *  all registrations with the current instance of the lookup discovery
     *  service.
     */
    private long curEventID = 0;

    /** random number generator for random lease ID generation */
    private final SecureRandom secRand = new SecureRandom();
    /** 64-bit buffer for use with secRand */
    private final byte[] secRandBuf8 = new byte[8];

    /** Earliest expiration time over all active registrations */
    private long minExpiration = Long.MAX_VALUE;

    /** The lookup discovery manager the join manager will use */
    private LookupDiscoveryManager joinMgrLookupDiscoveryMgr;
    /** Manager for discovering and registering with lookup services */
    private JoinManager joinMgr;
    /** Task manager for sending remote discovery events */
    private final TaskManager taskMgr = new TaskManager(10, 1000 * 15, 1.0f);
    /** Registration lease expiration thread */
    private final Thread leaseExpireThread = new LeaseExpireThread();
    /** Snapshot-taking thread */
    private final Thread snapshotThread = new SnapshotThread();

    /** Concurrent object to control read and write access */
    private final ReadersWriter concurrentObj = new ReadersWriter();
    /** Object for synchronizing with the registration expire thread */
    private final Object leaseExpireThreadSyncObj = new Object();
    /** Object on which the snapshot-taking thread will synchronize */
    private final Object snapshotThreadSyncObj = new Object();

    /** Reliable log object to hold persistent state of the service */
    private ReliableLog log;
    /** Flag indicating whether system is in a state of recovery */
    private boolean inRecovery;
    /** Current number of records in the Log File since the last snapshot */
    private int logFileSize = 0;
    /** The name of the directory to which the persistent state of this
     *  instance of the lookup discovery service will be logged (using
     *  ReliableLog).
     */
    private String logDirname;
    /** least upper bound applied to all granted lease durations */
    private long leaseDurationBound = 1000 * 60 * 30;
    /** Weight factor applied to snapshotSize when deciding to take snapshot */
    private float snapshotWt = 10;
    /** Log File must contain this many records before snapshot allowed */
    private int logToSnapshotThresh = 200;

    /** Groups whose members are lookup services this service should join */
    private String[] thisServicesGroups = {""}; // default to public group
    /** Locators of specific lookup services this service should join */
    private LookupLocator[] thisServicesLocators = {};
    /** The attributes to use when joining lookup service(s) */
    private Entry[] thisServicesAttrs
            = new Entry[]
                 { new ServiceInfo(PRODUCT,MANUFACTURER,VENDOR,VERSION,"",""),
                   new BasicServiceType("Lookup Discovery Service")
                 };
    /* Simple object supporting selective debug output to files or display */
    private Debug debugObj;
    /* Debug PrintWriters for the various debug options */
    private PrintWriter debugAllWriter     = null;//all debug info written
    private PrintWriter debugStartupWriter = null;//only startup info written
    private PrintWriter debugTasksWriter   = null;//tasks' run method info
    private PrintWriter debugEventsWriter  = null;//info on sent events
    private PrintWriter debugGroupsWriter  = null;//group info over all regs
    private PrintWriter debugLocsWriter    = null;//locator info over all regs
    private PrintWriter debugDiscardWriter = null;//info on discard processing
    private PrintWriter debugLeaseWriter   = null;//info on lease processing
    private PrintWriter debugRegWriter     = null;//info on registrations
    private PrintWriter debugLogWriter     = null;//info on logging process

    /* ************************* BEGIN Constructors ************************ */
    /* Constructor for the Activatable case */
    /**
     * Constructs a new instance of FiddlerImpl. This version of the
     * constructor is used to create an activatable, persistent instance 
     * of the lookup discovery service.
     * <p>
     * A constructor having this signature is required for the class to be
     * activatable. This constructor is automatically called by the 
     * activation group when the lookup discovery service is activated.
     * 
     * @param activationID the activation ID generated by the activation
     *                     system and assigned to the instance of the lookup
     *                     discovery service being constructed
     * @param data         the data structure (represented as a 
     *                     <code>MarshalledObject</code>) which contains the
     *                     following data:
     *                     <ul>
     *                     <li> the name of the directory to which the 
     *                          persistent state of this service will be logged
     *                     <li> the set containing the names of the groups
     *                          whose members are the lookup services that
     *                          should be discovered and joined
     *                     <li> the set containing the locators corresponding
     *                          to the specific lookup services to discover
     *                          and join
     *                     </ul>
     *
     * @throws java.io.IOException this exception can occur when there is a
     *         failure during the export process
     */
    public FiddlerImpl(ActivationID activationID,
                       MarshalledObject data) throws IOException
    {
        this.activationID = activationID;
        Activatable.exportObject(this, activationID, 0);
        try {
            FiddlerStartParams params = (FiddlerStartParams)data.get();
            logDirname           = params.logDirname;
            thisServicesGroups   = params.groups;
            thisServicesLocators = getLocatorsFromStartParams(params.locators);
        } catch (Exception e) {
            throw new IllegalArgumentException
                     ("bad values (logDirname,groups,locators) in "
                      +"MarshalledObject data structure");
        }
        init();
    }

    /* Constructor for the NON-Activatable, NON-persistent (transient) case */
    /**
     * Constructs a new instance of FiddlerImpl. This version of the
     * constructor is used to create a NON-activatable, NON-persistent
     * instance of the lookup discovery service, often referred to as a
     * <i>transient</i> lookup discovery service.
     *
     * @param  logDirname <code>String</code> containing the absolute path
     *         of the directory directory in which the state of the service
     *         will be recorded
     * @param  groups <code>String</code> array containing the names of the
     *         groups whose members are the lookup services that should be
     *         discovered and joined
     * @param  locators <code>LookupLocator</code> array containing elements
     *         that correspond to specific lookup services to discover and
     *         join
     *
     * @throws java.io.IOException this exception can occur when there is a
     *         failure during the export process
     */
    FiddlerImpl(String          logDirname,
                String[]        groups,
                LookupLocator[] locators) throws IOException
    {
        constructTransientInstance(logDirname,groups,locators);
    }

    /**
     * Constructs a new instance of FiddlerImpl. This version of the
     * constructor is used to create a NON-activatable, NON-persistent
     * instance of the lookup discovery service, often referred to as a
     * <i>transient</i> lookup discovery service.
     *
     * @param  logDirname <code>String</code> containing the absolute path
     *         of the directory directory in which the state of the service
     *         will be recorded
     * @param  groups <code>String</code> array containing the names of the
     *         groups whose members are the lookup services that should be
     *         discovered and joined
     * @param  locatorHostnames <code>String</code> array containing the names
     *         of the hosts on which the specific lookup services to discover
     *         and join are running
     *
     * @throws java.io.IOException this exception can occur when there is a
     *         failure during the export process
     */
    FiddlerImpl(String   logDirname,
                String[] groups,
                String[] locatorHostnames) throws IOException
    {
        LookupLocator[] locators = new LookupLocator[0];
        if ( (locatorHostnames != null) && (locatorHostnames.length > 0) ) {
            locators = new LookupLocator[locatorHostnames.length];
            for(int i=0;i<locatorHostnames.length;i++) {
                try {
                    locators[i] = new LookupLocator
                                      ("jini://"+locatorHostnames[i]+":4160/");
                } catch (MalformedURLException e) {
                    System.out.println("MalformedURLException from "+
                                       "LookupLocatorDiscovery constructor:\n"
                                       +e.toString());
                }
            }
        }
        constructTransientInstance(logDirname,groups,locators);
    }
    /* ************************* END Constructors ************************** */

    /* ******************* BEGIN Inner Class Definitions ******************* */
    /** Class which is used to communicate the status of this service to
     *  interested entities. In particular, when certain errors occur during
     *  operation, an instance of this class will be registered as an 
     *  attribute in all of the lookup services with which this service
     *  is registered. By registering for notification (from the lookup
     *  services) of the existence of this attribute, interested entities
     *  such as administrative clients and clients wishing to use this
     *  service will be informed when this service can not proceed with its
     *  processing, and can take appropriate action.
     */
    private class FiddlerStatus extends Status
    {
        public FiddlerStatus(StatusType severity) {
            super(severity);
        }
    }//end class FiddlerStatus

    /** Class which listens for the announcement of a service ID assignment
     *  by the JoinManager
     */
    private class SrvcIDListener implements ServiceIDListener
    {
        public SrvcIDListener() {
            super();
        }
        /** The JoinManager will invoke this method when it receives a
         *  valid ServiceID from a lookup service; passing in that 
         *  ServiceID so that it may be stored for persistence.
         */
        public void serviceIDNotify(ServiceID serviceID) {
            FiddlerImpl.this.serviceID = serviceID;
            /* Take a snapshot so the serviceID is recorded. This should occur
             * only once during the life of any one particular instance of
             * this service: when the service is instantiated for the very
             * first time. That is, once a lookup service assigns a serviceID
             * to such a service instance, that instance should use the
             * assigned serviceID forever, until the service is formally
             * destroyed (through the invocation of the <code>destroy</code>
             * method of the <code>DestroyAdmin</code> interface).
             */
            try {
	        log.snapshot();
            } catch(IOException e) {
                writeInfoLog(debugAllWriter,debugLogWriter,
                             "IOException -- while taking snapshot in "
                             +"SrvcIDListener: "+e.toString());
            }
            writeInfoServiceID();
        }
    }//end class SrvcIDListener

    /** Class whose discovered() method is invoked by threads in the 
     *  LookupDiscovery class whenever a new lookup service is discovered
     *  on behalf of a client registration
     */
    private class LookupDiscoveryListener implements DiscoveryListener
    {
        public LookupDiscoveryListener() {
            super();
        }
        public void discovered(DiscoveryEvent event) {
            taskMgr.add(new DiscoveredEventTask(event));
        }
        public void discarded(DiscoveryEvent event) {
            taskMgr.add(new DiscardedEventTask(event));
        }
    }//end class LookupDiscoveryListener

    /** This class acts as a record of one registration with the lookup
     *  discovery service; containing all of the information about that
     *  registration.
     */
    private final static class RegistrationInfo
                                          implements Comparable, Serializable
    {
        private static final long serialVersionUID = -42790777303907567L;

        /** The unique identifier assigned to the registration to which the
         *  data in the current implementation of this class corresponds.
         *  This identifier is unique across all other active registrations
         *  generated with the current instance of the lookup discovery
         *  service.
         *  @serial
         */
        public final UUID registrationID;
        /** Map from the set of instances of the <code>ServiceRegistrar</code>
         *  interface, to the set of marshalled instances of the
         *  <code>ServiceRegistrar</code> interface, where each key and
         *  each value (which is the marshalled form of its corresponding
         *  key) is a proxy to one of lookup service(s) that have been
         *  discovered for the current registration. The contents of
         *  this set represents the 'remote state' of the registration's
         *  currently discovered lookup service(s).
         *  @serial
         */
        public HashMap discoveredRegsMap;
        /** The managed set containing the names of the groups whose
         *  members are the lookup services the lookup discovery service
         *  should attempt to discover for the current registration.
         *  (HashSet is used to prevent duplicates.)
         *  @serial
         */
        public HashSet groups;
        /** The managed set containing the locators of the specific lookup
         *  services the lookup discovery service should attempt to discover
         *  for the current registration. (HashSet is used to prevent
         *  duplicates.)
         *  @serial
         */
        public HashSet locators;
        /** The ID of the lease placed on the current registration.
         *  @serial
         */
        public final long leaseID;
        /** The absolute expiration time of the current lease.
         *  @serial
         */
        public long leaseExpiration;
        /** The identifier that maps the current registration to the remote
         *  event listener and the managed set of groups and locators.
         */
        public long eventID;
        /** The current sequence number of the set of remote discovery events
         *  sent to the current registration's listener. When a registration
         *  is granted, this class is instantiated to contain the information
         *  related to that particular registration. The event sequence
         *  number is initialized to 0 upon instantiation because the
         *  remote discovery events are sent to the listeners of each 
         *  separate registration. Thus, each registration has its own
         *  sequence of events.
         *  @serial
         */
        public long seqNum;
        /** The handback object returned with every remote discovery event
         *  sent to the current registration's listener.
         *  @serial
         */
        public final MarshalledObject handback;
        /** When the lookup discovery service discards a registrar as a 
         *  result of some internal condition (such as multicast announcements
         *  ceasing) and not as a result of a request from a registration,
         *  every registration configured for discovery of that discarded
         *  registrar will be sent a remote discarded event. On the other
         *  hand, for the case where a registrar is discarded as a result
         *  of a request from a registration, only those registrations that
         *  actually request that the registrar be discarded will be sent
         *  a remote discarded event. This flag is used to determine whether
         *  to send a remote discarded event to one or multiple listeners.
         */
        public boolean discardFlag;
        /** The remote event listener registered by the client. This field
         *  is transient because it is marshalled separately from the rest
         *  of this class when being serialized. (See the description for
         *  <code>writeObject</code> below.)
         */
        public transient RemoteEventListener listener;

        /** Constructs an instance of this class and stores the information
         *  related to the current registration: IDs, managed sets, lease
         *  information, and event registration information.
         */
        public RegistrationInfo(UUID registrationID,
                                String[] groups,
                                LookupLocator[] locators,
                                long leaseID,
                                long leaseExpiration,
                                long eventID,
                                MarshalledObject handback,
                                RemoteEventListener listener)
        {
            this.registrationID = registrationID;
            /* Initialize the groups field, removing nulls and duplicates */
            if(groups != null) {
                this.groups = new HashSet();
                for(int i=0;i<groups.length;i++) {
                    if(groups[i] == null) continue;
                    this.groups.add(groups[i]);
                }
            }
            /* Initialize the locators field, removing nulls and duplicates */
            this.locators = new HashSet();
            if( (locators != null) && (locators.length > 0) ) {
                for(int i=0;i<locators.length;i++) {
                    if(locators[i] == null) continue;
                    this.locators.add(locators[i]);
                }
            }
            this.discoveredRegsMap = new HashMap(11);
            this.leaseID = leaseID;
            this.leaseExpiration = leaseExpiration;

            this.eventID = eventID;
            this.seqNum = 0; // initialize to 0
            this.handback = handback;
            this.discardFlag = false;//set true only on first discard request
            this.listener = listener;
        }//end constructor

        /** Determines which of the registrars from the input set are also in
         *  the set of desired registrars associated with the current instance
         *  of this class.
         */
        public ServiceRegistrar[] intersectDesiredRegs(ServiceRegistrar[] regs)
        {
            /* If configured for ALL_GROUPS, then every registrar is desired */
            if(groups == null) {
                return regs;
            }
            /* Not configured for ALL_GROUPS, find the finite intersection */
            HashSet regSet = new HashSet(); //no duplicates
            /* Loop thru the already-discovered registrars, retrieving 
             * each's locator and the groups each is a member of
             */
            nextRegistrar:
                for(int i=0;i<regs.length;i++) {
                    /* Look for a locator match first */
                    LookupLocator curLoc = null;
                    try {
                        curLoc = regs[i].getLocator();
                    } catch (RemoteException e) { 
                        continue nextRegistrar; //comm problem, next registrar
                    }
                    /* Loop thru the registration's locators of interest */
                    for( Iterator itr=locators.iterator();itr.hasNext(); ) {
                        LookupLocator regLoc = (LookupLocator)itr.next();
                        if( regLoc.equals(curLoc) ) {
                            regSet.add(regs[i]);
                            continue nextRegistrar; // match found, next reg
                        }//endif
                    }//end loop
                    /* Determine if the current registrar is a member of any of
                     * this class' desired groups.
                     */
                    String[] curGroups = null;
                    try {
                        curGroups = regs[i].getGroups();
                    } catch (RemoteException e) { 
                        continue nextRegistrar; //comm problem, next registrar
                    }
                    for(int j=0;j<curGroups.length;j++) {
                        /* Loop thru desired groups until match is found */
                        for( Iterator itr=groups.iterator();itr.hasNext(); ) {
                            String regGroup = (String)itr.next();
                            if(regGroup.equals(curGroups[j]) ) {
                                regSet.add(regs[i]);
                                continue nextRegistrar;//match found, next reg
                            }//endif
                        }//end loop(itr)
                    }//end loop(j)
                }//end loop(i)
            //end label: nextRegistrar
            return ( (ServiceRegistrar[])(regSet).toArray
                                      (new ServiceRegistrar[regSet.size()]) );
        }//end intersectDesiredRegs

        /** Determines which of the registrars from the input set are also in
         *  the set of discovered registrars associated with the current
         *  instance of this class.
         */
        public ServiceRegistrar[] intersectDiscoveredRegs
                                                     (ServiceRegistrar[] regs)
        {
            /* Get the array of discovered regs */
            Set regKeys = discoveredRegsMap.keySet();
            ServiceRegistrar[] dRegs =
                  (ServiceRegistrar[])(regKeys).toArray
                                        (new ServiceRegistrar[regKeys.size()]);
            HashSet regSet = new HashSet(); //no duplicates
            /* Compare each input registrar to each registrar in the set of
             * currently discovered registrars, storing matches along the way
             */
            nextRegistrar:
                for(int i=0;i<regs.length;i++) {
                    for(int j=0;j<dRegs.length;j++) {
                        if( (regs[i]).equals(dRegs[j]) ) {
                            regSet.add(regs[i]);
                            continue nextRegistrar; // match found, next reg
                        }//endif
                    }//end loop(j)
                }//end loop(i)
            //end label: nextRegistrar
            return ( (ServiceRegistrar[])(regSet).toArray
                                      (new ServiceRegistrar[regSet.size()]) );
        }//end intersectDiscoveredRegs

        /** Determines the complement of the set of currently discovered 
         *  registrars with the set of desired registrars, both corresponding
         *  to the current instance of this class. That is, this method
         *  will return a set containing only those registrars that have
         *  been discovered but which are no longer in the set of "desired"
         *  registrars (desired through both locator and group discovery).
         *  <p>
         *  For example, suppose a client requests a registration configured
         *  for locator discovery for a set of locators corresponding to
         *  the registrars {r0,r1}, and for group discovery of the groups
         *  {g0,g1}; where registrars r0 and r2 are members of g0, and
         *  registrar r3 is a member of g1. This means that the registration's
         *  "desired" set of registrars is {r0,r1,r2,r3}.
         *  <p>
         *  Next suppose registrars r0, r2 and r3 are discovered, but not r1.
         *  This means that the registration's "discovered" set of registrars
         *  is {r0,r2,r3}. Now suppose the registration requests the removal
         *  of the locator corresponding to r0 as well as both group g0 and
         *  g1 from the "desired" sets. After the removal is complete, the
         *  set of "desired" registrars becomes {r3}, while the set of
         *  discovered registrars is still {r1,r2,r3}.
         *  <p>
         *  If this method is invoked after the removal is performed, the
         *  set {r2,r3} will be returned since r2 and r3 are both in the set
         *  of discovered registrars but neither is in the set of desired 
         *  registrars.
         */
        public ServiceRegistrar[] complementDiscoveredWithDesired() {
            /* If configured for ALL_GROUPS, then return the empty set since
             * the group(s) to which every discovered registrar is a member
             * is included by ALL_GROUPS
             */
            if(groups == null) {
                return new ServiceRegistrar[0];
            }
            /* Get the array of discovered regs */
            Set regKeys = discoveredRegsMap.keySet();
            ServiceRegistrar[] regs =
                  (ServiceRegistrar[])(regKeys).toArray
                                        (new ServiceRegistrar[regKeys.size()]);
            /* Not configured for ALL_GROUPS, find the finite intersection */
            HashSet regSet = new HashSet(); //no duplicates
            /* Loop thru the already-discovered registrars, retrieving 
             * each's locator and the groups each is a member of
             */
            nextRegistrar:
                for(int i=0;i<regs.length;i++) {
                    /* Look for a locator match first */
                    LookupLocator curLoc = null;
                    try {
                        curLoc = regs[i].getLocator();
                    } catch (RemoteException e) { 
                        continue nextRegistrar; //comm problem, next registrar
                    }
                    /* Loop thru the registration's locators of interest */
                    for( Iterator itr=locators.iterator();itr.hasNext(); ) {
                        LookupLocator regLoc = (LookupLocator)itr.next();
                        if( regLoc.equals(curLoc) ) {
                            continue nextRegistrar; // match found, next reg
                        }//endif
                    }//end loop
                    /* Determine if the current registrar is a member of any of
                     * this class' desired groups.
                     */
                    String[] curGroups = null;
                    try {
                        curGroups = regs[i].getGroups();
                    } catch (RemoteException e) { 
                        continue nextRegistrar; //comm problem, next registrar
                    }
                    for(int j=0;j<curGroups.length;j++) {
                        /* Loop thru desired groups until match is found */
                        for( Iterator itr=groups.iterator();itr.hasNext(); ) {
                            String regGroup = (String)itr.next();
                            if(regGroup.equals(curGroups[j]) ) {
                                continue nextRegistrar;//match found, next reg
                            }//endif
                        }//end loop(itr)
                    }//end loop(j)
                    regSet.add(regs[i]);// no match found, add it to return set
                }//end loop(i)
            //end label: nextRegistrar
            return ( (ServiceRegistrar[])(regSet).toArray
                                      (new ServiceRegistrar[regSet.size()]) );
        }//end complementDiscoveredWithDesired

        /**
         * Attempts to marshal each element of the input set of instances of
         * the <code>ServiceRegistrar</code> interface and then map the
         * registrar to its marshalled form, and store the mapping in the
         * registration's <code>discoveredRegsMap</code> field.
         * <p>
         * This method is typically invoked to handle discovered (as opposed
         * to discarded) registrars. Note that if a particular registrar
         * cannot be serialized (marshalled), it is not included in the
         * mapping; nor is it included in the return set.
         * 
         * @param regs the registrars to serialize and store
         * 
         * @return an array containing the instances of ServiceRegistrar that
         *         were successfully serialized (marshalled) and stored in
         *         the registration's <code>discoveredRegsMap</code> field
         * <p>     
         *         If there are no such registrars, this method returns an
         *         array of zero length.
         */
        public ServiceRegistrar[] addToDiscoveredRegs(ServiceRegistrar[] regs){
            HashSet regSet = new HashSet(); //no duplicates
            if( (regs != null) && (regs.length > 0) ) {
                nextRegistrar:
                    for(int i=0;i<regs.length;i++) {
                        /* If already in map, go to next registrar */
                        if( discoveredRegsMap.containsKey(regs[i]) ) {
                            continue nextRegistrar;
                        }
                        /* It doesn't contain it, try to marshal it */
                        MarshalledObject mReg = null;
                        try {
                            mReg = new MarshalledObject(regs[i]);
                        } catch(IOException e) {
                            continue nextRegistrar;//failed, go to next reg
                        }
                        /* Succeeded, map registrar to its marshalled form */
                        discoveredRegsMap.put(regs[i],mReg);
                        regSet.add(regs[i]);// add the reg to the return set
                    }//end loop
                //end label: nextRegistrar
            }//endif
            return ( (ServiceRegistrar[])(regSet).toArray
                                      (new ServiceRegistrar[regSet.size()]) );
        }//end addToDiscoveredRegs

        /**
         * For each key of the registration's <code>discoveredRegsMap</code>
         * field that equals an element of the <code>regs</code> parameter,
         * removes from the map the corresponding (key,value) pair.
         */
        public void removeFromDiscoveredRegs(ServiceRegistrar[] regs) {
            if( (regs != null) && (regs.length > 0) ) {
                nextRegistrar:
                    for(int i=0;i<regs.length;i++) {
                        discoveredRegsMap.remove(regs[i]);
                    }//end loop
                //end label: nextRegistrar
            }//endif
        }//end removeFromDiscoveredRegs

        /** Performs a primary sort by leaseExpiration, and a secondary sort
         *  by registrationID. The secondary sort is immaterial, except to
         *  ensure a total order (required by <code>TreeMap</code>).
         */
        public int compareTo(Object obj) {
            RegistrationInfo regInfo = (RegistrationInfo)obj;
            if (this == regInfo) {
                return 0;
            }
            if (    (leaseExpiration < regInfo.leaseExpiration)
                 || (    (leaseExpiration == regInfo.leaseExpiration)
		      && (eventID < regInfo.eventID) )  )
            {
                return -1;
            }
            return 1;
        }//end compareTo

        /** When a registration is granted to a client, the client registers
         *  a remote listener with the lookup discovery service so that the
         *  lookup discovery service may send remote discovery events to the
         *  client. The client typically annotates the listener with an RMI
         *  codebase from which the back-end server can download the remote
         *  listener's proxy (stub). When the current registration is logged
         *  to persistent storage (for example, a snapshot is taken), the
         *  listener is written to the output snapshot or log file through
         *  an <code>ObjectOutputStream</code> which only serializes the
         *  listener; it does not marshal the listener. Thus, when the
         *  listener field of this class is logged, the codebase from which
         *  to retrieve the listener is not included in the output. 
         *
         *  In order to include the codebase with the listener when saving
         *  state, the following custom <code>writeObject</code> method
         *  is provided which first serializes the current instance of
         *  this class (excluding the transient <code>listener</code> field),
         *  and then explicitly marshals the listener to preserve the
         *  codebase upon writing to the file.
         *  This
         *  is why the <code>listener</code> field of this class 
         *  in the standard; that is, will not 
         */
        private void writeObject(ObjectOutputStream stream) throws IOException
        {
            stream.defaultWriteObject();
            stream.writeObject(new MarshalledObject(listener));
        }

        /** When this class is deserialized, this method is invoked. This
         *  method first deserializes the non-transient elements of this
         *  class, and then unmarshals the remote event listener. (See the
         *  description for <code>writeObject</code> above.)
         */
        private void readObject(ObjectInputStream stream)
                                    throws IOException, ClassNotFoundException
        {
            stream.defaultReadObject();
            MarshalledObject mo = (MarshalledObject)stream.readObject();
            try {
                listener = (RemoteEventListener)mo.get();
            } catch (Throwable e) {
                if (     (e instanceof Error)
                     && !(    (e instanceof LinkageError)
                           || (e instanceof OutOfMemoryError)
                           || (e instanceof StackOverflowError) )  )
                {
                    throw (Error)e;
                }
            }
        }
    }//end class RegistrationInfo

    /** This class represents a <code>Task</code> object that is placed
     *  in the <code>TaskManager</code> queue for processing in the thread
     *  pool. Instances of this class are placed on the task queue when
     *  registrations are granted. 
     *  <p>
     *  The <code>run</code> method of this class will determine if any of
     *  the new registration's desired lookup service(s) have already been
     *  discovered and will send the appropriate remote discovery event to
     *  the registration's listener.
     */
    private final class NewRegistrationTask implements TaskManager.Task {
        /** The data structure record corresponding to the new registration */
        public final RegistrationInfo regInfo;
        /** Constructs an instance of this class and stores the registration
         *  information.
         */
        public NewRegistrationTask(RegistrationInfo regInfo) {
            this.regInfo = regInfo;
        }//end constructor
        /** This method processes the information associated with the new
         *  registration and determines, based on the current state of the
         *  set of 'already-discovered' lookup service(s), whether to send
         *  a <code>RemoteDiscoveryEvent</code> to the new registration's
         *  listener.
         */
        public void run() {
            concurrentObj.writeLock();
            try {
                writeInfoTasks("NewRegistrationTask.run(): "
                               +"new Registration added ...");
                /* Get discovered registrars across all registrations */
                ServiceRegistrar[] curRegs = discoveryMgr.getRegistrars();
                /* Build and send the "discovered event" if appropriate */
                sendDiscoveredEvent(regInfo,curRegs);
            } finally {
                concurrentObj.writeUnlock();
            }
        }//end run

        /** This method returns true if the current instance of this class
         *  must be run after at least one task in the input task list with
         *  an index less than the <code>size</code> parameter (size may be
         *  less than tasks.size()).
         *  <p>
         *  Note that using List.get will be more efficient than List.iterator.
         *
         *  @param tasks the tasks to consider.  A read-only List, with all
         *         elements being an instanceof Task.
         *  @param size elements with index less than size should be considered
         */
        public boolean runAfter(List tasks, int size) {
            return false;
        }//end runAfter
    }//end class NewRegistrationTask

    /** This class represents a <code>Task</code> object that is placed
     *  in the <code>TaskManager</code> queue for processing in the thread
     *  pool.
     *  <p>
     *  The <code>run</code> method of this class will process discovery
     *  event information and determines to which active registrations the
     *  appropriate <code>RemoteDiscoveryEvent</code> should be sent; and
     *  then then will send that event.
     */
    private final class DiscoveredEventTask implements TaskManager.Task {
        /** The local event sent by the discovery manager. */
        public final DiscoveryEvent event;
        /** Constructs an instance of this class and stores the event*/
        public DiscoveredEventTask(DiscoveryEvent event) {
            this.event = event;
        }//end constructor
        /** This method processes the local discovery event information and
         *  determines, based on the current state of each active 
         *  registration, to which such registration the appropriate
         *  <code>RemoteDiscoveryEvent</code> should be sent. After making
         *  the determination, the remote event appropriate for each 
         *  registration is constructed and sent.
         */
        public void run() {
            writeInfoTasks("DiscoveredEventTask.run(): "
                           +"processing the discovery event ...");
            concurrentObj.writeLock();
            try {
                /* Get the just-discovered registrars */
                ServiceRegistrar[] discoveredRegs = event.getRegistrars();
                /* Loop thru regInfo's, adding only those not already known */
                for( Iterator itr=registrationByID.values().iterator();
                                                              itr.hasNext(); )
                {
                    RegistrationInfo regInfo = (RegistrationInfo)itr.next();
                    /* Build and send the "discovered event" if appropriate */
                    sendDiscoveredEvent(regInfo,discoveredRegs);
                }//end loop
            } finally {
                concurrentObj.writeUnlock();
            }
        }//end run

        /** This method returns true if the current instance of this class
         *  must be run after at least one task in the input task list with
         *  an index less than the <code>size</code> parameter (size may be
         *  less than tasks.size()).
         *  <p>
         *  Note that using List.get will be more efficient than List.iterator.
         *
         *  @param tasks the tasks to consider.  A read-only List, with all
         *         elements being an instanceof Task.
         *  @param size elements with index less than size should be considered
         */
        public boolean runAfter(List tasks, int size) {
            return false;
        }//end runAfter

    }//end class DiscoveredEventTask

    /** This class represents a <code>Task</code> object that is placed
     *  in the <code>TaskManager</code> queue for processing in the thread
     *  pool. Instances of this class are placed on the task queue when
     *  registrations are granted. 
     *  <p>
     *  The <code>run</code> method of this class will determine if any of
     *  the registrars currently identified as discovered for the referenced
     *  registration have been removed from the registration's set of desired
     *  registrars. (Typically, elements of the registration's set of desired
     *  registrars will be removed through the invocation of either
     *  <code>setLocators</code> or <code>removeLocators</code>).
     *  <p>
     *  If any registrars were removed from the registration's set of desired
     *  registrars, this task will send the appropriate remote discovery
     *  event to the registration's listener.
     */
    private final class RegistrarRemovedTask implements TaskManager.Task {
        /** Data structure record corresponding to the registration that had
         *  registrars removed from its set of desired registrars
         */
        public final RegistrationInfo regInfo;
        /** Constructs an instance of this class and stores the registration
         *  information.
         */
        public RegistrarRemovedTask(RegistrationInfo regInfo) {
            this.regInfo = regInfo;
        }//end constructor
        /** This method processes the information associated with the 
         *  registration and determines, based on the current state of that
         *  registration's set of 'desired' lookup service(s), whether to send
         *  a <code>RemoteDiscoveryEvent</code> to the registration's listener.
         *  After making the determination, the appropriate remote event is
         *  constructed and sent.
         */
        public void run() {
            concurrentObj.writeLock();
            try {
                writeInfoTasks("RegistrarRemovedTask.run(): registrar "
                               +"removed from the set of desired lookups ...");
                /* Get discovered registrars that are no longer desired */
                ServiceRegistrar[] regs
                                   = regInfo.complementDiscoveredWithDesired();
                /* Get discovered registrars across all registrations */
                ServiceRegistrar[] curRegs = discoveryMgr.getRegistrars();
                /* Get the intersection of the two sets from above */
                ServiceRegistrar[] commonRegs = intersectRegSets(regs,curRegs);
                /* The registrars from above should be "discarded" for the
                 * current registration. But because they are also in the
                 * discovery manager's set of discovered registrars, they
                 * are desired by other registrations. This means that the
                 * discovery manager will not send a local discard event 
                 * which would have resulted in a remote discard event. Thus,
                 * the common registrars from above must be removed from the
                 * registration's set of discovered registrars, and a discard
                 * event must be sent to the registration's listener.
                 */
                regInfo.removeFromDiscoveredRegs(commonRegs);
                RemoteDiscoveryEvent event = buildEvent(regInfo,
                                                        commonRegs,
                                                        true); //discarded
                if(event != null) {
                    sendEvent(regInfo,event);
                    writeInfoTasks("RegistrarRemovedTask.run(): "
                                   +"DISCARDED Event was SENT\n");
                }
            } finally {
                concurrentObj.writeUnlock();
            }
        }//end run

        /** This method returns true if the current instance of this class
         *  must be run after at least one task in the input task list with
         *  an index less than the <code>size</code> parameter (size may be
         *  less than tasks.size()).
         *  <p>
         *  Note that using List.get will be more efficient than List.iterator.
         *
         *  @param tasks the tasks to consider.  A read-only List, with all
         *         elements being an instanceof Task.
         *  @param size elements with index less than size should be considered
         */
        public boolean runAfter(List tasks, int size) {
            return false;
        }//end runAfter
    }//end class RegistrarRemovedTask

    /** This class represents a <code>Task</code> object that is placed
     *  in the <code>TaskManager</code> queue for processing in the thread
     *  pool. 
     *  <p>
     *  The <code>run</code> method of this class will process event
     *  information resulting from the "discarding" of one or more
     *  lookup services (registrars), and determines to which active 
     *  registrations the appropriate <code>RemoteDiscoveryEvent</code>
     *  should be sent; and then then will send that event.
     */
    private final class DiscardedEventTask implements TaskManager.Task {
        /** The local event sent by the discovery manager. */
        public final DiscoveryEvent event;
        /** Constructs an instance of this class and stores the event*/
        public DiscardedEventTask(DiscoveryEvent event) {
            this.event = event;
        }//end constructor
        /** This method processes the local discovery event information and
         *  determines, based on the current state of each active 
         *  registration, to which such registration the appropriate
         *  <code>RemoteDiscoveryEvent</code> should be sent. After making
         *  the determination, the remote event appropriate for each 
         *  registration is constructed and sent.
         */
        public void run() {
            writeInfoTasks("DiscardedEventTask.run(): "
                           +"processing the DISCARD event ...");
            concurrentObj.writeLock();
            try {
                /* Get the registrars that were just discarded */
                ServiceRegistrar[] discardedRegs = event.getRegistrars();

                RegistrationInfo regInfo = externalDiscardRequest();
                if(regInfo != null) {
                    /* Send discard event to only this one registration */
                    sendDiscardedEvent(regInfo,discardedRegs);
                } else {
                    /* Send discard event to each "eligible" registration */
                    for( Iterator itr=registrationByID.values().iterator();
                                                              itr.hasNext(); )
                    {
                        regInfo = (RegistrationInfo)itr.next();
                        sendDiscardedEvent(regInfo,discardedRegs);
                    }//end loop
                }//endif
            } finally {
                concurrentObj.writeUnlock();
            }
        }//end run

        /** This method returns true if the current instance of this class
         *  must be run after at least one task in the input task list with
         *  an index less than the <code>size</code> parameter (size may be
         *  less than tasks.size()).
         *  <p>
         *  Note that using List.get will be more efficient than List.iterator.
         *
         *  @param tasks the tasks to consider.  A read-only List, with all
         *         elements being an instanceof Task.
         *  @param size elements with index less than size should be considered
         */
        public boolean runAfter(List tasks, int size) {
            return false;
        }//end runAfter

        /** This method determines, based on the current state of the input
         *  registration, whether or not to send a remote discarded event
         *  to the registration's listener, and then builds and sends the
         *  event if appropriate.
         */
        private void sendDiscardedEvent(RegistrationInfo regInfo,
                                        ServiceRegistrar[] discardedRegs)
        {
            /* Determine which of the discarded registrars are also in the
             * registration's set of currently discovered registrars
             */
            ServiceRegistrar[] removeRegs
                              = regInfo.intersectDiscoveredRegs(discardedRegs);
            /* Remove the discarded registrars from the registration's map of
             * currently discovered registrars.
             */
            regInfo.removeFromDiscoveredRegs(removeRegs);
            /* Build and send the "discarded event" */
            RemoteDiscoveryEvent event = buildEvent(regInfo,removeRegs,true);
            if(event != null) {
                sendEvent(regInfo,event);
                writeInfoEvents("DiscardedEventTask.run(): "
                                +"DISCARDED Event was SENT\n");
            }
        }//end sendDiscardedEvent
    }//end class DiscardedEventTask

    /** This class represents a <code>Task</code> object that is placed
     *  in the <code>TaskManager</code> queue for processing in the thread
     *  pool. Instances of this class are placed on the task queue when
     *  registrations are granted. 
     *  <p>
     *  The <code>run</code> method of this class will remove the indicated
     *  registrar from the registration's set of discovered registrars and
     *  if successfully removed, will build and send a remote discarded event
     *  to the registration's listener..
     */
    private final class DiscardRegistrarTask implements TaskManager.Task {
        /** Data structure record corresponding to the registration that has
         *  requested to have one of its discovered registrars discarded
         */
        public final RegistrationInfo regInfo;
        /** The registrar to discard */
        public final ServiceRegistrar registrar;
        /** Constructs an instance of this class and stores the registration
         *  information.
         */
        public DiscardRegistrarTask(RegistrationInfo regInfo,
                                    ServiceRegistrar registrar)
        {
            this.regInfo   = regInfo;
            this.registrar = registrar;
        }//end constructor
        /** This method attempts to remove the indicated registrar from
         *  the registration's set of discovered registrars. If successful,
         *  this method builds and sends a remote discarded event to the
         *  registration's listener.
         */
        public void run() {
            concurrentObj.writeLock();
            try {
                writeInfoTasks("DiscardRegistrarTask.run(): "
                               +"registrar requested to be discarded ...");
                /* Remove registrar from the set of discovered registrars */
                Object previousValue = 
                                (regInfo.discoveredRegsMap).remove(registrar);
                if(previousValue != null) { // successfully removed, send event
                    ServiceRegistrar[] discardedReg = new ServiceRegistrar[1];
                    discardedReg[0] = registrar;
                    RemoteDiscoveryEvent event = buildEvent(regInfo,
                                                            discardedReg,
                                                            true); //discarded
                    if(event != null) {
                        sendEvent(regInfo,event);
                        writeInfoEvents("DiscardRegistrarTask.run(): "
                                        +"DISCARDED Event was SENT\n");
                    }
                }
            } finally {
                concurrentObj.writeUnlock();
            }
        }//end run

        /** This method returns true if the current instance of this class
         *  must be run after at least one task in the input task list with
         *  an index less than the <code>size</code> parameter (size may be
         *  less than tasks.size()).
         *  <p>
         *  Note that using List.get will be more efficient than List.iterator.
         *
         *  @param tasks the tasks to consider.  A read-only List, with all
         *         elements being an instanceof Task.
         *  @param size elements with index less than size should be considered
         */
        public boolean runAfter(List tasks, int size) {
            return false;
        }//end runAfter
    }//end class DiscardRegistrarTask

    /**
     * Handler class for the persistent storage facility.
     * <p>
     * At any point during processing in this service, there will exist
     * both a 'snapshot' of the service's state and a set of records
     * detailing each significant change that has occurred to the state
     * since the snapshot was taken. The snapshot information and the
     * incremental change information will be stored in separate files
     * called, respectively, the snapshot file and the log file. Together,
     * these files are used to recover the state of the service after a
     * crash or a network outage (or if the service or its ActivationGroup
     * is un-registered and then re-registered through the Activation Daemon).
     * <p>
     * This class contains the methods that are used to record and recover
     * the snapshot of the service's state; as well as the method used to
     * apply the state changes that were recorded in the log file.
     * <p>
     * When the ReliableLog class is instantiated, a new instance of this
     * class is passed to its constructor so that the methods of this
     * class may be invoked by the methods defined in the ReliableLog.
     * Because this class extends the LogHandler class associated with
     * the ReliableLog class, this class must provide implementations of
     * the abstract methods declared in the LogHandler. Also, some of the
     * methods defined in this class override the methods of the LogHandler
     * in order to customize the handling of snapshot creation and
     * retrieval.
     * <p>
     * Each significant change to the service's state is written to the
     * log file as an individual record (when addLogRecord() is invoked).
     * After the number of records logged exceeds a pre-defined threshold,
     * a snapshot of the state is recorded by invoking -- through the
     * ReliableLog and its LogHandler -- the snapshot() method defined in
     * this class. After the snapshot is taken, the log file is cleared
     * and the incremental log process starts over.
     * <p>
     * The contents of the snapshot file reflect the DATA contained in
     * the fields making up the current state of the service. That data
     * represents many changes -- over time -- to the service's state.
     * On the other hand, each record written to the log file is an object
     * that reflects both the data used and the ACTIONS taken to make one
     * change to the service's state at a particular point in time.
     * <p>
     * During recovery, the state of the service at the time of a crash
     * or outage is re-constructed by first retrieving the 'base' state from
     * the snapshot file; and then modifying that base state according to
     * the records retrieved from the log file. The reconstruction of the
     * base state is achieved by invoking the recover() method defined in
     * this class. The modifications recorded in the log file are then
     * applied to the base state by invoking the applyUpdate() method
     * defined in this class. Both recover() and applyUpdate() are invoked
     * through the ReliableLog and its associated Loghandler.
     * <p>
     * NOTE: The following lines must be added to the service's policy file
     * <pre>
     *     permission java.io.FilePermission "dirname",   "read,write,delete";
     *     permission java.io.FilePermission "dirname/-", "read,write,delete";
     * </pre>
     *     where 'dirname' is the name of the directory path (relative or
     *     absolute) where the snapshot and log file will be maintained.
     */
    private class LocalLogHandler extends LogHandler {

        /** No-arg public constructor */
        public LocalLogHandler() { }

        /* Overrides snapshot() defined in ReliableLog's LogHandler class. */
        public void snapshot(OutputStream out) throws IOException {
            takeSnapshot(out);
        }

        /* Overrides recover() defined in ReliableLog's LogHandler class. */
        public void recover(InputStream in)
                            throws IOException, ClassNotFoundException
        {
            recoverSnapshot(in);
        }

        /**
         * Required method that implements the abstract applyUpdate()
         * defined in ReliableLog's associated LogHandler class.
         * <p>
         * During state recovery, the recover() method defined in the
         * ReliableLog class is invoked. That method invokes the method
         * recoverUpdates() which invokes the method readUpdates(). Both
         * of those methods are defined in ReliableLog. The method
         * readUpdates() retrieves a record from the log file and then
         * invokes this method.
         * <p>
         * This method invokes the version of the method apply() that
         * corresponds to the particular type of 'log record' object
         * that is input as the first argument. The log record object and its
         * corresponding apply() method are defined in one of the so-called
         * LogObj classes. Any instance of one the LogObj classes is an
         * implementation of the LogRecord interface. The particular
         * implementation that is input to this method is dependent on the
         * type of record that was originally logged. The apply() method
         * will then modify the state of the service in a way dictated
         * by the type of record that was retrieved.
         */
        public void applyUpdate(Object logRecObj) {
            ((LogRecord)logRecObj).apply(FiddlerImpl.this);
        }
    }//end class LocalLogHandler
    /* ******************* END Inner Class Definitions ********************* */

    /* ******************* BEGIN Thread Class Definitions ****************** */
    /** Thread which is used to monitor the current leases in effect and
     *  cancel (expire) those leases with expiration times that have exceeded
     *  the current time.
     */
    private class LeaseExpireThread extends Thread {
        /** Create a daemon thread */
        public LeaseExpireThread() {
            super("lease expire");
            setDaemon(true);
        }

        public void run() {
            try {
                concurrentObj.writeLock();
            } catch (ConcurrentLockException e) {
                return;
            }
            try {
                while (!isInterrupted()) {
                    long curTime  = System.currentTimeMillis();
                    minExpiration = Long.MAX_VALUE;
                    /* Loop through registrationByTime removing registrations
                     * with expiration times that are earlier than the current
                     * time. The logic of this loop relies on the fact that 
                     * registrationByTime is a TreeMap in which the elments
                     * are ordered (in ascending order) by the lease expiration
                     * times. Thus, when one registration is encountered with
                     * an expiration time that is later than the current time,
                     * it can be assumed that all remaining registrations have
                     * expiration times that are also later than the current
                     * time; and the loop can be exited. Until such a
                     * registration is encountered, each registration is
                     * removed from its various storage locations.
                     */
                    while (!registrationByTime.isEmpty()) {
                        RegistrationInfo regInfo
                             = (RegistrationInfo)registrationByTime.firstKey();
                        if (regInfo.leaseExpiration > curTime) {
                            minExpiration = regInfo.leaseExpiration;
                            break;
                        }
	                /* The removal of a registration typically involves the
                         * the modification of the managed sets in the
                         * discovery manager, which usually involves starting
                         * the discovery protocol. An IOException can occur
                         * when the discovery protocol fails to start. When
                         * such an exception does occur, register an ERROR
                         * status attribute (along with a Comment attribute
                         * describing the nature of the problem) to all lookup
                         * services with which this service is registered. 
                         *
                         * Administrative clients, as well as clients that use
                         * this service should have registered for notification
                         * of the existence of this attribute.
	                 */
                        try {
                            removeRegistration(regInfo);
                        } catch(IOException e) {
                            Entry[] errorAttrs
                                    = new Entry[]
                                        { new FiddlerStatus(StatusType.ERROR),
                                          new Comment("Failure while removing "
                                                      +"registration (ID = "
                                                      +regInfo.registrationID
                                                      +") from service state")
                                        };
                            joinMgr.addAttributes(errorAttrs,true);
	                    e.printStackTrace();
                        }
                    }//end while
                    try {
                        concurrentObj.writerWait(leaseExpireThreadSyncObj,
                                                 (minExpiration - curTime));
                    } catch (ConcurrentLockException e) {
                        return;
                    }
                }//end while
            } finally {
                concurrentObj.writeUnlock();
            }
        }
    }//end class LeaseExpireThread

    /**
     * Snapshot-taking thread. 
     * <p>
     * A snapshot is taken when -- after writing a new record to the 
     * log file -- it is determined that the size of the log file has 
     * exceeded a certain threshold. The code which adds the new record 
     * to the log file and which, in turn, decides that a snapshot
     * must be taken is "wrapped" in a writer mutex. That is, synchronization
     * of processing is achieved in this service through a "reader/writer"
     * mutex construct. This construct allows only one writer at any one
     * time; but allows an unlimited number of simultaneous readers as
     * long as no writer has locked the mutex. During steady-state, it is
     * anticipated that far more "read actions" will occur (e.g. discovery
     * events being sent) than "write actions" (e.g. modifying the managed
     * sets). Since the process of taking a snapshot can be time-consuming,
     * if the whole snapshot-taking process occupies that single writer
     * mutex, then a significant number of read actions will be un-necessarily
     * blocked; possibly resulting in an unacceptable degradation in
     * response time. 
     * <p>
     * It is for the above reason that the process of taking a snapshot is
     * performed in a separate thread. The thread waits on the monitor
     * belonging to the snapshotThreadSyncObj instance until it is notified
     * (or "signalled") that a snapshot must be taken. The notification
     * is sent by another thread, created by this service, which determines
     * when the conditions are right for a snapshot. The notification takes
     * the form of an interrupt indicating that the snapshot monitor is
     * available. Although the interrupt is sent while the writer mutex is
     * locked, the act of sending the notification is less time-consuming
     * than the act of taking the snapshot itself. When the thread receives
     * a notification, it awakens and requests a lock on the reader mutex
     * (this is all done in the readerWait() method). Because a reader -- not
     * a writer -- mutex is locked, read-only processes still have access
     * to the system state, so discovery events can be sent and the service's
     * state can be queried; but the reader mutex prevents changes to the
     * state while the snapshot is in progress.  
     * <p>
     * Note that the current snapshot is guaranteed to complete before the
     * next snapshot request is received. This is because even though
     * the act of taking a snapshot can be viewed as a writer process, 
     * the fact that the next snapshot notification will be wrapped in a
     * writer mutex, combined with the fact that a writer mutex can not
     * be locked while a reader mutex is locked, allows the snapshot to
     * be treated as a reader process.
     */
    private class SnapshotThread extends Thread {
        /** Create a daemon thread */
        public SnapshotThread() {
            super("snapshot thread");
            setDaemon(true);
        }

        public void run() {
            try {
                concurrentObj.readLock();
            } catch (ConcurrentLockException e) {
                return;
            }
            try {
                while (!isInterrupted()) {
                    try {
                        concurrentObj.readerWait(snapshotThreadSyncObj,
                                                 Long.MAX_VALUE);
                        try {
                            log.snapshot();
                            logFileSize = 0;
                        } catch (InterruptedIOException e) {
                            return;
                        } catch (Exception e) {
                            if (    (e instanceof LogException)
                                 && (((LogException)e).detail instanceof
				                 InterruptedIOException)  )
                            {
                                return;
                            }
	                    /* If taking the snapshot fails for any reason,
                             * then register an ERROR status attribute (along
                             * with a Comment attribute describing the nature
                             * of the problem) to all lookup services with
                             * which this service is registered. 
                             *
                             * Administrative clients, as well as clients that
                             * use this service should have registered for
                             * notification of the existence of this attribute.
	                     */
                            Entry[] errorAttrs
                                    = new Entry[]
                                        { new FiddlerStatus(StatusType.ERROR),
                                          new Comment("Failure while taking a "
                                                      +"snapshot of the "
                                                      +"service state")
                                        };
                            joinMgr.addAttributes(errorAttrs,true);
	                    e.printStackTrace();
                        }
                    } catch (ConcurrentLockException e) {
                        return;
                    }
                }//end while
            } finally {
                concurrentObj.readUnlock();
            }
        }//end run
    }//end class SnapshotThread

    /** Thread which is used to terminate the current executing instance
     *  of the Fiddler implementation of the lookup discovery service. 
     *  Termination processing is performed in a separate thread (that is,
     *  in an instance of this class) in order to avoid deadlock that 
     *  can occur because Activatable.inactive will block until all 
     *  in-progress RMI calls have completed.
     */
    private class DestroyThread extends Thread {
	/** Constructor that creates a non-daemon thread */
	public DestroyThread() {
	    super("destroy");
	    /* override inheritance from RMI daemon thread */
	    setDaemon(false);
	}

	public void run() {
	    /* must unregister before unexport */
	    if (activationID != null) {
		try {
		    Activatable.unregister(activationID);
		} catch (RemoteException e) {
		    /* give up until we can at least unregister */
		    return;
		} catch (ActivationException e) {
		}
	    }
	    try {
		while (!UnicastRemoteObject.unexportObject(FiddlerImpl.this,
							   false))
		{
		    Thread.yield();
		}
	    } catch (NoSuchObjectException e) {
	    }
	    /* all daemons must terminate before deleting persistent store */
	    leaseExpireThread.interrupt();
	    snapshotThread.interrupt();
	    taskMgr.terminate();
	    joinMgr.terminate();
	    try {
		leaseExpireThread.join();
		snapshotThread.join();
	    } catch (InterruptedException e) {
	    }
	    log.deletePersistentStore();
	    if (activationID != null) {
		/* inactive will set current group ID to null */
		ActivationGroupID gid = ActivationGroup.currentGroupID();
		try {
		    while (!Activatable.inactive(activationID)) {
			Thread.yield();
		    }
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
		try {
		    ActivationGroup.getSystem().unregisterGroup(gid);
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
	    }
	}
    }//end class DestroyThread

    /** Thread which is used to exit the virtual machine (VM) in which 
     *  the terminate the current instance of this service is running.
     *  No clean up is performed. That is, unlike the thread class
     *  <code>DestroyThread</code>, no threads are terminated and
     *  the all storage in which this service's persistence state is
     *  logged (the "log files") are not removed.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#killVM
     */
    private class KillVMThread extends Thread {
        long delay;
	/** Creates thread that inherits its daemon status from main thread */
	public KillVMThread(long delay) {
	    super("killVM");
            this.delay = delay;
	}
	public void run() {
            try {
                Thread.sleep(delay);
            } catch (InterruptedException e) { }
            System.exit(0);
	}
    }//end class KillVMThread
    /* ******************* END Thread Class Definitions ******************** */

    /* ************************ BEGIN Public Methods *********************** */
    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler --> net.jini.admin.Administrable
     */
    /** 
     * Returns a proxy to the current instance of this class through which
     * a client may administer the lookup discovery service
     *
     * @return a proxy object through which the lookup discovery service
     *         may be administered.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getAdmin
     * @see net.jini.Administrable#getAdmin
     */
    public Object getAdmin() throws RemoteException {
        return (new FiddlerAdminProxy(this, proxyID));
    }
    /*  END com.sun.jini.fiddler.Fiddler --> net.jini.admin.Administrable   */
    /* -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     *                                --> com.sun.jini.fiddler.FiddlerAdmin
     *                                         --> net.jini.admin.JoinAdmin
     */
    /** 
     * Returns the current attribute sets for the lookup discoverey service. 
     * 
     * @return array of net.jini.core.entry.Entry containing the current
     *         attribute sets for the lookup discoverey service
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getLookupAttributes
     * @see net.jini.admin.JoinAdmin#getLookupAttributes
     */
    public Entry[] getLookupAttributes() throws RemoteException {
        concurrentObj.readLock();
        try {
            return thisServicesAttrs;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /** 
     * Adds attribute sets to the current set of attributes associated
     * with the lookup discovery service. The resulting set will be used
     * for all future registrations with lookup services. The new attribute
     * sets are also added to the lookup discovery service's attributes
     * on each lookup service with which the lookup discovery service
     * is currently registered.
     *
     * @param  attrSets array of net.jini.core.entry.Entry containing the
     *         attribute sets to add
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         attributes may or may not have been added successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#addLookupAttributes
     * @see net.jini.admin.JoinAdmin#addLookupAttributes
     */
    public void addLookupAttributes(Entry[] attrSets) throws RemoteException {
        concurrentObj.writeLock();
        try {
            joinMgr.addAttributes(attrSets, true);
            thisServicesAttrs = joinMgr.getAttributes();
	    addLogRecord(new LookupAttrsAddedLogObj(this,attrSets));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Modifies the current set of attributes associated with the lookup
     * discovery service. The resulting set will be used for all future
     * registrations with lookup services. The same modifications are 
     * also made to the lookup discovery service's attributes on each
     * lookup service with which the lookup discovery service is currently
     * registered.
     *
     * @param  attrSetTemplates  array of net.jini.core.entry.Entry containing
     *         the templates to use for selecting the attributes (contained
     *         within the set of existing attributes) that are to be
     *         modified
     * @param  attrSets array of net.jini.core.entry.Entry containing the
     *         modifications to make to matching sets
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         attributes may or may not have been modified successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#modifyLookupAttributes
     * @see net.jini.admin.JoinAdmin#modifyLookupAttributes
     */
    public void modifyLookupAttributes(Entry[] attrSetTemplates,
				       Entry[] attrSets) throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            joinMgr.modifyAttributes(attrSetTemplates, attrSets, true);
            thisServicesAttrs = joinMgr.getAttributes();
            addLogRecord
               (new LookupAttrsModifiedLogObj(this,attrSetTemplates,attrSets));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Get the names of the groups whose members are lookup services the
     * lookup discovery services wishes to register with (join).
     * 
     * @return String array containing the names of the groups whose members
     *         are lookup services the lookup discovery service wishes to
     *         join.
     * <p>
     *         If the array returned is empty, the lookup discovery service
     *         is configured to join no groups. If null is returned, the
     *         lookup discovery service is configured to join all groups.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getLookupGroups
     * @see net.jini.admin.JoinAdmin#getLookupGroups
     */
    public String[] getLookupGroups() throws RemoteException {
        concurrentObj.readLock();
        try {
            return thisServicesGroups;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /** 
     * Add new names to the set consisting of the names of groups whose
     * members are lookup services the lookup discovery service wishes
     * to register with (join). Any lookup services belonging to the
     * new groups that the lookup discovery service has not yet registered
     * with, will be discovered and joined.
     *
     * @param  String array containing the names of the groups to add
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been added successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#addLookupGroups
     * @see net.jini.admin.JoinAdmin#addLookupGroups
     */
    public void addLookupGroups(String[] groups) throws RemoteException {
        concurrentObj.writeLock();
        try {
            try {
                joinMgrLookupDiscoveryMgr.addGroups(groups);
            } catch (IOException e) {
                throw new RuntimeException(e.toString());
            }
            thisServicesGroups = joinMgrLookupDiscoveryMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(thisServicesGroups));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Remove a set of group names from lookup discovery service's managed
     * set of groups (the set consisting of the names of groups whose
     * members are lookup services the lookup discovery service wishes
     * to join). Any leases granted to the lookup discovery service by
     * lookup services that are not members of the groups whose names 
     * remain in the managed set will be cancelled at those lookup services.
     *
     * @param  String array containing the names of the groups to remove
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been removed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#removeLookupGroups
     * @see net.jini.admin.JoinAdmin#removeLookupGroups
     */
    public void removeLookupGroups(String[] groups) throws RemoteException {
        concurrentObj.writeLock();
        try {
            joinMgrLookupDiscoveryMgr.removeGroups(groups);
            thisServicesGroups = joinMgrLookupDiscoveryMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(thisServicesGroups));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Replace the lookup discovery service's managed set of groups with a
     * new set of group names. Any leases granted to the lookup discovery
     * service by lookup services that are not members of the groups whose
     * names are in the new managed set will be cancelled at those lookup
     * services. Lookup services that are members of groups reflected in
     * the new managed set will be discovered and joined.
     *
     * @param  String array containing the names of the new groups
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been replaced successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setLookupGroups
     * @see net.jini.admin.JoinAdmin#setLookupGroups
     */
    public void setLookupGroups(String[] groups) throws RemoteException {
        concurrentObj.writeLock();
        try {
            try {
                joinMgrLookupDiscoveryMgr.setGroups(groups);
            } catch (IOException e) {
                throw new RuntimeException(e.toString());
            }
            thisServicesGroups = joinMgrLookupDiscoveryMgr.getGroups();
            addLogRecord(new LookupGroupsChangedLogObj(thisServicesGroups));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Get the lookup discovery service's managed set of locators. The
     * managed set of locators is the set of LookupLocator objects
     * corresponding to the specific lookup services with which the lookup
     * discovery service wishes to register (join).
     * 
     * @return array of objects of type net.jini.core.discovery.LookupLocator,
     *         each of which corresponds to a specific lookup service the
     *         lookup discovery service wishes to join.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getLookupLocators
     * @see net.jini.admin.JoinAdmin#getLookupLocators
     */
    public LookupLocator[] getLookupLocators() throws RemoteException {
        concurrentObj.readLock();
        try {
            return thisServicesLocators;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /** 
     * Add a set of LookupLocator objects to the lookup discovery service's
     * managed set of locators. The managed set of locators is the set of
     * LookupLocator objects corresponding to the specific lookup services
     * with which the lookup discovery service wishes to register (join).
     * <p>
     * Any lookup services corresponding to the new locators that the lookup
     * discovery service has not yet joined, will be discovered and joined.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects to add
     *         to the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new locators may or may not have been added successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#addLookupLocators
     * @see net.jini.admin.JoinAdmin#addLookupLocators
     */
    public void addLookupLocators(LookupLocator[] locators)
                                                      throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            joinMgrLookupDiscoveryMgr.addLocators(locators);
            thisServicesLocators = joinMgrLookupDiscoveryMgr.getLocators();
            addLogRecord
                     (new LookupLocatorsChangedLogObj(thisServicesLocators));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Remove a set of LookupLocator objects from the lookup discovery
     * service's managed set of locators. The managed set of locators is the
     * set of LookupLocator objects corresponding to the specific lookup
     * services with which the lookup discovery service wishes to register
     * (join).
     * <p>
     * Note that any leases granted to the lookup discovery service by
     * lookup services that do not correspond to any of the locators
     * remaining in the managed set will be cancelled at those lookup
     * services.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects to
     *         remove from the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new locators may or may not have been removed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#removeLookupLocators
     * @see net.jini.admin.JoinAdmin#removeLookupLocators
     */
    public void removeLookupLocators(LookupLocator[] locators)
                                                       throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            joinMgrLookupDiscoveryMgr.removeLocators(locators);
            thisServicesLocators = joinMgrLookupDiscoveryMgr.getLocators();
            addLogRecord
                     (new LookupLocatorsChangedLogObj(thisServicesLocators));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /** 
     * Replace the lookup discovery service's managed set of locators with
     * a new set of locators. The managed set of locators is the set of
     * LookupLocator objects corresponding to the specific lookup services
     * with which the lookup discovery service wishes to register (join).
     * <p>
     * Note that any leases granted to the lookup discovery service by
     * lookup services whose corresponding locator is removed from the
     * managed setwill be cancelled at those lookup services. The lookup
     * services corresponding to the new locators in the managed set
     * will be discovered and joined.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects with
     *         which to replace the current managed set of locators
     *         remove from the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         locators in the managed set may or may not have been replaced
     *         successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setLookupLocators
     * @see net.jini.admin.JoinAdmin#setLookupLocators
     */
    public void setLookupLocators(LookupLocator[] locators)
                                                       throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            joinMgrLookupDiscoveryMgr.setLocators(locators);
            thisServicesLocators = joinMgrLookupDiscoveryMgr.getLocators();
            addLogRecord
                     (new LookupLocatorsChangedLogObj(thisServicesLocators));
        } finally {
            concurrentObj.writeUnlock();
        }
    }
    /* END com.sun.jini.fiddler.Fiddler --> com.sun.jini.fiddler.FiddlerAdmin
     *                                           --> net.jini.admin.JoinAdmin
     * -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     *                    --> com.sun.jini.fiddler.FiddlerAdmin
     *                          --> com.sun.jini.admin.StorageLocationAdmin
     */
    /**
     * Returns the location of the lookup discovery service's persistent
     * storage. Typically returns a directory pathname unless the lookup
     * discovery service specifies otherwise.
     * 
     * @return String that contains the current location of the lookup
     *         discovery service's persistent storage.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getStorageLocation
     * @see com.sun.jini.admin.StorageLocationAdmin#getStorageLocation
     */
    public String getStorageLocation() throws RemoteException {
        concurrentObj.readLock();
        try {
            return logDirname;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /**
     * Sets the location of the lookup discovery service's persistent
     * storage, moving all current persistent storage from the current
     * location to the specified new location.
     * 
     * @param String that contains the new location of the lookup discovery
     *         service's persistent storage.
     *
     * @throws java.io.IOException typically, this exception occurs when
     *         there is a failure while moving the persistent storage
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new location for the lookup discovery service's persistent
     *         storage may or may not have been replaced successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getStorageLocation
     * @see com.sun.jini.admin.StorageLocationAdmin#getStorageLocation
     */
    public void setStorageLocation(String location)
                                         throws IOException, RemoteException
    {
        if (location.equals(logDirname)) {
            return;
        }
        concurrentObj.writeLock();
        try {
            ReliableLog newLog = new ReliableLog(location,
                                                 new LocalLogHandler());
            newLog.snapshot();
            ReliableLog oldLog = this.log;
            this.log = newLog;
            logDirname = location;

            if (activationID != null) {
                try {
                    ActivationSystem sys = ActivationGroup.getSystem();
                    ActivationDesc desc = sys.getActivationDesc(activationID);
                    desc = new ActivationDesc(desc.getGroupID(),
                                              desc.getClassName(),
                                              desc.getLocation(),
                                              new MarshalledObject(location),
                                              desc.getRestartMode());
                    sys.setActivationDesc(activationID, desc);
                } catch (ActivationException e) {
                    throw new LogException("activation system problem", e);
                }
            }
            oldLog.deletePersistentStore();
        } finally {
            concurrentObj.writeUnlock();
        }
    }
    /* END com.sun.jini.fiddler.Fiddler
                              --> com.sun.jini.fiddler.FiddlerAdmin
     *                                  --> com.sun.jini.admin.DestroyAdmin
     * -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     *                            --> com.sun.jini.fiddler.FiddlerAdmin
     *                                  --> com.sun.jini.admin.DestroyAdmin
     */
    /**
     * Destroy the lookup discovery service, if possible, including its
     * persistent storage. This method will typically spawn a separate
     * thread to do the actual work asynchronously, so a successful
     * return from this method usually does not mean that the service
     * has been destroyed.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         lookup discovery service may or may not have been successfully
     *         destroyed.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#destroy
     * @see com.sun.jini.admin.DestroyAdmin#destroy
     */
    public void destroy() throws RemoteException {
        destroyDo();
    }
    /* END com.sun.jini.fiddler.Fiddler
                                --> com.sun.jini.fiddler.FiddlerAdmin
     *                                  --> com.sun.jini.admin.DestroyAdmin
     * -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     *                                --> com.sun.jini.fiddler.FiddlerAdmin
     */
    /**
     * Changes the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     * <p>
     * This method is a mechanism for an entity with the appropriate
     * privileges to administratively change the value of the least upper
     * bound that will be applied by the Fiddler implementation of the lookup
     * discovery service when determining the duration to assign to the lease
     * on a requested registration.
     *
     * @param newBound <code>long</code> value representing the new least
     *        upper bound (in milliseconds) on the set of all possible
     *        lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         bound value may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setLeaseDurationBound
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLeaseDurationBound
     */
    public void setLeaseDurationBound(long newBound) throws RemoteException {
        concurrentObj.writeLock();
        try {
	    if (newBound > MAX_LEASE) {
                throw new IllegalArgumentException("max duration exceeded");
            }
            leaseDurationBound = newBound;
            addLogRecord(new LeaseDurationBoundSetLogObj(newBound));
        } finally {
            concurrentObj.writeUnlock();
        }
    }
    /**
     * Retrieves the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     *
     * @return <code>long</code> value representing the current least
     *         upper bound (in milliseconds) on the set of all possible
     *         lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setLeaseDurationBound
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLeaseDurationBound
     */
    public long getLeaseDurationBound() throws RemoteException {
        concurrentObj.readLock();
        try {
            return leaseDurationBound;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /**
     * Change the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not
     * to take a "snapshot" of the system state.
     *
     * @param weight weight factor for snapshot size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         weight factor may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setSnapshotWeight
     * @see com.sun.jini.fiddler.FiddlerAdmin#setSnapshotWeight
     */
    public void setSnapshotWeight(float weight) throws RemoteException {
        concurrentObj.writeLock();
        try {
            snapshotWt = weight;
            addLogRecord(new SnapshotWeightSetLogObj(weight));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /**
     * Retrieve the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not to
     * take a "snapshot" of the system state.
     * 
     * @return float value corresponding to the weight factor for snapshot
     *         size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getSnapshotWeight
     * @see com.sun.jini.fiddler.FiddlerAdmin#getSnapshotWeight
     */
    public float getSnapshotWeight() throws RemoteException {
        concurrentObj.readLock();
        try {
            return snapshotWt;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /**
     * Change the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     *
     * @param threshold size threshold for taking a snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         threshold may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setLogToSnapshotThreshold
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLogToSnapshotThreshold
     */
    public void setLogToSnapshotThreshold(int threshold) throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            logToSnapshotThresh = threshold;
            addLogRecord(new LogToSnapshotThresholdSetLogObj(threshold));
        } finally {
            concurrentObj.writeUnlock();
        }
    }

    /**
     * Retrieve the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     * 
     * @return int value corresponding to the size threshold of the snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getLogToSnapshotThreshold
     * @see com.sun.jini.fiddler.FiddlerAdmin#getLogToSnapshotThreshold
     */
    public int getLogToSnapshotThreshold() throws RemoteException {
        concurrentObj.readLock();
        try {
	    return logToSnapshotThresh;
        } finally {
            concurrentObj.readUnlock();
        }
    }

    /**
     * Changes the value of the <code>com.sun.jini.fiddler.server.debug</code>
     * system property. The value of this property configures the debug
     * facility of the fiddler implementation of the lookup discovery service
     * to write debug information (to a file or standard output) selectively;
     * depending on the values the debug property is given through this
     * method.
     *
     * @param newValues <code>String</code> consisting of a comma-separated set
     *                  of  valid debug property values
     *
     * @return <code>boolean</code> value that indicates whether or not the
     *         debug property was successfully set
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         debug property may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setDebugProperty
     */
    public boolean setDebugProperty(String newValues) throws RemoteException {
        boolean propSet = false;
        concurrentObj.writeLock();
        try {
            propSet = setDebugPropertyDo(newValues);
            if(propSet == true) { // only log if property successfully changed
                addLogRecord(new DebugPropertySetLogObj(newValues));
            }
        } finally {
            concurrentObj.writeUnlock();
        }
	return propSet;
    }

    /**
     * Returns the current value of the system property used to configure the
     * debug facility of the fiddler implementation of the lookup discovery
     * service (<code>com.sun.jini.fiddler.server.debug</code>).
     *
     * @return <code>String</code> consisting of a comma-separated set
     *         of valid debug property values
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     * @see com.sun.jini.fiddler.FiddlerAdmin#getDebugProperty
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#getDebugProperty
     */
    public String getDebugProperty() throws RemoteException {
        concurrentObj.readLock();
        try {
            String propVals = System.getProperty
                                        ("com.sun.jini.fiddler.server.debug");
	    return propVals;
        } catch(Throwable e) { 
            //problems here should not prevent it from continuing to operate
            return new String("Failure retrieving debug property: "
                              +"check security policy file -- "+e.toString());
        } finally {
            concurrentObj.readUnlock();
        }
    }
    /* END com.sun.jini.fiddler.Fiddler --> com.sun.jini.fiddler.FiddlerAdmin
     * -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     *                                  --> com.sun.jini.mahout.KillVMAdmin
     */
    /**
     * This method is intended to allow for the simulation of a system
     * crash by exiting the virtual machine (VM) in which the current 
     * instance of this service is running. No clean up is performed
     * (just as no clean up would be performed in a real system crash).
     * That is, unlike the <code>destroy</code> method of the
     * <code>DestroyAdmin</code> interface, no threads are terminated and
     * the storage locations in which this service's persistence state is
     * logged (the "log files") are not removed.
     * <p>
     * This method should be used to test the persistent state recovery 
     * mechanism of the service. As such, it should only be invoked on
     * versions of the service which actually persist their state; that is,
     * it should not be invoked on versions of the service which are 
     * typically referred to as "transient". If the service is activatable
     * and "restartable", the activation daemon (RMID) will automatically
     * restart the service after the "crash". If the service is activatable
     * but not "restartable", then RMID will restart the service upon the
     * first invocation of any remote method of the service. When such a
     * service is restarted, it typically attempts to recover any state
     * it may have persisted before the "crash".
     * <p>
     * Note that because calls to this method are asynchronous, certain
     * requirements are imposed on the implementation. If the VM is exited
     * from the current thread, rather than from a separate thread, 
     * entities which invoke this method will always receive a 
     * <code>RemoteException</code> because exiting the VM in the current
     * thread will break the communication link between the frontend 
     * and the backend of the service. Thus, the VM must be exited from
     * a separate thread.
     * <p>
     * Another implementation requirement, related to the thread requirement,
     * involves preventing <code>RemoteException</code>s through the
     * introduction of timing delays. If the VM is exited before the remote
     * call has had a chance to complete, just as with the case of exiting
     * the VM from the current thread, the communication link between the
     * service's frontend and backend will be broken, resulting in a
     * <code>RemoteException</code>. Thus, not only must the VM be exited
     * from a separate thread, but it also must be exited only after the
     * remote call to this method has completed. By instructing the thread
     * that exits the VM to delay a certain amount of time before actually
     * exiting the VM, this method attempts to "guarantee" that the VM will
     * not be exited until the remote call has completed. Note that the
     * amount of time to delay is simply a rough estimate of the amount
     * of time a typical remote call should take to complete. Thus, a
     * <code>RemoteException</code> may occur because the time delay is
     * too small.
     * <p>
     * Finally, note that the use of this method also imposes a related
     * timing requirement on the client-side entity. That is, when an
     * entity invokes this method, it must also delay a certain amount
     * time before attempting to again interact with the service. This
     * delay is necessary to allow the service to be re-activated and
     * to recover its persisted state. This method returns a <code>long</code>
     * value that represents the amount of time the client-side entity
     * should delay before attempting to interact with the service again.
     * Once again, this time value is a rough estimate, and may need to
     * be adjusted if the entity experiences <code>RemoteException</code>s
     * on a regular basis when the entity attempts to interact with the
     * service after invoking this method.
     *
     * @return <code>long</code> value that represents an estimate of the
     *         amount of time (in milliseconds) the entity that invokes this
     *         method should delay before attempting to interact with the
     *         service again.
     *                      
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the VM of the service
     *         may or may not have been exited successfully.
     *
     * @see com.sun.jini.mahout.KillVMAdmin#killVM
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#killVM
     */
    public long killVM() throws RemoteException {
        long delay = 20; //thread should wait N milliseconds before exiting VM
        (new KillVMThread(delay)).start();
        return (delay*1000); //just an estimate
    }
    /* END com.sun.jini.fiddler.Fiddler --> com.sun.jini.mahout.KillVMAdmin
     * -------------------------------------------------------------------- */

    /* -------------------------------------------------------------------- 
     * BEGIN com.sun.jini.fiddler.Fiddler
     */
    /**
     * Returns the unique identifier generated (or recovered) by the back-end
     * implementation of the lookup discovery service when an instance of
     * that service is constructed. This ID is typically used to determine
     * equality between the proxies of any two instances of the lookup
     * discovery service.
     * 
     * @return the unique ID that was generated (or recovered) by the
     *         back-end implementation of the lookup discovery service
     *         at creation time
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     */
    public UUID getProxyID() throws RemoteException {
        return proxyID;
    }

    /**
     * Registers with the lookup discovery service. When a client invokes
     * this method, it requests that the lookup discovery service perform
     * discovery processing on its behalf.
     *
     * @param groups        String array, none of whose elements may be null,
     *                      consisting of zero or more names of groups to
     *                      which lookup services to discover belong.
     *                      A null value or an empty array
     *                      (DiscoveryGroupManagement.ALL_GROUPS or 
     *                      DiscoveryGroupManagement.NO_GROUPS) are both
     *                      acceptable.
     * @param locators      array of zero or more non-null LookupLocator
     *                      objects, each corresponding to a specific lookup
     *                      service to discover. If either the empty array
     *                      or null is passed to this argument, then no
     *                      locator discovery will be performed for the
     *                      associated registration.
     * @param listener      a non-null instance of RemoteEventListener. This 
     *                      argument specifies the entity that will receive
     *                      events notifying the registration that a lookup
     *                      service of interest has been discovered. A 
     *                      non-null value must be passed to this argument,
     *                      otherwise a NullPointerException will be thrown
     *                      and the registration.
     * @param handback      null or an instance of MarshalledObject. This
     *                      argument specifies an object that will be 
     *                      included in the notification event that the
     *                      lookup discovery service sends to the registered
     *                      listener.
     * @param leaseDuration long value representing the amount of time (in
     *                      milliseconds) for which the resources of the
     *                      lookup discovery service are being requested.
     *
     * @return an instance of FiddlerRegistration which implements the
     *         LookupDiscoveryRegistration interface, and acts as a proxy
     *         to the registration-related methods of the back-end server
     *         of the Fiddler implementation of the lookup discovery service
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the registration may
     *         or may not have completed successfully.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         null is input to the <code>listener</code> parameter, as well
     *         as when one or more of the elements of the <code>groups</code>
     *         parameter is null.
     *
     * @throws java.lang.IllegalArgumentException this exception occurs when
     *         the value input to the <code>leaseDuration</code> parameter
     *         is neither positive, Lease.FOREVER, nor Lease.ANY.
     *
     * @see net.jini.discovery.LookupDiscoveryService
     */
    public LookupDiscoveryRegistration register(String[] groups,
                                                LookupLocator[] locators,
                                                RemoteEventListener listener,
                                                MarshalledObject handback,
                                                long leaseDuration)
                                                        throws RemoteException
    {
        if(containsNullElement(groups)) {
            throw new RemoteException
                          (" on call to register() method",
                           new NullPointerException
                                      ("at least one null element in groups"));
        } else if (containsNullElement(locators)) {
            throw new RemoteException
                          (" on call to register() method",
                           new NullPointerException
                                    ("at least one null element in locators"));
        } else if (listener == null) {
            throw new RemoteException
                          (" on call to register() method",
                           new NullPointerException("null listener input"));
        }
        concurrentObj.writeLock();
        LookupDiscoveryRegistration reg = null;
        try {
            /*  Grant the registration request and add the registration to
             *  to this service's state.
             *
             *  The addition of a registration to this service's state
             *  typically involves the modification of the managed sets in 
             *  the discovery manager, which usually involves starting the 
             *  discovery protocol. An IOException can occur when the 
             *  discovery protocol fails to start. When such an exception 
             *  does occur, register an ERROR status attribute (along with 
             *  a Comment attribute describing the nature of the problem) to 
             *  all lookup services with which this service is registered. 
             *
             *  Administrative clients, as well as clients that use this 
             *  service should have registered for notification of the 
             *  existence of this attribute.
             */
            reg = registerDo(groups,locators,listener,handback,leaseDuration);
        } catch(IOException e) {
            Entry[] errorAttrs 
                  = new Entry[] { new FiddlerStatus(StatusType.ERROR),
                                  new Comment("Failure during registration")
                                };
            joinMgr.addAttributes(errorAttrs,true);
	    e.printStackTrace();
        } finally {
            concurrentObj.writeUnlock();
        }
        return reg;
    }//end register

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method returns an array consisting of proxies to the lookup
     * service(s) that have already been discovered for the registration
     * corresponding to the <code>registrationID</code> input parameter.
     * Each element of the return set is a marshalled instance of the
     * <code>ServiceRegistrar</code> interface.
     *
     * @param registrationID unique identifier assigned to the registration
     *                       from which the set of registrars is being 
     *                       retrieved
     * 
     * @return an array of MarshalledObject objects where each element is
     *         is a marshalled instance of ServiceRegistrar.
     * 
     * @throws net.jini.discovery.LookupUnmarshalException this exception
     *         is thrown when failure occurs while attempting to unmarshal
     *         one or more of the marshalled instances of ServiceRegistrar
     *         received from the lookup discovery service.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service, or when an invalid or non-existent
     *         <code>registrationID</code> parameter.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#getRegistrars
     * @see net.jini.discovery.LookupDiscoveryRegistration#getRegistrars
     */
    public MarshalledObject[] getRegistrars(UUID registrationID)
                                                        throws RemoteException
    {
        concurrentObj.readLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to getRegistrars() method",
                                           new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            Collection mVals = (regInfo.discoveredRegsMap).values(); 
            return ( (MarshalledObject[])(mVals).toArray
                                        (new MarshalledObject[mVals.size()]) );
        } finally {
            concurrentObj.readUnlock();
        }
    }//end getRegistrars

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method returns an array consisting of the names of the groups
     * whose members are lookup services the lookup discovery service will
     * attempt to discover for the registration corresponding to the current
     * instance of this class. This set of group names is referred to as the
     * registration's 'managed set of groups'.
     * <p>
     * If the registration's managed set of groups is currently empty, then
     * the empty array is returned. If the lookup discovery service currently
     * has no managed set of groups for the registration through which the
     * request is being made, then null will be returned.
     *
     * @param registrationID unique identifier assigned to the registration
     *                       from which the set of groups is being retrieved
     * 
     * @return a String array containing the elements of the managed set of
     *         groups for the registration.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#getGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#getGroups
     */
    public String[] getGroups(UUID registrationID)  throws RemoteException {
        concurrentObj.readLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to getGroups() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            String[] groups = null;
            if(regInfo.groups == null) {
                groups = DiscoveryGroupManagement.ALL_GROUPS;
            } else {
                groups = (String[])(regInfo.groups).toArray
                                           (new String[regInfo.groups.size()]);
            }//endif (regInfo.groups == null)
            return groups;
        } finally {
            concurrentObj.readUnlock();
        }
    }//end getGroups

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method returns an array consisting of the the LookupLocator
     * objects corresponding to specific lookup services the lookup discovery
     * service will attempt to discover for for the registration
     * corresponding to the current instance of this class. This set of
     * locators is referred to as the registration's 'managed set of locators'.
     * <p>
     * If the registration's managed set of locators is currently empty, then
     * the empty array is returned. If the lookup discovery service currently
     * has no managed set of locators for the registration through which the
     * request is being made, then null will be returned.
     *
     * @param registrationID unique identifier assigned to the registration
     *                       from which the set of locators is being retrieved
     * 
     * @return array consisting of net.jini.core.discovery.LookupLocator
     *         objects corresponding to the elements of the managed set of
     *         locators for the registration.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#getLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#getLocators
     */
    public LookupLocator[] getLocators(UUID registrationID)
                                                      throws RemoteException
    {
        concurrentObj.readLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to getLocators() method",
                                           new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            return (LookupLocator[])(regInfo.locators).toArray
                                  (new LookupLocator[regInfo.locators.size()]);
        } finally {
            concurrentObj.readUnlock();
        }
    }//end getLocators

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of group names for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then adds the input set of group names to the managed set of groups
     * associated with the registration.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of groups being augmented
     *                       corresponds
     * @param groups         a String array, none of whose elements may be
     *                       null, consisting of the group names with which to
     *                       augment the registration's managed set of groups.
     * <p>
     *                       If any element of this parameter duplicates any
     *                       other element of this parameter, the duplicate
     *                       will be ignored. If any element of this parameter
     *                       duplicates any element of the registration's
     *                       current managed set of groups, the duplicate will
     *                       be ignored.
     * <p>
     *                       If the empty set is input, then the registration's
     *                       managed set of groups will not change. If null is
     *                       input, this method will throw a
     *                       <code>NullPointerException</code>.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>addGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when the registration corresponding to the
     *         <code>registrationID</code> parameter has no managed set of
     *         groups to which to add the elements of the input parameter.
     *         That is, the registration's current managed set of groups is
     *         null. When a registration's managed set of groups is null,
     *         it means that all groups are being discovered for that 
     *         registration; thus, requesting that a set of groups be added
     *         to the set of all groups makes no sense.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the <code>groups</code> parameter, 
     *         or one or more of the elements of the <code>groups</code> 
     *         parameter is null. If a null <code>groups</code> parameter 
     *         is input, the registration is requesting that all groups be 
     *         added to its current managed set of groups; which is not 
     *         allowed. (Note that if a registration wishes to change its 
     *         managed set of groups from a finite set of names to "all 
     *         groups", it should invoke setGroups with a null input.)
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully augmented.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#addGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#addGroups
     */
    public void addGroups(UUID     registrationID,
                          String[] groups)  throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to addGroups() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(groups == null) { // asking that all groups be added
                throw new RemoteException
                         (" on call to addGroups() method",
                          new NullPointerException
                              ("cannot add 'ALL_GROUPS' (the null set) to a\n"
                               +"                                        "
                               +"registration's set of groups to discover"));
            } else if(containsNullElement(groups)) { // null element
                throw new RemoteException
                        (" on call to addGroups() method",
                         new NullPointerException
                            ("at least one null element in groups parameter"));
            } else if (regInfo.groups == null) { // all groups being discovered
                throw new RemoteException
                        (" on call to addGroups() method",
                         new UnsupportedOperationException
                           ("cannot add a set of groups to a set already\n"
                           +"                                                 "
                           +"configured for 'ALL_GROUPS' (the null set)"));
            }//endif
            /* Augment the current set of groups with the input set */
            try {
                addGroupsDo(regInfo, groups);
            } catch (IOException e) {
                throw new RemoteException
                        (" on call to addGroups() method",
                         new IOException
                            ("multicast request protocol failed to start"));
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's addGroups() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to addGroups() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"addGroups() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoGroups("\nAfter Group Addition --");
            addLogRecord(new GroupsAddedToRegistrationLogObj
                                             (regInfo.registrationID,groups));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end addGroups

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of group names for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then replaces the registration's current managed set of groups
     * with the input set of group names.
     * <p>
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of groups being replaced
     *                       corresponds
     * @param groups         a String array, none of whose elements may be 
     *                       null, consisting of the group names with which to 
     *                       replace the names in this registration's managed 
     *                       set of groups.
     * <p>
     *                       If any element of this parameter duplicates any 
     *                       other element of this parameter, the duplicate 
     *                       will be ignored.
     * <p>
     *                       If the empty set is input, then group discovery 
     *                       for the registration will cease. If null is input,
     *                       the lookup discovery service will attempt to 
     *                       discover all as yet undiscovered lookup services 
     *                       located within its multicast radius and, upon 
     *                       discovery of any such lookup service, will send 
     *                       to the registration's listener an event signaling
     *                       that discovery.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>setGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.NullPointerException this exception occurs when one
     *         or more of the elements of the groups parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully replaced.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#setGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#setGroups
     */
    public void setGroups(UUID     registrationID,
                          String[] groups)  throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to setGroups() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(containsNullElement(groups)) { // null element
                throw new RemoteException
                        (" on call to addGroups() method",
                         new NullPointerException
                            ("at least one null element in groups parameter"));
            } else if ((groups == null) && (regInfo.groups == null)) {
                /* null input, but already set to ALL groups; do nothing */
                return;
            }//endif
            /* Replace the current groups with the current groups */
            try {
                setGroupsDo(regInfo, groups);
            } catch (IOException e) {
                throw new RemoteException
                        (" on call to setGroups() method",
                         new IOException
                            ("multicast request protocol failed to start"));
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's setGroups() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to setGroups() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"setGroups() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoGroups("\nAfter Group Replacement --");
            addLogRecord(new GroupsSetInRegistrationLogObj
                                             (regInfo.registrationID,groups));
            /* Queue a task to determine if a discard event should be sent */
            taskMgr.add(new RegistrarRemovedTask(regInfo));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end setGroups

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of group names for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then deletes the input set of group names from the managed set of
     * groups associated with the registration.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of groups being removed
     *                       corresponds
     * @param groups         a String array, none of whose elements may be
     *                       null, consisting of the group names to delete 
     *                       from the registration's managed set of groups.
     * <p>
     *                       If any element of this parameter duplicates any 
     *                       other element of this parameter, the duplicate 
     *                       will be ignored. If any element of this parameter
     *                       is not currently contained in the registration's
     *                       managed set, no action is taken with respect to
     *                       that element.
     * <p>
     *                       If the empty set is input, the registration's 
     *                       managed set of groups will not change. If null is 
     *                       input, this method will throw a 
     *                       <code>NullPointerException</code>.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>removeGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when the registration corresponding to the
     *         <code>registrationID</code> parameter has no managed set 
     *         of groups from which to remove elements of the input parameter.
     *         That is, the registration's current managed set of groups is
     *         null. Thus, requesting that a set of groups be removed from
     *         null set makes no sense. 
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the <code>groups</code> parameter, 
     *         or one or more of the elements of the <code>groups</code> 
     *         parameter is null. If a null <code>groups</code> parameter 
     *         is input, the registration is requesting that all groups be 
     *         removed from its current managed set of groups; which is not 
     *         allowed. (Note that if a registration wishes to change its 
     *         managed set of groups from "all groups" to "no groups", it
     *         it should invoke setGroups with a zero length 
     *         <code>String</code> input.)
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of groups may or may not have been
     *         successfully modified.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeGroups
     */
    public void removeGroups(UUID     registrationID,
                             String[] groups)  throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to removeGroups() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(groups == null) { // asking that all groups be removed
                throw new RemoteException
                     (" on call to removeGroups() method",
                      new NullPointerException
                        ("cannot remove 'ALL_GROUPS' (the null set) from a\n"
                         +"                                        "
                         +"registration's set of groups to discover"));
            } else if(containsNullElement(groups)) { // null element
                throw new RemoteException
                        (" on call to removeGroups() method",
                         new NullPointerException
                            ("at least one null element in groups parameter"));
            } else if (regInfo.groups == null) { // all groups being discovered
                throw new RemoteException
                        (" on call to removeGroups() method",
                         new UnsupportedOperationException
                          ("cannot remove a set of groups from a set\n"
                           +"                                                 "
                           +"configured for 'ALL_GROUPS' (the null set)"));
            }//endif
            /* Remove the requested groups */
            try {
                removeGroupsDo(regInfo, groups);
            } catch (IOException e) {
                throw new RemoteException
                        (" on call to removeGroups() method",
                         new IOException
                            ("multicast request protocol failed to start"));
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's removeGroups() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to removeGroups() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"removeGroups() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoGroups("\nAfter Group Removal --");
            addLogRecord(new GroupsRemovedFromRegistrationLogObj
                                             (regInfo.registrationID,groups));
            /* Queue a task to determine if a discard event should be sent */
            taskMgr.add(new RegistrarRemovedTask(regInfo));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end removeGroups

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of group names for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then adds the input set of LookupLocator objects to the managed set
     * of locators associated with the registration.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of locators being augmented
     *                       corresponds
     * @param locators       an array, none of whose elements may be null, 
     *                       consisting of the LookupLocator objects with 
     *                       which to augment the registration's managed set 
     *                       of locators.
     * <p>
     *                       If any element of this parameter duplicates any 
     *                       other element of this parameter, the duplicate 
     *                       will be ignored. If any element of this parameter 
     *                       duplicates any element of the registration's 
     *                       managed set of locators, the duplicate will be 
     *                       ignored.
     * <p>
     *                       If the empty set is input, then the registration's
     *                       managed set of locators will not change. If null
     *                       is input, this method will throw a
     *                       <code>NullPointerException</code>.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>addLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when the registration corresponding to the
     *         <code>registrationID</code> parameter has no managed set of
     *         locators to which to add the elements of the input parameter.
     *         That is, the registration's current managed set of locators is
     *         null.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the <code>locators</code> parameter, 
     *         or one or more of the elements of the <code>locators</code> 
     *         parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully augmented.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#addLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#addLocators
     */
    public void addLocators(UUID            registrationID,
                            LookupLocator[] locators) throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to addLocators() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(locators == null) {
                throw new RemoteException
                         (" on call to addLocators() method",
                          new NullPointerException
                                  ("cannot add null to a registration's set \n"
                                   +"                                        "
                                   +"of locators to discover"));
            } else if(containsNullElement(locators)) { // null element
                throw new RemoteException
                        (" on call to addLocators() method",
                         new NullPointerException
                          ("at least one null element in locators parameter"));
            }//endif(locators == null)
            /* Augment the current set of locators with the input set */
            try {
                addLocatorsDo(regInfo, locators);
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's addLocators() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to addLocators() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"addLocators() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoLocators("\nAfter Locator Addition --");
            addLogRecord(new LocsAddedToRegistrationLogObj
                                           (regInfo.registrationID,locators));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end addLocators

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of locators for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then replaces the registration's current managed set of locators
     * with the input set of locators.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of locators being replaced
     *                       corresponds
     * @param locators       an array, none of whose elements may be null,
     *                       consisting of the LookupLocator objects with 
     *                       which to replace the locators in the 
     *                       registration's managed set of locators.
     * <p>
     *                       If any element of this parameter duplicates any 
     *                       other element of this parameter, the duplicate 
     *                       will be ignored.
     * <p>
     *                       If the empty array is input, then locator 
     *                       discovery for the registration will cease. If 
     *                       null is input, this method will throw a 
     *                       <code>NullPointerException</code>.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>setLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.NullPointerException this exception occurs when one
     *         or more of the elements of the locators parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully replaced.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#setLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#setLocators
     */
    public void setLocators(UUID            registrationID,
                            LookupLocator[] locators) throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException(" on call to setLocators() method",
                                          new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(locators == null) {
                throw new RemoteException
                         (" on call to setLocators() method",
                          new NullPointerException
                                  ("cannot replace a registration's current \n"
                                   +"                                        "
                                   +"set of locators with null"));
            } else if(containsNullElement(locators)) { // null element
                throw new RemoteException
                        (" on call to setLocators() method",
                         new NullPointerException
                          ("at least one null element in locators parameter"));
            }//endif(locators == null)
            /* Replace the current set of locators with the input set */
            try {
                setLocatorsDo(regInfo, locators);
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's setLocators() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to setLocators() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"setLocators() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoLocators("\nAfter Locator Replacement --");
            addLogRecord(new LocsSetInRegistrationLogObj
                                           (regInfo.registrationID,locators));
            /* Queue a task to determine if a discard event should be sent */
            taskMgr.add(new RegistrarRemovedTask(regInfo));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end setLocators

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method first tests the input set of locators for validity and 
     * throws the appropriate exception should any irregularities be found.
     * It then deletes the input set of locators from the managed set of
     * locators associated with the registration.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       to which the set of locators being removed
     *                       corresponds
     * @param locators       an array, none of whose elements may be null,
     *                       consisting of the LookupLocator objects to remove
     *                       from the registration's managed set of locators.
     * <p>
     *                       If any element of this parameter duplicates any
     *                       other element of this parameter, the duplicate
     *                       will be ignored.
     * <p>
     *                       If the empty set is input, the managed set of
     *                       locators will not change. If null is input,
     *                       this method will throw a 
     *                       <code>NullPointerException</code>.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>removeLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called. When this happens, in addition to
     *         propagating this exception, this method also registers an 
     *         ERROR status attribute (along with a Comment attribute
     *         describing the nature of the problem) with all lookup services
     *         with which this service is registered. Administrative clients,
     *         as well as clients that use this service should register
     *         for notification of the existence of this attribute.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when the registration corresponding to the
     *         <code>registrationID</code> parameter has no managed set of
     *         locators from which to remove the elements of the input
     *         parameter. That is, the registration's current managed set
     *         of locators is null.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either null is input to the <code>locators</code> parameter, 
     *         or one or more of the elements of the <code>locators</code> 
     *         parameter is null.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         registration's managed set of locators may or may not have
     *         been successfully modified.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeLocators
     */
    public void removeLocators(UUID            registrationID,
                               LookupLocator[] locators) throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException
                                        (" on call to removeLocators() method",
                                         new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            /* Check the input for validity */
            if(locators == null) {
                throw new RemoteException
                         (" on call to removeLocators() method",
                          new NullPointerException
                                 ("cannot remove null from a registration's \n"
                                  +"                                        "
                                  +"current set of locators"));
            } else if(containsNullElement(locators)) { // null element
                throw new RemoteException
                        (" on call to setLocators() method",
                         new NullPointerException
                          ("at least one null element in locators parameter"));
            }//endif(locators == null)
            /* Remove the requested set of locators from the current set */
            try {
                removeLocatorsDo(regInfo, locators);
            } catch(IllegalStateException e) {
                String errStr = "IllegalStateException: discovery "
                               +"manager's removeLocators() method was called "
                               +"after the manager was terminated";
                Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment(errStr)
                 };
                joinMgr.addAttributes(errorAttrs,true);
                throw new RemoteException
                        (" on call to setLocators() method",
                         new IllegalStateException
                                        ("discovery manager's "
                                         +"removeLocators() method was called "
                                         +"after the manager was terminated"));
            }//end try
            writeInfoLocators("\nAfter Locator Removal --");
            addLogRecord(new LocsRemovedFromRegistrationLogObj
                                           (regInfo.registrationID,locators));
            /* Queue a task to determine if a discard event should be sent */
            taskMgr.add(new RegistrarRemovedTask(regInfo));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end removeLocators

    /**
     * This method is the "back-end" server counterpart to the method of
     * the same name provided by the <code>LookupDiscoveryRegistration</code> 
     * proxy (an instance of <code>FiddlerRegistration</code>) that is
     * returned by this service when a client requests a registration.
     * <p>
     * This method informs the lookup discovery service of the existence of
     * an unavailable lookup service and requests that the lookup discovery
     * service discard the unavailable lookup service and make it eligible
     * to be re-discovered.
     * 
     * @param registrationID unique identifier assigned to the registration
     *                       making the current discard request
     * @param registrar      a reference to the lookup service that the lookup
     *                       discovery service is being asked to discard.
     * <p>
     *                       If this parameter equals none of the lookup
     *                       services contained in the managed set of lookup
     *                       services for this registration, no action will
     *                       be taken.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         null is input to the registrar parameter.
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, 
     *         the lookup service may or may not have been successfully
     *         discarded.
     *
     * @see com.sun.jini.fiddler.FiddlerRegistration#discard
     * @see net.jini.discovery.LookupDiscoveryRegistration#discard
     */
    public void discard(UUID registrationID,
                        ServiceRegistrar registrar)  throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            writeInfoDiscard("\ndiscard: ",registrationID);
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException
                                        (" on call to discard() method",
                                         new NoSuchObjectException
                                                  ("Invalid registration ID"));
            }//endif
            if(registrar == null) {
                throw new RemoteException
                                  (" on call to discard() method",
                                   new NullPointerException
                                      ("null input for registrar to discard"));
            }//endif

            /* Get discovered registrars across all registrations */
            ServiceRegistrar[] curRegs = discoveryMgr.getRegistrars();
            if( regIsElementOfRegSet(registrar,curRegs) ) {
                /* This must be the first discard request for this registrar 
                 * because the discovery manager has not discarded it yet.
                 * When the discovery manager discards the registrar, a
                 * local discarded event is sent to the listener.discarded
                 * method which queues a DiscardedEventTask which will 
                 * remove the discarded registrar from the the registration's
                 * set of discovered registrars and then send a remote
                 * discarded event to the registration's listener.
                 */
                writeInfoDiscard("  Registrar IS an element of Mgr's "
                                 +"discovered registrars ... discarding "
                                 +"from discovery manager");
                regInfo.discardFlag = true;
                discoveryMgr.discard(registrar);
            } else {
                writeInfoDiscard("  Registrar NOT an element of Mgr's "
                                 +"discovered registrars ... queuing "
                                 +"new DiscardRegistrarTask");
                /* For all subsequent discard requests, remove the registrar
                 * from the registration's set of discovered registrars and
                 * send a remote discarded event, but don't ask the discovery
                 * manager to discard the registrar.
                 */
                taskMgr.add(new DiscardRegistrarTask(regInfo,registrar));
            }
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end discard

    /**
     * This method is the "back-end" server counterpart to the 
     * <code>renew</code> method specified by the <code>Lease</code> interface,
     * implemented in the <code>com.sun.jini.lease.AbstractLease</code> class,
     * and invoked by way of the <code>doRenew</code> method of the
     * <code>FiddlerLease</code> class; an instance of which is
     * returned by the <code>getLease</code> method of the
     * <code>LookupDiscoveryRegistration</code> proxy (an instance of
     * <code>FiddlerRegistration</code>) that is returned by this service
     * when a client requests a registration.
     * <p>
     * This method renews the lease corresponding to the given 
     * <code>registrationID</code> and <code>leaseID</code> parameters,
     * granting a new duration that is less than or equal to the requested
     * duration value contained in the <code>duration</code> parameter.
     *
     * @param registrationID unique identifier assigned to the registration
     *                       to which the lease being renewed corresponds
     * @param leaseID        identifier assigned by the lease grantor to the
     *                       lease being renewed
     * @param duration       the requested duration for the lease being renewed
     *
     * @return <code>long</code> value representing the actual duration that
     *         was granted for the renewed lease. Note that the actual
     *         duration granted and returned by this method may be less than
     *         the duration requested.
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed does not exist, or is unknown
     *         to the lease grantor; typically because the lease has expired.
     *         
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been renewed successfully.
     *
     * @see net.jini.core.lease.Lease#renew
     * @see com.sun.jini.lease.AbstractLease#renew
     * @see com.sun.jini.lease.AbstractLease#doRenew
     * @see com.sun.jini.fiddler.FiddlerLease#doRenew
     */
    public long renewLease(UUID registrationID,
                           long leaseID,
                           long duration)
                                throws UnknownLeaseException, RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException
                       (" on call to renewLease() method",
                        new UnknownLeaseException
                        ("Invalid registration ID\n"
                         +"                                                   "
                         +"The lease may have expired or been cancelled"));
            }//endif
            /* Renew the lease */
            long newDuration = 0;
            try {
                newDuration = renewLeaseDo(regInfo, leaseID, duration);
                writeInfoLease("Renewed lease: ",registrationID,leaseID);
            } catch(UnknownLeaseException e) {
                throw new RemoteException
                          (" on call to renewLease() method",
                           new UnknownLeaseException
                           ("Invalid registration -- the lease has expired"));
            }
            /* The call to addLogRecord is in renewLeaseDo */
            return newDuration;
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end renewLease

    /**
     * This methods renews all leases from a <code>LeaseMap</code>, 
     * where each element of the map is a lease on a registration with
     * ID corresponding to an element of the <code>registrationIDs</code> 
     * parameter.
     * <p>
     * This method is the "back-end" server counterpart to the 
     * <code>renewAll</code> method specified by the
     * <code>LeaseMap</code> interface, implemented in the 
     * <code>com.sun.jini.lease.AbstractLeaseMap</code> class, and 
     * invoked by way of the <code>renewAll</code> method of the
     * <code>FiddlerLease</code> class; an instance of which is
     * returned by the <code>getLease</code> method of the
     * <code>LookupDiscoveryRegistration</code> proxy (an instance of
     * <code>FiddlerRegistration</code>) that is returned by this service
     * when a client requests a registration.
     *
     * @param registrationIDs array containing the unique identifiers assigned
     *                        to the each registration to which each lease 
     *                        to be renewed corresponds
     * @param leaseIDs        array containing the identifiers assigned by the
     *                        lease grantor to each lease being renewed
     * @param durations       array containing the requested durations for 
     *                        each lease being renewed
     * 
     * @return an instance of FiddlerRenewResults containing data corresponding
     *         to the results (granted durations or exceptions) of each
     *         renewal attempt
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, this method may or
     *         may not have complete its processing successfully.
     *
     * @see net.jini.core.lease.LeaseMap#renewAll
     */
    public FiddlerRenewResults renewLeases(UUID[] registrationIDs,
                                           long[] leaseIDs,
                                           long[] durations)
                                                        throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            return renewLeasesDo(registrationIDs, leaseIDs, durations);
            /* The call to addLogRecord is in renewLeasesDo */
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end renewLeases

    /**
     * This method is the "back-end" server counterpart to the 
     * <code>cancel</code> method specified by the <code>Lease</code>
     * interface and implemented in the <code>FiddlerLease</code> class; an
     * instance of which is returned by the <code>getLease</code> method
     * of the <code>LookupDiscoveryRegistration</code> proxy (an instance of
     * <code>FiddlerRegistration</code>) that is returned by this service
     * when a client requests a registration.
     * <p>
     * This method cancels the lease corresponding to the given 
     * <code>registrationID</code> and <code>leaseID</code> parameters.
     *
     * The cancelation of a lease typically involves the modification of the
     * managed sets in the discovery manager, which usually involves starting
     * the discovery protocol.  An IOException can occur when the discovery
     * protocol fails to start. When such an exception does occur, this 
     * method registers an ERROR status attribute (along with a Comment
     * attribute describing the nature of the problem) to all lookup services
     * with which this service is registered. 
     *
     * Administrative clients, as well as clients that use this service should
     * have registered for notification of the existence of this attribute.
     *
     * @param registrationID unique identifier assigned to the registration
     *                       to which the lease being cancelled corresponds
     * @param leaseID        identifier assigned by the lease grantor to the
     *                       lease that is to be cancelled
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being cancelled is unknown to the lease grantor.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been cancelled successfully.
     *
     * @see net.jini.core.lease.Lease#cancel
     */
    public void cancelLease(UUID registrationID,
                            long leaseID)
                             throws UnknownLeaseException, RemoteException
    {
        concurrentObj.writeLock();
        try {
            RegistrationInfo regInfo
                   = (RegistrationInfo)(registrationByID.get(registrationID));
            if(regInfo == null) {
                throw new RemoteException
                       (" on call to cancelLease() method",
                        new UnknownLeaseException
                        ("Invalid registration ID\n"
                         +"                                                   "
                         +"The lease may have expired or been cancelled"));
            }//endif
            /* Cancel the lease */
            try {
                cancelLeaseDo(regInfo, leaseID);
                writeInfoLease("Cancelled lease: ",registrationID,leaseID);
            } catch(UnknownLeaseException e) {
                throw new RemoteException
                          (" on call to cancelLease() method",
                           new UnknownLeaseException
                           ("Invalid registration -- the lease has expired"));
            } catch(IOException e) {
                Entry[] errorAttrs
                 = new Entry[] { new FiddlerStatus(StatusType.ERROR),
                                 new Comment("Failure while cancelling the "
                                            +"lease on registration with ID = "
                                            +registrationID)
                                        };
                joinMgr.addAttributes(errorAttrs,true);
	        e.printStackTrace();
            }
            addLogRecord(new LeaseCancelledLogObj
                                          (regInfo.registrationID, leaseID));
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end cancelLease

    /**
     * Cancels all leases from a <code>LeaseMap</code>.
     * <p>
     * For each element in the <code>registrationIDs</code> parameter,
     * this method will cancel the corresponding element in the
     * <code>leaseIDs</code> parameter.
     *
     * @param registrationIDs array containing the unique identifiers assigned
     *                        to the each registration to which each lease 
     *                        to be cancelled corresponds
     * @param leaseIDs        array containing the identifiers assigned by the
     *                        lease grantor to each lease being cancelled
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, this method may or
     *         may not have complete its processing successfully.
     * 
     * @return array consisting of any exceptions that may have occurred 
     *         while attempting to cancel one of the leases in the map.
     *
     * @see net.jini.core.lease.LeaseMap#cancelAll
     */
    public Exception[] cancelLeases(UUID[] registrationIDs,
                                    long[] leaseIDs)  throws RemoteException
    {
        concurrentObj.writeLock();
        try {
            /* don't bother to weed out unknown leases, so log first */
            addLogRecord( new LeasesCancelledLogObj(registrationIDs,leaseIDs));
            return cancelLeasesDo(registrationIDs, leaseIDs);
        } finally {
            concurrentObj.writeUnlock();
        }
    }//end cancelLeases
    /* END com.sun.jini.fiddler.Fiddler ----------------------------------- */
    /* ************************* END Public Methods *********************** */

    /* **************** BEGIN Private Static Utility Methods ************** */
    /**
     * Return a new array containing the elements of the input array parameter
     * with the input element parameter appended to the end of the array.
     */
    private static Object[] appendArray(Object[] array, Object elt) {
	int len = array.length;
	Object[] newArray =
	    (Object[])Array.newInstance(array.getClass().getComponentType(),
					len + 1);
	System.arraycopy(array, 0, newArray, 0, len);
	newArray[len] = elt;
	return newArray;
    }//end appendArray

    /**
     * Bounds the duration by the <code>bound</code>, and checks for negative
     * value.
     */
    private static long applyBoundToLeaseDuration(long leaseDuration,
                                                  long bound)
    {
        long newLeaseDuration = leaseDuration;
        if ( (leaseDuration == Lease.ANY) || (leaseDuration > bound) ) {
            newLeaseDuration = bound;
        } else if (leaseDuration < 0) {
            throw new IllegalArgumentException("negative lease duration");
        }
        return newLeaseDuration;
    }//end applyBoundToLeaseDuration

    /** Generates a random lease id */
    private long getRandomLeaseID() {
        secRand.nextBytes(secRandBuf8);
        long id = 0;
        for (int i = 8; --i >= 0; ) {
            id = (id << 8) + (secRandBuf8[i] & 0xFF);
        }
        return id;
    }//end getRandomLeaseID

    /**
     * Determines if any element in the input array is null.
     * @param arr Object array to examine for null elements
     * @return true if any element is found to be null, false otherwise.
     *         Note that this means that if the array itself is null
     *         or has a non-positive length, false is returned (because
     *         the input parameter still does not contain a null element).
     */
    private static boolean containsNullElement(Object[] arr) {
        if( (arr == null) || (arr.length == 0) ) return false;
        for(int i=0;i<arr.length;i++) {
            if(arr[i] == null) return true;
        }
        return false;
    }//end containsNullElement

    /**
     * Marshals each element of the <code>Entry[]</code> array parameter.
     * This method is <code>static</code> so that it may called from
     * the <code>static</code> <code>LogRecord</code> classes when a set
     * of attributes is being logged to persistent storage.
     *
     * @param fiddlerImpl reference to the current instance of this service
     * @param attrs       <code>Entry[]</code> array consisting of the
     *                    attributes to marshal
     * @return array of <code>MarshalledObject[]</code>, where each element
     *         corresponds to an attribute in marshalled form 
     */
    private static MarshalledObject[] marshalAttributes
                                                   (FiddlerImpl fiddlerImpl,
                                                    Entry[] attrs)
    {
        if(attrs == null) return new MarshalledObject[0];
        ArrayList marshalledAttrs = new ArrayList();
        for(int i=0;i<attrs.length;i++) {
            /* Do not let an attribute problem prevent the service from
             * continuing to operate
             */
            try {
                marshalledAttrs.add(new MarshalledObject(attrs[i]));
            } catch(Throwable e) {
                writeInfoLog(fiddlerImpl.debugAllWriter,
                             fiddlerImpl.debugLogWriter,
                             "Error while marshalling attribute["+i+"] ("
                             +attrs[i]+"): "+e.toString());
            }
        }
        return ((MarshalledObject[])(marshalledAttrs.toArray
                             (new MarshalledObject[marshalledAttrs.size()])));
    }//end marshalAttributes

    /**
     * Unmarshals each element of the <code>MarshalledObject[]</code> array
     * parameter. This method is <code>static</code> so that it may called
     * from the <code>static</code> <code>LogRecord</code> classes when a
     * set of attributes is being recovered from persistent storage.
     *
     * @param fiddlerImpl     reference to the current instance of this service
     * @param marshalledAttrs <code>MarshalledObject[]</code> array consisting
     *                        of the attributes to unmarshal
     * @return array of <code>Entry[]</code>, where each element corresponds
     *         to an attribute that was successfully unmarshalled
     */
    private static Entry[] unmarshalAttributes
                                          (FiddlerImpl fiddlerImpl,
                                           MarshalledObject[] marshalledAttrs)
    {
        if(marshalledAttrs == null) return new Entry[0];
        ArrayList attrs = new ArrayList();
        for(int i=0;i<marshalledAttrs.length;i++) {
            /* Do not let an attribute problem prevent the service from
             * continuing to operate
             */
            try {
                attrs.add( (Entry)( marshalledAttrs[i].get() ) );
            } catch(Throwable e) {
                writeInfoLog(fiddlerImpl.debugAllWriter,
                             fiddlerImpl.debugLogWriter,
                             "Error while unmarshalling attribute["+i+"]: "
                             +e.toString());
            }
        }
        return ((Entry[])(attrs.toArray(new Entry[attrs.size()])));
    }//end unmarshalAttributes
    /* **************** END Private Static Utility Methods ***************** */

    /* ************** BEGIN Private NON-Static Utility Methods ************* */
    /* BEGIN Private Startup Methods --------------------------------------- */
    /**
     * Extracts a set of instances of <code>LookupLocator</code> from the
     * input <code>Object</code> array parameter.
     *
     * @param  locators <code>Object</code> array in which either all elements
     *         are instances of <code>LookupLocator</code> corresponding to
     *         a specific lookup service to discover and join, or all elements
     *         is an instance of <code>String</code> containing the name of 
     *         a host on which a specific lookup service to discover and join
     *         is running.
     *
     * @return array containing instances of <code>LookupLocator</code> in 
     *         which element corresponds to a specific lookup service to
     *         discover and join
     */
    private LookupLocator[] getLocatorsFromStartParams(Object[] locators) {
        if(locators == null)    return new LookupLocator[0];
        if(locators.length == 0) return new LookupLocator[0];

        LookupLocator[] locs = new LookupLocator[locators.length];
        if(locators[0] instanceof String) {
            try {
                for(int i=0;i<locators.length;i++) {
                    String locHost = (String)(locators[i]);
                    try {
                        locs[i] = new LookupLocator
                                               ("jini://"+locHost+":4160/");
                    } catch (MalformedURLException e) {
                        System.out.println("MalformedURLException from "+
                                        "LookupLocatorDiscovery constructor");
                    }
                }//end loop
            } catch(Exception e) {
                throw new IllegalArgumentException
                         ("invalid locators MarshalledObject data structure");
            }
        } else if(locators[0] instanceof LookupLocator) {
            try {
                for(int i=0;i<locators.length;i++) {
                    locs[i] = (LookupLocator)(locators[i]);
                }
            } catch(Exception e) {
                throw new IllegalArgumentException
                     ("invalid locators MarshalledObject data structure");
            }
        } else {
            throw new IllegalArgumentException
                         ("invalid locators MarshalledObject data structure");
        }//endif(instanceof)
        return locs;
    }//end class getLocatorsFromStartParams

    /**
     * Constructs and exports a new transient instance of FiddlerImpl.
     *
     * @param  logDirname <code>String</code> containing the absolute path
     *         of the directory directory in which the state of the service
     *         will be recorded
     * @param  groups <code>String</code> array containing the names of the
     *         groups whose members are the lookup services that should be
     *         discovered and joined
     * @param  locators <code>LookupLocator</code> array containing elements
     *         that correspond to specific lookup services to discover and
     *         join
     *
     * @throws java.io.IOException this exception can occur when there is a
     *         failure during the export process
     */
    private void constructTransientInstance(String   logDirname,
                                            String[] groups,
                                            LookupLocator[] locators)
                                                            throws IOException
    {
	UnicastRemoteObject.exportObject(this);
	this.logDirname           = logDirname;
        this.thisServicesGroups   = groups;
        this.thisServicesLocators = locators;
        init();
    }//end class constructTransientInstance

    /** Initialization common to both activatable and non-activatable 
     * (transient) instances.
     */
    private void init() throws IOException {
        setDebugWriters();
        writeInfoStartup();

        String host = System.getProperty("java.rmi.server.hostname");
        if (host == null) {
            host = InetAddress.getLocalHost().getHostName();
        }
        /* Create the discovery mechanisms, initially discovering no groups
         * and no locators. These will be populated during recovery, and
         * will be started after recovery.
         */
        discoveryMgr = new LookupDiscoveryManager
                                          (DiscoveryGroupManagement.NO_GROUPS,
                                           new LookupLocator[0],
                                           null);
        /* Recover any prior State */
        log = new ReliableLog(logDirname, new LocalLogHandler());
        inRecovery = true;
        log.recover();
        inRecovery = false;
        /* Recovery complete. Set non-dynamic info about this service */
	if (proxyID == null) {
            proxyID = (new UUIDFactory()).newUUID();
        }
        fiddlerProxy = new FiddlerProxy((Fiddler)RemoteObject.toStub(this),
                                         proxyID);
        /* Take snapshot of current state -- "cleans up" the log file */
	log.snapshot();
        /* Start discovery mechanisms for all recovered registrations */
        discoveryMgr.addDiscoveryListener(discoveryListener);
        writeInfoJoinManager();
        /* Create a join manager to manage this service's join state with
         * the lookup services with which this service is configured to
         * register.
         * 
         * First, create the lookup discovery manager that the join manager
         * should use to manage its discovery duties. Then, if the current
         * instance of this service does not already have a service ID,
         * instantiate a join manager with a ServiceIDListener, which will
         * cause one of the lookup services to generate a service ID on
         * behalf of this service. If the current instance of this service
         * already has a service ID that it has recovered from persistent
         * storage, instantiate a join manager with that service ID.
         */
        joinMgrLookupDiscoveryMgr
                           = new LookupDiscoveryManager(thisServicesGroups,
                                                        thisServicesLocators,
                                                        null);
	if (serviceID == null) {
            joinMgr = new JoinManager(fiddlerProxy, thisServicesAttrs,
                                      new SrvcIDListener(),
                                      joinMgrLookupDiscoveryMgr, null);
        } else {
            joinMgr = new JoinManager(fiddlerProxy, thisServicesAttrs,
                                      serviceID,
                                      joinMgrLookupDiscoveryMgr, null);
        }//endif

	/* start up all the daemon threads */
	leaseExpireThread.start();
        snapshotThread.start();

    }//end init
    /* END Private Startup Methods ----------------------------------------- */

    /* BEGIN Private Shutdown Methods -------------------------------------- */
    /**
     * Called by the public method <code>destroy</code> as well as by the
     * <code>apply</code> method in the various LogObj classes that,
     * during recovery, modify the managed sets of the discovery manager;
     * and, while doing so, experience an IOException when the multicast
     * request protocol fails to start during recovery (an un-recoverable
     * exception).
     *
     * This method destroys the lookup discovery service, if possible,
     * including its persistent storage. This method spawns a separate
     * thread to do the actual work asynchronously, so a successful
     * return from this method usually does not mean that the service
     * has been destroyed.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#destroy
     */
    private void destroyDo() {
        (new DestroyThread()).start();
    }//end destroyDo
    /* END Private Shutdown Methods ---------------------------------------- */

    /* BEGIN Private Registration Methods ---------------------------------- */
    /**
     * This method is called by the public method <code>register</code>.
     * This method creates a registration object that is an instance of
     * <code>FiddlerRegistration</code> which implements the interface
     * <code>LookupDiscoveryRegistration</code>, and acts as a proxy to the
     * registration-related methods of the back-end server of the Fiddler
     * implementation of the lookup discovery service.
     * <p>
     * This method also associates with the registration all information
     * needed by the registration to paricipate in the lookup discovery
     * service, information such as: IDs, lease information, event information.
     * 
     * @param groups        names of groups whose members are the lookup
     *                      services to discover.
     * @param locators      instances of LookupLocator, each corresponding to
     *                      a specific lookup service to discover.
     * @param listener      the entity that will receive events notifying the
     *                      registration that a lookup service of interest has
     *                      been discovered.
     * @param handback      the object that will be included in every
     *                      notification event sent to the registered listener.
     * @param leaseDuration long value representing the amount of time (in
     *                      milliseconds) for which the resources of the
     *                      lookup discovery service are being requested.
     *                      
     * @return an instance of FiddlerRegistration which implements the
     *         LookupDiscoveryRegistration interface, and acts as a proxy
     *         to the registration-related methods of the back-end server
     *         of the Fiddler implementation of the lookup discovery service
     *
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#register
     */
    private LookupDiscoveryRegistration registerDo
                                             (String[] groups,
                                              LookupLocator[] locators,
                                              RemoteEventListener listener,
                                              MarshalledObject handback,
                                              long leaseDuration)
                                                             throws IOException
    {
        /* Input okay. Create the registration and associated information */
	long curTime    = System.currentTimeMillis();
        leaseDuration   = applyBoundToLeaseDuration(leaseDuration,
                                                    leaseDurationBound);
        UUID regID      = (new UUIDFactory()).newUUID();
        long leaseID    = getRandomLeaseID();
        long expiration = curTime + leaseDuration;

        RegistrationInfo regInfo = new RegistrationInfo( regID,
                                                         groups,locators,
                                                         leaseID,expiration,
                                                         curEventID,handback,
                                                         listener );
        curEventID++;
	addRegistration(regInfo);
        writeInfoRegistration("\nAdded registration: ",regID);
        addLogRecord(new RegistrationGrantedLogObj(regInfo));
        /* Queue task for sending a discovered event */
        taskMgr.add(new NewRegistrationTask(regInfo));

	/* see if the expire thread needs to wake up earlier */
	if (expiration < minExpiration) {
	    minExpiration = expiration;
	    concurrentObj.waiterNotify(leaseExpireThreadSyncObj);
	}
        FiddlerLease regLease
                         = new FiddlerLease(this, regID, leaseID, expiration);
        EventRegistration eventReg = new EventRegistration( regInfo.eventID,
                                                            fiddlerProxy,
                                                            regLease,
                                                            regInfo.seqNum );
        writeInfoGroups();
        writeInfoLocators();
        return new FiddlerRegistration(this, regID, eventReg, regLease);
    }//end registerDo

    /** 
     * Places the registration corresponding to the <code>regInfo</code> 
     * parameter in both the <code>registrationByID</code> map and the 
     * <code>registrationByTime</code> map. This method also updates the 
     * managed sets in the discovery manager in the appropriate way. This 
     * should be called whenever a new registration has been created 
     * and needs to be added to the data base (for example, the methods
     * <code>registerDo</code>, <code>recoverSnapshot</code> and the 
     * <code>apply</code> method of the <code>RegistrationGrantedLogObj</code>
     * class all call this method).
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration that is to be added to the data base
     *
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#registerDo
     * @see com.sun.jini.fiddler.FiddlerImpl#recoverSnapshot
     */
    private void addRegistration(RegistrationInfo regInfo) throws IOException {
        if (regInfo.listener == null) {
            return; /* failed to recover from log */
        }
        /* First add the indicated registration */
        registrationByID.put(regInfo.registrationID, regInfo);
        registrationByTime.put(regInfo,regInfo);
        /* Update the set of groups managed by the discovery manager */
        String[] groupsAcrossAllRegs = getGroupsFromAllRegs();
        if(groupsAcrossAllRegs == DiscoveryGroupManagement.ALL_GROUPS) {
            if(discoveryMgr.getGroups()!=DiscoveryGroupManagement.ALL_GROUPS){
                discoveryMgr.setGroups(DiscoveryGroupManagement.ALL_GROUPS);
            }
        } else { // no registration desires ALL groups be discovered
            discoveryMgr.setGroups( groupsAcrossAllRegs );
        }
        /* Update the set of locators managed by the discovery manager */
        LookupLocator[] locsAcrossAllRegs = getLocatorsFromAllRegs();
        discoveryMgr.setLocators( locsAcrossAllRegs );
    }//end addRegistration

    /** 
     * Removes the registration corresponding to the <code>regInfo</code> 
     * parameter from this service's state. Removes the registration from 
     * both the <code>registrationByID</code> map and the 
     * <code>registrationByTime</code> map. This method also updates the 
     * managed sets in the discovery manager in the appropriate way. This 
     * should be called whenever a current registration needs to be removed
     * from the data base (for example, when the registration's lease is
     * is expired in the <code>LeaseExpireThread</code>, and when the
     * registration's lease is cancelled in the <code>cancelLeaseDo</code>
     * method).
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration that is to be removed from the data base
     *
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#cancelLeaseDo
     */
    private void removeRegistration(RegistrationInfo regInfo)
                                                         throws IOException
    {
        /* First remove the current registration */
        registrationByID.remove(regInfo.registrationID);
        registrationByTime.remove(regInfo);
        writeInfoRegistration("\nRemoved registration: ",
                              regInfo.registrationID);
        /* Update the set of groups managed by the discovery manager */
        String[] groupsAcrossAllRegs = getGroupsFromAllRegs();
        if(groupsAcrossAllRegs == DiscoveryGroupManagement.ALL_GROUPS) {
            if(discoveryMgr.getGroups()!=DiscoveryGroupManagement.ALL_GROUPS){
                discoveryMgr.setGroups(DiscoveryGroupManagement.ALL_GROUPS);
            }
        } else { // no registration desires ALL groups be discovered
            discoveryMgr.setGroups( groupsAcrossAllRegs );
        }
        /* Update the set of locators managed by the discovery manager */
        LookupLocator[] locsAcrossAllRegs = getLocatorsFromAllRegs();
        discoveryMgr.setLocators( locsAcrossAllRegs );
        writeInfoGroups();
        writeInfoLocators();
    }//end removeRegistration
    /* END Private Registration Methods ------------------------------------ */

    /* BEGIN Private Group Management Methods ------------------------------ */
    /**
     * Called by the public method <code>addGroups</code>. This method adds
     * new group names to the current managed set of groups for the
     * registration corresponding to the <code>regInfo</code> data structure
     * parameter. Note that only elements that do not duplicate elements
     * already in the set are added.
     * <p>
     * In addition to augmenting the current managed set of groups, this
     * method adds the input group names to the set of all groups to
     * discover (across all registrations); also with duplicates not added.
     * <p>
     * Note that if a group name input to this method is already in the
     * registration's set of groups, it must also be in the set of all
     * groups. But if that input group name is not in the registration's
     * current managed set of groups, it still may be a member of the set
     * of all groups (because it was added by another registration). Thus,
     * this method attempts to add the group name to both sets, but includes
     * the group name in the set to add to the discovery manager only
     * if it was successfully added to the set of all groups.
     * 
     * @param regInfo the data structure record corresponding to the
     *                registration whose managed set of groups is to be
     *                augmented
     * @param groups  a String array, none of whose elements may be null,
     *                consisting of the group names with which to augment the
     *                registration's managed set of groups.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>addGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#addGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#addGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#addGroups
     */
    private void addGroupsDo(RegistrationInfo regInfo,
                             String[] groups)   throws IOException
    {
        /* Build a HashSet (removes duplicates) from the input set of groups */
        HashSet newGroupSet = new HashSet();
        for(int i=0;i<groups.length;i++) {
            newGroupSet.add(groups[i]);
        }//end loop
        /* If the input set was not empty, add the new groups to the 
         * registration's managed set of groups. And if the discovery
         * manager is not currently configured to discover ALL groups, add
         * the new groups to the set of groups managed by that manager.
         */
        if( newGroupSet.size() > 0 ) {
            (regInfo.groups).addAll(newGroupSet);
            if(discoveryMgr.getGroups()!=DiscoveryGroupManagement.ALL_GROUPS){
                String[] newGroups = (String[])newGroupSet.toArray
                                              (new String[newGroupSet.size()]);
                discoveryMgr.addGroups( newGroups );
            }
        }
    }//end addGroupsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>GroupsAddedToRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method adds new group names to the
     * current managed set of groups for the registration corresponding to
     * the <code>registrationID</code> parameter. Note that only elements
     * that do not duplicate elements already in the set are added.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         managed set of groups is to be augmented
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param groups           a <code>String</code> array, none of whose 
     *                         elements may be null, consisting of the group
     *                         names with which to augment the registration's
     *                         managed set of groups.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>addGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#addGroupsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#addGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#addGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#addGroups
     */
    private void addGroupsDo(UUID registrationID,
                             HashMap registrationByID,
                             String[] groups)   throws IOException
    {
        addGroupsDo((RegistrationInfo)(registrationByID.get(registrationID)),
                     groups);
    }//end addGroupsDo

    /**
     * Called by the public method <code>setGroups</code>. This method
     * replaces the current managed set of groups for the registration
     * corresponding to the <code>regInfo</code> data structure
     * parameter.
     * <p>
     * In addition to replacing the registration's current managed set of
     * groups, this method constructs the set of group names with which to
     * replace the set of all groups to discover (across all registrations).
     * That set is made up of both the set of group names that are placed 
     * in the registration's current set (if they were placed in the
     * registration's set of groups, they must also be new to the set 
     * of all groups), as well as the set of all group names associated
     * with the remaining registrations actively registered with this service.
     * After constructing this set of group names, the set of all groups
     * managed by the discovery manager is replaced with this newly
     * constructed set.
     * <p>
     * Note that when the set of all groups is replaced with this constructed
     * set, the effect on that set is the following: the elements from 
     * the input set that are not already in the set of all groups are added
     * to the set of all groups, and any elements that are removed from
     * the registration's current managed set of groups that are not also
     * in any of the managed sets associated with the other registrations
     * actively registered with this service are also removed from the set of
     * all groups. But note also that if the set of all groups is currently
     * set to "discover all groups" (null), then it will not change; unless
     * the current input registration is the only registration currently set
     * to null, and that registration is requesting that its managed set of
     * groups be replaced with a finite set of group names.
     * 
     * @param regInfo the data structure record corresponding to the
     *                registration whose managed set of groups is to be
     *                replaced
     * @param groups  a String array, none of whose elements may be null, 
     *                consisting of the group names with which to replace the
     *                registration's managed set of groups.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>setGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#setGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#setGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#setGroups
     */
    private void setGroupsDo(RegistrationInfo regInfo,
                             String[] groups)   throws IOException
    {
        /* Replace the registration's managed set of groups with the new groups
         * (even if the new set of groups is empty -- this just means group
         * discovery will be "turned off" for this registration. After 
         * replacing the registration's managed set, modify the discovery 
         * manager's managed set of groups in the appropriate way.
         */
        if(groups == DiscoveryGroupManagement.ALL_GROUPS) {
            regInfo.groups = null;
        } else {
            /* Build a HashSet (removes duplicates) from the input set */
            HashSet newGroups = new HashSet();
            for(int i=0;i<groups.length;i++) {
                newGroups.add(groups[i]);
            }//end loop
            /* Prepare the registration's managed set for replacement */
            if(regInfo.groups == null) {
                regInfo.groups = new HashSet();
            } else {
                (regInfo.groups).clear();
            }
            /* Replace the registration's managed set with the new set */
            (regInfo.groups).addAll(newGroups);
        }//end if (groups == DiscoveryGroupManagement.ALL_GROUPS)

        /* Build the set containing the groups from all registrations. */
        String[] groupsAcrossAllRegs = getGroupsFromAllRegs();
        /* 1. If at least one registration desires that ALL groups be
         *    discovered, and if the discovery manager is not currently 
         *    configured to discover ALL groups, replace the discovery 
         *    manager's managed set of groups with ALL groups.
         *
         * 2. If no registrations want ALL groups discovered, replace the 
         *    discovery manager's managed set of groups with the new 
         *    set of groups constructed above; even if that manager is 
         *    currently discovering ALL groups.
         */
        if(groupsAcrossAllRegs == DiscoveryGroupManagement.ALL_GROUPS) {
            if(discoveryMgr.getGroups()!=DiscoveryGroupManagement.ALL_GROUPS){
                discoveryMgr.setGroups(DiscoveryGroupManagement.ALL_GROUPS);
            }
        } else { // no registration desires ALL groups be discovered
            discoveryMgr.setGroups( groupsAcrossAllRegs );
        }
    }//end setGroupsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>GroupsSetInRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method This method replaces the
     * current managed set of groups for the registration corresponding to
     * the <code>registrationID</code> parameter with the input set of group
     * names.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         managed set of groups is to be replaced
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param groups           a <code>String</code> array, none of whose 
     *                         elements may be null, consisting of the group
     *                         names with which to replace the registration's
     *                         managed set of groups.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>setGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#setGroupsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#setGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#setGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#setGroups
     */
    private void setGroupsDo(UUID registrationID,
                             HashMap registrationByID,
                             String[] groups)   throws IOException
    {
        setGroupsDo((RegistrationInfo)(registrationByID.get(registrationID)),
                     groups);
    }//end setGroupsDo

    /**
     * Called by the public method <code>removeGroups</code>. This method
     * removes group names from the current managed set of groups for the
     * registration corresponding to the <code>regInfo</code> data structure
     * parameter. Note that elements from the input set are removed from the
     * registration's managed set of groups only if they are currently
     * contained the registration's managed set of groups.
     * <p>
     * In addition to removing group names from the registration's current
     * managed set of groups, this method removes the input group names from
     * the set of all groups to discover (across all registrations); but only
     * if those group names belong to no other registration's managed set.
     * 
     * @param regInfo the data structure record corresponding to the
     *                registration from whose managed set of groups the input
     *                set of groups is to be removed
     * @param groups  a String array, none of whose elements may be null,
     *                consisting of the group names to remove from the 
     *                registration's managed set of groups
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>removeGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#removeGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeGroups
     */
    private void removeGroupsDo(RegistrationInfo regInfo,
                                String[] groups)   throws IOException
    {
        /* Build a HashSet (removes duplicates) from the input set of groups */
        HashSet removeSet = new HashSet();
        for(int i=0;i<groups.length;i++) {
            removeSet.add(groups[i]);
        }//end loop
        /* If the set to remove is not empty, remove its contents from the 
         * registration's managed set of groups. And if the discovery
         * manager is not currently configured to discover ALL groups, remove
         * each group name from that manager's set; but only if those groups
         * are not in the managed set of any of the other registrations
         * managed by this service.
         */
        if( removeSet.size() > 0 ) {
            /* Remove the groups from the registration's managed set */
            (regInfo.groups).removeAll(removeSet);
            /* Build the set containing the groups from all registrations. */
            String[] groupsAcrossAllRegs = getGroupsFromAllRegs();
            /* Update the discovery manager's managed set of groups, but
             * only if every registration is configured for a finite set 
             * of groups -- not ALL groups (this means that the discovery 
             * manager's managed set is also not set to ALL groups).
             *
             * Note that a replacement operation is applied to the discovery 
             * manager's set rather than a removal operation in order to 
             * prevent the removal of groups that may be in both the managed
             * set of the input registration and the managed set of at least
             * one of the other registrations.
             */
            if(groupsAcrossAllRegs != DiscoveryGroupManagement.ALL_GROUPS) {
                discoveryMgr.setGroups( groupsAcrossAllRegs );
            }
        }//endif(removeSet.size > 0)
    }//end removeGroupsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>GroupsRemovedFromRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method removes group names from the
     * current managed set of groups for the registration corresponding to
     * the <code>registrationID</code> parameter. Note that elements from the
     * input set are removed from the registration's managed set of groups
     * only if they are currently contained the registration's managed set 
     * of groups.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration from whose
     *                         managed set of groups the input set of groups
     *                         is to be removed
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param groups           a <code>String</code> array, none of whose 
     *                         elements may be null, consisting of the group
     *                         names to remove from the registration's managed
     *                         set of groups
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol fails to start.
     * 
     * @throws java.lang.IllegalStateException this exception occurs when
     *         the <code>removeGroups</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#removeGroupsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#removeGroups
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeGroups
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeGroups
     */
    private void removeGroupsDo(UUID registrationID,
                                HashMap registrationByID,
                                String[] groups)   throws IOException
    {
        removeGroupsDo
                 ((RegistrationInfo)(registrationByID.get(registrationID)),
                   groups);
    }//end removeGroupsDo


    /** Builds a set containing the groups from all registrations. 
     *
     *  Iterates through all the active registrations, retrieving the groups
     *  to discover for each registration (minus duplicates). If at least one
     *  registration has requested that all groups be discovered, stops
     *  iterating (because the set must be ALL_GROUPS) and returns null.
     */
    private String[] getGroupsFromAllRegs() {
        HashSet  groupSet = new HashSet();
        for(Iterator itr=registrationByID.values().iterator();itr.hasNext();) {
            RegistrationInfo rInfo = (RegistrationInfo)itr.next();
            if(rInfo.groups == null) {
                return DiscoveryGroupManagement.ALL_GROUPS;
            } else {
                for( Iterator jtr=(rInfo.groups).iterator(); jtr.hasNext();) {
                    String group = (String)jtr.next();
                    groupSet.add(group);
                }//end loop
            }//end if
        }//end loop
        if(groupSet.isEmpty()) {
            return DiscoveryGroupManagement.NO_GROUPS;
        } else {
            return (String[])groupSet.toArray(new String[groupSet.size()]);
        }
    }//end getGroupsFromAllRegs
    /* END Private Group Management Methods -------------------------------- */

    /* BEGIN Private Locator Management Methods ---------------------------- */
    /**
     * Called by the public method <code>addLocators</code>. This method adds
     * new locators to the current managed set of locators for the
     * registration corresponding to the <code>regInfo</code> parameter.
     * Note that only elements that do not duplicate elements already in the
     * set are added.
     * <p>
     * In addition to augmenting the current managed set of locators, this
     * method adds the input locators to the set of all locators to
     * discover (across all registrations); also with duplicates not added.
     * <p>
     * Note that if a locator input to this method is already in the
     * registration's set of locators, it must also be in the set of all
     * locators. But if that input locator is not in the registration's
     * current managed set of locators, it still may be a member of the set
     * of all locators (because it was added by another registration). Thus,
     * this method attempts to add the locator to both sets, but includes
     * the locator in the set to add to the discovery manager only if it
     * was successfully added to the set of all locators.
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration whose managed set of locators is to be
     *                 augmented
     * @param locators an array, none of whose elements may be null,
     *                 consisting of the <code>LookupLocator</code> objects
     *                 with which to augment the registration's managed set
     *                 of locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>addLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#addLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#addLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#addLocators
     */
    private void addLocatorsDo(RegistrationInfo regInfo,
                               LookupLocator[]  locators)
    {
        /* Build a HashSet (removes duplicates) from the input set of locs */
        HashSet newLocSet = new HashSet();
        for(int i=0;i<locators.length;i++) {
            newLocSet.add(locators[i]);
        }//end loop
        /* If the input set to add is not empty, add the new locators to the 
         * registration's managed set of locators; and add the new locators
         * to the set of locators managed by that manager.
         */
        if( newLocSet.size() > 0 ) {
            (regInfo.locators).addAll(newLocSet);
            LookupLocator[] newLocs = (LookupLocator[])newLocSet.toArray
                                         (new LookupLocator[newLocSet.size()]);
            discoveryMgr.addLocators( newLocs );
        }
    }//end addLocatorsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>LocsAddedToRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method adds new locators to the
     * current managed set of locators for the registration corresponding to
     * the <code>registrationID</code> parameter. Note that only elements
     * that do not duplicate elements already in the set are added.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         managed set of locators is to be augmented
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param locators         an array, none of whose elements may be null,
     *                         consisting of the <code>LookupLocator</code>
     *                         objects with which to augment the
     *                         registration's managed set of locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>addLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#addLocatorsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#addLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#addLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#addLocators
     */
    private void addLocatorsDo(UUID registrationID,
                               HashMap registrationByID,
                               LookupLocator[] locators)
    {
        addLocatorsDo((RegistrationInfo)(registrationByID.get(registrationID)),
                      locators);
    }//end addLocatorsDo

    /**
     * Called by the public method <code>setLocators</code>. This method
     * replaces the current managed set of locators for the registration
     * corresponding to the <code>regInfo</code> parameter.
     * <p>
     * In addition to replacing the registration's current managed set of
     * locators, this method constructs the set of locators with which to
     * replace the set of all locators to discover (across all registrations).
     * That set is made up of both the set of locators that are placed 
     * in the registration's current set (if they were placed in the
     * registration's set of locators, they must also be new to the set 
     * of all locators), as well as the set of all locators associated
     * with the remaining registrations actively registered with this service.
     * After constructing this set of locators, the set of all locators
     * managed by the discovery manager is replaced with this newly
     * constructed set.
     * <p>
     * Note that when the set of all locators is replaced with this constructed
     * set, the effect on that set is the following: the elements from 
     * the input set that are not already in the set of all locators are added
     * to the set of all locators, and any elements that are removed from
     * the registration's current managed set of locators that are not also
     * in any of the managed sets associated with the other registrations
     * actively registered with this service are also removed from the set of
     * all locators.
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration whose managed set of locators is to be
     *                 replaced
     * @param locators an array, none of whose elements may be null,
     *                 consisting of the <code>LookupLocator</code> objects
     *                 with which to replace the registration's managed set
     *                 of locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>setLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#setLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#setLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#setLocators
     */
    private void setLocatorsDo(RegistrationInfo regInfo,
                               LookupLocator[]  locators)
    {
        /* Replace the registration's managed set of locators with the new
         * locators (even if the new set of locators is empty -- this just
         * means locator discovery will be "turned off" for this registration.
         * After replacing the registration's managed set, modify the  
         * discovery manager's managed set of locators in the appropriate way.
         */

        /* Build a HashSet (removes duplicates) from the input set of locs */
        HashSet newLocSet = new HashSet();
        for(int i=0;i<locators.length;i++) {
            newLocSet.add(locators[i]);
        }//end loop
        /* Prepare the registration's managed set for replacement */
        if(regInfo.locators == null) {
            regInfo.locators = new HashSet();
        } else {
            (regInfo.locators).clear();
        }
        /* Replace the registration's managed set with the new set */
        (regInfo.locators).addAll(newLocSet);
        /* Build the set containing the locators from all registrations. */
        LookupLocator[] locsAcrossAllRegs = getLocatorsFromAllRegs();
        /* Replace the discovery manager's managed set with the new set
         * derived from all the registrations.
         */
        discoveryMgr.setLocators( locsAcrossAllRegs );
    }//end setLocatorsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>LocsSetInRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method This method replaces the current
     * managed set of locators for the registration corresponding to the
     *  <code>registrationID</code> parameter.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         managed set of locators is to be replaced
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param locators         an array, none of whose elements may be null,
     *                         consisting of the <code>LookupLocator</code>
     *                         objects with which to replace the
     *                         registration's managed set of locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>setLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#setLocatorsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#setLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#setLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#setLocators
     */
    private void setLocatorsDo(UUID registrationID,
                               HashMap registrationByID,
                               LookupLocator[] locators)
    {
        setLocatorsDo((RegistrationInfo)(registrationByID.get(registrationID)),
                      locators);
    }//end setLocatorsDo

    /**
     * Called by the public method <code>removeLocators</code>. This method
     * removes locators from the current managed set of locators for the
     * registration corresponding to the<code>regInfo</code> parameter. 
     * Note that elements from the input set are removed from the
     * registration's managed set of locators only if they are currently
     * contained the registration's managed set of locators.
     * <p>
     * In addition to removing locators from the registration's current
     * managed set of locators, this method removes the input locators from
     * the set of all locators to discover (across all registrations); but only
     * if those locators belong to no other registration's managed set.
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration from whose managed set of locators the
     *                 input set of locators is to be removed
     * @param locators an array, none of whose elements may be null,
     *                 consisting of the <code>LookupLocator</code> objects
     *                 to remove from the registration's managed set of 
     *                 locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>removeLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#removeLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeLocators
     */
    private void removeLocatorsDo(RegistrationInfo regInfo,
                                  LookupLocator[]  locators)
    {
        /* Build a HashSet (removes duplicates) from the input set of locs */
        HashSet removeSet = new HashSet();
        for(int i=0;i<locators.length;i++) {
            removeSet.add(locators[i]);
        }//end loop
        /* If the set to remove is not empty, remove its contents from the 
         * registration's managed set of locators. And if the discovery
         * manager is not currently configured to discover ALL locators, remove
         * each locator from that manager's set; but only if those locators
         * are not in the managed set of any of the other registrations
         * managed by this service.
         */
        if( removeSet.size() > 0 ) {
            /* Remove the locators from the registration's managed set */
            (regInfo.locators).removeAll(removeSet);
            /* Build the set containing the locators from all registrations. */
            LookupLocator[] locsAcrossAllRegs = getLocatorsFromAllRegs();
            /* Update the discovery manager's managed set of locators.
             *
             * Note that a replacement operation is applied to the discovery 
             * manager's set rather than a removal operation in order to 
             * prevent the removal of locators that may be in both the managed
             * set of the input registration and the managed set of at least
             * one of the other registrations.
             */
            discoveryMgr.setLocators( locsAcrossAllRegs );
        }//endif(removeSet.size > 0)
    }//end removeLocatorsDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>LocsRemovedFromRegistrationLogObj</code> (which is invoked 
     * during state recovery). This method removes locators from the
     * current managed set of locators for the registration corresponding to
     * the <code>registrationID</code> parameter. Note that elements from
     * the input set are removed from the registration's managed set of
     * locators only if they are currently contained the registration's
     * managed set of locators.
     * <p>
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration from whose
     *                         managed set of groups the input set of locators
     *                         is to be removed
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param locators         an array, none of whose elements may be null,
     *                         consisting of the <code>LookupLocator</code>
     *                         objects to remove from the registration's
     *                         managed set of locators.
     * 
     * @throws java.lang.IllegalStateException this exception may occur when
     *         the <code>removeLocators</code> method of the discovery
     *         manager is invoked after the <code>terminate</code> method 
     *         of that manager is called.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#removeLocatorsDo
     * @see com.sun.jini.fiddler.FiddlerImpl#removeLocators
     * @see com.sun.jini.fiddler.FiddlerRegistration#removeLocators
     * @see net.jini.discovery.LookupDiscoveryRegistration#removeLocators
     */
    private void removeLocatorsDo(UUID registrationID,
                                  HashMap registrationByID,
                                  LookupLocator[] locators)
    {
        removeLocatorsDo
                   ((RegistrationInfo)(registrationByID.get(registrationID)),
                    locators);
    }//end removeLocatorsDo

    /** Builds a set containing the locators from all registrations. 
     *
     *  Iterates through all the active registrations, retrieving the locators
     *  to discover for each registration (minus duplicates).
     */
    private LookupLocator[] getLocatorsFromAllRegs() {
        HashSet locatorSet = new HashSet();
        for(Iterator itr=registrationByID.values().iterator();itr.hasNext();) {
            RegistrationInfo rInfo = (RegistrationInfo)itr.next();
            if(rInfo.locators == null) {
                return null;
            } else {
                for(Iterator jtr=(rInfo.locators).iterator(); jtr.hasNext();) {
                    LookupLocator locator = (LookupLocator)jtr.next();
                    locatorSet.add(locator);
                }//end loop
            }//end if
        }//end loop
        if(locatorSet.isEmpty()) {
           return new LookupLocator[0];
        } else {
            return (LookupLocator[])locatorSet.toArray
                                        (new LookupLocator[locatorSet.size()]);
        }
    }//end getLocatorsFromAllRegs
    /* END Private Locator Management Methods ------------------------------ */

    /* BEGIN Private Lease Renewal Methods --------------------------------- */
    /**
     * Called by the public method <code>renewLease</code>. This method renews
     * the lease corresponding to the <code>registrationID</code> and 
     * <code>leaseID</code> parameters, granting a new duration that is
     * less than or equal to the requested duration value contained in
     * the <code>duration</code> parameter.
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration whose lease is to be renewed
     * @param leaseID  identifier assigned by the lease grantor to the lease
     *                 that is to be renewed
     * @param duration the requested duration for the lease being renewed
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed is unknown to the lease grantor.
     *
     * @return <code>long</code> value representing the actual duration that
     *         was granted for the renewed lease. Note that the actual
     *         duration granted and returned by this method may be less than
     *         the duration requested.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLease
     */
    private long renewLeaseDo(RegistrationInfo regInfo,
                              long leaseID,
                              long duration) throws UnknownLeaseException
    {
        long curTime       = System.currentTimeMillis();
        long newExpiration = renewLeaseInt(regInfo,leaseID,duration,curTime);
        addLogRecord(new LeaseRenewedLogObj
                           (regInfo.registrationID, leaseID, newExpiration));
        return (newExpiration - curTime);
    }//end renewLeaseDo

    /**
     * Called by the method <code>renewLeaseDo</code>. This method takes
     * a requested duration for the lease corresponding to the
     * <code>registrationID</code> and <code>leaseID</code> parameters,
     * converts that duration to an absolute expiration (based on the local
     * clock) with the appropriate bound applied, and attempts to renew 
     * the lease, granting the new expiration time on the lease.
     * 
     * @param regInfo  the data structure record corresponding to the
     *                 registration whose lease is to be renewed
     * @param leaseID  identifier assigned by the lease grantor to the lease
     *                 that is to be renewed
     * @param duration the requested duration for the lease being renewed
     *
     * @return <code>long</code> value representing the new absolute
     *         expiration time granted for the renewed lease.
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed is unknown to the lease grantor.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeaseDo
     */
    private long renewLeaseInt(RegistrationInfo regInfo,
                               long leaseID,
                               long duration,
                               long curTime)  throws UnknownLeaseException
    {
        duration = applyBoundToLeaseDuration(duration, leaseDurationBound);
        long newExpiration = curTime + duration;
        renewLeaseAbs(regInfo, leaseID, newExpiration, curTime);
        /* see if the expire thread needs to wake up earlier */
        if (newExpiration < minExpiration) {
            minExpiration = newExpiration;
            concurrentObj.waiterNotify(leaseExpireThreadSyncObj);
        }
        return newExpiration;
    }//end renewLeaseInt

    /**
     * Called by the method <code>renewLeaseInt</code>. This method performs
     * the final steps in the process of renewing the lease on the registration
     * corresponding to the <code>regInfo</code> and <code>leaseID</code>
     * parameters, granting a requested absolute expiration time for that
     * lease.
     * 
     * @param regInfo    the data structure record corresponding to the
     *                   registration whose lease is to be renewed
     * @param leaseID    identifier assigned by the lease grantor to the lease
     *                   that is to be renewed
     * @param expiration the requested absolute expiration time for the lease
     *                   being renewed
     * @param curTime    the current time based on the local clock
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed is unknown to the lease grantor.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeaseInt
     */
    private void renewLeaseAbs(RegistrationInfo regInfo,
                               long leaseID,
                               long expiration,
                               long curTime)  throws UnknownLeaseException
    {
        if (    (regInfo == null) 
             || (regInfo.leaseID != leaseID)
             || (regInfo.leaseExpiration <= curTime) ) {
            throw new UnknownLeaseException();
        }
        /* The act of renewing a registration's lease simply involves 
         * changing the expiration time stored in the registration's
         * data structure. But because the registrationByTime sorts its
         * elements by time, changing the value of the leaseExpiration field
         * of the regInfo parameter "in place" would typically result in 
         * the registrationByTime map being out of order. Contrast this with
         * the modification of the registration's non-time-dependent fields
         * (for example, adding, replacing or removing groups or locators),
         * where modifying the input regInfo parameter will modify the
         * appropriate field of the corresponding element of both the
         * registrationByID and registrationByTime maps; and the order 
         * will still be valid in the registrationByTime map because the
         * time-dependent data has not changed.
         * 
         * Thus, in order to maintain a valid ordering in the 
         * registrationByTime map, a re-sort is forced by first removing
         * from that map the element corresponding to regInfo, resetting
         * the leaseExpiration field of that object, and then re-inserting
         * the modified regInfo into the map, which forces another sort.
         */
        registrationByTime.remove(regInfo);       // force first re-sort
        regInfo.leaseExpiration = expiration;     // modify outside of map
        registrationByTime.put(regInfo, regInfo); // insert & force new sort
    }//end renewLeaseAbs

    /**
     * Called by the <code>apply</code> method of the class
     * <code>LeaseRenewedLogObj</code> (which is invoked during state
     * recovery). This method performs the final steps in the process
     * of renewing the lease on the registration corresponding to the
     * <code>registrationID</code> and <code>leaseID</code> parameters,
     * granting a requested absolute
     * expiration time for that lease.
     * 
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         lease is to be renewed
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param leaseID          identifier assigned by the lease grantor to
     *                         the lease that is to be renewed
     * @param expiration       the requested absolute expiration time for the
     *                         lease being renewed
     * @param curTime          the current time based on the local clock
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being renewed is unknown to the lease grantor.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeaseDo
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeaseInt
     */
    private void renewLeaseAbs(UUID registrationID,
                               HashMap registrationByID,
                               long leaseID,
                               long expiration,
                               long curTime)  throws UnknownLeaseException
    {
        renewLeaseAbs((RegistrationInfo)(registrationByID.get(registrationID)),
                      leaseID,expiration,curTime);
    }//end renewLeaseAbs

    /**
     * Called by the public method renewLeases to renew all of
     * the leases from a <code>LeaseMap</code> that correspond to the
     * elements of the <code>registrationIDs</code> and <code>leaseIDs</code>
     * parameters, granting as new durations the corresponding elements
     * of the <code>duration</code> parameter.
     * <p>
     * Note that all of the input parameters must be the same length.
     *
     * @param registrationIDs array containing the unique identifiers assigned
     *                        to the each registration to which each lease 
     *                        to be renewed corresponds
     * @param leaseIDs        array containing the identifiers assigned by the
     *                        lease grantor to each lease being renewed
     * @param durations       array containing the requested durations for 
     *                        each lease being renewed
     * 
     * @return an instance of FiddlerRenewResults containing data corresponding
     *         to the results (granted durations or exceptions) of each
     *         renewal attempt
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeases
     */
    private FiddlerRenewResults renewLeasesDo(UUID[] registrationIDs,
				              long[] leaseIDs,
				              long[] durations)
    {
        long curTime = System.currentTimeMillis();
        Exception[] exceptions = null;
        for (int i = 0; i < registrationIDs.length; i++) {
            RegistrationInfo regInfo
                = (RegistrationInfo)(registrationByID.get(registrationIDs[i]));
            try {
                durations[i] = renewLeaseInt(regInfo, leaseIDs[i],
                                             durations[i], curTime);
            } catch (Exception e) {
                durations[i] = -1;
                if (exceptions == null) {
                    exceptions = new Exception[]{e};
                } else {
                    exceptions = (Exception[])appendArray(exceptions, e);
                }
            }
	}
        /* don't bother to weed out problem leases */
        addLogRecord
             ( new LeasesRenewedLogObj(registrationIDs,leaseIDs,durations) );
        for (int i = registrationIDs.length; --i >= 0; ) {
            if (durations[i] >= 0) {
                durations[i] -= curTime;
            }
        }
        return new FiddlerRenewResults(durations, exceptions);
    }//end renewLeasesDo

    /**
     * Using the absolute expiration times contained in the
     * <code>expirations</code> parameter, renews all of the leases 
     * from a <code>LeaseMap</code> that correspond to the elements 
     * of the <code>registrationIDs</code> and <code>leaseIDs</code>
     * parameters; skipping any negative expiration times.
     * <p>
     * Note that all of the input parameters must be the same length.
     *
     * @param registrationIDs array containing the unique identifiers assigned
     *                        to the each registration to which each lease 
     *                        to be renewed corresponds
     * @param leaseIDs        array containing the identifiers assigned by the
     *                        lease grantor to each lease being renewed
     * @param expirations     array containing the absolute expiration times
     *                        to which to renew each lease
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeases
     * @see com.sun.jini.fiddler.FiddlerImpl#renewLeaseAbs
     */
    private void renewLeasesAbs(UUID[] registrationIDs,
                                long[] leaseIDs,
                                long[] expirations)
    {
        for (int i = registrationIDs.length; --i >= 0; ) {
            long expiration = expirations[i];
            if (expiration < 0) {
                continue;
            }
            try {
                RegistrationInfo regInfo
                = (RegistrationInfo)(registrationByID.get(registrationIDs[i]));
                /* Set curTime to 0 to guarantee no UnknownLeaseException */
                renewLeaseAbs(regInfo, leaseIDs[i], expiration, 0);
            } catch (UnknownLeaseException e) {
                /* this exception should never occur (when recovering) */
            }
        }
    }//end renewLeasesDo
    /* END Private Lease Renewal Methods ----------------------------------- */

    /* BEGIN Private Lease Cancellation Methods ---------------------------- */
    /**
     * Called by the public method <code>cancelLease</code>. This method
     * cancels the lease on the registration corresponding to the 
     * <code>regInfo</code> and <code>leaseID</code> parameters.
     * 
     * @param regInfo the data structure record corresponding to the
     *                registration whose lease is to be cancelled
     * @param leaseID identifier assigned by the lease grantor to the lease
     *                that is to be cancelled
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being cancelled is unknown to the lease grantor.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol experiences a failure.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#cancelLease
     */
    private void cancelLeaseDo(RegistrationInfo regInfo, long leaseID)
                                     throws UnknownLeaseException, IOException
    {
        long curTime = System.currentTimeMillis();
        if ( (regInfo == null) || (regInfo.leaseExpiration <= curTime) ) {
            throw new UnknownLeaseException();
        }
        removeRegistration(regInfo);
        /* wake up thread if this might be the (only) earliest time */
        if (regInfo.leaseExpiration == minExpiration) {
            concurrentObj.waiterNotify(leaseExpireThreadSyncObj);
        }
    }//end cancelLeaseDo

    /**
     * Called by the <code>apply</code> method of the class
     * <code>LeaseCancelledLogObj</code> (which is invoked during state
     * recovery). This method cancels the lease on the registration
     * corresponding to the <code>registrationID</code> and
     * <code>leaseID</code> parameters.
     * 
     * @param regInfo the data structure record corresponding to the
     *                registration whose lease is to be cancelled
     * @param registrationID   the ID of the data structure record
     *                         corresponding to the registration whose
     *                         lease is to be cancelled
     * @param registrationByID the map containing all active registrations
     *                         managed by this service
     * @param leaseID          identifier assigned by the lease grantor to
     *                         the lease that is to be cancelled
     *
     * @throws net.jini.core.lease.UnknownLeaseException this exception occurs
     *         when the lease being cancelled is unknown to the lease grantor.
     * 
     * @throws java.io.IOException this exception occurs when the multicast
     *         request protocol experiences a failure.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#cancelLeaseDo
     * @see com.sun.jini.fiddler.FiddlerImpl#cancelLease
     */
    private void cancelLeaseDo(UUID registrationID,
                               HashMap registrationByID,
                               long leaseID)
                                     throws UnknownLeaseException, IOException
    {
        cancelLeaseDo((RegistrationInfo)(registrationByID.get(registrationID)),
                      leaseID);
    }//end cancelLeaseDo

    /**
     * Called by the public method cancelLeases to the cancel all of
     * the leases from a <code>LeaseMap</code> that correspond to the
     * elements of the <code>registrationIDs</code> and <code>leaseIDs</code>
     * parameters.
     * <p>
     * The two input arrays must be the same length. If there are no
     * exceptions, the return value is null. Otherwise, the return value
     * has the same length as the <code>registrationIDs</code> parameter,
     * and has nulls for the leases that were successfully renewed.
     * 
     * @return an array containing the instances of Exception (or null for
     *         successfully renewed leases), each corresponding to the 
     *         exception that occurred upon attempting renewal.
     *
     * @see com.sun.jini.fiddler.FiddlerImpl#cancelLeases
     */
    private Exception[] cancelLeasesDo(UUID[] registrationIDs,
                                       long[] leaseIDs)
    {
        Exception[] exceptions = null;
        for (int i = registrationIDs.length; --i >= 0; ) {
            try {
                RegistrationInfo regInfo =
                     (RegistrationInfo)(registrationByID.get
                                                         (registrationIDs[i]));
                cancelLeaseDo(regInfo, leaseIDs[i]);
            } catch (Exception e) {
                if (exceptions == null) {
                    exceptions = new Exception[registrationIDs.length];
                }
                exceptions[i] = e;
            }
        }
        return exceptions;
    }//end cancelLeasesDo
    /* END Private Lease Cancellation Methods ------------------------------ */

    /* BEGIN Private Event-Related Methods --------------------------------- */
    /** Returns the set of registrars common to both input parameters */
    private ServiceRegistrar[] intersectRegSets(ServiceRegistrar[] regs0,
                                                ServiceRegistrar[] regs1)
    {
        HashSet regSet = new HashSet(); //no duplicates
        /* Compare each input element of regs0 with each element of regs1,
         * storing matches along the way
         */
            nextRegistrar:
                for(int i=0;i<regs0.length;i++) {
                    for(int j=0;j<regs1.length;j++) {
                        if( (regs0[i]).equals(regs1[j]) ) {
                            regSet.add(regs0[i]);
                            continue nextRegistrar; // match found, next reg
                        }//endif
                    }//end loop(j)
                }//end loop(i)
            //end label: nextRegistrar
            return ( (ServiceRegistrar[])(regSet).toArray
                                      (new ServiceRegistrar[regSet.size()]) );
    }//end intersectRegSets

    /** Returns true if the input registrar is an element of the given array */
    private boolean regIsElementOfRegSet(ServiceRegistrar   reg,
                                         ServiceRegistrar[] regSet)
    {
        for(int i=0;i<regSet.length;i++) {
            if( (regSet[i]).equals(reg) ) {
                return true;
            }
        }
        return false;
    }//end regIsElementOfRegSet

    /**
     * This method determines which of the registrars in the <code>regs</code>
     * parameter belong to the set of desired registrars corresponding to the
     * <code>regInfo</code> parameter, it then adds those registrars to
     * the registration's map of discovered registrars (including only
     * those registrars that can be successfully serialized), and then builds
     * and sends an instance of <code>RemoteDiscoveryEvent</code> - indicating
     * that the included registrars have been recently discovered - to the
     * registration's listener.
     * 
     * @param regInfo the data structure record corresponding to the 
     *                registration whose listener will receive the event
     * 
     * @param regs    the set of registrars to consider sending in the event
     */
    private void sendDiscoveredEvent(RegistrationInfo regInfo,
                                     ServiceRegistrar[] regs)
    {
        /* Determine the registrars that are common to both the input set of
         * registrars (the currently or newly discovered registrars) and the
         * registration's set of desired registrars. These are the registrars
         * to send in the event.
         */
        ServiceRegistrar[] regsToAdd = regInfo.intersectDesiredRegs(regs);
        /* Add the registrars that are both desired and discovered to the
         * registration's map of discovered registrars. If any registrars
         * cannot be serialized, drop them.
         */
        ServiceRegistrar[] regsAdded = regInfo.addToDiscoveredRegs(regsToAdd);
        /* Build and send a "discovered event" */
        RemoteDiscoveryEvent event = buildEvent(regInfo,regsAdded,false);
        if(event != null) {
            sendEvent(regInfo,event);
            writeInfoEvents("NewReg/Discovered EventTask.run(): "
                            +"DISCOVERED Event was SENT\n");
        }
   }//end sendDiscoveredEvent

    /** Examines the <code>discardFlag</code> for each active registration
     *  until a value of <code>true</code> is encountered or the set of
     *  registrations is exhausted. If that flag is <code>false</code> for
     *  all registrations, this method will return <code>null</code>;
     *  otherwise, it will return the registration found, but with the
     *  the value of the flag set back to <code>false</code>.
     */
    private RegistrationInfo externalDiscardRequest()
    {
        /* Loop thru regInfo's, until one is found with a true flag*/
        for(Iterator itr=registrationByID.values().iterator();itr.hasNext();) {
            RegistrationInfo regInfo = (RegistrationInfo)itr.next();
            if(regInfo.discardFlag == true) {
                writeInfoDiscard("\nexternalDiscardRequest: "
                                 +"discardFlag == true\n");
                regInfo.discardFlag = false;
                return regInfo;
            }
        }//end loop
        return null;
    }//end externalDiscardRequest

    /**
     * This method constructs the appropriate remote discovery event from the
     * information contained in the input parameters. This method encapsulates
     * common exception-handling functionality. If the remote discovery
     * event cannot be successfully constructed, null will be returned.
     * 
     * @param regInfo   the data structure record corresponding to the 
     *                  registration whose listener will receive the event
     * @param regs      the set of registrars to include in the event
     * @param discarded flag indicating whether the registrars included in
     *                  the event have been discarded or discovered
     *
     * @return instance of <code>RemoteDiscoveryEvent</code> containing the
     *         appropriate information corresponding to the input registration
     *         record
     */
    private RemoteDiscoveryEvent buildEvent(RegistrationInfo regInfo,
                                            ServiceRegistrar[] regs,
                                            boolean discarded)
    {
        RemoteDiscoveryEvent newEvent = null;
        if( (regInfo != null) && (regs != null) && (regs.length > 0) ) {
            try {
                newEvent = new RemoteDiscoveryEvent(fiddlerProxy,
                                                    regInfo.eventID,
                                                    ++regInfo.seqNum,
                                                    regInfo.handback,
                                                    discarded,
                                                    regs);
                writeInfoEvents(regs,regInfo.eventID,regInfo.seqNum,
                                regInfo.handback,discarded);
            } catch (IOException e) {
                /* The constructor for <code>RemoteDiscoveryEvent</code>
                 * throws an <code>IOException</code> if the constructor
                 * fails to successfully serialize even one registrar from
                 * the input set (which means there is no need to send the
                 * event since there are no marshalled registrars to send).
                 * When such a situation occurs, register an ERROR status
                 * attribute (along with a Comment attribute describing the
                 * nature of the problem) to all lookup services with which
                 * this service is registered. 
                 *
                 * Administrative clients, as well as clients that use this 
                 * service should have registered for notification of the 
                 * existence of this attribute.
                 */
                Entry[] errorAttrs =
                new Entry[] { new FiddlerStatus(StatusType.WARNING),
                              new Comment("Failed to serialize ALL registrars "
                                         +"during event construction "
                                         +"... could not send event")
                            };
                joinMgr.addAttributes(errorAttrs,true);
	        e.printStackTrace();
            }//end try
        }//endif
        return newEvent;
    }//end buildEvent

    /**
     * This method sends a <code>RemoteDiscoveryEvent</code> to the listener
     * of the registration that corresponds to the <code>regInfo</code> 
     * parameter. This method handles all exceptions and error conditions
     * in the appropriate manner.
     * 
     * @param regInfo the data structure record corresponding to the 
     *                registration whose listener will receive the event
     * 
     * @param regs    the instance of <code>RemoteDiscoveryEvent</code> to
     *                send registration's listener
     */
    private void sendEvent(RegistrationInfo regInfo,
                           RemoteDiscoveryEvent event)
    {
        try {
            regInfo.listener.notify(event);
        } catch (Throwable e) {
            writeInfoEvents("Exception in sendEvent: "+e.toString()+"\n");
            if ( !(   (      (e instanceof RemoteException)
                         && !(e instanceof NoSuchObjectException) )
                    || (e instanceof LinkageError)
                    || (e instanceof OutOfMemoryError) ) )
            {
                if (     (e instanceof Error)
                     && !(e instanceof StackOverflowError) )
                {
                    throw (Error)e;
                }
                /* If the listener throws UnknownEvent or some other
                 * non-remote exception, or the listener is gone, it's
                 * okay to cancel the lease. The lease is not cancelled
                 * on LinkageError, since it's conceivable the problem
                 * is transient.
                 */
                try {
                    writeInfoEvents("  Cancelling lease on registration: "
                                    +" (registrationID,leaseID) = ("
                                   +regInfo.registrationID+", "
                                   +regInfo.leaseID+")");
                    cancelLeaseDo(regInfo,regInfo.leaseID);
                    addLogRecord(new LeaseCancelledLogObj
                                                    (regInfo.registrationID,
                                                     regInfo.leaseID));
                } catch (UnknownLeaseException ee) {
                } catch (IOException ee) { }
            }//endif
        }//end try
    }//end sendEvent
    /* END Private Event-Related Methods ----------------------------------- */

    /* BEGIN Persistent State Logging Interfaces & Classes ----------------- */
    /**
     * Writes the current state of this service to persistent storage.
     * <p>
     * A 'snapshot' of the service's current state is represented by
     * the data contained in certain fields defined in the service. 
     * That data represents many changes -- over time -- to the service's
     * state. This method will record that data to a file referred to as
     * the snapshot file.
     * <p>
     * The data written by this method to the snapshot file -- as well as
     * the format of the file -- is shown below:
     * <ul>
     * <li> the service's class name
     * <li> log format version number
     * <li> the service's service ID assigned by the lookup service
     * <li> the unique ID of the proxy to the current instance of this service
     * <li> the current event ID
     * <li> the current join state of this service
     * <li> the current configuration parameters of this service
     * <li> contents of the container holding the current set of registrations
     * <li> null (termination 'marker' for the set of registrations)
     * </ul>
     * Each data item is written to the snapshot file in serialized form.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private void takeSnapshot(OutputStream  out) throws IOException {
        ObjectOutputStream stream = new ObjectOutputStream(out);
        /* Stable information about the current instance of this service */
        stream.writeUTF(FiddlerImpl.class.getName());
        stream.writeInt(LOG_VERSION);
        stream.writeObject(serviceID);
        stream.writeObject(proxyID);
        /* The current event ID assigned to the discovery events being sent */
        stream.writeLong(curEventID);
        /* The current join state of this service */
        stream.writeObject(thisServicesGroups);
        stream.writeObject(thisServicesLocators);
        stream.writeObject( marshalAttributes(this,thisServicesAttrs) );
        /* The current configuration parameters of this service */
        stream.writeLong(leaseDurationBound);
        stream.writeInt(logToSnapshotThresh);
        stream.writeFloat(snapshotWt);
        /* The current set of registrations */
        for(Iterator iter=registrationByID.values().iterator();iter.hasNext();)
        {
            stream.writeObject(iter.next());
        }
        stream.writeObject(null); //termination marker
        stream.flush();
    }//end takeSnapshot

    /**
     * Retrieve the contents of the snapshot file and reconstitute the 'base'
     * state of this service from the retrieved data.
     * <p>
     * Refer to <code>takeSnapshot</code> for a description of the data
     * retrieved by this method from the snapshot file.
     * <p>
     * During recovery, the state of the service at the time of a crash
     * or outage is re-constructed by first reconstituting the 'base state'
     * from the snapshot file; and then modifying that base state according
     * to the records retrieved from the log file. This method is invoked to
     * perform the first step in that reconstruction. As each registered
     * service or event is retrieved, it is inserted into its appropriate
     * container object.
     * <p>
     * Because events can be generated before the next snapshot is taken,
     * the event sequence numbers must be incremented. This is because the
     * event specification requires that a set of event sequence numbers
     * corresponding to an particular event type (a particular event ID)
     * be monotonically increasing. Since the number of events that might
     * have been sent is arbitrary, each sequence number will be incremented
     * by a 'large' number so as to guarantee adherence to the specification.
     * 
     * @see FiddlerImpl.LocalLogHandler#takeSnapshot
     * @see FiddlerImpl.LocalLogHandler
     */
    private void recoverSnapshot(InputStream in)
                                 throws IOException, ClassNotFoundException
    {
        ObjectInputStream stream = new ObjectInputStream(in);
        /* Retrieve the service's stable information */
        if (!FiddlerImpl.class.getName().equals(stream.readUTF())) {
            throw new IOException("log from wrong implementation");
        }
        if (stream.readInt() != LOG_VERSION) {
            throw new IOException("wrong log format version");
        }
        serviceID  = (ServiceID)stream.readObject();
        proxyID    = (UUID)stream.readObject();
        /* Retrieve the current event ID */
        curEventID = stream.readLong();
        /* Retrieve the current join state of this service */
        thisServicesGroups   = (String[])stream.readObject();
        thisServicesLocators = (LookupLocator[])stream.readObject();
        MarshalledObject[] marshalledAttrs
                                    = (MarshalledObject[])stream.readObject();
        thisServicesAttrs = unmarshalAttributes(this, marshalledAttrs);
        /* Retrieve the current configuration parameters of this service */
        leaseDurationBound  = stream.readLong();
        logToSnapshotThresh = stream.readInt();
        snapshotWt          = stream.readFloat();
        RegistrationInfo regInfo;
        while ((regInfo = (RegistrationInfo)stream.readObject()) != null) {
            regInfo.seqNum += Integer.MAX_VALUE;
            addRegistration(regInfo);
        }
    }//end recoverSnapshot

    /**
     * Add a state-change record to persistent storage.
     * <p>
     * Whenever a significant change occurs to Fiddler's state, this
     * method is invoked to record that change in a file called a log file.
     * Each record written to the log file is an object reflecting both
     * the data used and the ACTIONS taken to make one change to that
     * state at a particular point in time. If the number of records
     * contained in the log file exceeds the pre-defined threshold,
     * a snapshot of Fiddler's current state will be recorded.
     * <p>
     * Whenever one of the following state changes occurs, this method
     * will be invoked with the appropriate implementation of the
     * LogRecord interface as the input argument.
     * <ul>
     * <li> new attributes were added to Fiddler's attributes with lookup
     * <li> Fiddler's existing attributes with lookup were modified
     *
     * <li> new groups were added to the set of groups Fiddler should join
     * <li> Fiddler's existing groups to join were replaced with a new set
     * <li> Fiddler's existing groups to join were removed
     *
     * <li> new locators were added to the set of locators Fiddler should join
     * <li> Fiddler's existing locators to join were replaced with a new set
     * <li> Fiddler's existing locators to join were removed
     * <li> Fiddler's debug properties were changed
     *
     * <li> the bound on granted lease durations was set to a new value
     * <li> weight factor used to determine when to take a snapshot was changed
     * <li> threshold used to determine when to take a snapshot was changed
     *
     * <li> a registration request for use of Fiddler's resources was granted
     *
     * <li> groups were added to a registration's set of groups to discover
     * <li> the set of groups to discover for a registration was replaced
     * <li> the set of groups to discover for a registration was removed
     *
     * <li> locators were added to a registration's set of locators to discover
     * <li> the set of locators to discover for a registration was replaced
     * <li> the set of locators to discover for a registration was removed
     *
     * <li> a lease on a registration with Fiddler was renewed
     * <li> a set of registration leases with Fiddler were renewed
     * <li> a lease on a registration with Fiddler was cancelled
     * <li> a set of registration leases with Fiddler were cancelled
     *
     * </ul>
     * 
     * @see RegistrarImpl.LocalLogHandler
     */
    private void addLogRecord(LogRecord rec) {
        writeInfoAddLogRecord(rec);
        try {
            log.update(rec, true);
            if (++logFileSize >= logToSnapshotThresh) {
                int snapshotSize = registrationByID.size();
                if ((float)logFileSize >= snapshotWt*((float)snapshotSize)) {
                    /* take snapshot */
                    concurrentObj.waiterNotify(snapshotThreadSyncObj);
                }
            }
        } catch (Exception e) {
            /* If log updating fails, then register an ERROR status attribute
             * (along with a Comment attribute describing the nature of the
             * problem) to all lookup services with which this service is
             * registered. 
             *
             * Administrative clients, as well as clients that use this
             * service should have registered for notification of the
             * existence of this attribute.
             */
            Entry[] errorAttrs = new Entry[]
                 { new FiddlerStatus(StatusType.ERROR),
                   new Comment( "Failure while updating the persistent "
                               +"log containing the service state")
                 };
            joinMgr.addAttributes(errorAttrs,true);
            e.printStackTrace();
        }
    }//end addLogRecord

    /**
     * Interface defining the method(s) that must be implemented by each of
     * the concrete LogObj classes. This allows for the definition of
     * object-dependent invocations of the appropriate implementation of
     * the method(s) declared in this interface.
     */
    private static interface LogRecord extends Serializable {
        void apply(FiddlerImpl fiddlerImpl);
    }

    /* 1. Definitions related to state changes generated by JoinAdmin */

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * new attributes are added to this service's existing set of attributes.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LookupAttrsAddedLogObj implements LogRecord {
        private static final long serialVersionUID = 4983778026976938585L;
        /** The attributes that were added to each lookup service with which
         *  this service is registered, written out in marshalled form.
         *  @serial
         */
        private MarshalledObject[] marshalledAttrs;
        /** Constructs this class and stores the attributes that were added */
        public LookupAttrsAddedLogObj(FiddlerImpl fiddlerImpl, Entry[] attrs) {
            this.marshalledAttrs = marshalAttributes(fiddlerImpl,attrs);
        }
        /** Modifies this service's state by adding (after unmarshalling) the
         *  elements of marshalledAttrs to the service's existing set of
         *  attributes.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                         "Log recovery: apply adding lookup attributes ...");
            Entry[] attrs = unmarshalAttributes(fiddlerImpl, marshalledAttrs);
            fiddlerImpl.thisServicesAttrs 
                  = LookupAttributes.add(fiddlerImpl.thisServicesAttrs,attrs);
        }
    }//end LookupAttrsAddedLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the attributes currently associated with this service in the lookup
     * services with which it is registered are modified.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LookupAttrsModifiedLogObj implements LogRecord {
        private static final long serialVersionUID = 2671139518188470469L;
	/** The attribute set templates used to select the attributes (from
         *  the service's existing set of attributes) that were modified,
         *  written out in marshalled form.
	 *  @serial
	 */
        private MarshalledObject[] marshalledAttrTmpls;
        /** The attributes with which this service's existing attributes 
         *  were modified, written out in marshalled form.
         *  @serial
         */
        private MarshalledObject[] marshalledModAttrs;
        /** Constructs this class and stores the modified attributes */
        public LookupAttrsModifiedLogObj(FiddlerImpl fiddlerImpl,
                                         Entry[] attrTmpls,
                                         Entry[] modAttrs)
        {
            this.marshalledAttrTmpls = marshalAttributes
                                                       (fiddlerImpl,attrTmpls);
            this.marshalledModAttrs = marshalAttributes(fiddlerImpl,modAttrs);
        }
        /** Modifies this service's state by modifying (after unmarshalling)
         *  the service's existing attributes according to the contents of
         *  marshalledAttrTmpls and marshalledModAttrs.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                        "Log recovery: apply modifying lookup attributes ...");
            Entry[] attrTmpls = unmarshalAttributes
                                            (fiddlerImpl, marshalledAttrTmpls);
            Entry[] modAttrs  = unmarshalAttributes
                                            (fiddlerImpl, marshalledModAttrs);
            fiddlerImpl.thisServicesAttrs 
                       = LookupAttributes.modify(fiddlerImpl.thisServicesAttrs,
                                                 attrTmpls, modAttrs);
        }
    }//end LookupAttrsModifiedLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set containing the names of the groups whose members are lookup
     * services the lookup discovery service wishes to register with (join)
     * is modified in some way; for example, through the invocation of:
     * <code>JoinAdmin.addLookupGroups</code>,
     * <code>JoinAdmin.removeLookupGroups</code> or
     * <code>JoinAdmin.setLookupGroups</code>.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LookupGroupsChangedLogObj implements LogRecord {
        private static final long serialVersionUID = -6973676200404539919L;
        /** The new groups whose members are the lookup services this
         *  service should join.
         *  @serial
         */
        private String[] groups;
	/** Constructs this class and stores the new groups */
        public LookupGroupsChangedLogObj(String[] groups) {
            this.groups = groups;
        }
        /** Modifies this service's state by modifying this service's existing
         *  set of 'groups to join'.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                     "Log recovery: apply changing lookup groups to join ...");
            fiddlerImpl.thisServicesGroups = groups; 
        }
    }//end LookupGroupsChangedLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set containing the instances of <code>LookupLocator</code> that
     * correspond to specific lookup services the lookup discovery service
     * wishes to register with (join) is modified in some way; for example,
     * through the invocation of:
     * <code>JoinAdmin.addLookupLocators</code>,
     * <code>JoinAdmin.removeLookupLocators</code> or
     * <code>JoinAdmin.setLookupLocators</code>.
     *
     *  @see FiddlerImpl.LocalLogHandler
     */
    private static class LookupLocatorsChangedLogObj implements LogRecord {
        private static final long serialVersionUID = 6448427261140043291L;
        /** The locators that correspond to the new lookup services this
         *  service should join.
         *  @serial
         */
        private LookupLocator[] locators;
        /** Constructs this class and stores the new locators */
        public LookupLocatorsChangedLogObj(LookupLocator[] locators) {
            this.locators = locators;
        }
        /** Modifies this service's state by modifying this service's existing
         *  set of 'locators to join'.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                   "Log recovery: apply changing lookup locators to join ...");
            fiddlerImpl.thisServicesLocators = locators; 
        }
    }//end LookupLocatorsChangedLogObj

    /* 2. Definitions related to state changes generated by FiddlerAdmin */

    /**
     * LogObj class whose instances are recorded to the log file whenever a
     * new value is assigned to the least upper bound applied during the
     * determination of the duration of the lease on a requested registration.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LeaseDurationBoundSetLogObj implements LogRecord {
        private static final long serialVersionUID = 6084059678114714281L;
        /** The new least upper bound used to determine a lease's duration.
         *  @serial
         */
        private long newLeaseDurationBound;
        /** Constructs this class and stores the new lease duration bound */
        public LeaseDurationBoundSetLogObj(long newLeaseDurationBound) {
            this.newLeaseDurationBound = newLeaseDurationBound;
        }
        /** Modifies this service's state by setting the value of the private
         *  leaseDurationBound field to the value of the new bound.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                     "Log recovery: apply changing lease duration bound ...");
            fiddlerImpl.leaseDurationBound = newLeaseDurationBound; 
        }
    }//end LeaseDurationBoundSetLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever a
     * new value is assigned to the Weight Factor applied to the Snapshot File
     * size used in the "time-to-take-a-snapshot determination" expression.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class SnapshotWeightSetLogObj implements LogRecord {
        private static final long serialVersionUID = -4079318959709953285L;
        /** The new snapshot weight factor.
         * @serial
         */
        private float newWeight;
	/** Constructs this class and stores the new weight factor */
        public SnapshotWeightSetLogObj(float newWeight) {
            this.newWeight = newWeight;
        }
        /** Modifies this service's state by setting the value of the private
         *  snapshotWt field to the value of the new weight factor.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                    "Log recovery: apply changing snapshot weight factor ...");
            fiddlerImpl.snapshotWt = newWeight; 
        }
    }//end SnapshotWeightSetLogObj

    /**
     * LogObj class whose instances are recorded to the log file 
     * whenever a new value is assigned to the Threshold used in the
     * "time-to-take-a-snapshot determination" expression.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LogToSnapshotThresholdSetLogObj implements LogRecord {
        private static final long serialVersionUID = 2952948867001823559L;
        /** The new snapshot threshold.
         *  @serial
         */
        private int newThreshold;
        /** Constructs this class and stores the new snapshot threshold */
        public LogToSnapshotThresholdSetLogObj(int newThreshold) {
            this.newThreshold = newThreshold;
        }
        /** Modifies this service's state by setting the value of the private
         *  logToSnapshotThresh field to the value of the new threshold.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                 "Log recovery: apply changing log-to-snapshot threshold ...");
            fiddlerImpl.logToSnapshotThresh = newThreshold; 
        }
    }//end LogToSnapshotThresholdSetLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the <code>com.sun.jini.fiddler.server.debug</code> property is changed
     * through the administrative interface.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class DebugPropertySetLogObj implements LogRecord {
        private static final long serialVersionUID = -5061186405988644704L;
        /** Comma-separated set of valid debug property values.
         *  @serial
         */
        private String newValues;
        /** Constructs this class and stores the <code>String</code> 
         *  containing the new values of the debug property.
         */
        public DebugPropertySetLogObj(String newValues) {
            this.newValues = newValues;
        }
        /** Modifies this service's state by changing the current value of
         *  the debug property.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                         "Log recovery: apply changing debug property ...");
            fiddlerImpl.setDebugPropertyDo(newValues);
        }
    }//end DebugPropertySetLogObj

    /* 3. Definitions related to state changes made to the set of active
     *    registrations, generated by client requests made through the
     *    <code>LookupDiscoveryRegistration</code> interface.
     */

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is created and returned to a client.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class RegistrationGrantedLogObj implements LogRecord {
        private static final long serialVersionUID = 3983963008572188308L;
        /** Object which acts as a record of the current registration with the
         *  lookup discovery service; containing all of the information about
         *  that registration: IDs, managed sets, lease information, and
         *  event registration information.
         *  @serial
         */
        private RegistrationInfo regInfo;
        /** Constructs this class and stores the registration information */
        public RegistrationGrantedLogObj(RegistrationInfo regInfo) {
            this.regInfo = regInfo;
        }
        /** Modifies this service's state by registering the information
         *  stored in the regInfo parameter; and by updating both the event
         *  sequence number and the event ID for the registration.
         *
         *  Note that the granting of a registration to a client typically
         *  involves the modification of the managed sets in the discovery
         *  manager, which usually involves starting the discovery protocol.
         *  Since an IOException can occur when the discovery protocol fails
         *  to start, and since such a situation is un-recoverable, this
         *  method does the following: catches the exception, informs this
         *  service's administrator by displaying the stack trace, and exits.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                         "Log recovery: apply recording granted registration "
                         +"info\n              (ID = "
                         +regInfo.registrationID+") ...");
            regInfo.seqNum += Integer.MAX_VALUE;
            try {
                fiddlerImpl.addRegistration(regInfo);
            } catch(IOException e) {
                System.out.println
                             ("During log recovery (apply addRegistration) -- "
                              +"multicast request protocol failed to start\n"
                              +e.toString());
                fiddlerImpl.destroyDo();
            }
            fiddlerImpl.curEventID++;
        }
    }//end RegistrationGrantedLogObj

    /* 4. Definitions related to state changes made to the set of active
     *    registrations, generated by client requests made through the
     *    <code>LookupDiscoveryService</code> interface.
     */

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the managed set of groups corresponding to a registration is 
     * is augmented with new elements.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class GroupsAddedToRegistrationLogObj implements LogRecord {
        private static final long serialVersionUID = -2133211272536572504L;
        /** The ID of the data structure record corresponding to the
         *  registration whose managed set of groups was augmented.
         *  @serial
         */
        private UUID registrationID;
        /** The set of groups added to the registration's managed set of groups
         *  @serial
         */
        private String[] groups;
        /** Constructs this class and stores the ID and new set of groups */
        public GroupsAddedToRegistrationLogObj(UUID registrationID,
                                               String[] groups)
        {
            this.registrationID = registrationID;
            this.groups         = groups;
        }
        /** Modifies this service's state by adding the set of group names to
         *  registration's managed set of groups, as well as by updating the
         *  set of all groups (across all registrations) to discover.
         *
         *  Note that the augmentation of a registration's set of groups
         *  typically involves the modification of the managed sets in the
         *  discovery manager, which usually involves starting the discovery
         *  protocol. Since an IOException can occur when the discovery
         *  protocol fails to start, and since such a situation is
         *  un-recoverable, this method does the following: catches the
         *  exception, informs this service's administrator by displaying
         *  the stack trace, and exits.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                        "Log recovery: apply adding to groups to discover for "
                        +"registration\n              (ID = "
                        +registrationID+") ...");
            try {
                fiddlerImpl.addGroupsDo(registrationID,
                                        fiddlerImpl.registrationByID,
                                        groups);
            } catch(IOException e) {
                System.out.println
                               ("During log recovery (apply addGroupsDO) -- "
                                +"multicast request protocol failed to start\n"
                                   +e.toString());
                fiddlerImpl.destroyDo();
            }
        }
    }//end GroupsAddedToRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a the managed set of groups corresponding to a registration is 
     * is replaced (set) with a new set of group names.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class GroupsSetInRegistrationLogObj implements LogRecord {
        /** The ID of the data structure record corresponding to the
         *  registration whose managed set of groups was replaced.
         *  @serial
         */
        private UUID registrationID;
        /** The set of groups that replaced the registration's current
         *  managed set of groups.
         *  @serial
         */
        private String[] groups;
        /** Constructs this class and stores the ID and new set of groups */
        public GroupsSetInRegistrationLogObj(UUID registrationID,
                                             String[] groups)
        {
            this.registrationID = registrationID;
            this.groups         = groups;
        }
        /** Modifies this service's state by replacing the registration's
         *  current managed set of groups with the set of group names 
         *  stored in this class by the constructor, as well as by updating
         *  the set of all groups (across all registrations) to discover.
         *
         *  Note that the replacement of a registration's set of groups
         *  typically involves the modification of the managed sets in the
         *  discovery manager, which usually involves starting the discovery
         *  protocol. Since an IOException can occur when the discovery
         *  protocol fails to start, and since such a situation is
         *  un-recoverable, this method does the following: catches the
         *  exception, informs this service's administrator by displaying
         *  the stack trace, and exits.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                        "Log recovery: apply replacing groups to discover for "
                        +"registration\n              (ID = "
                        +registrationID+") ...");
            try {
                fiddlerImpl.setGroupsDo(registrationID,
                                        fiddlerImpl.registrationByID,
                                        groups);
            } catch(IOException e) {
                System.out.println("During log recovery (apply setGroups) -- "
                                +"multicast request protocol failed to start\n"
                                   +e.toString());
                fiddlerImpl.destroyDo();
            }
        }
    }//end GroupsSetInRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * one or more elements of the managed set of groups corresponding to a
     * registration are removed.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class GroupsRemovedFromRegistrationLogObj
                                                        implements LogRecord
    {
        private static final long serialVersionUID = 6379639264267662740L;
        /** The ID of the data structure record corresponding to the
         *  registration with managed set from which groups were removed.
         *  @serial
         */
        private UUID registrationID;
        /** The set of groups removed from the registration's managed set 
         *  of groups.
         *  @serial
         */
        private String[] groups;
        /** Constructs this class and stores the ID and groups to remove */
        public GroupsRemovedFromRegistrationLogObj(UUID registrationID,
                                                   String[] groups)
        {
            this.registrationID = registrationID;
            this.groups         = groups;
        }
        /** Modifies this service's state by removing the set of group names
         *  from registration's managed set of groups, as well as by updating
         *  the set of all groups (across all registrations) to discover.
         *
         *  Note that the removal of one or more group names from a
         *  registration's set of groups typically involves the modification 
         *  of the managed sets in the discovery manager, which usually 
         *  involves starting the discovery protocol. Since an IOException 
         *  can occur when the discovery protocol fails to start, and since 
         *  such a situation is un-recoverable, this method does the 
         *  following: catches the exception, informs this service's 
         *  administrator by displaying the stack trace, and exits.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                        "Log recovery: apply removing from groups to discover "
                        +"for registration\n              (ID = "
                        +registrationID+") ...");
            try {
                fiddlerImpl.removeGroupsDo(registrationID,
                                           fiddlerImpl.registrationByID,
                                           groups);
            } catch(IOException e) {
                System.out.println
                                ("During log recovery (apply removeGroups) -- "
                                +"multicast request protocol failed to start\n"
                                   +e.toString());
                fiddlerImpl.destroyDo();
            }
        }
    }//end GroupsRemovedFromRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the managed set of locators corresponding to a registration is 
     * is augmented with new elements.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LocsAddedToRegistrationLogObj implements LogRecord {
        private static final long serialVersionUID = 7111619051211742297L;
        /** The ID of the data structure record corresponding to the
         *  registration whose managed set of locators was augmented.
         *  @serial
         */
        private UUID registrationID;
        /** The set of locators added to the registration's managed set
         *  @serial
         */
        private LookupLocator[] locators;
        /** Constructs this class and stores the ID and new set of locators */
        public LocsAddedToRegistrationLogObj(UUID registrationID,
                                             LookupLocator[] locators)
        {
            this.registrationID = registrationID;
            this.locators       = locators;
        }
        /** Modifies this service's state by adding the set of locators to
         *  registration's managed set of locators, as well as by updating the
         *  set of all locators (across all registrations) to discover.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                      "Log recovery: apply adding to locators to discover for "
                      +"registration\n              (ID = "
                      +registrationID+") ...");
            fiddlerImpl.addLocatorsDo(registrationID,
                                      fiddlerImpl.registrationByID,
                                      locators);
        }
    }//end LocsAddedToRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the managed set of locators corresponding to a registration is 
     * is replaced (set) with a new set of locators.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LocsSetInRegistrationLogObj implements LogRecord {
        private static final long serialVersionUID = -3504497106960878465L;
        /** The ID of the data structure record corresponding to the
         *  registration whose managed set of locators was replaced.
         *  @serial
         */
        private UUID registrationID;
        /** The set of locators that replaced the registration's current
         *  managed set of locators.
         *  @serial
         */
        private LookupLocator[] locators;
        /** Constructs this class and stores the ID and new set of locators */
        public LocsSetInRegistrationLogObj(UUID registrationID,
                                           LookupLocator[] locators)
        {
            this.registrationID = registrationID;
            this.locators       = locators;
        }
        /** Modifies this service's state by replacing the registration's
         *  current managed set of locators with the set of locators 
         *  stored in this class by the constructor, as well as by updating
         *  the set of all locators (across all registrations) to discover.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                      "Log recovery: apply replacing locators to discover for "
                      +"registration\n              (ID = "
                      +registrationID+") ...");
            fiddlerImpl.setLocatorsDo(registrationID,
                                      fiddlerImpl.registrationByID,
                                      locators);
        }
    }//end LocsSetInRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * one or more elements of the managed set of locators corresponding to a
     * registration are removed.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LocsRemovedFromRegistrationLogObj
                                                         implements LogRecord 
    {
        private static final long serialVersionUID = -548023639986297780L;
        /** The ID of the data structure record corresponding to the
         *  registration with managed set from which locators were removed.
         *  @serial
         */
        private UUID registrationID;
        /** The set of locators removed from the registration's managed set 
         *  of locators.
         *  @serial
         */
        private LookupLocator[] locators;
        /** Constructs this class and stores the ID and locators to remove */
        public LocsRemovedFromRegistrationLogObj(UUID registrationID,
                                                 LookupLocator[] locators)
        {
            this.registrationID = registrationID;
            this.locators       = locators;
        }
        /** Modifies this service's state by removing the set of locators
         *  from registration's managed set of locators, as well as by updating
         *  the set of all locators (across all registrations) to discover.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,fiddlerImpl.debugLogWriter,
                        "Log recovery: apply removing from locators to "
                        +"discover for registration\n              (ID = "
                        +registrationID+") ...");
            fiddlerImpl.removeLocatorsDo(registrationID,
                                         fiddlerImpl.registrationByID,
                                         locators);
        }
    }//end LocsRemovedFromRegistrationLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a lease on an existing registration (granted by the current back-end
     * server of the lookup discovery service -- the lease grantor) is renewed.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LeaseRenewedLogObj implements LogRecord {
        private static final long serialVersionUID = 3784114423855097981L;
        /** The ID of the data structure record corresponding to the
         *  registration whose lease was renewed.
         *  @serial
         */
        private UUID registrationID;
        /** The identifier assigned by the lease grantor to the lease that was
         *  renewed.
         *  @serial
         */
        private long leaseID;
        /** The new absolute time of expiration of the lease that was renewed.
         *  @serial
         */
        private long expiration;
        /** Constructs this class and stores the IDs and the expiration time */
        public LeaseRenewedLogObj(UUID registrationID,
                                  long leaseID,
                                  long expiration)
        {
            this.registrationID = registrationID;
            this.leaseID        = leaseID;
            this.expiration     = expiration;
        }
        /** Modifies this service's state by renewing the lease with ID equal
         *  to this class' leaseID field, and which corresponds to the 
         *  regInfo record. The lease will be renewed to have a new expiration
         *  time equal to the value of the expiration field.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            try {
                writeInfoLog(fiddlerImpl.debugAllWriter,
                             fiddlerImpl.debugLogWriter,
                             "Log recovery: apply renewing lease for "
                             +"registration\n              (ID = "
                             +registrationID+") ...");
                /* Set curTime to 0 to guarantee no UnknownLeaseException */
                fiddlerImpl.renewLeaseAbs(registrationID, 
                                          fiddlerImpl.registrationByID,
                                          leaseID, expiration, 0);
            } catch (UnknownLeaseException e) {
                /* this exception should never occur when recovering */
            }
        }
    }//end LeaseRenewedLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a set of leases from a <code>LeaseMap</code> are renewed.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LeasesRenewedLogObj implements LogRecord {
        private static final long serialVersionUID = 5229943723792437288L;
        /** The set of unique identifiers each assigned to a registration that
         *  corresponds to one of the leases that was renewed.
         *  @serial
         */
        private UUID[] registrationIDs;
        /** The set of identifiers each assigned by the lease grantor to one
         *  of the leases that was renewed.
         *  @serial
         */
        private long[] leaseIDs;
        /** The set of new absolute expiration times of each lease that was
         *  renewed.
         *  @serial
         */
        private long[] expirations;
        /** Constructs this class and stores the sets of IDs and the set of
         *  expiration times.
         */
        public LeasesRenewedLogObj(UUID[] registrationIDs,
                                   long[] leaseIDs,
                                   long[] expirations)
        {
            this.registrationIDs = registrationIDs;
            this.leaseIDs        = leaseIDs;
            this.expirations     = expirations;
        }
        /** Modifies this service's state by renewing, with the corresponding
         *  expiration time, each of the leases specified by the stored IDs.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,
                         fiddlerImpl.debugLogWriter,
                         "Log recovery: apply renewing leases corresponding "
                         +"to "+registrationIDs.length
                         +" registration IDs ...");
            fiddlerImpl.renewLeasesAbs(registrationIDs, leaseIDs, expirations);
        }
    }//end LeasesRenewedLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a lease on an existing registration (granted by the current back-end
     * server of the lookup discovery service -- the lease grantor) is
     * cancelled.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LeaseCancelledLogObj implements LogRecord {
        private static final long serialVersionUID = -7005467028431146532L;
        /** The ID of the data structure record corresponding to the
         *  registration whose lease was cancelled.
         *  @serial
         */
        private UUID registrationID;
        /** The identifier assigned by the lease grantor to the lease that was
         *  cancelled.
         *  @serial
         */
        private long leaseID;
        /** Constructs this class and stores the IDs corresponding to the
         *  lease that was cancelled.
         */
        public LeaseCancelledLogObj(UUID registrationID,
                                    long leaseID) {
            this.registrationID = registrationID;
            this.leaseID        = leaseID;
        }
        /** Modifies this service's state by canceling the lease on the
         *  registration with ID equal to that stored by the constructor,
         *  as well as by updating the managed set of groups and locators 
         *  (across all registrations) in the appropriate way.
         *
         *  Note that the cancelation of a lease typically involves the
         *  modification of the managed sets in the discovery manager, which
         *  usually involves starting the discovery protocol. Since an
         *  IOException can occur when the discovery protocol fails to start,
         *  and since such a situation is un-recoverable, this method does
         *  the following: catches the exception, informs this service's
         *  administrator by displaying the stack trace, and exits.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            try {
                writeInfoLog(fiddlerImpl.debugAllWriter,
                             fiddlerImpl.debugLogWriter,
                             "Log recovery: apply cancelling lease for "
                             +"registration\n              (ID = "
                             +registrationID+") ...");
                fiddlerImpl.cancelLeaseDo(registrationID, 
                                          fiddlerImpl.registrationByID,
                                          leaseID);
            } catch(IOException e) {
                System.out.println
                               ("During log recovery (apply cancelLease) -- "
                                +"multicast request protocol failed to start\n"
                                +e.toString());
                fiddlerImpl.destroyDo();
            } catch (UnknownLeaseException e) {
                /* this exception should never occur when recovering */
            }
        }
    }//end LeaseCancelledLogObj

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a set of leases from a <code>LeaseMap</code> are cancelled.
     * 
     * @see FiddlerImpl.LocalLogHandler
     */
    private static class LeasesCancelledLogObj implements LogRecord {
        private static final long serialVersionUID = -3751964829217612505L;
        /** The set of unique identifiers each assigned to a registration that
         *  corresponds to one of the leases that was cancelled.
         *  @serial
         */
        private UUID[] registrationIDs;
        /** The set of identifiers each assigned by the lease grantor to one
         *  of the leases that was cancelled.
         *  @serial
         */
        private long[] leaseIDs;
        /** Constructs this class and stores the IDs corresponding to the
         *  leases that were cancelled.
         */
        public LeasesCancelledLogObj(UUID[] registrationIDs,
                                     long[] leaseIDs)
        {
            this.registrationIDs = registrationIDs;
            this.leaseIDs        = leaseIDs;
        }
        /** Modifies this service's state by canceling each of the leases
         *  specified by the stored IDs.
         *  @see FiddlerImpl.LocalLogHandler#applyUpdate
         */
        public void apply(FiddlerImpl fiddlerImpl) {
            writeInfoLog(fiddlerImpl.debugAllWriter,
                         fiddlerImpl.debugLogWriter,
                         "Log recovery: apply cancelling leases corresponding "
                         +"to "+registrationIDs.length
                         +" registration IDs ...");
            /* Because unknown leases were not weeded out, exceptions can
             * occur and be propagated upward, but they can be ignored.
             */
            fiddlerImpl.cancelLeasesDo(registrationIDs, leaseIDs);
        }
    }//end LeasesCancelledLogObj
    /* END Persistent State Logging Interfaces & Classes ------------------- */

    /* BEGIN Private Debug Facility Methods -------------------------------- */
    /* Note:
     * 
     * In order to query, modify and/or use the debug facility, the policy
     * file of the fiddler implementation of the lookup discovery service
     * must grant the following permissions to the appropriate codebases;
     * otherwise, a SecurityException will result.
     * <pre>
     * grant codebase "file:/<directory_path>/sun-util.jar" { 
     *    . . .
     *    permission java.util.PropertyPermission
     *                               "com.sun.jini.fiddler.server.*", "read";
     *    permission java.io.FilePermission "./*", "read,write,execute,delete";
     *    . . .
     * }
     * grant codebase "file:/<directory_path>/jars/fiddler.jar" { 
     *    . . .
     *    permission java.util.PropertyPermission "*", "read,write"; 
     *    . . .
     * }
     * </pre>
     */

    /**
     * Changes the value of the <code>com.sun.jini.fiddler.server.debug</code>
     * system property to the new value, and resets all of the debug
     * <code>PrintWriter</code>s to the appropriate value corresponding to
     * the input property values.
     *
     * @param newValues <code>String</code> consisting of a comma-separated set
     *                  of  valid debug property values
     *
     * @return <code>boolean</code> value that indicates whether or not the
     *         debug property was successfully set
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     * @see com.sun.jini.fiddler.FiddlerAdminProxy#setDebugProperty
     * @see com.sun.jini.fiddler.FiddlerImpl#setDebugProperty
     */
    private boolean setDebugPropertyDo(String newValues) {
        boolean propSet = false;
        try {
            Properties props = (Properties)System.getProperties();
            props.put("com.sun.jini.fiddler.server.debug",newValues);
            System.setProperties(props);
            setDebugWriters();
            propSet = true;
        } catch (Throwable e) { } // return false if anything goes wrong
	return propSet;
    }//end setDebugPropertyDo

    /* In order to use the debug facilities, the policy file must include the
     * following permissions for the codebase from which the debug facility
     * is loaded (sun-util.jar); otherwise, a SecurityException will be thrown:
     *
     * permission java.util.PropertyPermission
     *                               "com.sun.jini.fiddler.server.*", "read";
     * permission java.io.FilePermission "./", "read,write,execute,delete"; 
     */
    private void setDebugWriters() {
        debugObj = new Debug("com.sun.jini.fiddler.server.debug");
        debugAllWriter     = debugObj.getWriter("all");
        debugStartupWriter = debugObj.getWriter("startup");
        debugTasksWriter   = debugObj.getWriter("tasks");
        debugEventsWriter  = debugObj.getWriter("events");
        debugGroupsWriter  = debugObj.getWriter("groups");
        debugLocsWriter    = debugObj.getWriter("locators");
        debugDiscardWriter = debugObj.getWriter("discard");
        debugLeaseWriter   = debugObj.getWriter("lease");
        debugRegWriter     = debugObj.getWriter("registration");
        debugLogWriter     = debugObj.getWriter("log");
    }//end setDebugWriters

    /* Writes the elements of the array to a single line of output */
    private static void writeArrayElements(Object[] arr, PrintWriter pw) {
        if( (arr == null) || (pw == null) ) return;
        if(arr.length <= 0) {
            pw.println("{ }");
            return;
        }
        pw.print("{ "+arr[0]);
        for(int i=1;i<arr.length;i++){
            pw.print(", "+arr[i]);
        }
        pw.println(" }");
    }//end writeArrayElements

    /* Writes the group names to a single line of output */
    private static void writeGroupArray(String[] groups, PrintWriter pw) {
        if(pw == null) return;
        if(groups == null) {
            pw.println("{ ALL_GROUPS }");
            return;
        }
        if(groups.length <= 0) {
            pw.println("{ NO_GROUPS }");
            return;
        }
        if(groups[0].compareTo("") == 0) {
            pw.print("{ The PUBLIC Group");
        } else {
            pw.print("{ "+groups[0]);
        }
        for(int i=1;i<groups.length;i++) {
            if(groups[i].compareTo("") == 0) {
                pw.print(", The PUBLIC Group");
            }else{
                pw.print(", "+groups[i]);
            }
        }
        pw.println(" }");
    }//end writeGroupArray

    /* Writes each registrar's locator on a single line of output */
    private static void writeRegisrarsArray(ServiceRegistrar[] regs,
                                            PrintWriter pw) {
        if( (regs == null) || (pw == null) ) return;
        if(regs.length > 0) {
            try{
                LookupLocator loc = regs[0].getLocator();
                pw.print("{ "+loc);
            }catch(RemoteException e){
                pw.println("RemoteException retrieving registrar[0] " 
                                   +e.toString());
            }
            for(int i=1;i<regs.length;i++){
                try{
                    LookupLocator loc = regs[i].getLocator();
                    pw.print(", "+loc);
                }catch(RemoteException e){
                    pw.println("RemoteException retrieving registrar["
                                       +i+"] "+e.toString());
                }
            }
            pw.println(" }");
        }else{
            pw.println("NO REGISTRARS for Event");
        }
    }//end writeRegisrarsArray

    /* Writes data on a single attribute to a file or standard output */
    private static void writeAttribute(Entry attr, PrintWriter pw) {
        if( (attr == null) || (pw == null) ) return;
        if(attr instanceof BasicServiceType) {
            pw.println("  attribute = BasicServiceType");
            pw.println("    Display Name = "
                               +((BasicServiceType)(attr)).getDisplayName());
            pw.println("    Description  = "
                          +((BasicServiceType)(attr)).getShortDescription());
        } else if(attr instanceof ServiceInfo) {
            pw.println("  attribute = ServiceInfo");
            pw.println("    Service Name         = "
                                       +((ServiceInfo)(attr)).name);
            pw.println("    Service Manufacturer = "
                                      +((ServiceInfo)(attr)).manufacturer);
            pw.println("    Service Vendor       = "
                                       +((ServiceInfo)(attr)).vendor);
            pw.println("    Service Version      = "
                                       +((ServiceInfo)(attr)).version);
            pw.println("    Service Model        = "
                                       +((ServiceInfo)(attr)).model);
            pw.println("    Service Serial #     = "
                                      +((ServiceInfo)(attr)).serialNumber);
        } else {
            pw.println("  attribute = "+attr);
        }
    }//end writeAttribute

    /* Writes data on each attribute in a set to a file or standard output */
    private static void writeAttributes(Entry[] attrs, PrintWriter pw) {
        if( (attrs == null) || (pw == null) ) return;
        for(int i=0;i<attrs.length;i++) {
            if(attrs[i] == null) continue;
            writeAttribute(attrs[i],pw);
        }
        pw.println("");
    }//end writeAttributes

    /* Writes startup data to file or standard output */
    private void writeInfoStartup() {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                     debugAllWriter : debugStartupWriter );
        if(pw == null) return;
        String propVal = System.getProperty
                                     ("com.sun.jini.fiddler.server.debug");
        pw.println("\nDebug property value(s): "
                   +"com.sun.jini.fiddler.server.debug = "+propVal);
        pw.println("\nPersistent state directory:         "+logDirname);
    }//end writeInfoStartup

    /* Writes startup information related to the JoinManager */
    private void writeInfoJoinManager() {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                     debugAllWriter : debugStartupWriter );
        if(pw == null) return;
        pw.print("Lookup service group(s) to join:    ");
        writeGroupArray(thisServicesGroups,pw);
        pw.print("Specific lookup service(s) to join: ");
        writeArrayElements(thisServicesLocators,pw);
        pw.println("Attributes to register in each lookup service: ");
        writeAttributes(thisServicesAttrs,pw);
        pw.println("");
	if (serviceID == null) {
            pw.println("Create JoinManager >>> request new serviceID");
        } else {
            pw.println("Create JoinManager >>> use current serviceID");
            writeInfoServiceID();
        }
    }//end writeInfoJoinManager

    /* Writes startup information related to this service's ID */
    private void writeInfoServiceID() {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                     debugAllWriter : debugStartupWriter );
        if(pw == null) return;
        pw.println("serviceID = "+(FiddlerImpl.this.serviceID).toString());
    }//end writeInfoServiceID

    /* Writes information related to the run method of a particular task */
    private void writeInfoTasks(String str) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugTasksWriter );
        if(pw == null) return;
        pw.println(str);
    }//end writeInfoTasks

    /* Writes information about events that are sent */
    private void writeInfoEvents(String str) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugEventsWriter );
        if(pw == null) return;
        pw.println(str);
    }//end writeInfoEvents

    /* Writes information about events that are sent */
    private void writeInfoEvents(ServiceRegistrar[] regs,
                                 long eventID,
                                 long seqNum,
                                 MarshalledObject handback,
                                 boolean discarded)
    {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugEventsWriter );
        if(pw == null) return;
        String discardedStr = (discarded == true?"DISCARDED":"DISCOVERED");
        Object hb = null;
        try {
            hb = handback.get();
        } catch (ClassNotFoundException e) {
            pw.println("ClassNotFoundException when unmarshalling handback: "
                       +e.toString());
        } catch (IOException e) {
            pw.println("IOException when unmarshalling handback: "
                       +e.toString());
        }
        pw.println("\n"+discardedStr+" Event:");
        pw.println("  EventID  = "+eventID);
        pw.println("  SeqNum   = "+seqNum);
        pw.println("  handback = "+hb);
        pw.print("  Registrars = ");
        writeRegisrarsArray(regs,pw);
        pw.println("");
    }//end writeInfoEvents

    /* Writes group state information over all active registrations */
    private void writeInfoGroups() {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugGroupsWriter );
        if(pw == null) return;
        String[] allGroups = discoveryMgr.getGroups();
        pw.print("  Group(s) over all registrations: ");
        writeGroupArray(allGroups,pw);
    }//end writeInfoGroups

    /* Writes group state information over all active registrations */
    private void writeInfoGroups(String headerStr) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugGroupsWriter );
        if(pw == null) return;
        pw.print(headerStr);
        writeInfoGroups();
    }//end writeInfoGroups

    /* Writes locator state information over all active registrations */
    private void writeInfoLocators() {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLocsWriter );
        if(pw == null) return;
        LookupLocator[] allLocators = discoveryMgr.getLocators();
        pw.print("  Locator(s) over all registrations: ");
        writeArrayElements(allLocators,pw);
    }//end writeInfoLocators

    /* Writes locator state information over all active registrations */
    private void writeInfoLocators(String headerStr) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLocsWriter );
        if(pw == null) return;
        pw.print(headerStr);
        writeInfoLocators();
    }//end writeInfoLocators

    /* Writes information useful to debugging the discard process */
    private void writeInfoDiscard(String str, UUID regID) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugDiscardWriter );
        if(pw == null) return;
        if((str != null) && (regID != null)) {
            pw.println(str+" registrationID = "+regID);
        }
    }//end writeInfoDiscard

    /* Writes information useful to debugging the discard process */
    private void writeInfoDiscard(String str) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugDiscardWriter );
        if(pw == null) return;
        pw.println(str);
    }//end writeInfoDiscard

    /* Writes information useful to debugging the leasing mechanism */
    private void writeInfoLease(String str, UUID regID, long leaseID) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLeaseWriter );
        if(pw == null) return;
        if((str != null) && (regID != null)) {
            pw.println(str+" (registrationID,leaseID) = ("
                          +regID+", "+leaseID+")");
        }
    }//end writeInfoLease

    /* Writes information useful to debugging the leasing mechanism */
    private void writeInfoLease(String str, UUID regID) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLeaseWriter );
        if(pw == null) return;
        if((str != null) && (regID != null)) {
            pw.println(str+" registrationID = "+regID);
        }
    }//end writeInfoLease

    /* Writes information useful to debugging the leasing mechanism */
    private void writeInfoLease(String str,long leaseID) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLeaseWriter );
        if(pw == null) return;
        if(str != null) {
            pw.println(str+" leaseID = "+leaseID);
        }
    }//end writeInfoLease

    /* Writes information useful to debugging the leasing mechanism */
    private void writeInfoLease(String str) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLeaseWriter );
        if(pw == null) return;
        pw.println(str);
    }//end writeInfoLease

    /* Writes information useful to debugging the registration mechanism */
    private void writeInfoRegistration(String str, UUID regID) {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugRegWriter );
        if(pw == null) return;
        if((str != null) && (regID != null)) {
            pw.println(str+" registrationID = "+regID);
        }
    }//end writeInfoRegistration

    /* Writes information useful to debugging the logging mechanism */
    private static void writeInfoLog(PrintWriter debugAllWriter,
                                     PrintWriter debugLogWriter,
                                     String str)
    {
        PrintWriter pw = ( (debugAllWriter != null) ?
                                         debugAllWriter : debugLogWriter );
        if(pw == null) return;
        pw.println(str);
    }//end writeInfoLog

    /* Writes information useful to debugging the addLogRecord method */
    private void writeInfoAddLogRecord(LogRecord rec) {
        if( (debugAllWriter == null) && (debugLogWriter == null) ) return;
        String logStr = "Logging a state change: Unknown log record instance";
        /* JoinAdmin */
        if(rec instanceof LookupAttrsAddedLogObj) {
            logStr = "Logging state change: lookup attributes added";
        } else if (rec instanceof LookupAttrsModifiedLogObj) {
            logStr = "Logging state change: lookup attributes modified";
        } else if (rec instanceof LookupGroupsChangedLogObj) {
            logStr = "Logging state change: groups to join changed";
        } else if (rec instanceof LookupLocatorsChangedLogObj) {
            logStr = "Logging state change: locators to join changed";
        /* FiddlerAdmin */
        } else if (rec instanceof LeaseDurationBoundSetLogObj) {
            logStr = "Logging state change: lease duration bound changed";
        } else if (rec instanceof SnapshotWeightSetLogObj) {
            logStr = "Logging state change: snapshot weight factor changed";
        } else if (rec instanceof LogToSnapshotThresholdSetLogObj) {
            logStr = "Logging state change: log-to-snapshot threshold changed";
        } else if (rec instanceof DebugPropertySetLogObj) {
            logStr = "Logging state change: debug property changed";
        /* LookupDiscoveryService */
        } else if (rec instanceof RegistrationGrantedLogObj) {
            logStr = "Logging state change: new registration granted";
        /* LookupDiscoveryRegistration */
        } else if (rec instanceof GroupsAddedToRegistrationLogObj) {
            logStr = "Logging state change: added new groups to "
                     +"registration's set of groups to discover";
        } else if (rec instanceof GroupsSetInRegistrationLogObj) {
            logStr = "Logging state change: replaced registration's set of "
                     +"groups to discover";
        } else if (rec instanceof GroupsRemovedFromRegistrationLogObj) {
            logStr = "Logging state change: removed groups from "
                     +"registration's set of groups to discover";
        } else if (rec instanceof LocsAddedToRegistrationLogObj) {
            logStr = "Logging state change: added new locators to "
                     +"registration's set of locators to discover";
        } else if (rec instanceof LocsSetInRegistrationLogObj) {
            logStr = "Logging state change: replaced registration's set of "
                     +"locators to discover";
        } else if (rec instanceof LocsRemovedFromRegistrationLogObj) {
            logStr = "Logging state change: removed locators from "
                     +"registration's set of locators to discover";
        } else if (rec instanceof LeaseRenewedLogObj) {
            logStr = "Logging state change: registration's lease renewed";
        } else if (rec instanceof LeasesRenewedLogObj) {
            logStr = "Logging state change: set of leases renewed for a "
                     +"set of registrations";
        } else if (rec instanceof LeaseCancelledLogObj) {
            logStr = "Logging state change: registration's lease cancelled";
        } else if (rec instanceof LeasesCancelledLogObj) {
            logStr = "Logging state change: set of leases cancelled for a "
                     +"set of registrations";
        }
        writeInfoLog(debugAllWriter,debugLogWriter,logStr);
    }//end writeInfoAddLogRecord
    /* END Private Debug Facility Methods ---------------------------------- */
    /* *************** END Private NON-Static Utility Methods ************** */

}/*end class FiddlerImpl */
